
/*
 * File FormulaBuilder.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __API_FormulaBuilder__
#define __API_FormulaBuilder__

#include <string>
#include <ostream>

#include "Helper.hpp"

namespace Api {

using namespace std;


class ApiException
{
public:
  ApiException(string msg)
  : _msg(msg) {}

 
  string msg() const { return _msg; }
protected:
  string _msg;
};

class FormulaBuilderException
: public ApiException
{
public:
  FormulaBuilderException(string msg)
  : ApiException(msg) {}
};


class SortMismatchException
: public FormulaBuilderException
{
public:
  SortMismatchException(string msg)
  : FormulaBuilderException(msg) {}
};

class InvalidTPTPNameException
: public FormulaBuilderException
{
public:
  InvalidTPTPNameException(string msg, string name)
  : FormulaBuilderException(msg), _name(name) {}

 
  string name() const { return _name; }
private:
  string _name;
};

typedef unsigned Var;
class Sort;
class Function;
class Predicate;
class Term;
class Formula;
class AnnotatedFormula;

class FormulaBuilder
{
public:
 
  FormulaBuilder(bool checkNames=true, bool checkBindingBoundVariables=true,
      bool allowImplicitlyTypedVariables=true, bool outputDummyNames=false);

  enum Connective {
    TRUE,
    FALSE,
    ATOM,
    AND,
    OR,
    IMP,
    IFF,
    XOR,
    NOT,
    FORALL,
    EXISTS,
   
    ITE
  };

  enum InterpretedPredicate {
    INT_GREATER,
    INT_GREATER_EQUAL,
    INT_LESS,
    INT_LESS_EQUAL,
  };

 
  enum Annotation {
   
    AXIOM,
   
    ASSUMPTION,
   
    CONJECTURE
  };

 
  Sort sort(const string& sortName);
 
  Sort integerSort();
 
  Sort rationalSort();
 
  Sort realSort();

 
  static Sort defaultSort();

 
  Var var(const string& varName);

 
  Var var(const string& varName, Sort varSort);

 
  Function function(const string& funName, unsigned arity, bool builtIn=false);

 
  Function function(const string& funName, unsigned arity, Sort rangeSort, Sort* domainSorts, bool builtIn=false);

 
  Function integerConstant(int i);
 
  Function integerConstant(string i);

 
  Predicate predicate(const string& predName, unsigned arity, bool builtIn=false);

 
  Predicate predicate(const string& predName, unsigned arity, Sort* domainSorts, bool builtIn=false);

 
  Predicate interpretedPredicate(InterpretedPredicate symbol);

 
  string getSortName(Sort s);
 
  string getPredicateName(Predicate p);
 
  string getFunctionName(Function f);
 
  string getVariableName(Var v);




  void addAttribute(Predicate p, string name, string value);
  unsigned attributeCount(Predicate p);
  string getAttributeName(Predicate p, unsigned index);
  string getAttributeValue(Predicate p, unsigned index);
  string getAttributeValue(Predicate p, string attributeName);

  void addAttribute(Function fn, string name, string value);
  unsigned attributeCount(Function fn);
  string getAttributeName(Function fn, unsigned index);
  string getAttributeValue(Function fn, unsigned index);
  string getAttributeValue(Function fn, string attributeName);

  void addAttribute(Sort s, string name, string value);
  unsigned attributeCount(Sort s);
  string getAttributeName(Sort s, unsigned index);
  string getAttributeValue(Sort s, unsigned index);
  string getAttributeValue(Sort s, string attributeName);

 
  Term varTerm(const Var& v);

 
  Term term(const Function& f,const Term* args);

 
  Formula atom(const Predicate& p, const Term* args, bool positive=true);

 
  Formula equality(const Term& lhs,const Term& rhs, bool positive=true);

 
  Formula equality(const Term& lhs,const Term& rhs, Sort sort, bool positive=true);

 
  Formula trueFormula();

 
  Formula falseFormula();

 
  Formula negation(const Formula& f);

 
  Formula formula(Connective c,const Formula& f1,const Formula& f2);

 
  Formula formula(Connective q,const Var& v,const Formula& f);

 
  Formula formula(Connective c,const Formula& cond,const Formula& thenBranch,const Formula& elseBranch);

  

 
  Term term(const Function& c);

 
  Term term(const Function& f,const Term& t);

 
  Term term(const Function& f,const Term& t1,const Term& t2);

 
  Term term(const Function& f,const Term& t1,const Term& t2,const Term& t3);

 
  Formula formula(const Predicate& p);

 
  Formula formula(const Predicate& p,const Term& t);

 
  Formula formula(const Predicate& p,const Term& t1,const Term& t2);

 
  Formula formula(const Predicate& p,const Term& t1,const Term& t2,const Term& t3);

 
  AnnotatedFormula annotatedFormula(Formula f, Annotation a, string name="");


 
  Term substitute(Term original, Var v, Term t);
 
  Formula substitute(Formula f, Var v, Term t);
  AnnotatedFormula substitute(AnnotatedFormula af, Var v, Term t);

 
  Term replaceConstant(Term original, Term replaced, Term target);

 
  Formula replaceConstant(Formula f, Term replaced, Term target);
  AnnotatedFormula replaceConstant(AnnotatedFormula f, Term replaced, Term target);
private:
  FBHelper _aux;

  friend class StringIterator;
  friend class Formula;
  friend class AnnotatedFormula;
};

}

std::ostream& operator<< (std::ostream& str,const Api::Sort& sort);
std::ostream& operator<< (std::ostream& str,const Api::Formula& f);
std::ostream& operator<< (std::ostream& str,const Api::AnnotatedFormula& f);

namespace Api
{

using namespace std;
using namespace Lib;

class Problem;

class OutputOptions
{
public:
 
  static bool sortedEquality() { return _sortedEquality; }
  static void setSortedEquality(bool newVal) { _sortedEquality = newVal; }

 
  static bool tffFormulas() { return _tffFormulas; }
  static void setTffFormulas(bool newVal) { _tffFormulas = newVal; }

 
  static bool assignFormulaNames() { return _assignFormulaNames; }
  static void setAssignFormulaNames(bool newVal);
private:
  static bool _sortedEquality;
  static bool _tffFormulas;
  static bool _assignFormulaNames;
};

class StringIterator
{
public:
  StringIterator() : _impl(0) {};
  explicit StringIterator(const VirtualIterator<string>& vit);
  ~StringIterator();
  StringIterator(const StringIterator& it);
  StringIterator& operator=(const StringIterator& it);

 
  bool hasNext();
 
  string next();

private:
  VirtualIterator<string>* _impl;
};

class Sort
{
public:
  Sort() {}
  explicit Sort(unsigned num) : _num(num) {}
  operator unsigned() const { return _num; }
private:
  unsigned _num;
};

class Function
{
public:
  Function() {}
  explicit Function(unsigned num) : _num(num) {}
  operator unsigned() const { return _num; }
private:
  unsigned _num;
};

class Predicate
{
public:
  Predicate() {}
  explicit Predicate(unsigned num) : _num(num) {}
  operator unsigned() const { return _num; }
private:
  unsigned _num;
};

class Term
{
public:
  Term() : content(0) {}

  string toString() const;

 
  bool isNull() const { return content==0; }

 
  bool isVar() const;

 
  Var var() const;

 
  Function functor() const;

 
  unsigned arity() const;

 
  Term arg(unsigned i);

 
  Sort sort() const;

  operator Kernel::TermList() const;
  explicit Term(Kernel::TermList t);
  explicit Term(Kernel::TermList t, ApiHelper aux);

  bool operator==(const Term& o) const {
    return toString()==o.toString();
  }
private:
  size_t content;
  ApiHelper _aux;

  friend class FormulaBuilder;
  friend class FBHelperCore;
};

class Formula
{
public:
  Formula() : form(0) {}

  string toString() const;

 
  bool isNull() const { return form==0; }

 
  bool isTrue() const;

 
  bool isFalse() const;

 
  bool isNegation() const;

 
  FormulaBuilder::Connective connective() const;

 
  Predicate predicate() const;

 
  bool atomPolarity() const;

 
  unsigned argCnt() const;

 
  Formula formulaArg(unsigned i);

 
  Term termArg(unsigned i);

 
  StringIterator freeVars();

 
  StringIterator boundVars();

  operator Kernel::Formula*() const { return form; }
  explicit Formula(Kernel::Formula* f) : form(f) {}
  explicit Formula(Kernel::Formula* f, ApiHelper aux) : form(f), _aux(aux) {}

  bool operator==(const Formula& o) const {
    return toString()==o.toString();
  }
private:
  Kernel::Formula* form;
  ApiHelper _aux;

  friend class FormulaBuilder;
  friend class FBHelperCore;
  friend class Problem;
};

class AnnotatedFormula
{
public:
  AnnotatedFormula() : unit(0) {}

  string toString() const;

 
  string name() const;

 
  bool isNull() const { return unit==0; }

 
  StringIterator freeVars();

 
  StringIterator boundVars();

 
  FormulaBuilder::Annotation annotation() const;

 
  Formula formula();

  operator Kernel::Unit*() const { return unit; }
  explicit AnnotatedFormula(Kernel::Unit* fu) : unit(fu) {}
  explicit AnnotatedFormula(Kernel::Unit* fu, ApiHelper aux) : unit(fu), _aux(aux) {}

  bool operator==(const AnnotatedFormula& o) const {
    return toString()==o.toString();
  }
private:
  static void assignName(AnnotatedFormula& form, string name);

  Kernel::Unit* unit;
  ApiHelper _aux;

  friend class FormulaBuilder;
  friend class Problem;
};

}

#endif 
