
/*
 * File Helper_Internal.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __Api_Helper_Internal__
#define __Api_Helper_Internal__

#include "Forwards.hpp"

#include "FormulaBuilder.hpp"

#include "Helper.hpp"

#include "Lib/Environment.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Connective.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"
#include "Kernel/Unit.hpp"

#include "Shell/TPTP.hpp"
#include "Shell/VarManager.hpp"

namespace Api {

using namespace Shell;

typedef Kernel::Formula::VarList VarList;

class DefaultHelperCore
{
public:
  DefaultHelperCore() {}
  virtual ~DefaultHelperCore() {
  }

  static DefaultHelperCore* instance();
  virtual string getVarName(Var v) const;
  string toString(Kernel::TermList t) const;
  string toString(const Kernel::Term* t0) const;
  string toString(const Kernel::Formula* f) const;
  string toString(const Kernel::Clause* clause) const;
  string toString (const Kernel::Unit* unit) const;

  virtual VarManager::VarFactory* getVarFactory() { return 0; };

  virtual bool isFBHelper() const { return false; }
  virtual bool outputDummyNames() const { return false; }
private:
  struct Var2NameMapper;
public:
  StringIterator getVarNames(VarList* l);

  static string getDummyName(bool pred, unsigned functor);
  static string getDummyName(const Kernel::Term* t);

  string getSymbolName(bool pred, unsigned functor) const;
  string getSymbolName(const Kernel::Term* t) const;
};

class FBHelperCore
: public DefaultHelperCore
{
public:
  FBHelperCore() : nextVar(0), refCtr(0), varFact(*this), _unaryPredicate(0)
  {
  }

  void incRef()
  {
    CALL("ApiHelperCore::incRef");

    refCtr++;
  }

 
  void decRef()
  {
    CALL("ApiHelperCore::decRef");
    ASS_G(refCtr,0);

    refCtr--;
    if(refCtr==0) {
      delete this;
    }
  }

  virtual bool isFBHelper() const { return true; }


  Term term(const Function& f,const Term* args, unsigned arity);
  Formula atom(const Predicate& p, bool positive, const Term* args, unsigned arity);
  virtual string getVarName(Var v) const;
  Sort getVarSort(Var v) const;
  Var getVar(string varName, Sort varSort);

  virtual VarManager::VarFactory* getVarFactory()
  { return &varFact; }

 
  bool _checkNames;
 
  bool _checkBindingBoundVariables;

  bool _allowImplicitlyTypedVariables;

  bool _outputDummyNames;

  virtual bool outputDummyNames() const { return _outputDummyNames; }

 
  unsigned getUnaryPredicate();

  Sort getSort(const Api::Term t);
  void ensureArgumentsSortsMatch(BaseType* type, const Api::Term* args);
  void ensureEqualityArgumentsSortsMatch(const Api::Term arg1, const Api::Term arg2);

  typedef pair<string,string> AttribPair;
  typedef Stack<AttribPair> AttribStack;

  AttribStack& getSortAttributes(unsigned srt)
  {
    CALL("ApiHelperCore::getSortAttributes");
    AttribStack* res;
    _sortAttributes.getValuePtr(srt, res);
    return *res;
  }

  AttribStack& getPredicateAttributes(unsigned pred)
  {
    CALL("ApiHelperCore::getPredicateAttributes");
    AttribStack* res;
    _predicateAttributes.getValuePtr(pred, res);
    return *res;
  }

  AttribStack& getFunctionAttributes(unsigned func)
  {
    CALL("ApiHelperCore::getFunctionAttributes");
    AttribStack* res;
    _functionAttributes.getValuePtr(func, res);
    return *res;
  }

  static void addAttribute(AttribStack& stack, string name, string value);
private:

  DHMap<unsigned,AttribStack > _sortAttributes;
  DHMap<unsigned,AttribStack > _predicateAttributes;
  DHMap<unsigned,AttribStack > _functionAttributes;

  struct FBVarFactory : public VarManager::VarFactory
  {
    explicit FBVarFactory(FBHelperCore& parent) : _parent(parent) {}
    virtual unsigned getVarAlias(unsigned var);
    virtual string getVarName(unsigned var);

    FBHelperCore& _parent;
  };

 
  Map<string,Var> vars;
 
  Map<Var,string> varNames;
 
  Map<Var,Sort> varSorts;
 
  Var nextVar;

  int refCtr;

  FBVarFactory varFact;

 
  unsigned _unaryPredicate;
};

class SingleVarApplicator
{
public:
  SingleVarApplicator(unsigned var, TermList term) : _srcVar(var), _tgtTerm(term) {}
  TermList apply(unsigned var)
  {
    if(var!=_srcVar) {
      return TermList(var, false);
    }
    return _tgtTerm;
  }
private:
  unsigned _srcVar;
  TermList _tgtTerm;
};


}

#endif 
