
/*
 * File Problem.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __API_Problem__
#define __API_Problem__

#include "FormulaBuilder.hpp"

namespace Api {

class Problem;

class AnnotatedFormulaIterator
{
public:
  bool hasNext();
 
  AnnotatedFormula next();

 
  void del();
private:
  bool ready;
  void* idata;

  friend class Problem;
};

class Problem
{
public:
  Problem();
  Problem(const Problem& p);
  Problem& operator=(const Problem&);
  ~Problem();

  enum InliningMode {
    INL_OFF = 0,
   
    INL_ON = 1,
    INL_AXIOMS_ONLY = 2,
   
    INL_EPR_RESTORING = 3,
   
    INL_PREDICATE_EQUIVALENCES_ONLY = 4,
   
    INL_NO_DISCOVERED_DEFS = 5,
   
    INL_NON_GROWING = 6
  };

  enum EquivalenceDiscovery {
    ED_NONE,
   
    ED_PREDICATE_EQUIVALENCES,
   
    ED_PREDICATE_DEFINITIONS,
   
    ED_ATOM_EQUIVALENCES,
   
    ED_FORMULA_EQUIVALENCES
  };

  enum PreprocessingMode {
   
    PM_SELECTION_ONLY,
   
    PM_EARLY_PREPROCESSING,
    PM_SKOLEMIZE,
    PM_CLAUSIFY
  };

 
  enum FixpointCheck {
    FC_NONE,
   
    FC_FORMULA_COUNT
  };

 
  struct PreprocessingOptions
  {
    PreprocessingOptions();
   
    explicit PreprocessingOptions(string spec);

    PreprocessingMode mode;
   
    int namingThreshold;
   
    bool preserveEpr;
    InliningMode predicateDefinitionInlining;
    bool unusedPredicateDefinitionRemoval;
    bool showNonConstantSkolemFunctionTrace;
   
    bool traceInlining;
    bool sineSelection;
   
    float sineTolerance;
   
    unsigned sineDepthLimit;
   
    bool variableEqualityPropagation;
    bool traceVariableEqualityPropagation;

   
    bool eprSkolemization;
    bool traceEPRSkolemization;

   
    bool predicateDefinitionMerging;
    bool tracePredicateDefinitionMerging;

   
    bool traceClausification;

   
    bool traceUnusedPredicateDefinitionRemoval;

   
    bool predicateIndexIntroduction;

   
    bool flatteningTopLevelConjunctions;

   
    EquivalenceDiscovery equivalenceDiscovery;
   
    bool equivalenceDiscoveryRetrievePremises;
   
    unsigned equivalenceDiscoverySatConflictLimit;

   
    bool equivalenceDiscoveryAddImplications;

   
    bool equivalenceDiscoveryRandomSimulation;

   
    bool aigInlining;

   
    bool aigBddSweeping;

   
    bool aigDefinitionIntroduction;

   
    bool aigConditionalRewriting;

   
    unsigned repetitionCount;
   
    FixpointCheck repetitionEarlyTermination;


   
    void addAsymmetricRewritingRule(Formula lhs, Formula posRhs, Formula negRhs, Formula dblRhs=Formula());

   
    void importAssymmetricRulesFrom(const PreprocessingOptions& src);
   
    void restrictPredicateEquivalenceDiscovery(size_t set1Sz, Formula* set1, size_t set2Sz, Formula* set2);

   
    void printOptionValues(ostream& out);
  private:
    friend class Problem;
    void validate() const;

    void prepareOptionsReader(OptionsReader& rdr);
    void setDefaults();

    struct Atom2LitFn;

    struct OptDataStore {
      OptDataStore();
      OptDataStore(const OptDataStore& o);
      OptDataStore& operator=(const OptDataStore& o);
      ~OptDataStore();

      void setDefaults();


      
      Stack<Formula>* lhs;
      Stack<Formula>* posRhs;
      Stack<Formula>* negRhs;
      Stack<Formula>* dblRhs;
      Stack<Kernel::Literal*>* pedSet1;
      Stack<Kernel::Literal*>* pedSet2;
    };

    bool _predicateEquivalenceDiscoveryRestricted;
    OptDataStore _ods;
  };

 
  Problem clone();

 
  void addFormula(AnnotatedFormula f);

 
  void addFromStream(istream& s, string includeDirectory="./", bool simplifySyntax=false);

 
  Problem clausify(int namingThreshold=8, bool preserveEpr=false, InliningMode predicateDefinitionInlining=INL_OFF,
      bool unusedPredicateDefinitionRemoval=true);

 
  Problem skolemize(int namingThreshold=8, bool preserveEpr=false, InliningMode predicateDefinitionInlining=INL_OFF,
      bool unusedPredicateDefinitionRemoval=true);

 
  Problem inlinePredicateDefinitions(InliningMode mode=INL_ON);

 
  Problem removeUnusedPredicateDefinitions();

 
  Problem preprocess(const PreprocessingOptions& options);

 
  void readStageSpecs(string stagesStr, size_t& stageCnt, PreprocessingOptions*& stageSpecs);

 
  Problem preprocessInStages(size_t stageCount, const PreprocessingOptions* stageSpecs);
  Problem preprocessInStages(string stagesStr);

 
  Problem performAsymetricRewriting(Formula lhs, Formula posRhs, Formula negRhs, Formula dblRhs=Formula());

 
  Problem performAsymetricRewriting(size_t cnt, Formula* lhsArray, Formula* posRhsArray, Formula* negRhsArray,
      Formula* dblRhsArray);

 
  AnnotatedFormulaIterator formulas();

 
  size_t size();

 
  bool empty();

 
  void outputTypeDefinitions(ostream& out, bool outputAllTypeDefs=false);

 
  void output(ostream& out, bool outputTypeDefs=false, bool outputAllTypeDefs=false);

 
  void outputStatistics(ostream& out);
private:

  Problem singlePreprocessingIteration(const PreprocessingOptions& options);
  bool fixpointReached(FixpointCheck fc, Problem& oldPrb, Problem& newPrb);

  class PData;
  class ProblemTransformer;

  class Preprocessor1;
  class TopLevelFlattener;
  class VariableEqualityPropagator;
  class BDDSweeper;
  class AIGInliner;
  class AIGDefinitionIntroducer;
  class AIGConditionalRewriter;
  class PredicateIndexIntroducer;
  class PredicateEquivalenceDiscoverer;
  class PredicateDefinitionMerger;
  class PredicateDefinitionInliner;
  class EPRRestoringInliner;
  class ConstantSkolemizer;
  class EPRSkolemizer;
  class UnusedPredicateDefinitionRemover;
  class Clausifier;
  class SineSelector;

  PData* _data;
};

}

#endif 
