
/*
 * File SimpleCongruenceClosure.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __SimpleCongruenceClosure__
#define __SimpleCongruenceClosure__

#include "Forwards.hpp"

#include "Lib/DArray.hpp"
#include "Lib/Deque.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/Term.hpp"

#include "DecisionProcedure.hpp"

namespace DP {

using namespace Lib;
using namespace Kernel;



class SimpleCongruenceClosure : public DecisionProcedure
{
public:
  SimpleCongruenceClosure();

  virtual void addLiterals(LiteralIterator lits);

  virtual Status getStatus(bool retrieveMultipleCores);
  virtual unsigned getUnsatCoreCount() { return _unsatEqs.size(); }
  virtual void getUnsatCore(LiteralStack& res, unsigned coreIndex);

  virtual void reset();

private:
 
  typedef pair<unsigned,unsigned> CPair;
 
  struct CEq
  {
    CEq() : c1(0), c2(0) {}
    CEq(unsigned c1, unsigned c2, Literal* lit)
     : c1(c1), c2(c2), foOrigin(true), foPremise(lit) {}
    CEq(unsigned c1, unsigned c2)
     : c1(c1), c2(c2), foOrigin(false) {}

    bool isInvalid() const { ASS_EQ(c1==0, c2==0); return c1==0; }
    string toString() const;
    string toString(SimpleCongruenceClosure& parent) const;

    unsigned c1;
    unsigned c2;
   
    bool foOrigin;
   
    Literal* foPremise;

  };

  unsigned getMaxConst() const { return _cInfos.size()-1; }
  unsigned getFreshConst();
  unsigned getSignatureConst(unsigned symbol, bool funct);
  unsigned getPairName(CPair p);


  struct FOConversionWorker;
  static bool isDistinctPred(Literal* l);
  CEq convertFOEquality(Literal* equality);
  unsigned convertFONonEquality(Literal* lit);
  unsigned convertFO(TermList trm);
  void readDistinct(Literal* lit);


  unsigned deref(unsigned c) const {
    CALL("SimpleCongruenceClosure::deref");
    unsigned repr = _cInfos[c].reprConst;
    unsigned res = (repr==0) ? c : repr;
    COND_LOG("bug", _cInfos[res].reprConst!=0, "res: "<<res);
    ASS_REP2(_cInfos[res].reprConst==0, _cInfos[res].reprConst, c);
    return res;
  }
  CPair deref(CPair p) const { return CPair(deref(p.first), deref(p.second)); }
  CPair deref(CEq p) const { return CPair(deref(p.c1), deref(p.c2)); }

  unsigned getClassSize(unsigned c) const {
    return _cInfos[c].classList.size();
  }

  bool checkPositiveDistincts(bool retrieveMultipleCores);
  Status checkNegativeDistincts(bool retrieveMultipleCores);

  void addPendingEquality(CEq eq);
  void makeProofRepresentant(unsigned c);
  void propagate();

  unsigned getProofDepth(unsigned c);
  void collectUnifyingPath(unsigned c1, unsigned c2, Stack<unsigned>& path);


  static const unsigned NO_SIG_SYMBOL;
  struct ConstInfo
  {
    void init();
    void resetEquivalences(SimpleCongruenceClosure& parent, unsigned selfIndex);

#ifdef VDEBUG
    void assertValid(SimpleCongruenceClosure& parent, unsigned selfIndex) const;
#endif


   
    unsigned sigSymbol;
    bool sigSymIsFunct;
   
    TermList term;
   
    Literal* lit;
   
    CPair namedPair;


   
    unsigned reprConst;

   
    unsigned proofPredecessor;
   
    CEq predecessorPremise;


   
    Stack<unsigned> classList;
   
    Stack<unsigned> useList;
  };

 
  DArray<ConstInfo> _cInfos;

 
  unsigned _posLitConst;
 
  unsigned _negLitConst;

 
  DHMap<pair<unsigned,bool>,unsigned> _sigConsts;

  typedef DHMap<CPair,unsigned> PairMap;
 
  PairMap _pairNames;

 
  DHMap<TermList,unsigned> _termNames;
 
  DHMap<Literal*,unsigned> _litNames;

 
  Stack<CEq> _unsatEqs;

  Deque<CEq> _pendingEqualities;

  Stack<CEq> _negEqualities;


  struct DistinctEntry
  {
    DistinctEntry(Literal* l) : _lit(l) {}
    Literal* _lit;
    Stack<unsigned> _consts;
  };
  typedef Stack<DistinctEntry> DistinctStack;
  DistinctStack _distinctConstraints;
 
  DistinctStack _negDistinctConstraints;

 
  bool _hadPropagated;
};

}

#endif 
