
/*
 * File ClauseSharing.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/InferenceStore.hpp"

#include "Saturation/SaturationAlgorithm.hpp"

#include "Shell/Statistics.hpp"

#include "ClauseSharing.hpp"

namespace Indexing
{
using namespace Lib;
using namespace Kernel;
using namespace Saturation;

void ClauseSharing::init(SaturationAlgorithm* sa)
{
  CALL("ClauseSharing::init");

  _sa=sa;
}

bool ClauseSharing::doSharing(Clause* cl)
{
  CALL("ClauseSharing::doSharing");

  ClauseSharing::InsertionResult res;
  Clause* shCl=insert(cl, res);

  if(cl!=shCl) {
    if(res==ClauseSharing::OLD_MODIFIED) {
      _sa->addNewClause(shCl);
    }
    _sa->onParenthood(shCl, cl);
    if(res==ClauseSharing::OLD) {
	if(shCl->store()==Clause::ACTIVE || shCl->store()==Clause::PASSIVE ||
	    shCl->store()==Clause::REACTIVATED || shCl->store()==Clause::SELECTED_REACTIVATED) {
	  _sa->onNonRedundantClause(shCl);
	}
    }
    ASS(res==ClauseSharing::OLD || res==ClauseSharing::OLD_MODIFIED);
    return true;
  }
  ASS(res==ClauseSharing::INSERTED || res==ClauseSharing::ALREADY_THERE);
  return false;

}

Clause* ClauseSharing::tryGet(Literal* const * lits, unsigned len)
{
  CALL("ClauseSharing::tryGet/2");

  ClauseIterator variants=_index.retrieveVariants(lits, len);
  if(!variants.hasNext()) {
    return false;
  }
  Clause* res=variants.next();
  ASS(!variants.hasNext());
  return res;
}

Clause* ClauseSharing::tryGet(Clause* c)
{
  CALL("ClauseSharing::tryGet/1");

  return tryGet(c->literals(), c->length());
}

void ClauseSharing::insertNew(Clause* c)
{
  CALL("ClauseSharing::insertNew");
  ASS_EQ(tryGet(c),0); 

  env.statistics->uniqueComponents++;

  c->incRefCnt();
  {
    TimeCounter tc(TC_SPLITTING_COMPONENT_INDEX_MAINTENANCE);
    _index.insert(c);
  }
}

Clause* ClauseSharing::insert(Clause* cl, InsertionResult& res)
{
  CALL("ClauseSharing::insert");

  ClauseIterator variants=_index.retrieveVariants(cl->literals(), cl->length());
  if(variants.hasNext()) {
    Clause* comp=variants.next();

    if(comp==cl) {
      res=ALREADY_THERE;
      return cl;
    }

    ASS(!variants.hasNext());

    BDDNode* oldCompProp=comp->prop();
    BDDNode* oldClProp=cl->prop();
    BDDNode* newCompProp=BDD::instance()->conjunction(oldCompProp, oldClProp);

    if(oldCompProp==newCompProp) {
      res=OLD;
      return comp;
    }

    comp->setProp( newCompProp );
    InferenceStore::instance()->recordMerge(comp, oldCompProp, cl, newCompProp);

    res=OLD_MODIFIED;
    return comp;

  } else {
    insertNew(cl);

    res=INSERTED;
    return cl;
  }

}

}
