
/*
 * File BDDMarkingSubsumption.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <math.h>

#include "Lib/DHSet.hpp"
#include "Lib/Environment.hpp"
#include "Lib/SharedSet.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"

#include "Saturation/SaturationAlgorithm.hpp"

#include "Shell/Statistics.hpp"

#include "BDDMarkingSubsumption.hpp"

namespace Inferences
{

using namespace Kernel;
using namespace Saturation;

void BDDMarkingSubsumption::init(SaturationAlgorithm *sa)
{
  CALL("BDDMarkingSubsumption::init");

  _sa=sa;
}

bool BDDMarkingSubsumption::subsumed(Clause* cl)
{
  CALL("BDDMarkingSubsumption::subsumed");

  bool res=BDD::instance()->isRefuted(cl->prop());
  if(res) {
    env.statistics->subsumedByMarking++;
  }
  return res;
}

void BDDMarkingSubsumption::onNewPropositionalClause(Clause* cl)
{
  CALL("BDDMarkingSubsumption::onNewPropositionalClause");

  BDD* bdd=BDD::instance();

  if(cl->isEmpty() && !bdd->isFalse(cl->prop())) {
    ASS(cl->noSplits());

    bdd->markRefuted(cl->prop());
  }
}

void BDDMarkingSubsumption::onAllProcessed()
{
  CALL("BDDMarkingSubsumption::onAllProcessed");

  static Stack<Clause*> toRemove(64);
  static DHSet<Clause*> checked;
  static BDD* bdd=BDD::instance();

  static int counter1=0;
  counter1++;
  if(counter1>sqrt((float)_sa->activeClauseCount())) {
    counter1=0;

    TimeCounter tc(TC_BDD_MARKING_SUBSUMPTION);

    
    checked.reset();
    ClauseIterator rit=_sa->activeClauses();
    while(rit.hasNext()) {
	Clause* cl=rit.next();
	if(checked.find(cl)) {
	  continue;
	}
	checked.insert(cl);
	if(bdd->isRefuted(cl->prop())) {
	  toRemove.push(cl);
	}
    }
  }

  static int counter2=0;
  counter2++;
  if(counter2>sqrt((float)_sa->passiveClauseCount())) {
    counter2=0;

    TimeCounter tc(TC_BDD_MARKING_SUBSUMPTION);

    
    checked.reset();
    ClauseIterator cit=_sa->passiveClauses();
    while(cit.hasNext()) {
	Clause* cl=cit.next();
	if(cl->store()!=Clause::PASSIVE || checked.find(cl)) {
	  continue;
	}
	checked.insert(cl);
	if(bdd->isRefuted(cl->prop())) {
	  toRemove.push(cl);
	}
    }
  }

  if(toRemove.isNonEmpty()) {
    TimeCounter tc(TC_BDD_MARKING_SUBSUMPTION);

    while(toRemove.isNonEmpty()) {
	Clause* cl=toRemove.pop();

	_sa->removeActiveOrPassiveClause(cl);
	cl->setProp(bdd->getTrue());
	env.statistics->subsumedByMarking++;
    }
  }

}

}



