
/*
 * File DistinctEqualitySimplifier.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Signature.hpp"

#include "DistinctEqualitySimplifier.hpp"

namespace Inferences
{

Clause* DistinctEqualitySimplifier::simplify(Clause* cl)
{
  CALL("DistinctEqualitySimplifier::simplify");

  if(!canSimplify(cl)) {
    return cl;
  }
  LOG("inf_des","can simplify: "<<cl->toString());
  static LiteralStack lits;
  static Stack<Unit*> prems;
  prems.reset();
  lits.reset();
  unsigned clen = cl->length();
  for(unsigned i=0; i<clen; i++) {
    Literal* lit = (*cl)[i];
    unsigned grp;
    if(!lit->isEquality() || !mustBeDistinct(*lit->nthArgument(0),*lit->nthArgument(1), grp)) {
      lits.push(lit);
      continue;
    }
    if(lit->isNegative()) {
      
      LOG_TAUT("inf_des",cl);
      return 0;
    }
    Unit* prem = env.signature->getDistinctGroupPremise(grp);
    if(prem) {
      prems.push(prem);
    }
    
  }

  
  
  ASS_L(lits.size(), clen);

  prems.push(cl);
  UnitList* premLst = 0;
  UnitList::pushFromIterator(Stack<Unit*>::Iterator(prems), premLst);
  Inference* inf;
  if(premLst) {
    inf = new InferenceMany(Inference::DISTINCT_EQUALITY_REMOVAL, premLst);
  }
  else {
    inf = new Inference(Inference::DISTINCT_EQUALITY_REMOVAL);
  }
  Unit::InputType inpType = cl->inputType();
  Clause* res = Clause::fromStack(lits, inpType, inf);
  LOG_SIMPL("inf_des",cl,res);
  return res;
}

bool DistinctEqualitySimplifier::mustBeDistinct(TermList t1, TermList t2)
{
  CALL("DistinctEqualitySimplifier::mustBeDistinct/2");

  unsigned aux;
  return mustBeDistinct(t1, t2, aux);
}

bool DistinctEqualitySimplifier::mustBeDistinct(TermList t1, TermList t2, unsigned& grp)
{
  CALL("DistinctEqualitySimplifier::mustBeDistinct/3");

  if(!t1.isTerm() || t1.term()->arity()!=0 || !t2.isTerm() || t2.term()->arity()!=0) {
    return false;
  }
  unsigned fn1 = t1.term()->functor();
  unsigned fn2 = t2.term()->functor();
  const List<unsigned>* dlst1 = env.signature->getFunction(fn1)->distinctGroups();
  if(!dlst1) {
    return false;
  }
  const List<unsigned>* dlst2 = env.signature->getFunction(fn2)->distinctGroups();
  if(!dlst2) {
    return false;
  }
  List<unsigned>::Iterator dl1it(dlst1);
  while(dl1it.hasNext()) {
    unsigned candGrp = dl1it.next(); 
    if(dlst2->member(candGrp)) {
      grp = candGrp;
      LOG("inf_des","must be distinct: "<<t1<<" and "<<t2<<" due to "<<grp);
      return true;
    }
  }
  return false;
}

bool DistinctEqualitySimplifier::canSimplify(Clause* cl)
{
  CALL("DistinctEqualitySimplifier::canSimplify");

  unsigned clen = cl->length();
  for(unsigned i=0; i<clen; i++) {
    Literal* lit = (*cl)[i];
    if(lit->isEquality() && mustBeDistinct(*lit->nthArgument(0),*lit->nthArgument(1))) {
      return true;
    }
  }
  return false;
}

}
