
/*
 * File ForwardDemodulation.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "Lib/DHSet.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Int.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/TimeCounter.hpp"
#include "Lib/Timer.hpp"
#include "Lib/VirtualIterator.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/EqHelper.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Ordering.hpp"
#include "Kernel/Renaming.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"

#include "Indexing/Index.hpp"
#include "Indexing/IndexManager.hpp"
#include "Indexing/TermIndex.hpp"

#include "Saturation/SaturationAlgorithm.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "ForwardDemodulation.hpp"

namespace Inferences {

using namespace Lib;
using namespace Kernel;
using namespace Indexing;
using namespace Saturation;

void ForwardDemodulation::attach(SaturationAlgorithm* salg)
{
  CALL("ForwardDemodulation::attach");
  ForwardSimplificationEngine::attach(salg);
  _index=static_cast<DemodulationLHSIndex*>(
	  _salg->getIndexManager()->request(DEMODULATION_LHS_SUBST_TREE) );

  _preorderedOnly=getOptions().forwardDemodulation()==Options::DEMODULATION_PREORDERED;
}

void ForwardDemodulation::detach()
{
  CALL("ForwardDemodulation::detach");
  _index=0;
  _salg->getIndexManager()->release(DEMODULATION_LHS_SUBST_TREE);
  ForwardSimplificationEngine::detach();
}


void ForwardDemodulation::perform(Clause* cl, ForwardSimplificationPerformer* simplPerformer)
{
  CALL("ForwardDemodulation::perform");

  TimeCounter tc(TC_FORWARD_DEMODULATION);

  Ordering& ordering = _salg->getOrdering();

  
  
  

  static DHSet<TermList> attempted;
  attempted.reset();

  unsigned cLen=cl->length();
  for(unsigned li=0;li<cLen;li++) {
    Literal* lit=(*cl)[li];
    NonVariableIterator nvi(lit);
    while(nvi.hasNext()) {
      TermList trm=nvi.next();
      if(!attempted.insert(trm)) {
	
	
	
	
	nvi.right();
	continue;
      }

      unsigned querySort = SortHelper::getTermSort(trm, lit);

      bool toplevelCheck=getOptions().demodulationRedundancyCheck() && lit->isEquality() &&
	  (trm==*lit->nthArgument(0) || trm==*lit->nthArgument(1));

      TermQueryResultIterator git=_index->getGeneralizations(trm, true);
      while(git.hasNext()) {
	TermQueryResult qr=git.next();
	ASS_EQ(qr.clause->length(),1);

	if(!simplPerformer->willPerform(qr.clause)) {
	  continue;
	}

	unsigned eqSort = SortHelper::getEqualityArgumentSort(qr.literal);

	if(querySort!=eqSort) {
	  continue;
	}

	TermList rhs=EqHelper::getOtherEqualitySide(qr.literal,qr.term);
	TermList rhsS;
	if(!qr.substitution->isIdentityOnQueryWhenResultBound()) {
	  
	  
	  
	  TermList lhsSBadVars=qr.substitution->applyToResult(qr.term);
	  TermList rhsSBadVars=qr.substitution->applyToResult(rhs);
	  Renaming rNorm, qNorm, qDenorm;
	  rNorm.normalizeVariables(lhsSBadVars);
	  qNorm.normalizeVariables(trm);
	  qDenorm.makeInverse(qNorm);
	  ASS_EQ(trm,qDenorm.apply(rNorm.apply(lhsSBadVars)));
	  rhsS=qDenorm.apply(rNorm.apply(rhsSBadVars));
	} else {
	  rhsS=qr.substitution->applyToBoundResult(rhs);
	}

	Ordering::Result argOrder = ordering.getEqualityArgumentOrder(qr.literal);
	bool preordered = argOrder==Ordering::LESS || argOrder==Ordering::GREATER;
#if VDEBUG
	if(preordered) {
	  if(argOrder==Ordering::LESS) {
	    ASS_EQ(rhs, *qr.literal->nthArgument(0));
	  }
	  else {
	    ASS_EQ(rhs, *qr.literal->nthArgument(1));
	  }
	}
#endif
	if(!preordered && (_preorderedOnly || ordering.compare(trm,rhsS)!=Ordering::GREATER) ) {
	  continue;
	}

	if(toplevelCheck) {
	  TermList other=EqHelper::getOtherEqualitySide(lit, trm);
	  Ordering::Result tord=ordering.compare(rhsS, other);
	  if(tord!=Ordering::LESS && tord!=Ordering::LESS_EQ) {
	    Literal* eqLitS=qr.substitution->applyToBoundResult(qr.literal);
	    bool isMax=true;
	    for(unsigned li2=0;li2<cLen;li2++) {
	      if(li==li2) {
		continue;
	      }
	      if(ordering.compare(eqLitS, (*cl)[li2])==Ordering::LESS) {
		isMax=false;
		break;
	      }
	    }
	    if(isMax) {
	      
	      
	      
	      
	      
	      
	      LOG("inf_fwd_tlcheck","prevented fw dem: "<<(*eqLitS)<<" in "<<(*lit)<<" of "<<(*cl));
	      continue;
	    }
	  }
	}

	Literal* resLit = EqHelper::replace(lit,trm,rhsS);
	if(EqHelper::isEqTautology(resLit)) {
	  env.statistics->forwardDemodulationsToEqTaut++;
	  simplPerformer->perform(qr.clause, 0);
	  if(!simplPerformer->clauseKept()) {
	    return;
	  }
	}

	Inference* inf = new Inference2(Inference::FORWARD_DEMODULATION, cl, qr.clause);
	Unit::InputType inpType = (Unit::InputType)
		Int::max(cl->inputType(), qr.clause->inputType());

	Clause* res = new(cLen) Clause(cLen, inpType, inf);

	(*res)[0]=resLit;

	unsigned next=1;
	for(unsigned i=0;i<cLen;i++) {
	  Literal* curr=(*cl)[i];
	  if(curr!=lit) {
	    (*res)[next++] = curr;
	  }
	}
	ASS_EQ(next,cLen);

	res->setAge(cl->age());
	env.statistics->forwardDemodulations++;

	simplPerformer->perform(qr.clause, res);
	if(!simplPerformer->clauseKept()) {
	  return;
	}

      }
    }
  }
}

}
