
/*
 * File ForwardLiteralRewriting.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Int.hpp"

#include "Kernel/Inference.hpp"
#include "Kernel/Ordering.hpp"

#include "Indexing/IndexManager.hpp"

#include "Saturation/SaturationAlgorithm.hpp"

#include "ForwardLiteralRewriting.hpp"

namespace Inferences
{

void ForwardLiteralRewriting::attach(SaturationAlgorithm* salg)
{
  CALL("ForwardLiteralRewriting::attach");
  ForwardSimplificationEngine::attach(salg);
  _index=static_cast<RewriteRuleIndex*>(
	  _salg->getIndexManager()->request(REWRITE_RULE_SUBST_TREE) );
}

void ForwardLiteralRewriting::detach()
{
  CALL("ForwardLiteralRewriting::detach");
  _index=0;
  _salg->getIndexManager()->release(REWRITE_RULE_SUBST_TREE);
  ForwardSimplificationEngine::detach();
}


void ForwardLiteralRewriting::perform(Clause* cl, ForwardSimplificationPerformer* simplPerformer)
{
  CALL("ForwardLiteralRewriting::perform");

  Ordering& ordering = _salg->getOrdering();

  TimeCounter tc(TC_FORWARD_LITERAL_REWRITING);

  unsigned clen=cl->length();

  for(unsigned i=0;i<clen;i++) {
    Literal* lit=(*cl)[i];
    SLQueryResultIterator git=_index->getGeneralizations(lit, lit->isNegative(), true);
    while(git.hasNext()) {
      SLQueryResult qr=git.next();
      Clause* counterpart=_index->getCounterpart(qr.clause);

      if(!simplPerformer->willPerform(qr.clause) || !simplPerformer->willPerform(counterpart)) {
	continue;
      }

      if(cl==qr.clause || cl==counterpart) {
	continue;
      }

      Literal* rhs0 = (qr.literal==(*qr.clause)[0]) ? (*qr.clause)[1] : (*qr.clause)[0];
      Literal* rhs = lit->isNegative() ? rhs0 : Literal::complementaryLiteral(rhs0);

      ASS(qr.substitution->isIdentityOnQueryWhenResultBound());

      
      
      ASS(qr.literal->containsAllVariablesOf(rhs));
      Literal* rhsS=qr.substitution->applyToBoundResult(rhs);

      if(ordering.compare(lit, rhsS)!=Ordering::GREATER) {
	continue;
      }

      Clause* premise=lit->isNegative() ? qr.clause : counterpart;
      Clause* reductionPremise=lit->isNegative() ? counterpart : qr.clause;
      if(reductionPremise==premise) {
	reductionPremise=0;
      }

      Inference* inf = new Inference2(Inference::FORWARD_LITERAL_REWRITING, cl, premise);
      Unit::InputType inpType = (Unit::InputType)
	Int::max(cl->inputType(), premise->inputType());

      Clause* res = new(clen) Clause(clen, inpType, inf);

      (*res)[0]=rhsS;

      unsigned next=1;
      for(unsigned i=0;i<clen;i++) {
	Literal* curr=(*cl)[i];
	if(curr!=lit) {
	  (*res)[next++] = curr;
	}
      }
      ASS_EQ(next,clen);

      res->setAge(cl->age());
      env.statistics->forwardLiteralRewrites++;

      simplPerformer->perform(premise, res, reductionPremise);
      if(!simplPerformer->clauseKept()) {
	return;
      }

    }
  }


}

};
