
/*
 * File GlobalSubsumption.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "Lib/Metaiterators.hpp"
#include "Lib/SharedSet.hpp"
#include "Lib/TimeCounter.hpp"
#include "Lib/VirtualIterator.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Grounder.hpp"
#include "Kernel/Inference.hpp"

#include "Indexing/Index.hpp"
#include "Indexing/IndexManager.hpp"
#include "Indexing/GroundingIndex.hpp"

#include "SAT/SATClause.hpp"
#include "SAT/SATSolver.hpp"

#include "Saturation/SaturationAlgorithm.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "GlobalSubsumption.hpp"

#undef LOGGING
#define LOGGING 0

namespace Inferences {

using namespace Lib;
using namespace Kernel;
using namespace Indexing;
using namespace Saturation;

void GlobalSubsumption::attach(SaturationAlgorithm* salg)
{
  CALL("GlobalSubsumption::attach");

  if(_allowExtraAttachment) {
    return;
  }
  ASS(!_index);

  ForwardSimplificationEngine::attach(salg);
  _index=static_cast<GroundingIndex*>(
	  _salg->getIndexManager()->request(GLOBAL_SUBSUMPTION_INDEX) );
}

void GlobalSubsumption::detach()
{
  CALL("GlobalSubsumption::detach");

  if(_allowExtraAttachment) {
    return;
  }

  _index=0;
  _salg->getIndexManager()->release(GLOBAL_SUBSUMPTION_INDEX);
  ForwardSimplificationEngine::detach();
}

void GlobalSubsumption::addClauseToIndex(Clause* cl)
{
  CALL("GlobalSubsumption::addClauseToIndex");

  SATSolver& solver = _index->getSolver();
  Grounder& grounder = _index->getGrounder();

  ASS_NEQ(solver.getStatus(),SATSolver::UNSATISFIABLE);

  SATClauseIterator sclIt = grounder.ground(cl);
  solver.ensureVarCnt(grounder.satVarCnt());
  solver.addClauses(sclIt, true);

  if(solver.getStatus()==SATSolver::UNSATISFIABLE) {
    
    Inference* inf = new Inference(Inference::TAUTOLOGY_INTRODUCTION);
    Clause* refutation = Clause::fromIterator(LiteralIterator::getEmpty(), Unit::CONJECTURE, inf);
    refutation->setAge(cl->age());

    Grounder::recordInference(cl, solver.getRefutation(), refutation);

    LOG("inf_gs","gs found refutation "<<(*refutation)<<" based on "<<(*cl));
    env.statistics->globalSubsumption++;
    throw MainLoop::RefutationFoundException(refutation);
  }
}

Clause* GlobalSubsumption::perform(Clause* cl)
{
  CALL("GlobalSubsumption::perform/1");

  TimeCounter tc(TC_GLOBAL_SUBSUMPTION);

  if(cl->splits() && cl->splits()->size()!=0) {
    
    
    return cl;
  }

  if(cl->prop() && !BDD::instance()->isFalse(cl->prop())) {
    
    return cl;
  }

  if(cl->color()==COLOR_LEFT) {
    return cl;
  }

  if(cl->length()==1) {
    addClauseToIndex(cl);
    return cl;
  }

  Grounder& grounder = _index->getGrounder();

  static SATLiteralStack slits;
  slits.reset();

  grounder.groundNonProp(cl, slits);

  addClauseToIndex(cl);

  unsigned clen = slits.size();


  for(unsigned resolvedIdx = 0; resolvedIdx<clen; resolvedIdx++) {
    Clause* replacement = tryResolvingAway(cl, resolvedIdx, slits);
    if(replacement!=0) {
      return replacement;
    }
  }
  return cl;
}

Clause* GlobalSubsumption::tryResolvingAway(Clause* cl, unsigned litIdx, SATLiteralStack& slits)
{
  CALL("GlobalSubsumption::tryResolvingAway");

  bool uprOnly = true;

  unsigned clen = cl->length();
  SATSolver& solver = _index->getSolver();

  for(unsigned i = 0; i<clen; i++) {
    if(i==litIdx) {
	continue;
    }
    solver.addAssumption(slits[i].opposite(), uprOnly);

    if(solver.getStatus()==SATSolver::UNSATISFIABLE) {
	static LiteralStack survivors;
	survivors.reset();

	for(unsigned j=0; j<=i; j++) {
	  if(j==litIdx) {
	    continue;
	  }
	  survivors.push((*cl)[j]);
	}


	
	Inference* inf = new Inference(Inference::TAUTOLOGY_INTRODUCTION);
	Clause* replacement = Clause::fromIterator(LiteralStack::BottomFirstIterator(survivors),
	    cl->inputType(), inf);
	replacement->setAge(cl->age());

	Grounder::recordInference(cl, solver.getRefutation(), replacement);
	LOG_SIMPL("inf_gs",cl,replacement);
	env.statistics->globalSubsumption++;
        ASS_L(replacement->length(), clen);

	solver.retractAllAssumptions();
	return replacement;
    }
  }
  solver.retractAllAssumptions();
  return 0;
}

void GlobalSubsumption::perform(Clause* cl, ForwardSimplificationPerformer* simplPerformer)
{
  CALL("GlobalSubsumption::perform/2");

  Clause* newCl = perform(cl);
  if(newCl==cl) {
    return;
  }

  ALWAYS(simplPerformer->willPerform(0));
  simplPerformer->perform(0, newCl);
  ALWAYS(!simplPerformer->clauseKept());
}

}
