
/*
 * File Clause.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __Clause__
#define __Clause__

#include <iosfwd>

#include "Forwards.hpp"

#include "Lib/Allocator.hpp"
#include "Lib/Event.hpp"
#include "Lib/InverseLookup.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/Reflection.hpp"
#include "Lib/Stack.hpp"

#include "Unit.hpp"


namespace Kernel {

using namespace Lib;

class Clause
  : public Unit
{
private:
 
  ~Clause() { ASSERTION_VIOLATION; }
 
  void operator delete(void* ptr) { ASSERTION_VIOLATION; }
public:
  typedef ArrayishObjectIterator<Clause> Iterator;

  DECL_ELEMENT_TYPE(Literal*);
  DECL_ITERATOR_TYPE(Iterator);

 
  enum Store {
   
    PASSIVE = 0u,
   
    ACTIVE = 1u,
   
    UNPROCESSED = 2u,
   
    NONE = 3u,
   
    BACKTRACKED = 4u,
   
    REACTIVATED = 5u,
   
    SELECTED = 6u,
   
    SELECTED_REACTIVATED = 7u
  };

  Clause(unsigned length,InputType it,Inference* inf);


  void* operator new(size_t,unsigned length);
  void operator delete(void* ptr,unsigned length);

  static Clause* fromStack(Stack<Literal*>& lits, InputType it, Inference* inf);

  template<class Iter>
  static Clause* fromIterator(Iter litit, InputType it, Inference* inf)
  {
    CALL("Clause::fromIterator");

    static Stack<Literal*> st;
    st.reset();
    st.loadFromIterator(litit);
    return fromStack(st, it, inf);
  }

  static Clause* fromClause(Clause* c);

 
  Literal*& operator[] (int n)
  { return _literals[n]; }
 
  Literal* operator[] (int n) const
  { return _literals[n]; }

 
  unsigned length() const { return _length; }
 
  unsigned size() const { return _length; }

 
  Literal** literals() { return _literals; }

 
  bool isEmpty() const { return _length == 0; }

  void destroy();
  void destroyExceptInferenceObject();
  string nonPropToString() const;
  string toString() const;
  string toTPTPString() const;
  string toNiceString() const;
  string toString(BDDNode* propPart) const;

 
  Store store() const { return _store; }

  void setStore(Store s);

 
  int age() const { return _age; }
 
  void setAge(int a) { _age = a; }

 
  unsigned selected() const { return _selected; }
 
  void setSelected(unsigned s)
  {
    ASS(s >= 0);
    ASS(s <= _length);
    _selected = s;
    notifyLiteralReorder();
  }

 
  int weight() const
  {
    if(!_weight) {
      computeWeight();
    }
    return _weight;
  }
  void computeWeight() const;

 
  Color color() const
  {
    if(static_cast<Color>(_color)==COLOR_INVALID) {
      computeColor();
    }
    return static_cast<Color>(_color);
  }
  void computeColor() const;

  bool skip() const;

  unsigned getLiteralPosition(Literal* lit);
  void notifyLiteralReorder();

  bool shouldBeDestroyed();
  void destroyIfUnnecessary();

  void incRefCnt() { _inferenceRefCnt++; }
  void decRefCnt()
  {
    CALL("Clause::decRefCnt");

    ASS_G(_inferenceRefCnt,0);
    _inferenceRefCnt--;
    destroyIfUnnecessary();
  }

  unsigned getReductionTimestamp() { return _reductionTimestamp; }
  void incReductionTimestamp()
  {
    _reductionTimestamp++;
    if(_reductionTimestamp==0) {
      INVALID_OPERATION("Clause reduction timestamp overflow!");
    }
  }

  ArrayishObjectIterator<Clause> getSelectedLiteralIterator()
  { return ArrayishObjectIterator<Clause>(*this,selected()); }

  bool isGround();
  bool isPropositional();
  bool isHorn();

  VirtualIterator<unsigned> getVariableIterator();

#if VDEBUG
  bool contains(Literal* lit);
  void assertValid();
#endif

 
  void markInput() { _input=1; }
 
  bool isInput() { return _input; }

 
  BDDNode* prop() const { return _prop; }
  void initProp(BDDNode* prop);
  void setProp(BDDNode* prop);
 
  bool noProp() const;

  SplitSet* splits() const { return _splits; }
  bool noSplits() const;
  void setSplits(SplitSet* splits) {
    CALL("Clause::setSplits");
    ASS(!_splits);
    _splits=splits;
  }

  VirtualIterator<string> toSimpleClauseStrings();


 
  void setAux(void* ptr)
  {
    ASS(_auxInUse);
    _auxTimestamp=_auxCurrTimestamp;
    _auxData=ptr;
  }
 
  template<typename T>
  bool tryGetAux(T*& ptr)
  {
    ASS(_auxInUse);
    if(_auxTimestamp==_auxCurrTimestamp) {
      ptr=static_cast<T*>(_auxData);
      return true;
    }
    return false;
  }
 
  template<typename T>
  T* getAux()
  {
    ASS(_auxInUse);
    ASS(_auxTimestamp==_auxCurrTimestamp);
    return static_cast<T*>(_auxData);
  }
  bool hasAux()
  {
    return _auxTimestamp==_auxCurrTimestamp;
  }

 
  static void requestAux()
  {
#if VDEBUG
    ASS(!_auxInUse);
    _auxInUse=true;
#endif
    _auxCurrTimestamp++;
    if(_auxCurrTimestamp==0) {
      INVALID_OPERATION("Auxiliary clause value timestamp overflow!");
    }
  }
 
  static void releaseAux()
  {
#if VDEBUG
    ASS(_auxInUse);
    _auxInUse=false;
#endif
  }

  unsigned propWeight() const;
  unsigned splitWeight() const;
  unsigned getNumeralWeight();
  float getEffectiveWeight(const Shell::Options& opt);

  void collectVars(DHSet<unsigned>& acc);
  unsigned varCnt();

  static ClauseEvent beforePropChange;
  static ClauseEvent afterPropChange;

protected:
 
  unsigned _length : 29;
 
  mutable unsigned _color : 2;
 
  unsigned _input : 1;
 
  unsigned _selected;
 
  unsigned _age;
 
  mutable unsigned _weight;
 
  Store _store;
 
  unsigned _inferenceRefCnt;
 
  unsigned _reductionTimestamp;
 
  InverseLookup<Literal>* _literalPositions;

 
  BDDNode* _prop;

  SplitSet* _splits;

  size_t _auxTimestamp;
  void* _auxData;

  static size_t _auxCurrTimestamp;
#if VDEBUG
  static bool _auxInUse;
#endif

 
  Literal* _literals[1];
}; 

}

#endif
