
/*
 * File ClauseQueue.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "Lib/Allocator.hpp"
#include "Lib/Random.hpp"
#include "Lib/Environment.hpp"

#if VDEBUG
#include "Clause.hpp"
#endif

#include "ClauseQueue.hpp"

using namespace Lib;
using namespace Kernel;

ClauseQueue::~ClauseQueue ()
{
  CALL("ClauseQueue::~ClauseQueue");

  removeAll();
} 

void ClauseQueue::insert(Clause* c)
{
  CALL("ClauseQueue::insert");

  
  unsigned h = 0;
  while (Random::getBit()) {
    h++;
  }
  if (h > _height) {
    if (_height < MAX_HEIGHT) {
      _height++;
    }
    h = _height;
    _left.nodes[h] = 0;
  }
  void* mem = ALLOC_KNOWN(sizeof(Node)+h*sizeof(Node*),
			  "ClauseQueue::Node");
  Node* newNode = reinterpret_cast<Node*>(mem);
  newNode->clause = c;

  
  
  
  Node* left = &_left;
  
  unsigned lh = _height;
  for (;;) {
    Node* next = left->nodes[lh];
    if (next == 0 || lessThan(c,next->clause)) {
      if (lh <= h) {
	left->nodes[lh] = newNode;
	newNode->nodes[lh] = next;
      }
      if (lh == 0) {
	return;
      }
      lh--;
      continue;
    }
    left = next;
  }
} 

bool ClauseQueue::remove(Clause* c)
{
  CALL("ClauseQueue::remove");

  unsigned h = _height;
  Node* left = &_left;

  for (;;) {
    Node* next = left->nodes[h];
    if (next && c == next->clause) {
      unsigned height = h;
      
      for (;;) {
	left->nodes[h] = next->nodes[h];
	if (h == 0) {
	  break;
	}
	h--;
	while (left->nodes[h] != next) {
	  left = left->nodes[h];
	}
      }
      
      DEALLOC_KNOWN(next,
		    sizeof(Node)+height*sizeof(Node*),
		    "ClauseQueue::Node");
      while (_height > 0 && ! _left.nodes[_height]) {
	_height--;
      }
      return true;
    }

    if (next == 0 || lessThan(c,next->clause)) {
      if(h==0) {
	return false;
      }
      h--;
    }
    else {
      left = next;
    }
  }
} 


Clause* ClauseQueue::pop()
{
  CALL("ClauseQueue::pop");
  ASS(_height >= 0);
  ASS(_left.nodes[0] != 0);

  Node* node = _left.nodes[0];
  unsigned h = 0;
  _left.nodes[0] = node->nodes[0];
  while (h < _height && _left.nodes[h+1] == node) {
    h++;
    _left.nodes[h] = node->nodes[h];
  }
  
  Clause* c = node->clause;

  
  DEALLOC_KNOWN(node,
		sizeof(Node)+h*sizeof(Node*),
		"ClauseQueue::Node");
  while (_height > 0 && ! _left.nodes[_height]) {
    _height--;
  }
  return c;
} 

void ClauseQueue::removeAll()
{
  CALL("ClauseQueue::removeAll");

  while (_left.nodes[0]) {
    pop();
  }
} 

#if VDEBUG
void ClauseQueue::output(ostream& str) const
{
  if (_height < 0) {
    return;
  }
  for (const Node* node = _left.nodes[0]; node; node=node->nodes[0]) {
    str << node->clause->toString() << '\n';
  }
} 
#endif
