
/*
 * File EqHelper.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Shell/Options.hpp"

#include "Ordering.hpp"
#include "SortHelper.hpp"
#include "TermIterators.hpp"

#include "EqHelper.hpp"

namespace Kernel {

using namespace Shell;

TermList EqHelper::getOtherEqualitySide(Literal* eq, TermList lhs)
{
  CALL("EqHelper::getOtherEqualitySide");
  ASS(eq->isEquality());

  if(*eq->nthArgument(0)==lhs) {
    return *eq->nthArgument(1);
  } else {
    ASS(*eq->nthArgument(1)==lhs);
    return *eq->nthArgument(0);
  }
}

Literal* EqHelper::replace(Literal* lit, TermList what, TermList by)
{
  CALL("EqHelper::replace(Literal*,...)");

  return static_cast<Literal*>(replace(static_cast<Term*>(lit), what, by));
}

Term* EqHelper::replace(Term* trm0, TermList tSrc, TermList tDest)
{
  CALL("EqHelper::replace(Term*,...)");
  ASS(trm0->shared());

  static Stack<TermList*> toDo(8);
  static Stack<Term*> terms(8);
  static Stack<bool> modified(8);
  static Stack<TermList> args(8);
  ASS(toDo.isEmpty());
  ASS(terms.isEmpty());
  modified.reset();
  args.reset();

  modified.push(false);
  toDo.push(trm0->args());

  for(;;) {
    TermList* tt=toDo.pop();
    if(tt->isEmpty()) {
      if(terms.isEmpty()) {
	
	
	ASS(toDo.isEmpty());
	break;
      }
      Term* orig=terms.pop();
      if(!modified.pop()) {
	args.truncate(args.length() - orig->arity());
	args.push(TermList(orig));
	continue;
      }
      
      
      
      TermList* argLst=&args.top() - (orig->arity()-1);
      args.truncate(args.length() - orig->arity());

      args.push(TermList(Term::create(orig,argLst)));
      modified.setTop(true);
      continue;
    } else {
      toDo.push(tt->next());
    }

    TermList tl=*tt;
    if(tl==tSrc) {
      args.push(tDest);
      modified.setTop(true);
      continue;
    }
    if(tl.isVar()) {
      args.push(tl);
      continue;
    }
    ASS(tl.isTerm());
    Term* t=tl.term();
    terms.push(t);
    modified.push(false);
    toDo.push(t->args());
  }
  ASS(toDo.isEmpty());
  ASS(terms.isEmpty());
  ASS_EQ(modified.length(),1);
  ASS_EQ(args.length(),trm0->arity());

  if(!modified.pop()) {
    
    
    ASSERTION_VIOLATION;
    return trm0;
  }

  
  
  
  TermList* argLst=&args.top() - (trm0->arity()-1);

  if(trm0->isLiteral()) {
    Literal* lit = static_cast<Literal*>(trm0);
    ASS_EQ(args.size(), lit->arity());
    return Literal::create(lit,argLst);
  }
  else {
    return Term::create(trm0,argLst);
  }
}

TermIterator EqHelper::getRewritableSubtermIterator(Literal* lit, const Ordering& ord)
{
  CALL("EqHelper::getRewritableSubtermIterator");





  if(lit->isEquality() && lit->isPositive()) {
    TermList sel;
    switch(ord.getEqualityArgumentOrder(lit))
    {
    case Ordering::INCOMPARABLE:
    {
      NonVariableIterator nvi(lit);
      return getUniquePersistentIteratorFromPtr(&nvi);
    }
    case Ordering::EQUAL:
    case Ordering::GREATER:
    case Ordering::GREATER_EQ:
      sel=*lit->nthArgument(0);
      break;
    case Ordering::LESS:
    case Ordering::LESS_EQ:
      sel=*lit->nthArgument(1);
#if VDEBUG
      break;
    default:
      ASSERTION_VIOLATION;
#endif
    }
    if(!sel.isTerm()) {
      return TermIterator::getEmpty();
    }
    return getUniquePersistentIterator(
	    getConcatenatedIterator(getSingletonIterator(sel),
	    vi( new NonVariableIterator(sel.term()) )) );
  } else {
    NonVariableIterator nvi(lit);
    return getUniquePersistentIteratorFromPtr(&nvi);
  }
}

TermIterator EqHelper::getLHSIterator(Literal* lit, const Ordering& ord)
{
  CALL("EqHelper::getLHSIterator");

  if(lit->isEquality()) {
    if(lit->isNegative()) {
      return TermIterator::getEmpty();
    }
    TermList t0=*lit->nthArgument(0);
    TermList t1=*lit->nthArgument(1);
    switch(ord.getEqualityArgumentOrder(lit))
    {
    case Ordering::INCOMPARABLE:
      return pvi( getConcatenatedIterator(getSingletonIterator(t0),
	      getSingletonIterator(t1)) );
    case Ordering::GREATER:
    case Ordering::GREATER_EQ:
      return pvi( getSingletonIterator(t0) );
    case Ordering::LESS:
    case Ordering::LESS_EQ:
      return pvi( getSingletonIterator(t1) );
#if VDEBUG
    case Ordering::EQUAL:
      
    default:
      ASSERTION_VIOLATION;
#endif
    }
    return TermIterator::getEmpty();
  } else {
    return TermIterator::getEmpty();
  }
}

struct EqHelper::IsNonVariable
{
  DECL_RETURN_TYPE(bool);
  bool operator()(TermList t)
  { return t.isTerm(); }
};

TermIterator EqHelper::getSuperpositionLHSIterator(Literal* lit, const Ordering& ord, const Options& opt)
{
  CALL("EqHelper::getSuperpositionLHSIterator");

  if(opt.superpositionFromVariables()) {
    return getLHSIterator(lit, ord);
  }
  else {
    return pvi( getFilteredIterator(getLHSIterator(lit, ord), IsNonVariable()) );
  }
}

TermIterator EqHelper::getDemodulationLHSIterator(Literal* lit, bool forward, const Ordering& ord, const Options& opt)
{
  CALL("EqHelper::getDemodulationLHSIterator");

  if(lit->isEquality()) {
    if(lit->isNegative()) {
      return TermIterator::getEmpty();
    }
    TermList t0=*lit->nthArgument(0);
    TermList t1=*lit->nthArgument(1);
    switch(ord.getEqualityArgumentOrder(lit))
    {
    case Ordering::INCOMPARABLE:
      if( forward ? (opt.forwardDemodulation()==Options::DEMODULATION_PREORDERED)
		  : (opt.backwardDemodulation()==Options::DEMODULATION_PREORDERED) ) {
	return TermIterator::getEmpty();
      }
      if(t0.containsAllVariablesOf(t1)) {
	if(t1.containsAllVariablesOf(t0)) {
	  return pvi( getConcatenatedIterator(getSingletonIterator(t0),
	      getSingletonIterator(t1)) );
	}
	return pvi( getSingletonIterator(t0) );
      }
      if(t1.containsAllVariablesOf(t0)) {
	return pvi( getSingletonIterator(t1) );
      }
      break;
    case Ordering::GREATER:
    case Ordering::GREATER_EQ:
      ASS(t0.containsAllVariablesOf(t1));
      return pvi( getSingletonIterator(t0) );
    case Ordering::LESS:
    case Ordering::LESS_EQ:
      ASS(t1.containsAllVariablesOf(t0));
      return pvi( getSingletonIterator(t1) );
#if VDEBUG
    case Ordering::EQUAL:
      
    default:
      ASSERTION_VIOLATION;
#endif
    }
    return TermIterator::getEmpty();
  } else {
    return TermIterator::getEmpty();
  }
}

TermIterator EqHelper::getEqualityArgumentIterator(Literal* lit)
{
  CALL("EqHelper::getEqualityArgumentIterator");
  ASS(lit->isEquality());

  return pvi( getConcatenatedIterator(
	  getSingletonIterator(*lit->nthArgument(0)),
	  getSingletonIterator(*lit->nthArgument(1))) );
}


}
