
/*
 * File Formula.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <string>

#include "Debug/Tracer.hpp"

#include "Lib/Exception.hpp"
#include "Lib/MultiCounter.hpp"

#include "BDD.hpp"
#include "Clause.hpp"
#include "Term.hpp"
#include "Formula.hpp"
#include "SubformulaIterator.hpp"
#include "FormulaVarIterator.hpp"

using namespace Lib;

namespace Kernel {


































void Formula::destroy ()
{
  CALL ("Formula::Data::destroy");
  ASS (this);

  switch ( connective() ) {
  case LITERAL:
    delete static_cast<AtomicFormula*>(this);
    return;

  case AND:
  case OR:
    delete static_cast<JunctionFormula*>(this);
    return;

  case IMP:
  case IFF:
  case XOR:
    delete static_cast<BinaryFormula*>(this);
    return;

  case NOT:
    delete static_cast<NegatedFormula*>(this);
    return;

  case FORALL:
  case EXISTS:
    delete static_cast<QuantifiedFormula*>(this);
    return;

  case ITE:
    delete static_cast<IteFormula*>(this);
    return;

  case TRUE:
  case FALSE:
    delete this;
    return;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
    return;
#endif
  }
} 































































































string Formula::toString (Connective c)
{
  static string names [] =

    { "", "&", "|", "=>", "<=>", "<~>", "~", "!", "?", "<ite>", "<formula_let>", "<term_let>", "$false", "$true"};
  ASS_EQ(sizeof(names)/sizeof(string), TRUE+1);

  return names[(int)c];
} 



string Formula::toString () const
{
  CALL("Formula::toString");

  Connective c = connective();
  string con = toString(c);

  switch (c) {
  case LITERAL:
    return literal()->toString();

  case AND:
  case OR:
    {
      ASS (args()->length() >= 2);
      string result = args()->head()->toStringInScopeOf(c);
      FormulaList::Iterator arg(args()->tail());
      while (arg.hasNext()) {
	result += ' ' + con + ' ' + arg.next()->toStringInScopeOf(c);
      }
      return result;
    }

  case IMP:
  case IFF:
  case XOR:
    return left()->toStringInScopeOf(c) + ' ' + con + ' ' +
           right()->toStringInScopeOf(c);

  case NOT:
    {
      const Formula* arg = uarg();
      if (arg->connective() == LITERAL) {
	return con + arg->literal()->toString();
      }
      return con + arg->toStringInScopeOf(c);
    }

  case FORALL:
  case EXISTS:
    {






      string result = con + " [";
      VarList::Iterator vs(vars());
      bool first=true;
      while (vs.hasNext()) {
	if(!first) {
	  result+= ",";
	}
	result += Term::variableToString(vs.next());
	first=false;
      }
      result += "] : ";
      return result + qarg()->toStringInScopeOf(c);
    }

  case ITE:
    return "$itef(" + condArg()->toStringInScopeOf(c) + "," + thenArg()->toStringInScopeOf(c) + "," +
           elseArg()->toStringInScopeOf(c) + ")";

  case FORMULA_LET:
    return "let "+formulaLetLhs()->toString() + " := " + formulaLetRhs()->toStringInScopeOf(c) +
	" in " + letBody()->toStringInScopeOf(c);

  case TERM_LET:
    return "let "+termLetLhs().toString() + " := " + termLetRhs().toString() +
	" in " + letBody()->toStringInScopeOf(c);

  case TRUE:
  case FALSE:
    return con;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 


bool Formula::parenthesesRequired (Connective outer) const
{
  CALL("Formula::parenthesesRequired");

  switch (connective())
    {
    case LITERAL:
    case NOT:
    case FORALL:
    case EXISTS:
    case TRUE:
    case FALSE:
      return false;

    case OR:
    case AND:
    case IMP:
    case IFF:
    case XOR:
    case ITE:
    case FORMULA_LET:
    case TERM_LET:
      return true;

#if VDEBUG
    default:
      ASSERTION_VIOLATION;
      return false;
#endif
    }
} 


string Formula::toStringInScopeOf (Connective outer) const
{
  return parenthesesRequired(outer) ?
         string("(") + toString() + ")" :
         toString();
} 









































































Formula::VarList* Formula::freeVariables () const
{
  CALL("Formula::freeVariables");

  FormulaVarIterator fvi(this);
  VarList* result = VarList::empty();
  VarList::FIFO stack(result);
  while (fvi.hasNext()) {
    stack.push(fvi.next());
  }
  return result;
} 

Formula::VarList* Formula::boundVariables () const
{
  CALL("Formula::boundVariables");

  VarList* res=0;
  SubformulaIterator sfit(const_cast<Formula*>(this));
  while(sfit.hasNext()) {
    Formula* sf=sfit.next();
    if(sf->connective()==FORALL || sf->connective()==EXISTS) {
      VarList* qvars=sf->vars();
      VarList* qvCopy=qvars->copy();
      res=VarList::concat(qvCopy, res);
    }
  }
  return res;
}

void Formula::collectAtoms(Stack<Literal*>& acc)
{
  CALL("Formula::collectPredicates");

  SubformulaIterator sfit(this);
  while(sfit.hasNext()) {
    Formula* sf = sfit.next();
    if(sf->connective()==LITERAL) {
      Literal* l = sf->literal();
      acc.push(Literal::positiveLiteral(l));
    }
  }
}

void Formula::collectPredicates(Stack<unsigned>& acc)
{
  CALL("Formula::collectPredicates");

  SubformulaIterator sfit(this);
  while(sfit.hasNext()) {
    Formula* sf = sfit.next();
    if(sf->connective()==LITERAL) {
      acc.push(sf->literal()->functor());
    }
  }
}

void Formula::collectPredicatesWithPolarity(Stack<pair<unsigned,int> >& acc, int polarity)
{
  CALL("Formula::collectPredicatesWithPolarity");

  switch (connective()) {
    case LITERAL:
    {
      Literal* l=literal();
      int pred = l->functor();
      acc.push(make_pair(pred, l->isPositive() ? polarity : -polarity));
      return;
    }

    case AND:
    case OR: {
      FormulaList::Iterator fs(args());
      while (fs.hasNext()) {
	fs.next()->collectPredicatesWithPolarity(acc,polarity);
      }
      return;
    }

    case IMP:
      left()->collectPredicatesWithPolarity(acc,-polarity);
      right()->collectPredicatesWithPolarity(acc,polarity);
      return;

    case NOT:
      uarg()->collectPredicatesWithPolarity(acc,-polarity);
      return;

    case IFF:
    case XOR:
      left()->collectPredicatesWithPolarity(acc,0);
      right()->collectPredicatesWithPolarity(acc,0);
      return;

    case FORALL:
    case EXISTS:
      qarg()->collectPredicatesWithPolarity(acc,polarity);
      return;

    case TRUE:
    case FALSE:
      return;

#if VDEBUG
    default:
      ASSERTION_VIOLATION;
      return;
#endif
  }
}


unsigned Formula::weight() const
{
  CALL("Formula::weight");

  unsigned result=0;

  SubformulaIterator fs(const_cast<Formula*>(this));
  while (fs.hasNext()) {
    const Formula* f = fs.next();
    switch (f->connective()) {
    case LITERAL:
      result += f->literal()->weight();
      break;
    default:
      result++;
      break;
    }
  }
  return result;
} 

Formula* JunctionFormula::generalJunction(Connective c, FormulaList* args)
{
  if(!args) {
    if(c==AND) {
      return new Formula(true);
    }
    else {
      ASS_EQ(c,OR);
      return new Formula(false);
    }
  }
  if(!args->tail()) {
    return FormulaList::pop(args);
  }
  return new JunctionFormula(c, args);
}

Color Formula::getColor()
{
  CALL("Formula::getColor");

  SubformulaIterator si(this);
  while(si.hasNext()) {
    Formula* f=si.next();
    if(f->connective()!=LITERAL) {
      continue;
    }

    if(f->literal()->color()!=COLOR_TRANSPARENT) {
      return f->literal()->color();
    }
  }
  return COLOR_TRANSPARENT;
}

bool Formula::getSkip()
{
  CALL("Formula::getColor");

  SubformulaIterator si(this);
  while(si.hasNext()) {
    Formula* f=si.next();
    if(f->connective()!=LITERAL) {
      continue;
    }

    if(!f->literal()->skip()) {
      return false;
    }
  }
  return true;
}

Formula* Formula::trueFormula()
{
  CALL("Formula::trueFormula");

  static Formula* res = new Formula(true);
  return res;
}

Formula* Formula::falseFormula()
{
  CALL("Formula::falseFormula");

  static Formula* res = new Formula(false);
  return res;
}

Formula* Formula::quantify(Formula* f)
{
  Set<unsigned> vars;
  FormulaVarIterator fvit( f );
  while(fvit.hasNext()) {
    vars.insert(fvit.next());
  }

  
  VarList* varLst=0;
  Set<unsigned>::Iterator vit(vars);
  while(vit.hasNext()) {
    VarList::push(vit.next(), varLst);
  }
  if(varLst) {
    f=new QuantifiedFormula(FORALL, varLst, f);
  }
  return f;
}

Formula* Formula::fromClause(Clause* cl)
{
  return fromClause(cl, cl->prop());
}

Formula* Formula::fromClause(Clause* cl, BDDNode* prop)
{
  CALL("Formula::fromClause");
  
  FormulaList* resLst=0;
  unsigned clen=cl->length();
  for(unsigned i=0;i<clen;i++) {
    Formula* lf=new AtomicFormula((*cl)[i]);
    FormulaList::push(lf, resLst);
  }

  if(prop && !BDD::instance()->isFalse(prop)) {
    FormulaList::push(BDD::instance()->toFormula(prop), resLst);
  }
  
  Formula* res=JunctionFormula::generalJunction(OR, resLst);
  
  Set<unsigned> vars;
  FormulaVarIterator fvit( res );
  while(fvit.hasNext()) {
    vars.insert(fvit.next());
  }

  
  VarList* varLst=0;
  Set<unsigned>::Iterator vit(vars);
  while(vit.hasNext()) {
    VarList::push(vit.next(), varLst);
  }
  if(varLst) {
    res=new QuantifiedFormula(FORALL, varLst, res);
  }

  return res;
}


std::ostream& operator<< (ostream& out, const Formula& f)
{
  CALL("operator <<(ostream&, const Formula&)");
  return out << f.toString();
}

}
