
/*
 * File Grounder.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <algorithm>

#include "Lib/Environment.hpp"
#include "Lib/SharedSet.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/Renaming.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/SubstHelper.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"

#include "SAT/SATClause.hpp"
#include "SAT/SATInference.hpp"
#include "SAT/SATLiteral.hpp"

#include "Grounder.hpp"

namespace Kernel
{

Grounder::Grounder()
: _nextSatVar(1)
{
  CALL("Grounder::Grounder");

}

SATClauseIterator Grounder::ground(Clause* cl)
{
  CALL("Grounder::ground(Clause*)");

  if(!cl->noSplits()) {
    NOT_IMPLEMENTED;
  }

  if(!cl->noProp()) {
    NOT_IMPLEMENTED;
  }

  SATClause* gndNonProp = groundNonProp(cl);


  SATInference* inf = new FOConversionInference(UnitSpec(cl));
  gndNonProp->setInference(inf);
  return pvi( getSingletonIterator(gndNonProp) );
}

void Grounder::groundNonProp(Clause* cl, SATLiteralStack& acc, Literal** normLits)
{
  CALL("Grounder::groundNonProp/2");

  static DArray<Literal*> lits;

  unsigned clen = cl->length();

  if(normLits) {
    std::copy(cl->literals(), cl->literals()+clen, normLits);
  }
  else {
    lits.initFromArray(clen, *cl);
    normLits = lits.array();
  }

  normalize(clen, normLits);

  for(unsigned i=0; i<clen; i++) {
    acc.push(groundNormalized(normLits[i]));
  }
}

SATClause* Grounder::groundNonProp(Clause* cl, Literal** normLits)
{
  CALL("Grounder::groundNonProp(Clause*,Literal**)");

  static SATLiteralStack gndLits;
  gndLits.reset();

  groundNonProp(cl, gndLits, normLits);

  SATClause* res = SATClause::fromStack(gndLits);
  return res;
}

SATLiteral Grounder::ground(Literal* lit)
{
  CALL("Grounder::ground(Literal*)");

  Literal* norm = lit;
  normalize(1, &norm);
  return groundNormalized(norm);
}

SATLiteral Grounder::groundNormalized(Literal* lit)
{
  CALL("Grounder::groundNormalized");

  bool isPos = lit->isPositive();
  Literal* posLit = Literal::positiveLiteral(lit);

  unsigned* pvar;
  if(_asgn.getValuePtr(posLit, pvar)) {
    *pvar = _nextSatVar++;
  }
  return SATLiteral(*pvar, isPos);
}

LiteralIterator Grounder::groundedLits()
{
  CALL("Grounder::groundedLits");

  return _asgn.domain();
}

void Grounder::recordInference(Clause* origClause, SATClause* refutation, Clause* resultClause)
{
  CALL("Grounder::recordInference");
  ASS(refutation);

  static Stack<UnitSpec> prems;
  prems.reset();

  if(origClause) {
    prems.push(UnitSpec(origClause));
  }
  SATInference::collectFOPremises(refutation, prems);

  unsigned premCnt = prems.size();

  InferenceStore::FullInference* inf = new(premCnt) InferenceStore::FullInference(premCnt);
  inf->rule = Inference::GLOBAL_SUBSUMPTION;

  for(unsigned i=0; i<premCnt; i++) {
    inf->premises[i] = prems[i];
  }

  InferenceStore::instance()->recordInference(UnitSpec(resultClause), inf);
}





struct GlobalSubsumptionGrounder::OrderNormalizingComparator
{
  Literal** _lits;
  OrderNormalizingComparator(Literal** lits) : _lits(lits) {}

  bool operator()(unsigned a, unsigned b) {
    Literal* la = _lits[a];
    Literal* lb = _lits[b];
    if(la==lb) { return false; }
    if(la->vars()!=lb->vars()) {
      
      
      return la->vars()<lb->vars();
    }
    if(la->weight()!=lb->weight()) {
      return la->weight()<lb->weight();
    }
    if(la->header()!=lb->header()) {
      return la->header()<lb->header();
    }
    
    static DisagreementSetIterator dsit;
    dsit.reset(la, lb, false);
    ALWAYS(dsit.hasNext());
    pair<TermList,TermList> da = dsit.next();
    if(da.first.isVar()!=da.second.isVar()) {
      return da.first.isVar();
    }
    if(da.first.isVar()) {
      ASS_NEQ(da.first.var(),da.second.var());
      return da.first.var()<da.second.var();
    }
    else {
      ASS_NEQ(da.first.term()->functor(),da.second.term()->functor());
      return da.first.term()->functor()<da.second.term()->functor();
    }
    return a<b; 
  }
};

void GlobalSubsumptionGrounder::normalize(unsigned cnt, Literal** lits)
{
  CALL("GlobalSubsumptionGrounder::normalize");

  if(!_doNormalization) { return; }

  if(cnt==0) { return; }
  if(cnt==1) {
    lits[0] = Renaming::normalize(lits[0]);
  }

  static Stack<unsigned> litOrder;
  litOrder.reset();
  litOrder.loadFromIterator(getRangeIterator(0u,cnt));

  std::sort(litOrder.begin(), litOrder.end(), OrderNormalizingComparator(lits));

  static Renaming normalizer;
  normalizer.reset();

  for(unsigned i=0; i<cnt; i++) {
    unsigned li = litOrder[i];
    normalizer.normalizeVariables(lits[li]);
    lits[li] = normalizer.apply(lits[li]);
  }
}






IGGrounder::IGGrounder()
{
  _tgtTerm = TermList(0, false);
  
}

class IGGrounder::CollapsingApplicator
{
  TermList _tgtTerm;
public:
  CollapsingApplicator(TermList tgtTerm) : _tgtTerm(tgtTerm) {}
  TermList apply(unsigned var)
  {

    return _tgtTerm;
  }
};

Literal* IGGrounder::collapseVars(Literal* lit)
{
  CALL("IGGrounder::collapseVars");

  CollapsingApplicator apl(_tgtTerm);
  return SubstHelper::apply(lit, apl);
}

void IGGrounder::normalize(unsigned cnt, Literal** lits)
{
  CALL("IGGrounder::normalize");

  for(unsigned i=0; i<cnt; i++) {
    lits[i] = collapseVars(lits[i]);
  }
}

}
