
/*
 * File Inference.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "Inference.hpp"

using namespace Kernel;

Inference::Inference(Rule r)
  : _rule(r)
{




}


void Inference::destroy()
{
  CALL ("Inference::destroy");
  delete this;
}

void Inference1::destroy()
{
  CALL ("Inference1::destroy");
  _premise1->decRefCnt();
  delete this;
}

void Inference2::destroy()
{
  CALL ("Inference2::destroy");
  _premise1->decRefCnt();
  _premise2->decRefCnt();
  delete this;
}

InferenceMany::InferenceMany(Rule rule,UnitList* premises)
  : Inference(rule),
    _premises(premises)
{
  CALL("InferenceMany::InferenceMany");
  UnitList* it=_premises;
  while(it) {
    it->head()->incRefCnt();
    it=it->tail();
  }
}

void InferenceMany::destroy()
{
  CALL ("InferenceMany::destroy");
  UnitList* it=_premises;
  while(it) {
    it->head()->decRefCnt();
    it=it->tail();
  }

  delete this;
}

Inference::Iterator InferenceMany::iterator()
{
  Iterator it;
  it.pointer = _premises;
  return it;
}

Inference::Iterator Inference1::iterator()
{
  Iterator it;
  it.integer = 1;
  return it;
}

Inference::Iterator Inference2::iterator()
{
  Iterator it;
  it.integer = 0;
  return it;
}

Inference::Iterator Inference::iterator()
{
  Iterator it;
#if VDEBUG
  it.integer=0;
#endif
  return it;
}

bool InferenceMany::hasNext(Iterator& it)
{
  return it.pointer;
}

bool Inference1::hasNext(Iterator& it)
{
  return it.integer;
}

bool Inference2::hasNext(Iterator& it)
{
  return it.integer < 2;
}

bool Inference::hasNext(Iterator&)
{
  return false;
}

Unit* InferenceMany::next(Iterator& it)
{
  ASS(it.pointer);
  UnitList* lst = reinterpret_cast<UnitList*>(it.pointer);
  it.pointer = lst->tail();
  return lst->head();
} 

Unit* Inference1::next(Iterator& it)
{
  ASS(it.integer);
  it.integer = 0;
  return _premise1;
} 

Unit* Inference2::next(Iterator& it)
{
  ASS(it.integer >= 0);
  ASS(it.integer < 2);
  return it.integer++ ? _premise2 : _premise1;
} 

Unit* Inference::next(Iterator&)
{
#if VDEBUG
  ASSERTION_VIOLATION;
#endif
  return 0;
} 



string Inference::ruleName(Rule rule)
{
  CALL("Inference::ruleName");

  switch (rule) {
  case INPUT:
    return "input";
  case NEGATED_CONJECTURE:
    return "negated conjecture";
  case ANSWER_LITERAL:
    return "answer literal";
  case RECTIFY:
    return "rectify";
  case CLOSURE:
    return "closure";
  case FLATTEN:
    return "flattening";
  case FORMULA_ITE_EXPANSION:
    return "formula ite expansion";
  case TERM_IF_THEN_ELSE_DEFINITION:
    return "term if then else definition";
  case SPECIAL_TERM_ELIMINATION:
    return "special term elimination";














  case REORDER_LITERALS:
    return "literal reordering";
  case ENNF:
    return "ennf transformation";
  case NNF:
    return "nnf transformation";










  case CLAUSIFY:
    return "cnf transformation";
  case REMOVE_DUPLICATE_LITERALS:
    return "duplicate literal removal";
  case SKOLEMIZE:
    return "skolemisation";
  case RESOLUTION:
    return "resolution";
  case EQUALITY_PROXY_REPLACEMENT:
    return "equality proxy replacement";
  case EQUALITY_PROXY_AXIOM1:
    return "equality proxy reflexivity axiom";
  case EQUALITY_PROXY_AXIOM2:
    return "equality proxy axiom";
  case DEFINITION_UNFOLDING:
    return "definition unfolding";
  case DEFINITION_FOLDING:
    return "definition folding";
  case SKOLEM_PREDICATE_INTRODUCTION:
  return "skolem predicate introduction";
  case PREDICATE_SKOLEMIZE:
    return "predicate skolemization";

  case PREDICATE_DEFINITION:
    return "predicate definition introduction";
  case PREDICATE_DEFINITION_UNFOLDING:
    return "predicate definition unfolding";
  case PREDICATE_DEFINITION_MERGING:
    return "predicate definition merging";
  case EQUIVALENCE_DISCOVERY:
    return "equivalence discovery";
  case FORMULA_SHARING:
    return "formula sharing";
  case REDUCE_FALSE_TRUE:
    return "true and false elimination";
  case LOCAL_SIMPLIFICATION:
    return "local simplification";
  case NORMALIZATION:
    return "normalization";
  case EQUALITY_PROPAGATION:
    return "equality propagation";
  case TRIVIAL_INEQUALITY_REMOVAL:
    return "trivial inequality removal";
  case FACTORING:
    return "factoring";
  case SUBSUMPTION_RESOLUTION:
    return "subsumption resolution";
  case SUPERPOSITION:
    return "superposition";
  case EQUALITY_FACTORING:
    return "equality factoring";
  case EQUALITY_RESOLUTION:
    return "equality resolution";
  case FORWARD_DEMODULATION:
    return "forward demodulation";
  case BACKWARD_DEMODULATION:
    return "backward demodulation";
  case FORWARD_LITERAL_REWRITING:
    return "forward literal rewriting";
  case CONDENSATION:
    return "condensation";
  case EVALUATION:
    return "evaluation";
  case INTERPRETED_SIMPLIFICATION:
    return "interpreted simplification";
  case UNUSED_PREDICATE_DEFINITION_REMOVAL:
    return "unused predicate definition removal";
  case PURE_PREDICATE_REMOVAL:
    return "pure predicate removal";
  case INEQUALITY_SPLITTING:
    return "inequality splitting";
  case INEQUALITY_SPLITTING_NAME_INTRODUCTION:
    return "inequality splitting name introduction";
  case GROUNDING:
    return "grounding";
  case EQUALITY_AXIOM:
    return "equality axiom";
  case THEORY:
    return "theory axiom";
  case BOOLEAN_TERM_ENCODING:
    return "boolean term encoding";
  case SPLITTING_COMPONENT:
    return "splitting component introduction";
  case SPLITTING:
    return "splitting";
  case BACKTRACKING_SPLITTING_COMPONENT:
    return "backtracking splitting component";
  case BACKTRACKING_SPLIT_REFUTATION:
    return "backtracking split refutation";
  case GENERAL_SPLITTING_COMPONENT:
    return "general splitting component introduction";
  case GENERAL_SPLITTING:
    return "general splitting";
  case COMMON_NONPROP_MERGE:
    return "merge";
  case PROP_REDUCE:
    return "prop reduce";
  case CLAUSE_NAMING:
    return "clause naming";
  case BDDZATION:
    return "bddzation";
  case TAUTOLOGY_INTRODUCTION:
    return "tautology introduction";
  case PROGRAM_ANALYSIS:
    return "program analysis";
  case COLOR_UNBLOCKING:
    return "color unblocking";
  case INSTANCE_GENERATION:
    return "instance generation";
  case UNIT_RESULTING_RESOLUTION:
    return "unit resulting resolution";
  case HYPER_SUPERPOSITION:
    return "hyper superposition";
  case GLOBAL_SUBSUMPTION:
    return "global subsumption";
  case DISTINCT_EQUALITY_REMOVAL:
    return "distinct equality removal";
  case EXTERNAL:
    return "external";
  case CLAIM_DEFINITION:
    return "claim definition";
  case BFNT_FLATTENING:
    return "BNFT flattening";
  case BFNT_DISTINCT:
    return "BNFT distinct";
  case BFNT_TOTALITY:
    return "BNFT totality";

  default:
    ASSERTION_VIOLATION;
    return "!UNKNOWN INFERENCE RULE!";
  }
} 



