
/*
 * File InterpretedLiteralEvaluator.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#include "Term.hpp"
#include "Theory.hpp"


#include "InterpretedLiteralEvaluator.hpp"

namespace Kernel
{

class InterpretedLiteralEvaluator::Evaluator
{
public:
  virtual ~Evaluator() {}

  bool canEvaluateFunc(unsigned func)
  {
    CALL("InterpretedLiteralEvaluator::Evaluator::canEvaluateFunc");

    if(!theory->isInterpretedFunction(func)) {
      return false;
    }
    Interpretation interp = theory->interpretFunction(func);
    return canEvaluate(interp);
  }

  bool canEvaluatePred(unsigned pred)
  {
    CALL("InterpretedLiteralEvaluator::Evaluator::canEvaluatePred");

    if(!theory->isInterpretedPredicate(pred)) {
      return false;
    }
    Interpretation interp = theory->interpretPredicate(pred);
    return canEvaluate(interp);
  }

  virtual bool canEvaluate(Interpretation interp) = 0;

  virtual bool tryEvaluateFunc(Term* trm, Term*& res) { return false; }
  virtual bool tryEvaluatePred(Literal* trm, bool& res)  { return false; }
};

class InterpretedLiteralEvaluator::ConversionEvaluator : public Evaluator
{
public:
  virtual bool canEvaluate(Interpretation interp)
  {
    CALL("InterpretedLiteralEvaluator::ConversionEvaluator::canEvaluate");

    return theory->isConversionOperation(interp);
  }

  virtual bool tryEvaluateFunc(Term* trm, Term*& res)
  {
    CALL("InterpretedLiteralEvaluator::ConversionEvaluator::tryEvaluateFunc");
    ASS(theory->isInterpretedFunction(trm));
    LOG("inf_ie","conv eval trying "<<trm->toString());

    try {
      Interpretation itp = theory->interpretFunction(trm);
      ASS(theory->isFunction(itp));
      ASS(theory->isConversionOperation(itp));
      ASS_EQ(theory->getArity(itp), 1);

      TermList argTrm = *trm->nthArgument(0);
      switch(itp) {
      case Theory::INT_TO_RAT:
      {
	IntegerConstantType arg;
	if(!theory->tryInterpretConstant(argTrm, arg)) { return false; }
	RationalConstantType resNum(arg,1);
	res = theory->representConstant(resNum);
	return true;
      }
      case Theory::INT_TO_REAL:
      {
	IntegerConstantType arg;
	if(!theory->tryInterpretConstant(argTrm, arg)) { return false; }
	RealConstantType resNum(RationalConstantType(arg,1));
	res = theory->representConstant(resNum);
	return true;
      }
      case Theory::RAT_TO_INT:
      {
	RationalConstantType arg;
	if(!theory->tryInterpretConstant(argTrm, arg)) { return false; }
	IntegerConstantType resNum = IntegerConstantType::floor(arg);
	res = theory->representConstant(resNum);
	return true;
      }
      case Theory::RAT_TO_REAL:
      {
	RationalConstantType arg;
	if(!theory->tryInterpretConstant(argTrm, arg)) { return false; }
	RealConstantType resNum(arg);
	res = theory->representConstant(resNum);
	return true;
      }
      case Theory::REAL_TO_INT:
      {
	RealConstantType arg;
	if(!theory->tryInterpretConstant(argTrm, arg)) { return false; }
	IntegerConstantType resNum = IntegerConstantType::floor(RationalConstantType(arg));
	res = theory->representConstant(resNum);
	return true;
      }
      case Theory::REAL_TO_RAT:
      {
	
	RealConstantType arg;
	if(!theory->tryInterpretConstant(argTrm, arg)) { return false; }
	RationalConstantType resNum(arg);
	res = theory->representConstant(resNum);
	return true;
      }

      default:
	ASSERTION_VIOLATION;
      }
    }
    catch(ArithmeticException&)
    {
      return false;
    }
  }
};

template<class T>
class InterpretedLiteralEvaluator::TypedEvaluator : public Evaluator
{
public:
  typedef T Value;

  TypedEvaluator() {}

  virtual bool canEvaluate(Interpretation interp)
  {
    CALL("InterpretedLiteralEvaluator::TypedEvaluator::canEvaluate");

    if(!theory->hasSingleSort(interp)) { return false; } 

    unsigned opSort = theory->getOperationSort(interp);
    return opSort==T::getSort();
  }

  virtual bool tryEvaluateFunc(Term* trm, Term*& res)
  {
    CALL("InterpretedLiteralEvaluator::tryEvaluateFunc");
    ASS(theory->isInterpretedFunction(trm));

    try {
      Interpretation itp = theory->interpretFunction(trm);
      ASS(theory->isFunction(itp));
      unsigned arity = theory->getArity(itp);

      if(arity!=1 && arity!=2) {
	INVALID_OPERATION("unsupported arity of interpreted operation: "+Int::toString(arity));
      }
      T resNum;
      TermList arg1Trm = *trm->nthArgument(0);
      T arg1;
      if(!theory->tryInterpretConstant(arg1Trm, arg1)) { return false; }
      if(arity==1) {
	if(!tryEvaluateUnaryFunc(itp, arg1, resNum)) { return false;}
      }
      else if(arity==2) {
	TermList arg2Trm = *trm->nthArgument(1);
	T arg2;
	if(!theory->tryInterpretConstant(arg2Trm, arg2)) { return false; }
	if(!tryEvaluateBinaryFunc(itp, arg1, arg2, resNum)) { return false;}
      }
      res = theory->representConstant(resNum);
      return true;
    }
    catch(ArithmeticException)
    {
      return false;
    }
  }

  virtual bool tryEvaluatePred(Literal* lit, bool& res)
  {
    CALL("InterpretedLiteralEvaluator::tryEvaluatePred");
    ASS(theory->isInterpretedPredicate(lit));

    try {
      Interpretation itp = theory->interpretPredicate(lit);
      ASS(!theory->isFunction(itp));
      unsigned arity = theory->getArity(itp);

      if(arity!=1 && arity!=2) {
	INVALID_OPERATION("unsupported arity of interpreted operation: "+Int::toString(arity));
      }
      TermList arg1Trm = *lit->nthArgument(0);
      T arg1;
      if(!theory->tryInterpretConstant(arg1Trm, arg1)) { return false; }
      if(arity==1) {
	if(!tryEvaluateUnaryPred(itp, arg1, res)) { return false;}
      }
      else {
	TermList arg2Trm = *lit->nthArgument(1);
	T arg2;
	if(!theory->tryInterpretConstant(arg2Trm, arg2)) { return false; }
	if(!tryEvaluateBinaryPred(itp, arg1, arg2, res)) { return false;}
      }
      if(lit->isNegative()) {
	res = !res;
      }
      return true;
    }
    catch(ArithmeticException&)
    {
      return false;
    }

  }
protected:

  virtual bool tryEvaluateUnaryFunc(Interpretation op, const T& arg, T& res)
  { return false; }
  virtual bool tryEvaluateBinaryFunc(Interpretation op, const T& arg1, const T& arg2, T& res)
  { return false; }

  virtual bool tryEvaluateUnaryPred(Interpretation op, const T& arg1, bool& res)
  { return false; }
  virtual bool tryEvaluateBinaryPred(Interpretation op, const T& arg1, const T& arg2, bool& res)
  { return false; }
};

class InterpretedLiteralEvaluator::IntEvaluator : public TypedEvaluator<IntegerConstantType>
{
protected:
  virtual bool tryEvaluateUnaryFunc(Interpretation op, const Value& arg, Value& res)
  {
    CALL("InterpretedLiteralEvaluator::IntEvaluator::tryEvaluateUnaryFunc");

    switch(op) {
    case Theory::INT_UNARY_MINUS:
      res = -arg;
      return true;
    case Theory::INT_SUCCESSOR:
      res = arg+1;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateBinaryFunc(Interpretation op, const Value& arg1,
      const Value& arg2, Value& res)
  {
    CALL("InterpretedLiteralEvaluator::IntEvaluator::tryEvaluateBinaryFunc");

    switch(op) {
    case Theory::INT_PLUS:
      res = arg1+arg2;
      return true;
    case Theory::INT_MINUS:
      res = arg1-arg2;
      return true;
    case Theory::INT_MULTIPLY:
      res = arg1*arg2;
      return true;
    case Theory::INT_DIVIDE:
      res = arg1/arg2;
      return true;
    case Theory::INT_MODULO:
      res = arg1%arg2;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateBinaryPred(Interpretation op, const Value& arg1,
      const Value& arg2, bool& res)
  {
    CALL("InterpretedLiteralEvaluator::IntEvaluator::tryEvaluateBinaryPred");

    switch(op) {
    case Theory::INT_GREATER:
      res = arg1>arg2;
      return true;
    case Theory::INT_GREATER_EQUAL:
      res = arg1>=arg2;
      return true;
    case Theory::INT_LESS:
      res = arg1<arg2;
      return true;
    case Theory::INT_LESS_EQUAL:
      res = arg1<=arg2;
      return true;
    case Theory::INT_DIVIDES:
      res = (arg1%arg2)==0;
      return true;
    default:
      return false;
    }
  }
};

class InterpretedLiteralEvaluator::RatEvaluator : public TypedEvaluator<RationalConstantType>
{
protected:
  virtual bool tryEvaluateUnaryFunc(Interpretation op, const Value& arg, Value& res)
  {
    CALL("InterpretedLiteralEvaluator::RatEvaluator::tryEvaluateUnaryFunc");

    switch(op) {
    case Theory::RAT_UNARY_MINUS:
      res = -arg;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateBinaryFunc(Interpretation op, const Value& arg1,
      const Value& arg2, Value& res)
  {
    CALL("InterpretedLiteralEvaluator::RatEvaluator::tryEvaluateBinaryFunc");

    switch(op) {
    case Theory::RAT_PLUS:
      res = arg1+arg2;
      return true;
    case Theory::RAT_MINUS:
      res = arg1-arg2;
      return true;
    case Theory::RAT_MULTIPLY:
      res = arg1*arg2;
      return true;
    case Theory::RAT_DIVIDE:
      res = arg1/arg2;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateBinaryPred(Interpretation op, const Value& arg1,
      const Value& arg2, bool& res)
  {
    CALL("InterpretedLiteralEvaluator::RatEvaluator::tryEvaluateBinaryPred");

    switch(op) {
    case Theory::RAT_GREATER:
      res = arg1>arg2;
      return true;
    case Theory::RAT_GREATER_EQUAL:
      res = arg1>=arg2;
      return true;
    case Theory::RAT_LESS:
      res = arg1<arg2;
      return true;
    case Theory::RAT_LESS_EQUAL:
      res = arg1<=arg2;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateUnaryPred(Interpretation op, const Value& arg1,
      bool& res)
  {
    CALL("InterpretedLiteralEvaluator::RatEvaluator::tryEvaluateBinaryPred");

    switch(op) {
    case Theory::RAT_IS_INT:
      res = arg1.isInt();
      return true;
    default:
      return false;
    }
  }
};

class InterpretedLiteralEvaluator::RealEvaluator : public TypedEvaluator<RealConstantType>
{
protected:
  virtual bool tryEvaluateUnaryFunc(Interpretation op, const Value& arg, Value& res)
  {
    CALL("InterpretedLiteralEvaluator::RealEvaluator::tryEvaluateUnaryFunc");

    switch(op) {
    case Theory::REAL_UNARY_MINUS:
      res = -arg;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateBinaryFunc(Interpretation op, const Value& arg1,
      const Value& arg2, Value& res)
  {
    CALL("InterpretedLiteralEvaluator::RealEvaluator::tryEvaluateBinaryFunc");

    switch(op) {
    case Theory::REAL_PLUS:
      res = arg1+arg2;
      return true;
    case Theory::REAL_MINUS:
      res = arg1-arg2;
      return true;
    case Theory::REAL_MULTIPLY:
      res = arg1*arg2;
      return true;
    case Theory::REAL_DIVIDE:
      res = arg1/arg2;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateBinaryPred(Interpretation op, const Value& arg1,
      const Value& arg2, bool& res)
  {
    CALL("InterpretedLiteralEvaluator::RealEvaluator::tryEvaluateBinaryPred");

    switch(op) {
    case Theory::REAL_GREATER:
      res = arg1>arg2;
      return true;
    case Theory::REAL_GREATER_EQUAL:
      res = arg1>=arg2;
      return true;
    case Theory::REAL_LESS:
      res = arg1<arg2;
      return true;
    case Theory::REAL_LESS_EQUAL:
      res = arg1<=arg2;
      return true;
    default:
      return false;
    }
  }

  virtual bool tryEvaluateUnaryPred(Interpretation op, const Value& arg1,
      bool& res)
  {
    CALL("InterpretedLiteralEvaluator::RealEvaluator::tryEvaluateBinaryPred");

    switch(op) {
    case Theory::REAL_IS_INT:
      res = arg1.isInt();
      return true;
    case Theory::REAL_IS_RAT:
      
      res = true;
      return true;
    default:
      return false;
    }
  }

};





InterpretedLiteralEvaluator::InterpretedLiteralEvaluator()
{
  CALL("InterpretedLiteralEvaluator::InterpretedLiteralEvaluator");

  _evals.push(new IntEvaluator());
  _evals.push(new RatEvaluator());
  _evals.push(new RealEvaluator());
  _evals.push(new ConversionEvaluator());

  _funEvaluators.ensure(0);
  _predEvaluators.ensure(0);

}

InterpretedLiteralEvaluator::~InterpretedLiteralEvaluator()
{
  CALL("InterpretedEvaluation::LiteralSimplifier::~LiteralSimplifier");

  while(_evals.isNonEmpty()) {
    delete _evals.pop();
  }
}

bool InterpretedLiteralEvaluator::evaluate(Literal* lit, bool& isConstant, Literal*& resLit, bool& resConst)
{
  CALL("InterpretedLiteralEvaluator::evaluate");

  resLit = TermTransformer::transform(lit);
  unsigned pred = resLit->functor();

  Evaluator* predEv = getPredEvaluator(pred);
  if(predEv) {
    if(predEv->tryEvaluatePred(resLit, resConst)) {
	isConstant = true;
	return true;
    }
  }
  if(resLit!=lit) {
    isConstant = false;
    return true;
  }
  return false;
}

TermList InterpretedLiteralEvaluator::transformSubterm(TermList trm)
{
  CALL("InterpretedLiteralEvaluator::transformSubterm");

  if(!trm.isTerm()) { return trm; }
  Term* t = trm.term();
  unsigned func = t->functor();

  Evaluator* funcEv = getFuncEvaluator(func);
  if(funcEv) {
    Term* res;
    if(funcEv->tryEvaluateFunc(t, res)) {
	return TermList(res);
    }
  }
  return trm;
}

InterpretedLiteralEvaluator::Evaluator* InterpretedLiteralEvaluator::getFuncEvaluator(unsigned func)
{
  CALL("InterpretedLiteralEvaluator::getFuncEvaluator");

  if(func>=_funEvaluators.size()) {
    unsigned oldSz = _funEvaluators.size();
    unsigned newSz = func+1;
    _funEvaluators.expand(newSz);
    for(unsigned i=oldSz; i<newSz; i++) {
	EvalStack::Iterator evit(_evals);
	while(evit.hasNext()) {
	  Evaluator* ev = evit.next();
	  if(ev->canEvaluateFunc(i)) {
	    ASS_EQ(_funEvaluators[i], 0); 
	    _funEvaluators[i] = ev;
	  }
	}
    }
  }
  return _funEvaluators[func];
}

InterpretedLiteralEvaluator::Evaluator* InterpretedLiteralEvaluator::getPredEvaluator(unsigned pred)
  {
    CALL("InterpretedLiteralEvaluator::getPredEvaluator");

    if(pred>=_predEvaluators.size()) {
      unsigned oldSz = _predEvaluators.size();
      unsigned newSz = pred+1;
      _predEvaluators.expand(newSz);
      for(unsigned i=oldSz; i<newSz; i++) {
	EvalStack::Iterator evit(_evals);
	while(evit.hasNext()) {
	  Evaluator* ev = evit.next();
	  if(ev->canEvaluatePred(i)) {
	    ASS_EQ(_predEvaluators[i], 0); 
	    _predEvaluators[i] = ev;
	  }
	}
      }
    }
    return _predEvaluators[pred];
  }

}
