
/*
 * File KBOForEPR.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"


#include "Lib/Environment.hpp"
#include "Lib/Comparison.hpp"
#include "Lib/Int.hpp"

#include "Indexing/TermSharing.hpp"

#include "Shell/Property.hpp"

#include "Problem.hpp"
#include "Term.hpp"
#include "Signature.hpp"

#include "KBOForEPR.hpp"

namespace Kernel {

using namespace Lib;


KBOForEPR::KBOForEPR(Problem& prb, const Options& opt)
 : KBOBase(prb, opt)
{
  CALL("KBOForEPR::KBOForEPR");
  ASS_EQ(prb.getProperty()->maxFunArity(),0);
}

Ordering::Result KBOForEPR::compare(Literal* l1, Literal* l2) const
{
  CALL("KBOForEPR::compare(Literal*...)");
  ASS(l1->shared());
  ASS(l2->shared());

  if (l1 == l2) {
    return EQUAL;
  }

  unsigned p1 = l1->functor();
  unsigned p2 = l2->functor();

  if( (l1->isNegative() ^ l2->isNegative()) && (p1==p2) &&
	  l1->weight()==l2->weight() && l1->vars()==l2->vars() &&
	  l1==env.sharing->tryGetOpposite(l2)) {
    return l1->isNegative() ? LESS : GREATER;
  }

  Result res;
  if (p1 != p2) {
    Comparison levComp=Int::compare(predicateLevel(p1),predicateLevel(p2));
    if(levComp!=Lib::EQUAL) {
      res = fromComparison(levComp);
      goto fin;
    }
  }

  if(l1->isEquality()) {
    ASS(l2->isEquality());
    return compareEqualities(l1, l2);
  }
  ASS(!l1->isEquality());
  ASS(!l2->isEquality());

  if (p1 != p2) {
    Comparison arComp=Int::compare(l1->arity(),l2->arity());
    if(arComp!=Lib::EQUAL) {
      
      
      
      res = fromComparison(arComp);
      goto fin;
    }

    Comparison prComp=Int::compare(predicatePrecedence(p1),predicatePrecedence(p2));
    ASS_NEQ(prComp, Lib::EQUAL); 
    res = fromComparison(prComp);
    goto fin;
  }

  {
    TermList* t1=l1->args();
    TermList* t2=l2->args();

    ASS(!t1->isEmpty());
    while(*t1==*t2) {
      t1=t1->next();
      t2=t2->next();
      ASS(!t1->isEmpty()); 
    }
    res = compare(*t1, *t2);
    goto fin;
  }

fin:
  if(_reverseLCM && (l1->isNegative() || l2->isNegative()) ) {
    if(l1->isNegative() && l2->isNegative()) {
      res = reverse(res);
    }
    else {
      res = l1->isNegative() ? LESS : GREATER;
    }
  }
  return res;
}

Ordering::Result KBOForEPR::compare(TermList tl1, TermList tl2) const
{
  CALL("KBOForEPR::compare(TermList)");
  ASS(!tl1.isTerm() || tl1.term()->arity()==0)
  ASS(!tl2.isTerm() || tl2.term()->arity()==0)

  if(tl1==tl2) {
    return EQUAL;
  }

  if(tl1.isOrdinaryVar() || tl2.isOrdinaryVar()) {
    return INCOMPARABLE;
  }

  
  
  ASS_NEQ(tl1.term()->functor(), tl2.term()->functor());

  return compareFunctionPrecedences(tl1.term()->functor(), tl2.term()->functor());
}

}
