
/*
 * File LiteralSelector.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Exception.hpp"

#include "Shell/Options.hpp"

#include "Clause.hpp"
#include "Signature.hpp"
#include "Term.hpp"

#include "LiteralSelector.hpp"

#include "MaximalLiteralSelector.hpp"
#include "BestLiteralSelector.hpp"
#include "LookaheadLiteralSelector.hpp"

#include "LiteralComparators.hpp"

#undef LOGGING
#define LOGGING 0

namespace Kernel
{

ZIArray<bool> LiteralSelector::_reversePredicate;


bool LiteralSelector::isPositiveForSelection(Literal* l) const
{
  if(l->isEquality()) {
    return l->isPositive(); 
  }
  unsigned pred = l->functor();
  return l->isPositive() ^ _reversePolarity ^ _reversePredicate[pred];
}

void LiteralSelector::reversePredicatePolarity(unsigned pred, bool reverse)
{
  CALL("reversePredicatePolarity");
  ASS(pred!=0 || !reverse); 

  _reversePredicate[pred] = reverse;
}

int LiteralSelector::getSelectionPriority(Literal* l) const
{
  CALL("LiteralSelector::getSelectionPriority");

  Signature::Symbol* psym=env.signature->getPredicate(l->functor());
  if(psym->swbName()) {
    if(l->isPositive() && _opt.splittingWithBlocking()) {
      return 1;
    }
    return -1;
  }
  if(psym->cfName() || psym->answerPredicate()) {
    return -2;
  }
  return 0;
}

LiteralSelector* LiteralSelector::getSelector(const Ordering& ordering, const Options& options, int selectorNumber)
{
  CALL("LiteralSelector::getSelector");

  using namespace LiteralComparators;

  typedef Composite<ColoredFirst,
	    Composite<MaximalSize,
	    LexComparator> > Comparator2;

  typedef Composite<ColoredFirst,
	    Composite<NoPositiveEquality,
	    Composite<LeastTopLevelVariables,
	    Composite<LeastDistinctVariables, LexComparator> > > > Comparator3;

  typedef Composite<ColoredFirst,
	    Composite<NoPositiveEquality,
	    Composite<LeastTopLevelVariables,
	    Composite<LeastVariables,
	    Composite<MaximalSize, LexComparator> > > > > Comparator4;

  typedef Composite<ColoredFirst,
	    Composite<NegativeEquality,
	    Composite<MaximalSize,
	    Composite<Negative, LexComparator> > > > Comparator10;

  int absNum = abs(selectorNumber);

  LiteralSelector* res;
  switch(absNum) {
  case 0: res = new TotalLiteralSelector(ordering, options); break;
  case 1: res = new MaximalLiteralSelector(ordering, options); break;
  case 2: res = new CompleteBestLiteralSelector<Comparator2>(ordering, options); break;
  case 3: res = new CompleteBestLiteralSelector<Comparator3>(ordering, options); break;
  case 4: res = new CompleteBestLiteralSelector<Comparator4>(ordering, options); break;
  case 10: res = new CompleteBestLiteralSelector<Comparator10>(ordering, options); break;

  case 11: res = new LookaheadLiteralSelector(true, ordering, options); break;

  case 1002: res = new BestLiteralSelector<Comparator2>(ordering, options); break;
  case 1003: res = new BestLiteralSelector<Comparator3>(ordering, options); break;
  case 1004: res = new BestLiteralSelector<Comparator4>(ordering, options); break;
  case 1010: res = new BestLiteralSelector<Comparator10>(ordering, options); break;

  case 1011: res = new LookaheadLiteralSelector(false, ordering, options); break;

  default:
    INVALID_OPERATION("Undefined selection function");
  }
  if(selectorNumber<0) {
    res->_reversePolarity = true;
  }
  return res;
}

void LiteralSelector::ensureSomeColoredSelected(Clause* c, unsigned eligible)
{
  CALL("LiteralSelector::ensureSomeColoredSelected");

  if(c->color()==COLOR_TRANSPARENT) {
    
    return;
  }

  unsigned selCnt=c->selected();

  for(unsigned i=0;i<selCnt;i++) {
    if((*c)[i]->color()!=COLOR_TRANSPARENT) {
      return;
    }
  }

  for(unsigned i=selCnt;i<eligible;i++) {
    if((*c)[i]->color()!=COLOR_TRANSPARENT) {
      swap((*c)[selCnt], (*c)[i]);
      c->setSelected(selCnt+1);
      return;
    }
  }
  
  ASS_L(eligible, c->length());
}

void LiteralSelector::select(Clause* c, unsigned eligibleInp)
{
  CALL("LiteralSelector::select");
  ASS_LE(eligibleInp, c->length());

  if(eligibleInp==0) {
    eligibleInp = c->length();
  }

  if(eligibleInp<=1) {
    c->setSelected(eligibleInp);
    return;
  }

  unsigned eligible=1;
  int maxPriority=getSelectionPriority((*c)[0]);
  bool modified=false;

  for(unsigned i=1;i<eligibleInp;i++) {
    int priority=getSelectionPriority((*c)[i]);
    if(priority==maxPriority) {
      if(eligible!=i) {
	swap((*c)[i],(*c)[eligible]);
	modified=true;
      }
      eligible++;
    }
    else if(priority>maxPriority) {
      maxPriority=priority;
      eligible=1;
      swap((*c)[i],(*c)[0]);
      modified=true;
    }
  }
  ASS_LE(eligible,eligibleInp);
  if(modified) {
    c->notifyLiteralReorder();
  }

  if(eligible==1) {
    c->setSelected(eligible);
    return;
  }

  ASS_G(eligible,1);
  doSelection(c, eligible);

  COND_TRACE("ls", isPositiveForSelection((*c)[0]),
      tout << "Selected "<<c->selected()<<" positive literals in " << (*c) << endl;
  );
}

void TotalLiteralSelector::doSelection(Clause* c, unsigned eligible)
{
  CALL("TotalLiteralSelector::doSelection");

  c->setSelected(eligible);
}

}
