
/*
 * File Ordering_Equality.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Term.hpp"

#include "Ordering.hpp"

namespace Kernel
{

class Ordering::EqCmp
{
public:
  EqCmp(Ordering* ordering) : _ordering(ordering)
  {
#if VDEBUG
  inUse=false;
#endif
  }

  Result compareEqualities(Literal* eq1, Literal* eq2) const;

#if VDEBUG
  mutable bool inUse;
#endif

private:

  Result compare(TermList trm1, TermList trm2) const
  {
    return _ordering->compare(trm1, trm2);
  }

  Result compare_s1Gt1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1GEt1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1It1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1It1_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s2Lt2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s2LEt2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1GEt1_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s1It2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s1GEt2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s1GEt2_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1GEt1_s1GEt2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1GEt1_s1It2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1GEt1_s2LEt2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1Gt1_s1GEt2_s2Lt2(TermList s1,TermList s2,TermList t1,TermList t2) const;
  Result compare_s1GEt1_s1GEt2_s2LEt1(TermList s1,TermList s2,TermList t1,TermList t2) const;

  mutable TermList s1,s2,t1,t2;

  Ordering* _ordering;
};

void Ordering::createEqualityComparator()
{
  CALL("Ordering::createEqualityComparator");

  _eqCmp=new EqCmp(this);
}

void Ordering::destroyEqualityComparator()
{
  CALL("Ordering::destroyEqualityComparator");

  delete _eqCmp;
#if VDEBUG
  _eqCmp=0;
#endif
}


Ordering::Result Ordering::compareEqualities(Literal* eq1, Literal* eq2) const
{
  CALL("Ordering::compareEqualities");
  ASS(eq1->isEquality());
  ASS(eq2->isEquality());

#if VDEBUG
  ASS(!_eqCmp->inUse);
  _eqCmp->inUse=true;
#endif

  Result res=_eqCmp->compareEqualities(eq1, eq2);

#if VDEBUG
  ASS(_eqCmp->inUse);
  _eqCmp->inUse=false;
#endif



  return res;
}

Ordering::Result Ordering::EqCmp::compareEqualities(Literal* eq1, Literal* eq2) const
{
  CALL("Ordering::EqCmp::compareEqualities");
  ASS(eq1->isEquality());
  ASS(eq2->isEquality());

  s1=*eq1->nthArgument(0);
  s2=*eq1->nthArgument(1);
  t1=*eq2->nthArgument(0);
  t2=*eq2->nthArgument(1);

  if (s1 == t1) {
    return compare(s2,t2);
  }
  if (s1 == t2) {
    return compare(s2,t1);
  }
  if (s2 == t1) {
    return compare(s1,t2);
  }
  if (s2 == t2) {
    return compare(s1,t1);
  }

  switch(compare(s1,t1)) {
  case GREATER:
    return compare_s1Gt1(s1,s2,t1,t2);
  case GREATER_EQ:
    return compare_s1GEt1(s1,s2,t1,t2);
  case INCOMPARABLE:
    return compare_s1It1(s1,s2,t1,t2);
  case LESS_EQ:
    return reverse(compare_s1GEt1(t1,t2,s1,s2));
  case LESS:
    return reverse(compare_s1Gt1(t1,t2,s1,s2));
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  CALL("Ordering::EqCmp::compare_s1Gt1");
  ASS_EQ(compare(s1,t1), GREATER);

  switch(compare(s2,t2)) {
  case GREATER:
  case GREATER_EQ:
    return GREATER;
  case INCOMPARABLE:
    return compare_s1Gt1_s2It2(s1,s2,t1,t2);
  case LESS_EQ:
    return compare_s1Gt1_s2LEt2(s1,s2,t1,t2);
  case LESS:
    return compare_s1Gt1_s2Lt2(s1,s2,t1,t2);
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1GEt1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER_EQ);

  switch(compare(s2,t2)) {
  case GREATER:
    return GREATER;
  case GREATER_EQ:
    return GREATER_EQ;
  case INCOMPARABLE:
    return compare_s1GEt1_s2It2(s1,s2,t1,t2);
  case LESS_EQ:
    return compare_s1GEt1_s2LEt2(s1,s2,t1,t2);
  case LESS:
    return reverse(compare_s1Gt1_s2LEt2(t2,t1,s2,s1));
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1It1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), INCOMPARABLE);

  switch(compare(s2,t2)) {
  case GREATER:
    return compare_s1Gt1_s2It2(s2,s1,t2,t1);
  case GREATER_EQ:
    return compare_s1GEt1_s2It2(s2,s1,t2,t1);
  case INCOMPARABLE:
    return compare_s1It1_s2It2(s1,s2,t1,t2);
  case LESS_EQ:
    return reverse(compare_s1GEt1_s2It2(t2,t1,s2,s1));
  case LESS:
    return reverse(compare_s1Gt1_s2It2(t2,t1,s2,s1));
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1It1_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), INCOMPARABLE);
  ASS_EQ(compare(s2,t2), INCOMPARABLE);

  switch(compare(s1,t2)) {
  case GREATER:
    return compare_s1Gt1_s1It2_s2It1(s1,s2,t2,t1);
  case GREATER_EQ:
    return compare_s1GEt1_s1It2_s2It1(s1,s2,t2,t1);
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    return reverse(compare_s1GEt1_s1It2_s2It1(t2,t1,s1,s2));
  case LESS:
    return reverse(compare_s1Gt1_s1It2_s2It1(t2,t1,s1,s2));
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s2,t2), INCOMPARABLE);

  switch(compare(s1,t2)) {
  case GREATER:
    return GREATER;
  case GREATER_EQ:
    return compare_s1Gt1_s1GEt2_s2It2(s1,s2,t1,t2);
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
  case LESS:
    ASS_NEQ(compare(s2,t1), LESS); 
    ASS_NEQ(compare(s2,t1), LESS_EQ); 
    ASS_EQ(compare(t1,t2), LESS); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1_s2Lt2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s2,t2), LESS);

  switch(compare(s1,t2)) {
  case GREATER:
    ASS_EQ(compare(s1,s2), GREATER); 
    return GREATER;
  case GREATER_EQ:
    ASS_NEQ(compare(t1,t2), GREATER); 
    ASS_EQ(compare(s1,s2), GREATER); 
    return compare_s1Gt1_s1GEt2_s2Lt2(s1,s2,t1,t2);

  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    ASS_NEQ(compare(s1,s2), LESS); 
    ASS_EQ(compare(t1,t2), LESS); 
    return reverse(compare_s1Gt1_s1GEt2_s2Lt2(t2,t1,s2,s1));

  case LESS:
    ASS_EQ(compare(t1,t2), LESS); 
    return LESS;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1_s2LEt2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s2,t2), LESS_EQ);

  switch(compare(s1,t2)) {
  case GREATER:
    ASS_EQ(compare(s1,s2), GREATER); 
    return GREATER;
  case GREATER_EQ:
    ASS_NEQ(compare(t1,t2), GREATER); 
    ASS(compare(s1,s2)==LESS || compare(s1,s2)==LESS_EQ); 
    return INCOMPARABLE;
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
  case LESS:
    ASS_EQ(compare(t1,t2), LESS); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1GEt1_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER_EQ);
  ASS_EQ(compare(s2,t2), INCOMPARABLE);

  switch(compare(s1,t2)) {
  case GREATER:
    return compare_s1Gt1_s1GEt2_s2It1(s1,s2,t2,t1);
  case GREATER_EQ:
    return compare_s1GEt1_s1GEt2_s2It1(s1,s2,t2,t1);
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    ASS(compare(t1,t2)==LESS || compare(t1,t2)==LESS_EQ); 
    ASS(compare(s2,t1) != LESS && compare(s2,t1) != LESS_EQ);
    return INCOMPARABLE;
  case LESS:
    ASS_EQ(compare(t1,t2), LESS); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1_s1It2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s1,t2), INCOMPARABLE);
  ASS_EQ(compare(s2,t1), INCOMPARABLE);

  switch(compare(s2,t2)) {
  case GREATER:
  case GREATER_EQ:
    return GREATER;
  case INCOMPARABLE:
  case LESS_EQ:
  case LESS:
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1_s1GEt2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s1,t2), GREATER_EQ);
  ASS_EQ(compare(s2,t1), INCOMPARABLE);

  switch(compare(s2,t2)) {
  case GREATER:
  case GREATER_EQ:
    return GREATER;
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    ASS(compare(s1,s2)==GREATER || compare(s1,s2)==GREATER_EQ); 
    return INCOMPARABLE;
  case LESS:
    ASS_EQ(compare(s1,s2), GREATER); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1Gt1_s1GEt2_s2It2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s1,t2), GREATER_EQ);
  ASS_EQ(compare(s2,t2), INCOMPARABLE);

  switch(compare(s2,t1)) {
  case GREATER:
    return GREATER;
  case GREATER_EQ:
    return GREATER_EQ;
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
  case LESS:
    ASS_EQ(compare(s1,s2), GREATER); 
    ASS_NEQ(compare(t1,t2), LESS); 
    ASS_NEQ(compare(t1,t2), LESS_EQ); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1GEt1_s1GEt2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER_EQ);
  ASS_EQ(compare(s1,t2), GREATER_EQ);
  ASS_EQ(compare(s2,t1), INCOMPARABLE);

  switch(compare(s2,t2)) {
  case GREATER:
    return GREATER;
  case GREATER_EQ:
    return GREATER_EQ;
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    ASS(compare(s1,s2)==GREATER || compare(s1,s2)==GREATER_EQ); 
    return INCOMPARABLE;
  case LESS:
    ASS_EQ(compare(s1,s2), GREATER); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1GEt1_s1It2_s2It1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER_EQ);
  ASS_EQ(compare(s1,t2), INCOMPARABLE);
  ASS_EQ(compare(s2,t1), INCOMPARABLE);

  ASS_NEQ(compare(t1,t2), GREATER); 
  ASS_NEQ(compare(t1,t2), GREATER_EQ); 
  ASS_NEQ(compare(s1,s2), LESS); 
  ASS_NEQ(compare(s1,s2), LESS_EQ); 

  switch(compare(s2,t2)) {
  case GREATER:
    ASS_NEQ(compare(s1,s2), GREATER); 
    ASS_NEQ(compare(s1,s2), GREATER_EQ); 
    ASS_NEQ(compare(t1,t2), LESS); 
    ASS_NEQ(compare(t1,t2), LESS_EQ); 
    return GREATER;
  case GREATER_EQ:
    ASS_NEQ(compare(s1,s2), GREATER); 
    ASS_NEQ(compare(s1,s2), GREATER_EQ); 
    ASS_NEQ(compare(t1,t2), LESS); 
    ASS_NEQ(compare(t1,t2), LESS_EQ); 
    return GREATER_EQ;
  case INCOMPARABLE:
  case LESS_EQ:
  case LESS:
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1GEt1_s2LEt2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER_EQ);
  ASS_EQ(compare(s2,t2), LESS_EQ);

  switch(compare(s1,t2)) {
  case GREATER:
    ASS_EQ(compare(s1,s2),GREATER); 
    ASS_NEQ(compare(t2,t1), GREATER);
    ASS_NEQ(compare(t2,t1), GREATER_EQ);
    ASS_NEQ(compare(s2,t1), GREATER_EQ);
    ASS_NEQ(compare(s2,t1), GREATER);
    return INCOMPARABLE;

  case GREATER_EQ:
    ASS(compare(s1,s2)==GREATER || compare(s1,s2)==GREATER_EQ); 
    return compare_s1GEt1_s1GEt2_s2LEt1(s1,s2,t2,t1);

  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    ASS(compare(t2,t1)==GREATER || compare(t2,t1)==GREATER_EQ); 
    return reverse(compare_s1GEt1_s1GEt2_s2LEt1(t2,t1,s1,s2));

  case LESS:
    ASS_EQ(compare(t2,t1),GREATER); 
    ASS_NEQ(compare(s2,s1), GREATER);
    ASS_NEQ(compare(s2,s1), GREATER_EQ);
    ASS_NEQ(compare(t2,s1), GREATER_EQ);
    ASS_NEQ(compare(t2,s1), GREATER);
    return INCOMPARABLE;

  default:
    ASSERTION_VIOLATION;
  }
}


Ordering::Result Ordering::EqCmp::compare_s1Gt1_s1GEt2_s2Lt2(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER);
  ASS_EQ(compare(s1,t2), GREATER_EQ);
  ASS_EQ(compare(s2,t2), LESS);

  ASS_EQ(compare(s1,s2), GREATER); 
  ASS_NEQ(compare(t1,t2), GREATER);
  ASS_NEQ(compare(t1,t2), GREATER_EQ);

  switch(compare(s2,t1)) {
  case GREATER:
    ASS_EQ(compare(t2,t1),GREATER); 
    return GREATER;
  case GREATER_EQ:
    ASS_EQ(compare(t2,t1),GREATER); 
    return GREATER_EQ;
  case INCOMPARABLE:
  case LESS_EQ:
  case LESS:
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}

Ordering::Result Ordering::EqCmp::compare_s1GEt1_s1GEt2_s2LEt1(TermList s1,TermList s2,TermList t1,TermList t2) const
{
  ASS_EQ(compare(s1,t1), GREATER_EQ);
  ASS_EQ(compare(s1,t2), GREATER_EQ);
  ASS_EQ(compare(s2,t1), LESS_EQ);

  switch(compare(s2,t2)) {
  case GREATER:
    ASSERTION_VIOLATION;
  case GREATER_EQ:
    ASS(compare(t1,t2)==GREATER || compare(t1,t2)==GREATER_EQ); 
    return GREATER_EQ;
  case INCOMPARABLE:
    return INCOMPARABLE;
  case LESS_EQ:
    ASS(compare(s1,s2)==GREATER || compare(s1,s2)==GREATER_EQ); 
    return INCOMPARABLE;
  case LESS:
    ASS_EQ(compare(s1,s2),GREATER); 
    return INCOMPARABLE;
  default:
    ASSERTION_VIOLATION;
  }
}


}
