
/*
 * File Signature.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Int.hpp"
#include "Shell/Options.hpp"
#include "Signature.hpp"

using namespace std;
using namespace Kernel;

const unsigned Signature::STRING_DISTINCT_GROUP = 0;
const unsigned Signature::INTEGER_DISTINCT_GROUP = 1;
const unsigned Signature::RATIONAL_DISTINCT_GROUP = 2;
const unsigned Signature::REAL_DISTINCT_GROUP = 3;
const unsigned Signature::LAST_BUILT_IN_DISTINCT_GROUP = 3;

Signature::Symbol::Symbol(const string& nm,unsigned arity, bool interpreted, bool stringConstant)
  : _name(nm),
    _arity(arity),
    _interpreted(interpreted ? 1 : 0),
    _introduced(0),
    _protected(0),
    _skip(0),
    _cfName(0),
    _swbName(0),
    _equalityProxy(0),
    _color(COLOR_TRANSPARENT),
    _stringConstant(stringConstant ? 1: 0),
    _answerPredicate(0),
    _type(0),
    _distinctGroups(0)
{
  CALL("Signature::Symbol::Symbol");
  ASS(!stringConstant || arity==0);

  if(symbolNeedsQuoting(_name, interpreted, arity, stringConstant)) {
    _name="'"+_name+"'";
  }
  if(_interpreted || isProtectedName(nm)) {
    markProtected();
  }
}

Signature::Symbol::~Symbol()
{
  CALL("Signature::Symbol::~Symbol");

  if(_type) {
    delete _type;
  }
}

void Signature::Symbol::destroyFnSymbol()
{
  CALL("Signature::Symbol::destroyFnSymbol");

  if(integerConstant()) {
    delete static_cast<IntegerSymbol*>(this);
  }
  else if(rationalConstant()) {
    delete static_cast<RationalSymbol*>(this);
  }
  else if(realConstant()) {
    delete static_cast<RealSymbol*>(this);
  }
  else if(interpreted()) {
    delete static_cast<InterpretedSymbol*>(this);
  }
  else {
    delete this;
  }
}

void Signature::Symbol::destroyPredSymbol()
{
  CALL("Signature::Symbol::destroyPredSymbol");

  if(interpreted()) {
    delete static_cast<InterpretedSymbol*>(this);
  }
  else {
    delete this;
  }
}

void Signature::Symbol::addToDistinctGroup(unsigned group)
{
  CALL("Signature::Symbol::addToDistinctGroup");

  ASS_EQ(arity(), 0);
  ASS(!_distinctGroups->member(group))

  List<unsigned>::push(group, _distinctGroups);
}

void Signature::Symbol::setType(BaseType* type)
{
  CALL("Signature::Symbol::setType");
  ASS(!_type);

  _type = type;
}

FunctionType* Signature::Symbol::fnType() const
{
  CALL("Signature::Symbol::fnType");

  if(!_type) {
    _type = new FunctionType(arity());
  }
  return static_cast<FunctionType*>(_type);
}

PredicateType* Signature::Symbol::predType() const
{
  CALL("Signature::Symbol::predType");

  if(!_type) {
    _type = new PredicateType(arity());
  }
  return static_cast<PredicateType*>(_type);
}


Signature::Signature ()
  : _funs(32),
    _preds(32),
    _nextFreshSymbolNumber(0),
    _strings(0),
    _integers(0),
    _rationals(0),
    _reals(0)
{
  CALL("Signature::Signature");

  
  addInterpretedPredicate(Theory::EQUAL, "=");
  ASS_EQ(predicateName(0), "="); 
  getPredicate(0)->markSkip();

  unsigned aux;
  aux = createDistinctGroup();
  ASS_EQ(STRING_DISTINCT_GROUP, aux);
  aux = createDistinctGroup();
  ASS_EQ(INTEGER_DISTINCT_GROUP, aux);
  aux = createDistinctGroup();
  ASS_EQ(RATIONAL_DISTINCT_GROUP, aux);
  aux = createDistinctGroup();
  ASS_EQ(REAL_DISTINCT_GROUP, aux);
} 

Signature::~Signature ()
{
  for (int i = _funs.length()-1;i >= 0;i--) {
    _funs[i]->destroyFnSymbol();
  }
  for (int i = _preds.length()-1;i >= 0;i--) {
    _preds[i]->destroyPredSymbol();
  }
} 

unsigned Signature::addInterpretedFunction(Interpretation interpretation, const string& name)
{
  CALL("Signature::addInterpretedFunction");
  ASS(Theory::isFunction(interpretation));

  unsigned res;
  if(_iSymbols.find(interpretation,res)) { 
    if(name!=functionName(res)) {
      USER_ERROR("Interpreted function '"+functionName(res)+"' has the same interpretation as '"+name+"' should have");
    }
    return res;
  }

  string symbolKey = name+"_i"+Int::toString(interpretation);

  ASS(!_funNames.find(symbolKey));

  unsigned fnNum = _funs.length();
  InterpretedSymbol* sym = new InterpretedSymbol(name, interpretation);
  _funs.push(sym);
  _funNames.insert(symbolKey, fnNum);
  ALWAYS(_iSymbols.insert(interpretation, fnNum));
  BaseType* fnType = Theory::getOperationType(interpretation);
  ASS(fnType->isFunctionType());
  sym->setType(fnType);

  return fnNum;
}

unsigned Signature::addInterpretedPredicate(Interpretation interpretation, const string& name)
{
  CALL("Signature::addInterpretedPredicate");
  ASS(!Theory::isFunction(interpretation));

  unsigned res;
  if(_iSymbols.find(interpretation,res)) { 
    if(name!=predicateName(res)) {
      USER_ERROR("Interpreted predicate '"+predicateName(res)+"' has the same interpretation as '"+name+"' should have");
    }
    return res;
  }

  string symbolKey = name+"_i"+Int::toString(interpretation);

  ASS(!_predNames.find(symbolKey));

  unsigned predNum = _preds.length();
  InterpretedSymbol* sym = new InterpretedSymbol(name, interpretation);
  _preds.push(sym);
  _predNames.insert(symbolKey,predNum);
  ALWAYS(_iSymbols.insert(interpretation, predNum));
  if(predNum!=0) {
    BaseType* predType = Theory::getOperationType(interpretation);
    ASS(!predType->isFunctionType());
    sym->setType(predType);
  }
  return predNum;
}



unsigned Signature::addIntegerConstant(const string& number)
{
  CALL("Signature::addIntegerConstant(string)");

  IntegerConstantType value(number);
  return addIntegerConstant(value);
}

unsigned Signature::addIntegerConstant(const IntegerConstantType& value)
{
  CALL("Signature::addIntegerConstant");

  
  string key = value.toString() + "_n";
  unsigned result;
  if(_funNames.find(key, result)) {
    return result;
  }
  _integers++;
  result = _funs.length();
  _funs.push(new IntegerSymbol(value));
  _funNames.insert(key, result);
  return result;
}

unsigned Signature::addRationalConstant(const string& numerator, const string& denominator)
{
  CALL("Signature::addRationalConstant(string,string)");

  RationalConstantType value(numerator, denominator);
  return addRationalConstant(value);
}

unsigned Signature::addRationalConstant(const RationalConstantType& value)
{
  CALL("Signature::addRationalConstant");

  string key = value.toString() + "_q";
  unsigned result;
  if(_funNames.find(key, result)) {
    return result;
  }
  _rationals++;
  result = _funs.length();
  _funs.push(new RationalSymbol(value));
  _funNames.insert(key, result);
  return result;
}

unsigned Signature::addRealConstant(const string& number)
{
  CALL("Signature::addRealConstant(string)");

  RealConstantType value(number);
  return addRealConstant(value);
}

unsigned Signature::addRealConstant(const RealConstantType& value)
{
  CALL("Signature::addRealConstant");

  string key = value.toString() + "_r";
  unsigned result;
  if(_funNames.find(key, result)) {
    return result;
  }
  _reals++;
  result = _funs.length();
  _funs.push(new RealSymbol(value));
  _funNames.insert(key, result);
  return result;
}

unsigned Signature::getInterpretingSymbol(Interpretation interp)
{
  CALL("Signature::getInterpretingSymbol");

  unsigned res;
  if(_iSymbols.find(interp, res)) {
    return res;
  }
  string name;
  switch(interp) {
  case Theory::INT_SUCCESSOR:
    
    name="$successor";
    break;
  case Theory::INT_DIVIDE:
  case Theory::RAT_DIVIDE:
  case Theory::REAL_DIVIDE:
    
    name="$divide";
    break;
  case Theory::INT_UNARY_MINUS:
  case Theory::RAT_UNARY_MINUS:
  case Theory::REAL_UNARY_MINUS:
    name="$uminus";
    break;
  case Theory::INT_PLUS:
  case Theory::RAT_PLUS:
  case Theory::REAL_PLUS:
    name="$sum";
    break;
  case Theory::INT_MINUS:
  case Theory::RAT_MINUS:
  case Theory::REAL_MINUS:
    name="$difference";
    break;
  case Theory::INT_MULTIPLY:
  case Theory::RAT_MULTIPLY:
  case Theory::REAL_MULTIPLY:
    name="$product";
    break;
  case Theory::INT_GREATER:
  case Theory::RAT_GREATER:
  case Theory::REAL_GREATER:
    name="$greater";
    break;
  case Theory::INT_GREATER_EQUAL:
  case Theory::RAT_GREATER_EQUAL:
  case Theory::REAL_GREATER_EQUAL:
    name="$greatereq";
    break;
  case Theory::INT_LESS:
  case Theory::RAT_LESS:
  case Theory::REAL_LESS:
    name="$less";
    break;
  case Theory::INT_LESS_EQUAL:
  case Theory::RAT_LESS_EQUAL:
  case Theory::REAL_LESS_EQUAL:
    name="$lesseq";
    break;
  case Theory::INT_IS_INT:
  case Theory::RAT_IS_INT:
  case Theory::REAL_IS_INT:
    name="$is_int";
    break;
  case Theory::INT_IS_RAT:
  case Theory::RAT_IS_RAT:
  case Theory::REAL_IS_RAT:
    name="$is_rat";
    break;
  case Theory::INT_IS_REAL:
  case Theory::RAT_IS_REAL:
  case Theory::REAL_IS_REAL:
    name="$is_real";
    break;
  case Theory::INT_TO_INT:
  case Theory::RAT_TO_INT:
  case Theory::REAL_TO_INT:
    name="$to_int";
    break;
  case Theory::INT_TO_RAT:
  case Theory::RAT_TO_RAT:
  case Theory::REAL_TO_RAT:
    name="$to_rat";
    break;
  case Theory::INT_TO_REAL:
  case Theory::RAT_TO_REAL:
  case Theory::REAL_TO_REAL:
    name="$to_real";
    break;
  default:
    ASSERTION_VIOLATION;
  }

  unsigned arity = Theory::getArity(interp);

  if(Theory::isFunction(interp)) {
    if(functionExists(name, arity)) {
      int i=0;
      while(functionExists(name+Int::toString(i), arity)) {
        i++;
      }
      name=name+Int::toString(i);
    }
    addInterpretedFunction(interp, name);
  }
  else {
    if(predicateExists(name, arity)) {
      int i=0;
      while(predicateExists(name+Int::toString(i), arity)) {
        i++;
      }
      name=name+Int::toString(i);
    }
    addInterpretedPredicate(interp, name);
  }

  
  return _iSymbols.get(interp);
}

bool Signature::functionExists(const string& name,unsigned arity) const
{
  CALL("Signature::functionExists");

  return _funNames.find(key(name, arity));
}

bool Signature::predicateExists(const string& name,unsigned arity) const
{
  CALL("Signature::predicateExists");

  return _predNames.find(key(name, arity));
}


unsigned Signature::addFunction (const string& name,
				 unsigned arity,
				 bool& added)
{
  CALL("Signature::addFunction");

  string symbolKey = key(name,arity);
#if VDEBUG
  unsigned result = 0;
#else
  unsigned result;
#endif
  if (_funNames.find(symbolKey,result)) {
    added = false;
    return result;
  }
  if (env.options->arityCheck()) {
    unsigned prev;
    if (_arityCheck.find(name,prev)) {
      unsigned prevArity = prev/2;
      bool isFun = prev % 2;
      USER_ERROR((string)"Symbol " + name +
		 " is used both as a function of arity " + Int::toString(arity) +
		 " and a " + (isFun ? "function" : "predicate") +
		 " of arity " + Int::toString(prevArity));
    }
    _arityCheck.insert(name,2*arity+1);
  }

  result = _funs.length();
  _funs.push(new Symbol(name,arity));
  _funNames.insert(symbolKey,result);
  added = true;
  return result;
} 

unsigned Signature::addStringConstant(const string& name)
{
  CALL("Signature::addStringConstant");

  string symbolKey = name + "_c";
#if VDEBUG
  unsigned result = 0;
#else
  unsigned result;
#endif
  if (_funNames.find(symbolKey,result)) {
    return result;
  }

  _strings++;
  string quotedName = "\"" + name + "\"";
  result = _funs.length();
  Symbol* sym = new Symbol(quotedName,0,false,true);
  sym->addToDistinctGroup(STRING_DISTINCT_GROUP);
  _funs.push(sym);
  _funNames.insert(symbolKey,result);
  return result;
}

unsigned Signature::addPredicate (const string& name,
				  unsigned arity,
				  bool& added)
{
  CALL("Signature::addPredicate");

  string symbolKey = key(name,arity);
#if VDEBUG
  unsigned result = 0;
#else
  unsigned result;
#endif
  if (_predNames.find(symbolKey,result)) {
    added = false;
    return result;
  }
  if (env.options->arityCheck()) {
    unsigned prev;
    if (_arityCheck.find(name,prev)) {
      unsigned prevArity = prev/2;
      bool isFun = prev % 2;
      USER_ERROR((string)"Symbol " + name +
		 " is used both as a predicate of arity " + Int::toString(arity) +
		 " and a " + (isFun ? "function" : "predicate") +
		 " of arity " + Int::toString(prevArity));
    }
    _arityCheck.insert(name,2*arity);
  }

  result = _preds.length();
  _preds.push(new Symbol(name,arity));
  _predNames.insert(symbolKey,result);
  added = true;
  return result;
} 

unsigned Signature::addNamePredicate(unsigned arity)
{
  CALL("Signature::addNamePredicate");
  return addFreshPredicate(arity,"sP");
} 

unsigned Signature::addFreshFunction(unsigned arity, const char* prefix, const char* suffix)
{
  CALL("Signature::addFreshFunction");

  string pref(prefix);
  string suf(suffix ? string("_")+suffix : "");
  bool added;
  unsigned result;
  
  
  


    do {
      result = addFunction(pref+Int::toString(_nextFreshSymbolNumber++)+suf,arity,added);
    }
    while (!added);

  Symbol* sym = getFunction(result);
  sym->markIntroduced();
  sym->markSkip();
  return result;
} 

unsigned Signature::addFreshPredicate(unsigned arity, const char* prefix, const char* suffix)
{
  CALL("Signature::addFreshPredicate");

  string pref(prefix);
  string suf(suffix ? string("_")+suffix : "");
  bool added = false;
  unsigned result;
  
  
  




    do {
      result = addPredicate(pref+Int::toString(_nextFreshSymbolNumber++)+suf,arity,added);
    }
    while (!added);

  Symbol* sym = getPredicate(result);
  sym->markIntroduced();
  sym->markSkip();
  return result;
} 

unsigned Signature::addSkolemFunction (unsigned arity, const char* suffix)
{
  CALL("Signature::addSkolemFunction");

  return addFreshFunction(arity, "sK", suffix);
} 

unsigned Signature::addIteFunction(unsigned arity, unsigned* argSorts, unsigned resSort)
{
  CALL("Signature::addIteFunction");

  unsigned res = addFreshFunction(arity, "sG");

  BaseType* type = BaseType::makeType(arity, argSorts, resSort);
  getFunction(res)->setType(type);

  return res;
}

string Signature::key(const string& name,int arity)
{
  return name + '_' + Int::toString(arity);
} 


void Signature::Symbol::addColor(Color color)
{
  ASS_L(color,3);
  ASS_G(color,0);
  ASS(env.colorUsed);

  if (_color && color != static_cast<Color>(_color)) {
    USER_ERROR("A symbol cannot have two colors");
  }
  _color = color;
} 

unsigned Signature::createDistinctGroup(Unit* premise)
{
  CALL("Signature::createDistinctGroup");

  unsigned res = _distinctGroupPremises.size();
  _distinctGroupPremises.push(premise);
  return res;
}

Unit* Signature::getDistinctGroupPremise(unsigned group)
{
  CALL("Signature::getDistinctGroupPremise");

  return _distinctGroupPremises[group];
}

void Signature::addToDistinctGroup(unsigned constantSymbol, unsigned groupId)
{
  CALL("Signature::addToDistinctGroup");

  Symbol* sym = getFunction(constantSymbol);
  sym->addToDistinctGroup(groupId);
}

bool Signature::isProtectedName(string name)
{
  CALL("Signature::isProtectedName");

  if(name=="$distinct") {
    
    return true;
  }

  string protectedPrefix = env.options->protectedPrefix();
  if(protectedPrefix.size()==0) {
    return false;
  }
  if(name.substr(0, protectedPrefix.size())==protectedPrefix) {
    return true;
  }
  return false;
}

bool Signature::symbolNeedsQuoting(string name, bool interpreted, unsigned arity, bool stringConstant)
{
  CALL("Signature::symbolNeedsQuoting");
  ASS_G(name.length(),0);

  if(stringConstant) { return false; }
  if(interpreted && (name=="=" || arity==0)) { return false; }

  const char* c = name.c_str();
  bool quote = false;
  bool first = true;
  if(*c=='$') {
    if(*(c+1)=='$') {
      c+=2; 
      first = false;
    }
    else if(interpreted) {
      c++; 
      first = false;
    }
  }
  while(!quote && *c) {
    quote |= charNeedsQuoting(*c, first);
    first = false;
    c++;
  }
  if(!quote) { return false; }
  if(name=="$distinct") {
    
    return false;
  }
  return true;
}

bool Signature::charNeedsQuoting(char c, bool first)
{
  switch (c) {
  case 'a':
  case 'b':
  case 'c':
  case 'd':
  case 'e':
  case 'f':
  case 'g':
  case 'h':
  case 'i':
  case 'j':
  case 'k':
  case 'l':
  case 'm':
  case 'n':
  case 'o':
  case 'p':
  case 'q':
  case 'r':
  case 's':
  case 't':
  case 'u':
  case 'v':
  case 'w':
  case 'x':
  case 'y':
  case 'z':

    return false;
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
  case 'G':
  case 'H':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
  case 'M':
  case 'N':
  case 'O':
  case 'P':
  case 'Q':
  case 'R':
  case 'S':
  case 'T':
  case 'U':
  case 'V':
  case 'W':
  case 'X':
  case 'Y':
  case 'Z':
  case '_':
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    return first;
  default:
    return true;
  }
}
