
/*
 * File SubformulaIterator.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "SubformulaIterator.hpp"

namespace Kernel {

class SubformulaIterator::Element {
public:
  Element (FormulaList* list, int polarity, Element* rest)
    : _isList(true),
      _list(list),
      _polarity(polarity),
      _rest(rest)
  {}
  Element (Formula* f, int polarity, Element* rest)
    : _isList(false),
      _formula(f),
      _polarity(polarity),
      _rest(rest)
  {}
  bool _isList;
  union{
    FormulaList* _list;
    Formula* _formula;
  };
  int _polarity;
  Element* _rest;

  CLASS_NAME(SubformulaIterator::Element);
  USE_ALLOCATOR(SubformulaIterator::Element);
};


SubformulaIterator::SubformulaIterator (Formula* f)
  : _current(f),
    _currentPolarity(1),
    _reserve(0)
{
} 


SubformulaIterator::SubformulaIterator (FormulaList* ts)
  : _current(0),
    _reserve(new Element(ts,1,0))
{
} 


bool SubformulaIterator::hasNext ()
{
  CALL("SubformulaIterator::hasNext");

  if (_current) {
    return true;
  }

  
  while (_reserve) {
    if (_reserve->_isList) {
      FormulaList* first = _reserve->_list;
      if (first->isEmpty()) {
	Element* rest = _reserve->_rest;
	delete _reserve;
	_reserve = rest;
      }
      else { 
	_current = first->head();
	_currentPolarity = _reserve->_polarity;
	_reserve->_list = first->tail();
	return true;
      }
    }
    else { 
      _current = _reserve->_formula;
      _currentPolarity = _reserve->_polarity;
      Element* rest = _reserve->_rest;
      delete _reserve;
      _reserve = rest;
      return true;
    }
  }
  
  return false;
} 


Formula* SubformulaIterator::next ()
{
  CALL("SubformulaIterator::next/0");

  int dummy;
  return next(dummy);
}

Formula* SubformulaIterator::next (int& resultPolarity)
{
  CALL("SubformulaIterator::next/1");

  Formula* result = _current;
  resultPolarity = _currentPolarity;

  switch (result->connective()) {
  case LITERAL:
  case TRUE:
  case FALSE:
    _current = 0;
    break;

  case AND:
  case OR:
    _reserve = new Element(result->args(), resultPolarity, _reserve);
    _current = 0;
    break;

  case IMP:
    _current = result->left();
    _currentPolarity = -resultPolarity;
    _reserve = new Element(result->right(), resultPolarity, _reserve);
    break;

  case IFF:
  case XOR:
    _current = result->left();
    _currentPolarity = 0;
    _reserve = new Element(result->right(), 0, _reserve);
    break;

  case NOT:
    _current = result->uarg();
    _currentPolarity = -resultPolarity;
    break;

  case FORALL:
  case EXISTS:
    _current = result->qarg();
    _currentPolarity = resultPolarity;
    break;

  case ITE:
    _current = result->condArg();
    _currentPolarity = 0;
    _reserve = new Element(result->elseArg(), resultPolarity, _reserve);
    _reserve = new Element(result->thenArg(), resultPolarity, _reserve);
    break;

  case TERM_LET:
  case FORMULA_LET:
    
    _current = result->letBody();
    _currentPolarity = resultPolarity;
    break;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }

  return result;
} 


SubformulaIterator::~SubformulaIterator ()
{
  while (_reserve) {
    Element* next = _reserve->_rest;
    delete _reserve;
    _reserve = next;
  }
} 

}
