
/*
 * File TermIterators.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __TermIterators__
#define __TermIterators__

#include "Forwards.hpp"

#include "Lib/Recycler.hpp"
#include "Lib/Stack.hpp"
#include "Lib/VirtualIterator.hpp"

#include "Term.hpp"


namespace Kernel {

class VariableIterator
: public IteratorCore<TermList>
{
public:
  DECL_ELEMENT_TYPE(TermList);
  VariableIterator() : _stack(8), _used(false) {}

  VariableIterator(const Term* term) : _stack(8), _used(false)
  {
    if(!term->shared() || !term->ground()) {
      _stack.push(term->args());
    }
  }

  VariableIterator(TermList t) : _stack(8), _used(false)
  {
    if(t.isVar()) {
      _aux[0].makeEmpty();
      _aux[1]=t;
      _stack.push(&_aux[1]);
    }
    else {
      Term* term=t.term();
      if(!term->shared() || !term->ground()) {
	_stack.push(term->args());
      }
    }
  }

  void reset(const Term* term)
  {
    _stack.reset();
    _used = false;
    if(!term->shared() || !term->ground()) {
      _stack.push(term->args());
    }
  }

  void reset(TermList t)
  {
    _stack.reset();
    _used = false;
    if(t.isVar()) {
      _aux[0].makeEmpty();
      _aux[1]=t;
      _stack.push(&_aux[1]);
    }
    else {
      Term* term=t.term();
      if(!term->shared() || !term->ground()) {
	_stack.push(term->args());
      }
    }
  }


  bool hasNext();
 
  TermList next()
  {
    ASS(!_used);
    ASS(_stack.top()->isVar());
    _used=true;
    return *_stack.top();
  }
private:
  Stack<const TermList*> _stack;
  bool _used;
  TermList _aux[2];
};

struct VariableIteratorFn
{
  DECL_RETURN_TYPE(VirtualIterator<TermList>);
  VirtualIterator<TermList> operator()(Term* t)
  {
    return vi( new VariableIterator(t) );
  }
  VirtualIterator<TermList> operator()(TermList t)
  {
    if(t.isVar()) {
      return pvi( getSingletonIterator(t) );
    }
    else {
      return (*this)(t.term());
    }
  }
};

struct OrdVarNumberExtractorFn
{
  DECL_RETURN_TYPE(unsigned);
  unsigned operator()(TermList t)
  {
    CALL("OrdVarNumberExtractorFn::operator()");
    ASS(t.isOrdinaryVar());

    return t.var();
  }
};

class SubtermIterator
: public IteratorCore<TermList>
{
public:
  SubtermIterator(const Term* term) : _used(false)
  {
    Recycler::get(_stack);
    _stack->reset();

    pushNext(term->args());
  }
  ~SubtermIterator()
  {
    Recycler::release(_stack);
  }

  bool hasNext();
 
  TermList next()
  {
    ASS(!_used && !_stack->isEmpty());
    _used=true;
    return *_stack->top();
  }

 
  void right();
protected:
  SubtermIterator() : _used(false)
  {
    Recycler::get(_stack);
    _stack->reset();
  }

  inline
  void pushNext(const TermList* t)
  {
    if(!t->isEmpty()) {
      _stack->push(t);
    }
  }

  Stack<const TermList*>* _stack;
  bool _used;
};

class ReversedCommutativeSubtermIterator
: public SubtermIterator
{
public:
  ReversedCommutativeSubtermIterator(const Term* trm)
  {
    CALL("Term::ReversedCommutativeSubtermIterator::ReversedCommutativeSubtermIterator");
    ASS(trm->commutative());
    ASS_EQ(trm->arity(),2);

    aux[0]=*trm->nthArgument(1);
    aux[1].makeEmpty();
    aux[2]=*trm->nthArgument(0);
    aux[3].makeEmpty();

    _stack->push(&aux[0]);
    _stack->push(&aux[2]);
  }
private:
  TermList aux[4];
};

class PolishSubtermIterator
: public IteratorCore<TermList>
{
public:
  PolishSubtermIterator(const Term* term) : _stack(8), _used(false)
  {
    pushNext(term->args());
  }

  bool hasNext();
 
  TermList next()
  {
    ASS(!_used && !_stack.isEmpty());
    _used=true;
    return *_stack.top();
  }
private:
  inline
  void pushNext(const TermList* t)
  {
    while(!t->isEmpty()) {
      _stack.push(t);
      if(!t->isTerm()) {
	return;
      }
      t=t->term()->args();
    }
  }
  Stack<const TermList*> _stack;
  bool _used;
};

class NonVariableIterator
: public IteratorCore<TermList>
{
public:
  NonVariableIterator(const Term* term) : _stack(8), _used(false)
  {
    pushNextNonVar(term->args());
  }

  bool hasNext();
 
  TermList next()
  {
    ASS(!_used && !_stack.top()->isVar());
    _used=true;
    return *_stack.top();
  }

  void right();
private:
  void pushNextNonVar(const TermList* t);

  Stack<const TermList*> _stack;
  bool _used;
};

class DisagreementSetIterator
: public IteratorCore<pair<TermList, TermList> >
{
public:
 
  DisagreementSetIterator()
  {
    _arg1.makeEmpty();
  }

 
  DisagreementSetIterator(TermList t1, TermList t2, bool disjunctVariables=true)
  : _stack(8)
  {
    CALL("Term::DisagreementSetIterator::DisagreementSetIterator(TermList...)");
    reset(t1, t2, disjunctVariables);
  }
 
  DisagreementSetIterator(Term* t1, Term* t2, bool disjunctVariables=true)
  : _stack(8), _disjunctVariables(disjunctVariables)
  {
    CALL("Term::DisagreementSetIterator::DisagreementSetIterator(Term*...)");
    reset(t1,t2,disjunctVariables);
  }

  void reset(TermList t1, TermList t2, bool disjunctVariables=true)
  {
    CALL("Term::DisagreementSetIterator::reset(TermList...)");
    ASS(!t1.isEmpty());
    ASS(!t2.isEmpty());

    _stack.reset();
    _disjunctVariables=disjunctVariables;
    if(!TermList::sameTop(t1,t2) || (t1.isVar() && disjunctVariables)) {
      _arg1=t1;
      _arg2=t2;
      return;
    }
    _arg1.makeEmpty();
    if(t1.isTerm() && t1.term()->arity()>0) {
      _stack.push(t1.term()->args());
      _stack.push(t2.term()->args());
    }
  }

  void reset(Term* t1, Term* t2, bool disjunctVariables=true)
  {
    CALL("Term::DisagreementSetIterator::reset(Term*...)");
    ASS_EQ(t1->functor(), t2->functor());

    _arg1.makeEmpty();

    if(t1->arity()>0) {
      _stack.push(t1->args());
      _stack.push(t2->args());
    }
  }

  bool hasNext();

 
  pair<TermList, TermList> next()
  {
    pair<TermList, TermList> res(_arg1,_arg2);
    _arg1.makeEmpty();
    return res;
  }
private:
  Stack<TermList*> _stack;
  bool _disjunctVariables;
  TermList _arg1;
  TermList _arg2;
};



class TermFunIterator
: public IteratorCore<unsigned>
{
public:
  TermFunIterator (const Term*);

  bool hasNext();
  unsigned next();
private:
 
  bool _hasNext;
 
  unsigned _next;
 
  Stack<const TermList*> _stack;
}; 


class TermVarIterator
: public IteratorCore<unsigned>
{
public:
  TermVarIterator (const Term*);
  TermVarIterator (const TermList*);

  bool hasNext ();
  unsigned next ();
private:
 
  bool _hasNext;
 
  unsigned _next;
 
  Stack<const TermList*> _stack;
}; 



}

#endif 
