
/*
 * File BinaryHeap.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __BinaryHeap__
#define __BinaryHeap__

#include <algorithm>
#include <limits>

#include "Debug/Assertion.hpp"
#include "Allocator.hpp"
#include "Exception.hpp"
#include "Comparison.hpp"
#include "Backtrackable.hpp"
#include "Metaiterators.hpp"

namespace Lib {

template <typename T, class Comparator>
class BinaryHeap
{
public:
 
  BinaryHeap()
  : _size(0), _capacity(0), _data(0), _data1(0)
  {
  }

 
  ~BinaryHeap()
  {
    CALL("BinaryHeap::~BinaryHeap");
    if(_data) {
      T* ep=_data+_size;
      while(ep!=_data1) {
	(--ep)->~T();
      }
      DEALLOC_KNOWN(_data,_capacity*sizeof(T),"BinaryHeap::T");
    }
  }

 
  void reset()
  {
    T* ep=_data+_size;
    while(ep!=_data1) {
	(--ep)->~T();
    }
    _size=0;
  }


 
  inline
  unsigned size() const
  {
    ASS(_size>=0);
    return _size;
  }

 
  inline
  bool isEmpty() const
  {
    ASS(_size>=0);
    return _size==0;
  }

 
  inline
  void insert(T obj)
  {
    CALL("BinaryHeap::insert");
    ensureAvaiablePosition();
    _size++;
    new(&_data1[_size]) T(obj);
    bubbleUp(_size);
  }

 
  inline
  const T& top()
  {
    ASS(!isEmpty());
    return _data[0];
  }

 
  inline
  T pop()
  {
    CALL("BinaryHeap::pop");
    ASS(!isEmpty());
    T res=_data[0];
    _size--;
    if(_size) {
      std::swap(_data[0],_data[_size]);
      bubbleDown(1);
    }
    _data[_size].~T();
    return res;
  }

 
  inline
  T popWithAllEqual()
  {
    CALL("BinaryHeap::pop");
    T res=pop();
    while(!isEmpty() && Comparator::compare(res, top())==EQUAL) {
      pop();
    }
    return res;
  }

  T backtrackablePop(unsigned& lastBubbleIndex)
  {
    CALL("BinaryHeap::backtrackablePop(unsigned&)");
    ASS(!isEmpty());
    T res=_data[0];
    _size--;
    if(_size) {
      std::swap(_data[0],_data[_size]);
      lastBubbleIndex=bubbleDown(1);
    } else {
      lastBubbleIndex=1;
    }
    _data[_size].~T();
    return res;
  }
  inline
  T backtrackablePop(BacktrackData& bd)
  {
    CALL("BinaryHeap::backtrackablePop(BacktrackData&)");
    unsigned lastBubbleIndex;
    T res=backtrackablePop(lastBubbleIndex);
    bd.addBacktrackObject(
	new BHPopBacktrackObject(this, res, lastBubbleIndex));
    return res;
  }

  unsigned backtrackableInsert(T obj)
  {
    CALL("BinaryHeap::backtrackableInsert(T,unsigned&)");
    ensureAvaiablePosition();
    _size++;
    new(&_data1[_size]) T(obj);
    return bubbleUp(_size);
  }
  inline
  void backtrackableInsert(T obj, BacktrackData& bd)
  {
    CALL("BinaryHeap::backtrackableInsert(T,BacktrackData&)");
    unsigned lastBubbleIndex=backtrackableInsert(obj);
    bd.addBacktrackObject(
	    new BHInsertBacktrackObject(this, lastBubbleIndex));
  }

  void backtrackInsert(unsigned lastBubbleIndex)
  {
    
    
    
    
    
    
    _data1[lastBubbleIndex]=Comparator::max();
    bubbleDown(lastBubbleIndex);
    ASS(_data1[_size]==Comparator::max());
    _data1[_size].~T();
    _size--;
  }

  void backtrackPop(T val, unsigned lastBubbleIndex)
  {
    
    
    
    
    
    
    _size++;
    new(&_data1[_size]) T(val);
    std::swap(_data1[_size], _data1[lastBubbleIndex]);
    
    
    
    
    
    
    
    bubbleUp(lastBubbleIndex);
  }

 
  struct Iterator
  : public PointerIterator<T>
  {
    Iterator(const BinaryHeap& obj)
    : PointerIterator<T>(obj._data, obj._data+obj._size) {}
  };

private:
  class BHPopBacktrackObject
  : public BacktrackObject
  {
  public:
    BHPopBacktrackObject(BinaryHeap* bh, T v, unsigned lastBubbleIndex)
    :_bh(bh), _val(v), _lastBubbleIndex(lastBubbleIndex) {}
    void backtrack()
    {
      _bh->backtrackPop(_val,_lastBubbleIndex);
    }
    CLASS_NAME(BinaryHeap::BHPopBacktrackObject);
    USE_ALLOCATOR(BHPopBacktrackObject);
  private:
    BinaryHeap* _bh;
    T _val;
    unsigned _lastBubbleIndex;
  };

  class BHInsertBacktrackObject
  : public BacktrackObject
  {
  public:
    BHInsertBacktrackObject(BinaryHeap* bh, unsigned lastBubbleIndex)
    :_bh(bh), _lastBubbleIndex(lastBubbleIndex) {}
    void backtrack()
    {
      _bh->backtrackInsert(_lastBubbleIndex);
    }
    CLASS_NAME(BinaryHeap::BHInsertBacktrackObject);
    USE_ALLOCATOR(BHInsertBacktrackObject);
  private:
    BinaryHeap* _bh;
    unsigned _lastBubbleIndex;
  };

 
  BinaryHeap(const BinaryHeap& obj);
 
  BinaryHeap& operator=(const BinaryHeap& obj);

 
  unsigned bubbleUp(unsigned index)
  {
    CALL("BinaryHeap::bubbleUp");
    ASS(index>0 && index<=_size);
    unsigned nextIndex=index>>1;
    while(nextIndex) {
      if(Comparator::compare(_data1[index], _data1[nextIndex])==LESS) {
	std::swap(_data1[index], _data1[nextIndex]);
      } else {
	return index;
      }
      index=nextIndex;
      nextIndex=index>>1;
    }
    return 1;
  }

 
  unsigned bubbleDown(unsigned index)
  {
    CALL("BinaryHeap::bubbleDown");
    ASS(index>0 && index<=_size);
    unsigned nextIndex=index<<1;
    while(nextIndex<=_size) {
      if(nextIndex!=_size && Comparator::compare(_data1[index], _data1[nextIndex|1])==GREATER) {
	if(Comparator::compare(_data1[nextIndex|1], _data1[nextIndex])==GREATER) {
	  std::swap(_data1[index], _data1[nextIndex]);
	} else {
	  std::swap(_data1[index], _data1[nextIndex|1]);
	  nextIndex|=1;
	}
      } else if(Comparator::compare(_data1[index], _data1[nextIndex])==GREATER) {
	std::swap(_data1[index], _data1[nextIndex]);
      } else {
	return index;
      }
      index=nextIndex;
      nextIndex=index<<1;
    }
    return index;
  }

 
  inline
  void ensureAvaiablePosition()
  {
    ASS(_capacity>=_size);
    if(_capacity==_size)
      expand();
  }

 
  void expand()
  {
    CALL("BinaryHeap::expand");

    ASS(_capacity==_size);

    unsigned oldCapacity=_capacity;
    T* oldData=_data;

    _capacity= _capacity ? _capacity*2 : 4;

    void* mem = ALLOC_KNOWN(_capacity*sizeof(T),"BinaryHeap::T");
    _data = static_cast<T*>(mem);
    _data1 = _data-1;

    if(_size) {
      T* otp = oldData+_size;
      T* ntp = _data+_size;
      do {
	new(--ntp) T(*(--otp));
	
	otp->~T();
      } while(ntp!=_data);
    }

    if(oldData) {
      DEALLOC_KNOWN(oldData,oldCapacity*sizeof(T),"BinaryHeap::T");
    }
  }

 
  unsigned _size;
 
  unsigned _capacity;

 
  T* _data;

 
  T* _data1;

}; 

};

#endif 


