
/*
 * File Hash.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __Hash__
#define __Hash__

#include <string>
#include <utility>

#include "Forwards.hpp"

namespace Lib {

struct HashUtils
{
 
  static unsigned combine(unsigned h1, unsigned h2) { return h1 ^ (h2 + 0x9e3779b9 + (h1 << 6) + (h1 >> 2)); }
};

template<class ElementHash>
struct StackHash {
  template<typename T>
  static unsigned hash(const Stack<T>& s) {
    unsigned res = 2166136261u;
    typename Stack<T>::ConstIterator it(s);
    while(it.hasNext()) {
      res = HashUtils::combine(res, ElementHash::hash(it.next()));
    }
    return res;
  }
};

class Hash
{
public:

 
  template<typename T>
  static bool equals(T o1, T o2)
  { return o1 == o2; }

  static unsigned hash(const char* str);
 
  static unsigned hash(const std::string& str)
  { return hash(str.c_str()); }

  static unsigned hash(Kernel::Unit* u);

  template<typename T>
  static unsigned hash(Stack<T> obj)
  { return StackHash<Hash>::hash(obj); }

  template<typename T>
  static unsigned hash(T obj)
  { return hash(reinterpret_cast<const unsigned char*>(&obj),sizeof(obj)); }

  template<typename T, typename U>
  static unsigned hash(std::pair<T,U> obj)
  {
    unsigned h[2];
    h[0]=hash(obj.first);
    h[1]=hash(obj.second);
    return hash(reinterpret_cast<const unsigned char*>(h), 2*sizeof(unsigned));
  }

  template<typename T>
  static unsigned hash(T obj, unsigned begin)
  { return hash(reinterpret_cast<const unsigned char*>(&obj),sizeof(obj), begin); }

  static unsigned hash(const unsigned char*,size_t length);
  static unsigned hash(const unsigned char*,size_t length,unsigned begin);
};

struct IdentityHash
{
  template<typename T>
  static bool equals(T o1, T o2)
  { return o1 == o2; }

  template<typename T>
  static unsigned hash(T val)
  { return static_cast<unsigned>(val); }
};

struct PtrIdentityHash
{
  static unsigned hash(const void* ptr) {
    return static_cast<unsigned>(reinterpret_cast<size_t>(ptr));
  }
};

struct PtrPairSimpleHash {
  template<typename T>
  static unsigned hash(T pp) {
    return static_cast<unsigned>(reinterpret_cast<size_t>(pp.first)^reinterpret_cast<size_t>(pp.second)^
	    (reinterpret_cast<size_t>(pp.first)>>3)^(reinterpret_cast<size_t>(pp.second)>>4));
  }
};

struct IntPairSimpleHash {
  template<typename T>
  static unsigned hash(T pp) {
    return static_cast<unsigned>(pp.first^pp.second^(pp.first<<1));
  }
};

template<typename T>
struct FirstHashTypeInfo {
  typedef Hash Type;
};

struct GeneralPairSimpleHash {
  template<typename T, typename U>
  static unsigned hash(std::pair<T,U> pp) {
    unsigned h1=FirstHashTypeInfo<T>::Type::hash(pp.first);
    unsigned h2=FirstHashTypeInfo<U>::Type::hash(pp.second);
    return static_cast<unsigned>(h1^h2^(h1<<1));
  }
};


template<typename T>
struct FirstHashTypeInfo<T*> {
  typedef PtrIdentityHash Type;
};

template<>
struct FirstHashTypeInfo<int> {
  typedef IdentityHash Type;
};
template<>
struct FirstHashTypeInfo<unsigned> {
  typedef IdentityHash Type;
};
#if ARCH_X64
template<>
struct FirstHashTypeInfo<size_t> {
  typedef IdentityHash Type;
};
#endif
template<>
struct FirstHashTypeInfo<char> {
  typedef IdentityHash Type;
};

template<>
struct FirstHashTypeInfo<std::pair<int,int> > {
  typedef IntPairSimpleHash Type;
};
template<>
struct FirstHashTypeInfo<std::pair<unsigned,unsigned> > {
  typedef IntPairSimpleHash Type;
};
template<typename T, typename U>
struct FirstHashTypeInfo<std::pair<T,U> > {
  typedef GeneralPairSimpleHash Type;
};

template<typename T>
struct FirstHashTypeInfo<Stack<T> > {
  typedef StackHash< typename FirstHashTypeInfo<T>::Type > Type;
};

}

#endif
