
/*
 * File List.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __list__
#  define __list__


#include "Forwards.hpp"

#include "Debug/Assertion.hpp"

#include "Allocator.hpp"
#include "VirtualIterator.hpp"

#if VDEBUG

#include <ostream>

#endif

namespace Lib {

template <class C>
class List
{
public:
  DECL_ELEMENT_TYPE(C);

 
  explicit inline List (C head)
    :
    _head(head),
    _tail(0)
  {
  } 

 
  inline List (C head, List* tail)
    :
    _head(head),
    _tail(tail)
  {
  } 

 
  inline List () {}

 
  inline static List* empty ()
  { return 0; }

 
  inline bool isEmpty () const
  {
    return this == 0;
  } 

 
  inline bool isNonEmpty () const
  {
    return this != 0;
  } 

 
  inline List* tail() const
  {
    ASS_NEQ(this,0);

    return _tail;
  } 

 
  inline List*& tailReference ()
  {
    ASS_NEQ(this,0);

    return _tail;
  } 

 
  inline List** tailPtr ()
  {
    ASS_NEQ(this,0);

    return &_tail;
  } 

 
  inline const C head () const
  {
    ASS_NEQ(this,0);

    return _head;
  } 

  inline C head ()
  {
    ASS_NEQ(this,0);

    return _head;
  } 

  inline C* headPtr ()
  {
    ASS_NEQ(this,0);
    return &(this->_head);
  }


 
  inline C second () const
  {
    return tail()->head();
  } 

 
  inline void setHead (C hd)
  {
    ASS_NEQ(this,0);

    _head = hd;
  } 


 
  inline void setTail(List* tl)
  {
    ASS_NEQ(this,0);

    _tail = tl;
  } 

 
  void destroy ()
  {
    CALL("List::destroy");

    if (isEmpty()) return;
    List* current = this;

    for (;;) {
      List* next = current->tail();
      delete current;
      if (next->isEmpty()) return;
      current = next;
    }
  } 

 
  void destroyWithDeletion()
  {
    CALL("List::destroyWithDeletion");

    if (isEmpty()) return;
    List* current = this;

    for (;;) {
      List* next = current->tail();
      delete current->head();
      delete current;
      if (next->isEmpty()) return;
      current = next;
    }
  } 

 
  List* copy () const
  {
    if (isEmpty()) return empty();

    List* result = new List;
    result->_head = _head;
    List* previous = result;
    List* rest = _tail;

    while (! rest->isEmpty()) {
      List* tmp = new List;
      tmp->_head = rest->_head;
      previous->_tail = tmp;
      previous = tmp;
      rest = rest->_tail;
    }

    previous->_tail = empty();
    return result;
  }  

 
  List* append (List* snd)
  {
    if (isEmpty()) return snd;

    List* result = new List;
    result->setHead(head());
    List* previous = result;
    List* rest = tail();

    while(rest->isNonEmpty()) {
      List* tmp = new List;
      tmp->setHead(rest->head());
      previous->setTail(tmp);
      previous = tmp;
      rest = rest->tail();
    }

    previous->setTail(snd);
    return result;
  }  

 
  inline List* cons(C elem)
  {
    return new List(elem,this);
  } 

 
  inline static void push(C elem,List* &lst)
  {
    lst = lst->cons(elem);
  } 

 
  template<class It>
  static void pushFromIterator(It it, List* &lst)
  {
    while(it.hasNext()) {
      push(it.next(), lst);
    }
  }

 
  bool hasLength(int length) const
  {
    ASS_GE(length,0);

    for (const List* lst = this;lst;lst = lst->tail()) {
      if (length == 0) return false;
      length--;
    }
    return length == 0;
  } 

 
  inline static C pop(List* &lst)
  {
    CALL("List::pop");
    ASS_NEQ(lst,0);

    List* tail = lst->tail();
    C result = lst->head();
    delete lst;
    lst = tail;

    return result;
  } 

 
  static List* concat(List* first,List* second)
  {
    if (first == 0) return second;
    if (second == 0) return first;

    List* current = first;
    for (;;) {
      List* next = current->tail();
      if (! next) {
        current->setTail(second);
        return first;
      }
      current = next;
    }
  } 

 
  List* reverse()
  {
    if (isEmpty()) return this;

    List* result = empty();
    List* lst = this;

    while (lst->isNonEmpty()) {
      List* tl = lst->tail();
      lst->setTail(result);
      result = lst;
      lst = tl;
    }

    return result;
  } 

 
  int length () const
  {
    int len = 0;

    for (const List* lst = this; lst->isNonEmpty(); lst = lst->tail()) {
      len ++;
    }

    return len;
  } 

 
  bool member (C elem) const
  {
    for (const List* lst = this; lst->isNonEmpty(); lst = lst->tail()) {
      if (lst->head() == elem) return true;
    }

    return false;
  } 

 
  List* remove (C elem)
  {
    if (isEmpty()) return this;

    if (head() == elem) {
      List* result = tail();
      delete this;
      return result;
    }
    if (tail()->isEmpty()) return this;

    List* current = this;
    List* next = tail();

    for (;;) {
      if (next->head() == elem) { 
        current->setTail(next->tail());
        delete next;
        return this;
      }
      current = next;
      next = next->tail();
      if (next->isEmpty()) return this;
    }
  } 

 
  C nth(int n) const
  {
    
    ASS_GE(n,0);

    const List* l = this;

    while (n != 0) {
      ASS_NEQ(l,0);

      l = l->tail();
      n--;
    }

    return l->head();
  } 

 
  static C deleteNth(List*& lst, int n)
  {
    
    ASS (n >= 0);

    C result;
    List* l = lst;
    ASS (lst->isNonEmpty());

    if (n == 0) {
      result = l->head();
      lst = l->tail();
      delete l;
      return result;
    }

    
    List* next = l->tail();

    while (--n != 0) {
      l = next;
      next = next->tail();
      ASS (next->isNonEmpty());
    }
    
    result = next->head();
    l->setTail(next->tail());
    delete next;

    return result;
  } 

 
  List* addLast (C elem)
  {
    if (! this) return new List (elem);

    
    List* current;
    for (current = this; current->_tail; current = current->_tail) {
    }

    current->setTail(new List(elem));

    return this;
  } 

 
  List* split (int n, List*& rest)
  {
    if (! this) {
      ASS_EQ(n,0);
      rest = empty();
      return empty();
    }

    if (n == 0) {
      rest = empty();
      return this;
    }

    List* nth = this;
    while (--n > 0) {
      ASS_NEQ(nth,0);
      nth = nth->_tail;
    }

    ASS_NEQ(nth,0);
    rest = nth->_tail;
    nth->_tail = empty();
    return this;
  } 

 
  class Iterator {
  public:
    DECL_ELEMENT_TYPE(C);

    inline Iterator() : _lst (0) {}

    inline explicit
    Iterator(List* l)
      : _lst (l)
    {}
    inline explicit
    Iterator(const List* l)
      : _lst (const_cast<List*>(l))
    {}

   
    inline C next()
    {
      ASS_NEQ(_lst,0);

      C result = _lst->head();
      _lst = _lst->tail();
      return result;
    }

   
    inline C peekAtNext()
    {
      return _lst->head();
    }

   
    inline bool hasNext() const
    {
      return _lst->isNonEmpty();
    }

    inline void reset(List* l) { _lst = l; }

   private:
   
    List* _lst;
  };

 
  class RefIterator {
   public:
     DECL_ELEMENT_TYPE(C&);

    inline explicit
    RefIterator(List* l)
      : _lst (l)
    {}
    inline explicit
    RefIterator(const List* l)
      : _lst (const_cast<List*>(l))
    {}

   
    inline C& next()
    {
      ASS_NEQ(_lst,0);

      C& result = _lst->_head;
      _lst = _lst->tail();
      return result;
    }

   
    inline bool hasNext() const
    { return _lst->isNonEmpty(); }

    inline void reset(List* l) { _lst = l; }

   private:
   
    List* _lst;
  };

  class PtrIterator
  {
  public:
    DECL_ELEMENT_TYPE(C*);
    inline
    PtrIterator(List* lst) : _l(lst) {}
    inline bool hasNext()
    { return _l->isNonEmpty(); }

    inline C* next()
    {
      C* res=_l->headPtr();
      _l=_l->tail();
      return res;
    }
  protected:
    List* _l;
  };


 
  class DelIterator {
   public:

    inline DelIterator (List*& l)
      :
      _lst(l),
      _prev(0),
      _cur(0)
    {}

   
    inline void reset()
    {
      _prev = 0;
      _cur = 0;
    } 

   
    inline C next()
    {
      if (_cur) { 
	_prev = _cur;
	_cur = _cur->tail();
	ASS_NEQ(_cur,0);
      }
      else { 
	_cur = _lst;
      }
      return _cur->head();
    }

   
    inline bool hasNext()
    {
      if (_cur) { 
	return _cur->tail() != 0;
      }
      return _lst->isNonEmpty();
    }

   
    void del()
    {
      
      ASS_NEQ(_cur,0);
      
      ASS_NEQ(_cur,_prev);

      if (_cur == _lst) { 
	_lst = _lst->tail();
	delete _cur;
	_cur = 0;
	return;
      }

      
      _prev->setTail(_cur->tail());
      delete _cur;
      _cur = _prev;
    } 

   
    inline void replace(C elem)
    {
      ASS_NEQ(_cur,0);
      _cur->setHead(elem);
    } 

   
    void insert (List* lst)
    {
      ASS_NEQ(_cur, _prev);
      if (! lst) return;

      List* last = lst;
      
      while (last->tail()) {
	last = last->tail();
      }

      if (_prev)
	_prev->setTail(lst);
      else _lst = lst;

      last->setTail(_cur);
      _prev = last;
    } 

   
    void insert (C elem)
    {
      ASS_NEQ(_cur, _prev);
      List* lst = new List(elem,_cur);
      if (_prev)
	_prev->setTail(lst);
      else _lst = lst;

      lst->setTail(_cur);
      _prev = lst;
    } 

  private:
   
    List*& _lst;
   
    List* _prev;
   
    List* _cur;
  };

 
  class DestructiveIterator {
  public:
    DECL_ELEMENT_TYPE(C);

    inline explicit
    DestructiveIterator(List* l)
      : _lst (l)
    {}

   
    inline C next()
    {
      ASS_NEQ(_lst,0);
      return List::pop(_lst);
    }

   
    inline bool hasNext() const
    {
      return _lst->isNonEmpty();
    }
   private:
   
    List* _lst;
  };

  
  CLASS_NAME(List);
  USE_ALLOCATOR(List);

 
  class FIFO {
  public:
   
    inline explicit FIFO(List* &lst)
      : _last(0),
	_initial(lst)
    {
      ASS_EQ(_initial,0);
    }
   
    inline void push(C elem)
    {
      List* newLast = new List(elem);
      if (_last) 
	_last->setTail(newLast);
      else _initial = newLast;

      _last = newLast;
    } 

  private:
   
    List* _last;
   
    List* &_initial;
  }; 

protected:  

 
  C _head;
 
  List* _tail;
};  




template<typename T>
typename List<T>::Iterator getContentIterator(List<T>* lst)
{
  return typename List<T>::Iterator(lst);
}







template<typename T>
struct ElementTypeInfo<List<T>* >
{
  typedef T Type;
};

template<typename T>
struct IteratorTypeInfo<List<T>* >
{
  typedef typename List<T>::Iterator Type;
};
template<typename T>
struct IteratorTypeInfo<const List<T>*>
{
  typedef typename List<T>::Iterator Type;
};



#if VDEBUG

template<typename T>
std::ostream& operator<< (ostream& out, const List<T>& lstr )
{
  const List<T>* lst=&lstr;
  out<<'[';

  while(lst) {
    out<<lst->head();
    lst=lst->tail();
    if (lst) out << ",\n";
  }

  return out<<']';
}

template<typename T>
std::ostream& operator<< (ostream& out, const List<T*>& lstr )
{
  const List<T*>* lst=&lstr;
  out<<'[';

  while(lst) {
    out<<(*lst->head());
    lst=lst->tail();
    if (lst) out << ",\n";
  }

  return out<<']';
}

#endif

}

#endif


