
/*
 * File MemoryLeak.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#if CHECK_LEAKS

#include "Hash.hpp"
#include "Stack.hpp"
#include "MemoryLeak.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"

using namespace Kernel;

bool MemoryLeak::_cancelled = false;

void MemoryLeak::release(Formula* f)
{
  CALL("MemoryLeak::release(Formula*)");

  if (_released.contains(f)) {
    return;
  }
  _released.insert(f);
  switch (f->connective()) {
  case LITERAL:
  case TRUE:
  case FALSE:
    break;
  case AND:
  case OR:
    {
      FormulaList* fs = f->args();
      while (fs) {
	if (_released.contains(fs)) {
	  break;
	}
	_released.insert(fs);
	FormulaList* next = fs->tail();
	release(fs->head());
	delete fs;
	fs = next;
      }
    }
    break;
  case IMP:
  case IFF:
  case XOR:
    release(f->left());
    release(f->right());
    break;
  case NOT:
    release(f->uarg());
    break;
  case FORALL:
  case EXISTS:
    {
      Formula::VarList* vs = f->vars();
      while (vs) {
	if (_released.contains(vs)) {
	  break;
	}
	_released.insert(vs);
	Formula::VarList* next = vs->tail();
	delete vs;
	vs = next;
      }
    }
    release(f->qarg());
    break;
#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
  f->destroy();
} 


void MemoryLeak::release(UnitList* units)
{
  Stack<Unit*> us(32);
  while (units) {
    UnitList* next = units->tail();
    Unit* u = units->head();
    if (_released.contains(u)) {
      continue;
    }
    _released.insert(u);
    us.push(u);
    delete units;
    units = next;
  }

  while (us.isNonEmpty()) {
    Unit* u = us.pop();
    Inference* inf = u->inference();
    Inference::Iterator ps = inf->iterator();
    while (inf->hasNext(ps)) {
      Unit* p = inf->next(ps);
      if (_released.contains(p)) {
	continue;
      }
      _released.insert(p);
      us.push(p);
    }

    if (! u->isClause()) {
      release(static_cast<FormulaUnit*>(u)->formula());
      u->destroy();
    }
  }
} 

#endif 
