
/*
 * File SCCAnalyzer.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __SCCAnalyzer__
#define __SCCAnalyzer__

#include "Forwards.hpp"

#include "DHMap.hpp"
#include "DHSet.hpp"
#include "Metaiterators.hpp"
#include "VirtualIterator.hpp"
#include "Stack.hpp"

namespace Lib {

template<typename T>
class MapToLIFOGraph
{
public:
  typedef MapToLIFO<T,T> Map;
  typedef T Node;
  typedef VirtualIterator<T> NodeIterator;

  MapToLIFOGraph(const Map& m) : _m(m)
  {
    CALL("MapToLIFOGraph::MapToLIFOGraph");

    typename Map::KeyIterator keys(_m);
    while(keys.hasNext()) {
      T n0 = keys.next();
      _nodes.push(n0);
      NodeIterator neighs = neighbors(n0);
      while(neighs.hasNext()) {
	_nodes.push(neighs.next());
      }
    }
    makeUnique(_nodes);
  }

  NodeIterator nodes() const
  {
    return pvi( typename Stack<T>::ConstIterator(_nodes) );
  }

  NodeIterator neighbors(T node) const
  {
    return pvi( _m.keyIterator(node) );
  }

 
  template<class It>
  void ensureNodes(It nodeIt) {
    _nodes.loadFromIterator(nodeIt);
    makeUnique(_nodes);
  }

  size_t size() const { return _nodes.size(); }

private:
  Stack<T> _nodes;
  const Map& _m;
};

template<typename Graph>
class Subgraph
{
public:
  typedef typename Graph::Node Node;


  struct Filter
  {
    Filter(DHSet<Node>& forbidden) : _forbidden(forbidden) {}
    bool operator()(Node n) const { return !_forbidden.find(n); }
  private:
    DHSet<Node>& _forbidden;
  };

  typedef FilteredIterator<typename Graph::NodeIterator,Filter> NodeIterator;

  template<class It>
  Subgraph(Graph& gr, It forbiddenNodes)
   : _filter(_forbidden), _gr(gr)
  {
    CALL("Subgraph::Subgraph");

    _forbidden.loadFromIterator(forbiddenNodes);

    _size = 0;
    NodeIterator nit = nodes();
    while(nit.hasNext()) {
      nit.next();
      _size++;
    }
  }

  NodeIterator nodes() const
  {
    return NodeIterator(_gr.nodes(), _filter);
  }

  NodeIterator neighbors(Node node) const
  {
    return NodeIterator(_gr.neighbors(node), _filter);
  }

  size_t size() const { return _size; }

private:
  DHSet<Node> _forbidden;
  Filter _filter;
  const Graph& _gr;
  size_t _size;
};

template<class Graph>
class SCCAnalyzer {
public:
  typedef typename Graph::Node Node;
  typedef Stack<Node> NodeStack;

  SCCAnalyzer(const Graph& gr)
  {
    CALL("SCCAnalyzer::SCCAnalyzer");
    analyze(gr);
  }


  const NodeStack& breakingNodes() const { return _breakingNodes; }
 
  const Stack<NodeStack>& components() const { return _components; }
private:

  typedef typename Graph::NodeIterator NodeIterator;

  
  

  size_t C;
  DHMap<Node,size_t> nums;
 
  DHMap<Node,size_t> scc;
  NodeStack S;
  NodeStack P;

  void analyze(const Graph& gr)
  {
    CALL("SCCAnalyzer::analyze");

    ASS_EQ(nums.size(),0);
    C = 0;
    NodeIterator nit = gr.nodes();
    while(nit.hasNext()) {
      Node v = nit.next();
      if(nums.find(v)) {
	continue;
      }
      traverse(v, gr);
    }
    makeUnique(_breakingNodes);
  }
  void traverse(Node v, const Graph& gr)
  {
    CALL("SCCAnalyzer::traverse");

    ALWAYS(nums.insert(v, C));
    C++;
    S.push(v);
    P.push(v);
    NodeIterator neigh = gr.neighbors(v);
    while(neigh.hasNext()) {
      Node w = neigh.next();
      size_t wNum;
      if(!nums.find(w,wNum)) {
	traverse(w, gr);
	continue;
      }
      if(scc.find(w)) { 
	continue;
      }
      
      _breakingNodes.push(w);
      while(nums.get(P.top())>wNum) {
	P.pop();
      }
    }
    if(v==P.top()) {
      size_t sccIndex = _components.size();
      
      
      _components.push(Stack<Node>());
      NodeStack& compNodes = _components[sccIndex];
      Node cNode;
      do {
	cNode = S.pop();
	compNodes.push(cNode);
	scc.insert(cNode, sccIndex);
      } while(cNode!=v);
      P.pop();
    }
  }

  NodeStack _breakingNodes;
  Stack<NodeStack> _components;
};

}

#endif 
