
/*
 * File SkipList.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#ifndef __SkipList__
#  define __SkipList__

#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"

#include "Allocator.hpp"
#include "Backtrackable.hpp"
#include "Comparison.hpp"
#include "List.hpp"
#include "Random.hpp"

#define SKIP_LIST_MAX_HEIGHT 32

namespace Lib
{

template <typename Value,class ValueComparator>
class SkipList
{
public:
  class Node {
  public:
    Value value;
    Node* nodes[1];
  };
 
  inline
  void insert(Value val)
  {
    CALL("SkipList::insert");
    Value* pval = insertPosition(val);
    *pval = val;
  } 

  template<class Iterator>
  inline
  void insertFromIterator(Iterator it)
  {
    CALL("SkipList::insertFromIterator");

    while(it.hasNext()) {
      insert(it.next());
    }
  }

 
  inline
  bool ensurePresent(Value val)
  {
    CALL("SkipList::ensurePresent");
    Value* pval;
    if(!getPosition(val, pval, true)) {
      *pval = val;
      return false;
    }
    return true;
  }

 
  template<typename Key>
  bool getPosition(Key key, Value*& pvalue, bool canCreate)
  {
    CALL("SkipList::getPosition");

    if(_top==0) {
      if(canCreate) {
	pvalue = insertPosition(key);
      }
      return false;
    }

    unsigned h = _top-1;

   
    
    
    Node* left = _left;
    for (;;) {
      Node* next = left->nodes[h];
      if (next == 0) {
	if (h == 0) {
	    if(canCreate) {
	      pvalue=insertPosition(key);
	    }
	    return false;
	}
	h--;
	continue;
      }
      
      switch (ValueComparator::compare(key,next->value))
	{
	case LESS:
	  
	  if (h == 0) {
	    if(canCreate) {
	      pvalue=insertPosition(key);
	    }
	    return false;
	  }
	  h--;
	  break;

	case EQUAL:
	  pvalue=&next->value;
	  return true;

	case GREATER:
	  left = next;
	  break;

#if VDEBUG
	default:
	  ASSERTION_VIOLATION;
#endif
	}
    }
  } 

 
  template<typename Key>
  Value* insertPosition(Key key)
  {
    CALL("SkipList::insertPosition");

    
    unsigned nodeHeight = 0;
    while (Random::getBit()) {
      nodeHeight++;
    }
    if (nodeHeight >= _top) {
      if (_top < SKIP_LIST_MAX_HEIGHT) {
	nodeHeight = _top;
	_top++;
      }
      else {
	ASS(_top == SKIP_LIST_MAX_HEIGHT);
	nodeHeight = _top - 1;
      }
    }
    Node* newNode = allocate(nodeHeight);
    new(&newNode->value) Value();


    unsigned h = _top - 1;


    
    
    
    Node* left = _left;
    for (;;) {
      Node* next = left->nodes[h];
      if (next == 0) {
	if(h<=nodeHeight) {
	  left->nodes[h] = newNode;
	  newNode->nodes[h] = 0;
	  if (h == 0) {
	    return &newNode->value;
	  }
	}
	h--;
	continue;
      }
      
      switch (ValueComparator::compare(key,next->value))
	{
	case LESS:
	  
	  if(h<=nodeHeight) {
	    newNode->nodes[h] = next;
	    left->nodes[h] = newNode;
	    if (h == 0) {
	      return &newNode->value;
	    }
	  }
	  h--;
	  break;

	case EQUAL: 
	case GREATER:
	  left = next;
	  break;

#if VDEBUG
	default:
	  ASSERTION_VIOLATION;
#endif
	}
    }
  } 

 
  template<typename Key>
  bool findLeastGreater(Key key, Value& value)
  {
    CALL("SkipList::findLeastGreater");

    if(_top==0) {
      return false;
    }

    unsigned h=_top-1;

   
    
    
    Node* left = _left;
    for (;;) {
      Node* next = left->nodes[h];
      if (next == 0) {
	if (h == 0) {
	  ASS_LE(max(),key);
	  return false;
	}
	h--;
	continue;
      }
      
      switch (ValueComparator::compare(key,next->value))
	{
	case LESS:
	  
	  if (h == 0) {
	    value=next->value;
	    return true;
	  }
	  h--;
	  break;

	case EQUAL:
	  while(next->nodes[0]) {
	    next=next->nodes[0];
	    if(ValueComparator::compare(key,next->value)==LESS) {
	      value=next->value;
	      return true;
	    }
	  }
	  ASS_LE(max(),key);
	  return false;

	case GREATER:
	  left = next;
	  break;

#if VDEBUG
	default:
	  ASSERTION_VIOLATION;
#endif
	}
    }
  } 


 
  int size() const
  {
    int res=0;
    Iterator it(*this);
    while(it.hasNext()) {
      it.next();
      res++;
    }
    return res;
  }

 
  inline
  bool isEmpty() const
  {
    return _left->nodes[0] == 0;
  } 

 
  inline
  bool isNonEmpty() const
  {
    return _left->nodes[0] != 0;
  } 

 
  inline
  const Value& top()
  {
    ASS(isNonEmpty());
    return _left->nodes[0]->value;
  }

 
  Value pop()
  {
    CALL("SkipList::pop");
    ASS(isNonEmpty());

    
    Node* node = _left->nodes[0];
    unsigned h;
    for (h = 1;h < _top;h++) {
      if (_left->nodes[h] != node) {
	break;
      }
    }
    
    for (unsigned i = 0;i < h;i++) {
      _left->nodes[i] = node->nodes[i];
    }
    Value val = node->value;
    deallocate(node,h-1);
    return val;
  } 

 
  const Value& max()
  {
    ASS(isNonEmpty());
    Node* node=_left;
    for(int h=_top-1;h>=0;h--) {
      while(node->nodes[h]) {
	node=node->nodes[h];
      }
    }
    ASS_NEQ(node, _left);
    return node->value;
  }

 
  template<typename Key>
  void remove(Key key)
  {
    CALL("SkipList::remove");
    ASS(_top > 0);

    Node* found = 0; 
#if VDEBUG
    unsigned foundHeight = 0; 
#else
    unsigned foundHeight; 
#endif

    Node* left = _left;
    unsigned h = _top-1;
    for (;;) {
      Node* next = left->nodes[h];
      if (next == 0) {
	ASS(h != 0); 
	h--;
	continue;
      }
      
      switch (ValueComparator::compare(key,next->value))
	{
	case LESS:
	  ASS(h != 0);
	  h--;
	  break;

	case GREATER:
	  left = next;
	  break;

	case EQUAL:
	  found = next;
	  foundHeight = h;
	  if(h>0 && found->nodes[0] && found->nodes[h]!=found->nodes[0] &&
		  ValueComparator::compare(key,found->nodes[0]->value)==EQUAL) {
	    
	    
	    
	    
	    
	    h=0;
	    while(found->nodes[0]==found->nodes[h+1]) {
	      h++;
	    }
	    left = found;
	    found = found->nodes[0];
	    foundHeight = h;
	  }
	  for(;;) {
	    left->nodes[h] = found->nodes[h];
	    if(h==0) {
	      break;
	    }
	    h--;
	    while(left->nodes[h]!=found) {
	      left=left->nodes[h];
	      ASS(ValueComparator::compare(key,left->value)!=LESS);
	    }
	  }

	  deallocate(found,foundHeight);
	  return;

#if VDEBUG
	default:
	  ASSERTION_VIOLATION;
#endif
	}
    }
  } 

  template<typename Key>
  inline
  bool find(Key key)
  {
    Value* pval;
    return getPosition(key,pval,false);
  }

  template<typename Key>
  inline
  bool find(Key key, Value& val)
  {
    Value* pval;
    bool res=getPosition(key,pval,false);
    val=*pval;
    return res;
  }


  inline
  void makeEmpty()
  {
    while (isNonEmpty()) {
      pop();
    }
  }

  inline
  List<Value>* toList()
  {
    

    
    
    
    
    
    
    
    if(_left->nodes[0]) {
      ASS_EQ(reinterpret_cast<List<Value>*&>(_left->nodes[0])->headPtr(), &_left->nodes[0]->value);
      ASS_EQ((void*)&(reinterpret_cast<List<Value>*&>(_left->nodes[0])->tailReference()),
	      (void*)&_left->nodes[0]->nodes[0]);
    }

    return reinterpret_cast<List<Value>*&>(_left->nodes[0]);
  }


 
  SkipList()
    : _left(allocate(SKIP_LIST_MAX_HEIGHT)),
      _top(0)
  {
    CALL("SkipList::SkipList");
    for (int h = SKIP_LIST_MAX_HEIGHT-1;h >= 0;h--) {
      _left->nodes[h] = 0;
    }
  }
 
  ~SkipList()
  {
    CALL("SkipList::~SkipList");

    makeEmpty();
    deallocate(_left,SKIP_LIST_MAX_HEIGHT);
  }
private:
 
  Node* _left;
 
  unsigned _top;

 
  inline
  static Node* allocate(unsigned h)
  {
    CALL("SkipList::allocate");

    void* memory = ALLOC_KNOWN(sizeof(Node)+h*sizeof(Node*),"SkipList::Node");

    return reinterpret_cast<Node*>(memory);
  }

 
  inline
  static void deallocate(Node* node,unsigned h)
  {
    CALL("SkipList::deallocate");
    DEALLOC_KNOWN(node,sizeof(Node)+h*sizeof(Node*),"SkipList::Node");
  }


  class SingleValBacktrackObject: public BacktrackObject
  {
  public:
    enum Action {
      REMOVE, INSERT
    };
    SingleValBacktrackObject(SkipList* sl, Action a, Value v): sl(sl), a(a), v(v) {}
    void backtrack()
    {
      switch(a) {
      case REMOVE:
	sl->insert(v);
	break;
      case INSERT:
	sl->remove(v);
	break;
      default:
	ASSERTION_VIOLATION;
      }
    }
  private:
    SkipList* sl;
    Action a;
    Value v;
  };
public:
  Value backtrackablePop(BacktrackData& bd)
  {
    Value v=pop();
    bd.addBacktrackObject(
	    new SingleValBacktrackObject(this, SingleValBacktrackObject::REMOVE, v));
    return v;
  }
  void backtrackableInsert(Value v, BacktrackData& bd)
  {
    insert(v);
    bd.addBacktrackObject(
	    new SingleValBacktrackObject(this, SingleValBacktrackObject::INSERT, v));
  }

public:

 
  class Iterator {
   public:
     DECL_ELEMENT_TYPE(Value);

    inline explicit
    Iterator(const SkipList& l)
      : _cur (l._left) {}

   
    inline Value next()
    {
      ASS(_cur->nodes[0]);
      _cur=_cur->nodes[0];
      return _cur->value;
    }

   
    inline bool hasNext() const
    { return _cur->nodes[0]; }

   private:
   
    Node* _cur;
  };

 
  class RefIterator {
  public:
    DECL_ELEMENT_TYPE(Value&);
    inline explicit
    RefIterator(const SkipList& l)
      : _cur (l._left) {}

   
    inline Value& next()
    {
      ASS(_cur->nodes[0]);
      _cur=_cur->nodes[0];
      return _cur->value;
    }

   
    inline bool hasNext() const
    { return _cur->nodes[0]; }

  private:
   
    Node* _cur;
  };

 
  class PtrIterator {
  public:
    DECL_ELEMENT_TYPE(Value*);

    inline explicit
    PtrIterator(const SkipList& l)
      : _cur (l._left)
    {}

   
    inline Value* next()
    {
      ASS(_cur->nodes[0]);
      _cur=_cur->nodes[0];
      return &_cur->value;
    }

   
    inline bool hasNext() const
    {
      return _cur->nodes[0];
    }

   private:
   
    Node* _cur;
  };


}; 


} 
#endif


