
/*
 * File Semaphore.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Portability.hpp"

#if !COMPILER_MSVC

#include <cerrno>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "Debug/Log.hpp"

#include "Lib/Exception.hpp"
#include "Lib/Int.hpp"
#include "Lib/List.hpp"
#include "Lib/System.hpp"

#include "Multiprocessing.hpp"

#include "Semaphore.hpp"

namespace Lib
{
namespace Sys
{

Semaphore::SemaphoreList* Semaphore::s_instances = 0;

Semaphore::Semaphore(int num)
:semCnt(num)
{
  CALL("Semaphore::Semaphore(int)");
  ASS_G(num,0);

  ensureEventHandlersInstalled();

  bool retrying=false;
get_retry:
  errno=0;
  semid=semget(IPC_PRIVATE, semCnt+2, 0600);
  if(semid==-1 && errno==ENOSPC && !retrying) {
    retrying=true;
    system("ipcs -s | grep 0x00000000 | cut -d' ' -f2|xargs -n 1 ipcrm -s");
    goto get_retry;
  }
  if(semid==-1) {
    SYSTEM_FAIL("Cannot create semaphore.",errno);
  }

  
  unsigned short semVals[semCnt+2];
  for(int i=0;i<=semCnt;i++) {
    semVals[i]=0;
  }
  semVals[semCnt+1]=1;

  semun arg;
  arg.array=semVals;

  errno=0;
  int res=semctl(semid, num, SETALL, arg);
  if(res==-1) {
    SYSTEM_FAIL("Cannot set initial semaphore values.",errno);
  }
  ASS_EQ(res,0);

  LOG("mp_sem","Semaphore: created");

  registerInstance();
}

Semaphore::~Semaphore()
{
  CALL("Semaphore::~Semaphore");

  deregisterInstance();
}

Semaphore::Semaphore(const Semaphore& s)
{
  CALL("Semaphore::Semaphore(const Semaphore&)");

  semid=s.semid;
  semCnt=s.semCnt;
  registerInstance();
}

const Semaphore& Semaphore::operator=(const Semaphore& s)
{
  CALL("Semaphore::operator=");

  deregisterInstance();
  semid=s.semid;
  semCnt=s.semCnt;
  registerInstance();
  return *this;
}

void Semaphore::doInc(int num)
{
  CALL("Semaphore::doInc");
  ASS(hasSemaphore());
  ASS_L(num, semCnt+2);

  sembuf buf;
  buf.sem_num=num;
  buf.sem_op=1;
  buf.sem_flg=SEM_UNDO;

  errno=0;
  int res=semop(semid, &buf, 1);
  if(res==-1) {
    SYSTEM_FAIL("Cannot increase semaphore.",errno);
  }
}

void Semaphore::doDec(int num)
{
  CALL("Semaphore::doDec");
  ASS(hasSemaphore());
  ASS_L(num, semCnt+2);

  sembuf buf;
  buf.sem_num=num;
  buf.sem_op=-1;
  buf.sem_flg=SEM_UNDO;

retry_decreasing:
  errno=0;
  int res=semop(semid, &buf, 1);
  if(res==-1 && errno==EINTR) {
    
    LOG("mp_sem","Semaphore: resuming waiting for decrease after a signal");
    goto retry_decreasing;
  }
  if(res==-1) {
    SYSTEM_FAIL("Cannot decrease semaphore.",errno);
  }
}

void Semaphore::doSet(int num, int val)
{
  CALL("Semaphore::doSet");
  ASS(hasSemaphore());
  ASS_L(num, semCnt+2);
  ASS_GE(val,0); 

  semun arg;
  arg.val=val;

  errno=0;
  int res=semctl(semid, num, SETVAL, arg);
  if(res==-1) {
    SYSTEM_FAIL("Cannot set the semaphore value.",errno);
  }
}

int Semaphore::doGet(int num)
{
  CALL("Semaphore::doGet");
  ASS(hasSemaphore());
  ASS_L(num, semCnt+2);

  semun unused_arg;

  errno=0;
  int res=semctl(semid, num, GETVAL, unused_arg);
  if(res==-1) {
    SYSTEM_FAIL("Cannot set the semaphore value.",errno);
  }
  ASS_GE(res,0);
  return res;
}

void Semaphore::inc(int num)
{
  CALL("Semaphore::inc");
  ASS(hasSemaphore());
  ASS_L(num, semCnt);

  doInc(num);
}

void Semaphore::dec(int num)
{
  CALL("Semaphore::dec");
  ASS(hasSemaphore());
  ASS_L(num, semCnt);

  doDec(num);
}

int Semaphore::get(int num)
{
  CALL("Semaphore::get");
  ASS(hasSemaphore());
  ASS_L(num, semCnt);

  return doGet(num);
}

void Semaphore::set(int num, int val)
{
  CALL("Semaphore::set");
  ASS(hasSemaphore());
  ASS_L(num, semCnt);
  ASS_GE(val, 0);

  doSet(num, val);
}

bool Semaphore::isLastInstance()
{
  CALL("Semaphore::isLastInstance");
  ASS(hasSemaphore());

  return doGet(semCnt)==1;
}
void Semaphore::registerInstance(bool addToInstanceList)
{
  CALL("Semaphore::registerInstance");

  if(!hasSemaphore()) {
    return;
  }

  acquireInstance();

  if(addToInstanceList) {
    SemaphoreList::push(this, s_instances);
  }
  LOG("mp_sem","Semaphore: instance registered");
}

void Semaphore::deregisterInstance()
{
  CALL("Semaphore::deregisterInstance");

  LOG("mp_sem","Semaphore: call to deregisterInstance");
  if(!hasSemaphore()) {
    return;
  }
  LOG("mp_sem","Semaphore: deregistering instance");
  ASS(s_instances->member(this));
  s_instances=s_instances->remove(this);

  releaseInstance();
}

void Semaphore::acquireInstance()
{
  CALL("Semaphore::registerInstance");
  ASS(hasSemaphore());

  doInc(semCnt);
}

void Semaphore::releaseInstance()
{
  CALL("Semaphore::deregisterInstance");
  ASS(hasSemaphore());

  
  
  
  
  doDec(semCnt+1);

  ASS_G(doGet(semCnt), 0);
  doDec(semCnt);

  if(doGet(semCnt)==0) {
    semun unused_arg;
    errno=0;
    int res=semctl(semid, 1, IPC_RMID, unused_arg);
    if(res==-1) {
      SYSTEM_FAIL("Cannot destroy semaphore.",errno);
    }
    ASS_EQ(res,0);
    LOG("mp_sem","Semaphore: sem destroyed");
  }
  else {
    
    doInc(semCnt+1);
  }
}

void Semaphore::releaseAllSemaphores()
{
  CALL("Semaphore::releaseAllSemaphores");

  LOG("mp_sem","Semaphore: releasing all");

  SemaphoreList* instIter=s_instances;
  while(instIter) {
    Semaphore* s=instIter->head();
    if(s->semid!=-1) {
      s->releaseInstance();
    }
    s->semid=-1;
    instIter=instIter->tail();
  }
}


void Semaphore::postForkInChild()
{
  CALL("Semaphore::postForkInChild");

  LOG("mp_sem","Semaphore: post-fork in child");

  SemaphoreList::Iterator sit(s_instances);
  while(sit.hasNext()) {
    Semaphore* s=sit.next();
    s->acquireInstance();
  }
}

void Semaphore::ensureEventHandlersInstalled()
{
  CALL("Semaphore::ensureForkHandlerInstalled");

  static bool installed=false;
  if(installed) {
    return;
  }
  Multiprocessing::instance()->registerForkHandlers(0,0,postForkInChild);
  
  
  System::addTerminationHandler(releaseAllSemaphores,1);
  installed=true;
}

}
}

#endif 

