
/*
 * File SyncPipe.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Portability.hpp"

#if !COMPILER_MSVC

#include <cerrno>
#include <unistd.h>

#include "Lib/Environment.hpp"
#include "Lib/Exception.hpp"
#include "Lib/fdstream.hpp"
#include "Lib/List.hpp"
#include "Lib/System.hpp"

#include "Multiprocessing.hpp"

#include "SyncPipe.hpp"

namespace Lib
{
namespace Sys
{

SyncPipe::PipeList* SyncPipe::s_instances = 0;

SyncPipe::SyncPipe()
: _isReading(false), _isWriting(false), _syncSemaphore(3)
{
  ensureEventHandlersInstalled();

  int fd[2];
  errno=0;
  int res=pipe(fd);
  if(res==-1) {
    SYSTEM_FAIL("Pipe creation.", errno);
  }

  _readDescriptor=fd[0];
  _writeDescriptor=fd[1];

  _istream=new fdstream(_readDescriptor);
  _ostream=new fdstream(_writeDescriptor);

  _istream->rdbuf()->pubsetbuf(0,0);

  
  _syncSemaphore.set(0,1);
  _syncSemaphore.set(1,1);
  
  _syncSemaphore.set(2,256);

  PipeList::push(this, s_instances);
}

SyncPipe::~SyncPipe()
{
  CALL("SyncPipe::~SyncPipe");

  releasePriviledges();
  ASS(s_instances->member(this));
  s_instances=s_instances->remove(this);

  if(canRead()) {
    neverRead();
  }
  if(canWrite()) {
    neverWrite();
  }
}

void SyncPipe::acquireRead()
{
  CALL("SyncPipe::acquireRead");
  ASS(canRead());
  ASS(!isReading());
  ASS(!isWriting()); 

  _syncSemaphore.dec(0);

  
  int preRead=_syncSemaphore.get(2);
  if(preRead==256) {
    preRead=-1;
  }
  ASS_LE(preRead,255);
  _istream->setPreReadChar(preRead);

  _isReading=true;
}

void SyncPipe::releaseRead()
{
  CALL("SyncPipe::releaseRead");
  ASS(isReading());

  _isReading=false;

  int preRead=_istream->getPreReadChar();
  if(preRead==-1) {
    preRead=256;
  }
  ASS_GE(preRead,0);
  ASS_LE(preRead,256);
  _syncSemaphore.set(2,preRead);

  _syncSemaphore.inc(0);
}

void SyncPipe::neverRead()
{
  CALL("SyncPipe::neverRead");
  ASS(canRead());  
  ASS(!isReading());

  int res=close(_readDescriptor);
  if(res==-1) {
    SYSTEM_FAIL("Closing read descriptor of a pipe.", errno);
  }
  ASS_EQ(res,0);
  delete _istream;
  _istream=0;
}


void SyncPipe::acquireWrite()
{
  CALL("SyncPipe::acquireWrite");
  ASS(canWrite());
  ASS(!isWriting());
  ASS(!isReading()); 

  _syncSemaphore.dec(1);
  _isWriting=true;
}

void SyncPipe::releaseWrite()
{
  CALL("SyncPipe::releaseWrite");
  ASS(isWriting());

  _ostream->flush();
  _isWriting=false;
  _syncSemaphore.inc(1);
}

void SyncPipe::neverWrite()
{
  CALL("SyncPipe::neverWrite");
  ASS(canWrite());  
  ASS(!isWriting());
  ASS(env.getOutputPipe()!=this); 

  int res=close(_writeDescriptor);
  if(res==-1) {
    SYSTEM_FAIL("Closing write descriptor of a pipe.", errno);
  }
  ASS_EQ(res,0);

  delete _ostream;
  _ostream=0;
}
void SyncPipe::releasePriviledges()
{
  CALL("SyncPipe::releasePriviledges");
  ASS(_syncSemaphore.hasSemaphore());

  if(isReading()) {
    releaseRead();
  }
  if(isWriting()) {
    releaseWrite();
  }
}

void SyncPipe::postForkChildHadler()
{
  CALL("SyncPipe::postForkChildHadler");

  PipeList::Iterator pit(s_instances);
  while(pit.hasNext()) {
    SyncPipe* p=pit.next();
    p->releasePriviledges();
  }
}

void SyncPipe::terminationHadler()
{
  CALL("SyncPipe::terminationHadler");

  PipeList* listIter=s_instances;
  while(listIter) {
    if(listIter->head()) {
      SyncPipe* p=listIter->head();
      p->releasePriviledges();
      listIter->setHead(0);
    }
    listIter=listIter->tail();
  }
}

void SyncPipe::ensureEventHandlersInstalled()
{
  CALL("SyncPipe::ensureEventHandlersInstalled");

  static bool installed=false;
  if(installed) {
    return;
  }
  Multiprocessing::instance()->registerForkHandlers(0,0,postForkChildHadler);
  
  
  System::addTerminationHandler(terminationHadler,0);
  installed=true;
}


}
}

#endif 
