
/*
 * File System.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Portability.hpp"

#include <stdlib.h>
#if COMPILER_MSVC
#  include <Winsock2.h>
#  include <process.h>
#else
#  include <unistd.h>
#  if !__APPLE__
#    include <sys/prctl.h>
#  endif
#endif

#include <cerrno>
#include <string>
#include <csignal>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>

#include "Debug/Tracer.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "Environment.hpp"
#include "Exception.hpp"
#include "Int.hpp"
#include "Stack.hpp"
#include "Timer.hpp"

#include "System.hpp"

bool outputAllowed()
{
#if VDEBUG
  return true;
#else
  return !Lib::env.options || Lib::env.options->mode()!=Shell::Options::MODE_SPIDER;
#endif
}

bool inSpiderMode()
{
  return Lib::env.options && Lib::env.options->mode()==Shell::Options::MODE_SPIDER;
}

void reportSpiderFail()
{
  reportSpiderStatus('!');
}

void reportSpiderStatus(char status)
{
  using namespace Lib;

  static bool headerPrinted=false;

  if(inSpiderMode() && !headerPrinted) {
    headerPrinted=true;

    env.beginOutput();
    env.out() << status << " "
      << (Lib::env.options ? Lib::env.options->problemName() : "unknown") << " "
      << (Lib::env.timer ? Lib::env.timer->elapsedDeciseconds() : 0) << " "
      << (Lib::env.options ? Lib::env.options->testId() : "unknown") << "\n";
    env.endOutput();
  }
}

#if COMPILER_MSVC

#include <windows.h>

long long Lib::System::getSystemMemory()
{
  MEMORYSTATUSEX status;
  GlobalMemoryStatusEx(&status);
  return status.ullTotalPhys;
}

unsigned Lib::System::getNumberOfCores()
{
  SYSTEM_INFO sysinfo;
  GetSystemInfo( &sysinfo );
  return sysinfo.dwNumberOfProcessors;
}

#else

#include <unistd.h>

long long Lib::System::getSystemMemory()
{
#if __APPLE__
  NOT_IMPLEMENTED;
#else
  long pages = sysconf(_SC_PHYS_PAGES);
  long page_size = sysconf(_SC_PAGE_SIZE);
  return static_cast<long long>(pages) * page_size;
#endif
}

unsigned Lib::System::getNumberOfCores()
{
#if __APPLE__
  NOT_IMPLEMENTED;
#else
  return sysconf( _SC_NPROCESSORS_ONLN );
#endif
}


#endif

namespace Lib {

using namespace std;
using namespace Shell;

bool System::s_initialized = false;
bool System::s_shouldIgnoreSIGINT = false;
const char* System::s_argv0 = 0;










const char* signalToString (int sigNum)
{
  switch (sigNum)
    {
    case SIGTERM:
      return "SIGTERM";
# ifndef _MSC_VER
    case SIGQUIT:
      return "SIGQUIT";
    case SIGHUP:
      return "SIGHUP";
    case SIGXCPU:
      return "SIGXCPU";
    case SIGBUS:
      return "SIGBUS";
    case SIGTRAP:
      return "SIGTRAP";
# endif
    case SIGINT:
      return "SIGINT";
    case SIGILL:
      return "SIGILL";
    case SIGFPE:
      return "SIGFPE";
    case SIGSEGV:
      return "SIGSEGV";
    case SIGABRT:
      return "SIGABRT";
    default:
      return "UNKNOWN SIGNAL";
    }
} 


void handleSignal (int sigNum)
{
  
  
  static bool handled = false;
  static bool haveSigInt = false;
  const char* signalDescription = signalToString(sigNum);

  switch (sigNum)
    {
    case SIGTERM:
# ifndef _MSC_VER
    case SIGQUIT:
      if (handled) {
	System::terminateImmediately(haveSigInt ? VAMP_RESULT_STATUS_SIGINT : VAMP_RESULT_STATUS_OTHER_SIGNAL);
      }
      handled = true;
      if(outputAllowed()) {
	if(env.options) {
	  env.beginOutput();
	  env.out() << "Aborted by signal " << signalDescription << " on " << env.options->inputFile() << "\n";
	  env.endOutput();
	} else {
	  cout << "Aborted by signal " << signalDescription << "\n";
	}
      }
      return;
    case SIGXCPU:
      if(outputAllowed()) {
	if(env.options) {
	  env.beginOutput();
	  env.out() << "External time out (SIGXCPU) on " << env.options->inputFile() << "\n";
	  env.endOutput();
	} else {
	  cout << "External time out (SIGXCPU)\n";
	}
      }
      System::terminateImmediately(VAMP_RESULT_STATUS_OTHER_SIGNAL);
      break;
# endif

    case SIGINT:
      if(System::shouldIgnoreSIGINT()) {
	return;
      }
      haveSigInt=true;



    case SIGILL:
    case SIGFPE:
    case SIGSEGV:

# ifndef _MSC_VER
    case SIGBUS:
    case SIGTRAP:
    case SIGHUP:
# endif
    case SIGABRT:
      {
	if (handled) {
	  System::terminateImmediately(haveSigInt ? VAMP_RESULT_STATUS_SIGINT : VAMP_RESULT_STATUS_OTHER_SIGNAL);
	}
	reportSpiderFail();
	handled = true;
	if(outputAllowed()) {
	  if(env.options && env.statistics) {
	    env.beginOutput();
	    env.out() << "Aborted by signal " << signalDescription << " on " << env.options->inputFile() << "\n";
	    env.statistics->print(env.out());
#if VDEBUG
	    Debug::Tracer::printStack(env.out());
#endif
	    env.endOutput();
	  } else {
	    cout << "Aborted by signal " << signalDescription << "\n";
#if VDEBUG
	    Debug::Tracer::printStack(cout);
#endif
	  }
	}
	System::terminateImmediately(haveSigInt ? VAMP_RESULT_STATUS_SIGINT : VAMP_RESULT_STATUS_OTHER_SIGNAL);
      }

    default:
      break;
    }
} 

void System::setSignalHandlers()
{
  signal(SIGTERM,handleSignal);
  signal(SIGINT,handleSignal);
  signal(SIGILL,handleSignal);
  signal(SIGFPE,handleSignal);
  signal(SIGSEGV,handleSignal);
  signal(SIGABRT,handleSignal);

#ifndef _MSC_VER
  signal(SIGQUIT,handleSignal);
  signal(SIGHUP,handleSignal);
  signal(SIGXCPU,handleSignal);
  signal(SIGBUS,handleSignal);
  signal(SIGTRAP,handleSignal);
#endif

  errno=0;
  int res=atexit(onTermination);
  if(res==-1) {
    SYSTEM_FAIL("Call of atexit() function in System::setSignalHandlers failed.", errno);
  }
  ASS_EQ(res,0);
}

ZIArray<List<VoidFunc>*>& System::initializationHandlersArray()
{
  CALL("System::initializationHandlersArray");

  static ZIArray<List<VoidFunc>*> arr;
  return arr;
}

void System::addInitializationHandler(VoidFunc proc, unsigned priority)
{
  CALL("System::addInitializationHandler");

  if(s_initialized) {
    proc();
  }
  else {
    VoidFuncList::push(proc, initializationHandlersArray()[priority]);
  }
}

void System::onInitialization()
{
  CALL("System::onTermination");
  ASS(!s_initialized); 

  s_initialized=true;

  int sz=initializationHandlersArray().size();
  for(int i=sz-1;i>=0;i--) {
    VoidFuncList::Iterator ihIter(initializationHandlersArray()[i]);
    while(ihIter.hasNext()) {
      VoidFunc func=ihIter.next();
      func();
    }
  }
}

void System::addTerminationHandler(VoidFunc proc, unsigned priority)
{
  CALL("System::addTerminationHandler");

  VoidFuncList::push(proc, s_terminationHandlers[priority]);
}

void System::onTermination()
{
  CALL("System::onTermination");

  static bool called=false;
  if(called) {
    return;
  }
  called=true;

  size_t sz=s_terminationHandlers.size();
  for(size_t i=0;i<sz;i++) {
    VoidFuncList::Iterator thIter(s_terminationHandlers[i]);
    while(thIter.hasNext()) {
      VoidFunc func=thIter.next();
      func();
    }
  }
}

void System::terminateImmediately(int resultStatus)
{
  CALL("System::terminateImmediately");

  onTermination();
  _exit(resultStatus);
}

void System::registerForSIGHUPOnParentDeath()
{
#if __APPLE__ || COMPILER_MSVC
  cerr<<"Death of parent process not being handled on Mac and Windows"<<endl;

#else
  prctl(PR_SET_PDEATHSIG, SIGHUP);
#endif
}

void System::readCmdArgs(int argc, char* argv[], StringStack& res)
{
  CALL("System::readCmdArgs");
  ASS_GE(argc,1);
  ASS(res.isEmpty()); 

  registerArgv0(argv[0]);
  for(int i=1; i<argc; i++) {
    res.push(argv[i]);
  }
}

string System::extractFileNameFromPath(string str)
{
  CALL("System::extractFileNameFromPath");

  size_t index=str.find_last_of("\\/")+1;
  if(index==string::npos) {
    return str;
  }
  return string(str, index);
}

bool System::extractDirNameFromPath(string path, string& dir)
{
  CALL("System::extractDirNameFromPath");

  size_t index=path.find_last_of("\\/");
  if(index==string::npos) {
    return false;
  }
  dir = path.substr(0, index);
  return true;
}

bool System::fileExists(string fname)
{
  CALL("System::fileExists");

  ifstream ifile(fname.c_str());
  return ifile;
}

string System::guessExecutableDirectory()
{
  CALL("System::guessExecutableDirectory");

  string res;

  if(s_argv0 && extractDirNameFromPath(s_argv0, res)) {
    return res;
  }

  return ".";
}

string System::guessExecutableName()
{
  CALL("System::guessExecutableName");

  string res;

  if(s_argv0) {
    return s_argv0;
  }

  return "./vampire";
}

pid_t System::getPID()
{
  CALL("System::getPID");

#if !COMPILER_MSVC
  return getpid();
#else
  
  return 0;
#endif
}

int System::executeCommand(string command, string input, Stack<string>& outputLines)
{
  CALL("System::executeCommand");

#if COMPILER_MSVC
  NOT_IMPLEMENTED;
#else
  string pidStr = Int::toString(getPID());
  string inFile  = "/tmp/vampire_executeCommand_"+pidStr+"_in";
  string outFile = "/tmp/vampire_executeCommand_"+pidStr+"_out";

  string cmdLine = command + " <" + inFile + " >" + outFile;

  {
    ofstream inpFile(inFile.c_str());
    inpFile << input;
    inpFile.close();
  }

  int resStatus=system(cmdLine.c_str());

  {
    outputLines.reset();
    string line;
    ifstream outpFile(outFile.c_str());
    while (getline(outpFile, line)) {
      outputLines.push(line);
    }
    outpFile.close();
  }








  remove(inFile.c_str());
  remove(outFile.c_str());

  if(WIFEXITED(resStatus)) {
    return WEXITSTATUS(resStatus);
  }
  else if(WIFSIGNALED(resStatus)) {
    return -WTERMSIG(resStatus);
  }
  else {
    return -0xffff;
  }

#endif
}

};
