
/*
 * File Vector.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __Vector__
#define __Vector__

#include <string>

#include "Forwards.hpp"

#include "Debug/Assertion.hpp"
#include "Allocator.hpp"

namespace Lib {

using namespace std;

template<typename C>
class Vector
{
public:
 
  inline C& operator[] (size_t n)
  {
    ASS(n < _length);
    return _array[n];
  } 

 
  inline const C& operator[](size_t n) const
  {
    ASS(n < _length);
    return _array[n];
  }

 
  size_t length() const { return _length; }

 
  static Vector* allocate(size_t length)
  {
    CALL("Vector::allocate");

    
    
    
    size_t sz=sizeof(Vector) + length*sizeof(C);
    sz-=sizeof(C);

    Vector* v = reinterpret_cast<Vector*>(ALLOC_KNOWN(sz, "Vector"));
    v->_length = length;
    C* arr = v->_array;
    array_new<C>(arr, length);
    return v;
  } 

 
  void deallocate()
  {
    CALL("Vector::deallocate");

    array_delete(_array, _length);

    
    
    
    size_t sz=sizeof(Vector) + _length*sizeof(C);
    sz-=sizeof(C);

    DEALLOC_KNOWN(this,sz,"Vector");
  } 


  string toString()
  {
    string res;
    for(size_t i=0;i<_length;i++) {
      if(i>0) {
	res+=",";
      }
      res+=(*this)[i].toString();
    }
    return res;
  }

  friend class Indexing::CodeTree;
  friend class Indexing::ClauseCodeTree;

 
  class DestructiveIterator
  {
  public:
    DECL_ELEMENT_TYPE(C);

    DestructiveIterator(Vector& v)
    : cur(v._array), afterLast(v._array+v.length()), vec(&v)
    {
      if(cur==afterLast) {
	vec->deallocate();
      }
    }

    bool hasNext()
    {
      return cur!=afterLast;
    }

    C next()
    {
      CALL("Vector::DestructiveIterator::next");
      ASS(hasNext());

      C res=*cur;
      cur++;
      if(cur==afterLast) {
	vec->deallocate();
      }
      return res;
    }
  private:
    C* cur;
    C* afterLast;
    Vector* vec;
  };
protected:
 
  size_t _length;
 
  C _array[1];
private:
  void* operator new(size_t,size_t length);
  void operator delete(void*)
  {
    ASSERTION_VIOLATION
  }
  Vector();
}; 

} 

#endif
