
/*
 * File TPTP.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __Parser_TPTP__
#define __Parser_TPTP__

#include <iostream>

#include "Lib/Array.hpp"
#include "Lib/Set.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Exception.hpp"
#include "Lib/IntNameTable.hpp"

#include "Kernel/Formula.hpp"
#include "Kernel/Unit.hpp"
#include "Kernel/Theory.hpp"

using namespace std;
using namespace Lib;
using namespace Kernel;

namespace Kernel {
  class Clause;
};

namespace Parse {

class TPTP;

class TPTP 
{
public:
 
  enum Tag {
   
    T_EOF,
   
    T_NAME,
   
    T_VAR,
   
    T_LPAR,
   
    T_RPAR,
   
    T_LBRA,
   
    T_RBRA,
   
    T_COMMA,
   
    T_COLON,
   
    T_NOT,
   
    T_AND,
   
    T_EQUAL,
   
    T_STRING,
   
    T_NEQ,
   
    T_FORALL,
   
    T_EXISTS,
   
    T_PI,
   
    T_SIGMA,
   
    T_IMPLY,
   
    T_XOR,
   
    T_IFF,
   
    T_REVERSE_IMP,
   
    T_DOT,
   
    T_REAL,
   
    T_RAT,
   
    T_INT,
   
    T_OR,
   
    T_ASS,
   
    T_LAMBDA,
   
    T_APP,
   
    T_STAR,
   
    T_UNION,
   
    T_ARROW,
   
    T_SUBTYPE,
   
    T_NOT_OR,
   
    T_NOT_AND,
   
    T_SEQUENT,
   
    T_THF_QUANT_ALL,
   
    T_THF_QUANT_SOME,
   
    T_APP_PLUS,
   
    T_APP_MINUS,
   
    T_TRUE,
   
    T_FALSE,
   
    T_TTYPE,
   
    T_BOOL_TYPE,
   
    T_DEFAULT_TYPE,
   
    T_INTEGER_TYPE,
   
    T_RATIONAL_TYPE,
   
    T_REAL_TYPE,
   
    T_FOT,
   
    T_FOF,
   
    T_TFF,
   
    T_THF,
   
    T_DOLLARS,
   
    T_ITEF,
   
    T_ITET,
   
    T_LETTT,
   
    T_LETTF,
   
    T_LETFT,
   
    T_LETFF,
  };

 
  enum State {
   
    UNIT_LIST,
   
    CNF,
   
    FOF,
   
    VAMPIRE,
   
    FORMULA,
   
    END_FOF,
   
    SIMPLE_FORMULA,
   
    END_FORMULA,
   
    VAR_LIST,
   
    ATOM,
   
    MID_ATOM,
   
    ARGS,
   
    TERM,
   
    END_TERM,
   
    TAG,
   
    INCLUDE,
   
    END_EQ,
   
    TFF,
   
    TYPE,
   
    END_TFF,
   
    END_TYPE,
   
    SIMPLE_TYPE,
   
    UNBIND_VARIABLES,
   
    ITEF,
   
    END_ITEF,
   
    END_ITET,
   
    END_ARGS,
   
    MID_EQ,
   
    LETTF,
   
    LETFF,
   
    END_LETTT,
   
    END_LETTF,
   
    END_LETFT,
   
    END_LETFF,
  };

 
  struct Token {
   
    Tag tag;
   
    int start;
   
    string content;
    string toString() const;
  };

 
  class Exception 
    : public ::Exception
  {
  public:
    Exception(string message,Token& tok);
    Exception(string message,int position);
    void cry(ostream&);
    ~Exception() {}
  protected:
    string _message;
  }; 
  friend class Exception;

  TPTP(istream& in);
  ~TPTP();
  void parse();
  static UnitList* parse(istream& str);
 
  inline UnitList* units() { return _units.list(); }
 
  bool containsConjecture() const { return _containsConjecture; }
  void addForbiddenInclude(string file);
  static bool findAxiomName(const Unit* unit, string& result);
  
  static void assignAxiomName(const Unit* unit, string& name);
private:
 
  const char* input() { return _chars.content(); }

  enum TypeTag {
    TT_ATOMIC,
    TT_PRODUCT,
    TT_ARROW
  };

 
  class Type {
  public:
    explicit Type(TypeTag tag) : _tag(tag) {}
   
    TypeTag tag() const {return _tag;}
  protected:
   
    TypeTag _tag;
  };

 
  class AtomicType
    : public Type
  {
  public:
    explicit AtomicType(unsigned sortNumber)
      : Type(TT_ATOMIC), _sortNumber(sortNumber)
    {}
   
    unsigned sortNumber() const {return _sortNumber;}
  private:
   
    unsigned _sortNumber;
  }; 

 
  class ArrowType
    : public Type
  {
  public:
    ArrowType(Type* lhs,Type* rhs)
      : Type(TT_ARROW), _lhs(lhs), _rhs(rhs)
    {}
   
    Type* argumentType() const {return _lhs;}
   
    Type* returnType() const {return _rhs;}
  private:
   
    Type* _lhs;
   
    Type* _rhs;
  }; 

 
  class ProductType
    : public Type
  {
  public:
    ProductType(Type* lhs,Type* rhs)
      : Type(TT_PRODUCT), _lhs(lhs), _rhs(rhs)
    {}
   
    Type* lhs() const {return _lhs;}
   
    Type* rhs() const {return _rhs;}
  private:
   
    Type* _lhs;
   
    Type* _rhs;
  }; 

 
  class UnitStack {
  public:
   
    inline explicit UnitStack()
      : _initial(0),
	_last(&_initial)
    {}

   
    inline void push(Unit* u)
    {
      UnitList* newList = new UnitList(u);
      *_last = newList;
      _last = reinterpret_cast<UnitList**>(&newList->tailReference());
    }

   
    UnitList* list() { return _initial; }

  private:
   
    UnitList* _initial;
   
    UnitList** _last;
  }; 

 
  bool _containsConjecture;
 
  Set<string>* _allowedNames;
 
  Stack<Set<string>*> _allowedNamesStack;
 
  Set<string> _forbiddenIncludes;
 
  istream* _in;
 
  Stack<istream*> _inputs;
 
  Array<char> _chars;
 
  int _gpos;
 
  int _cend;
 
  Array<Token> _tokens;
 
  int _tend;
 
  UnitStack _units;
 
  Stack<State> _states;
  
  int _lastInputType;
  
  int _isQuestion;
 
  Stack<string> _strings;
  
  Stack<int> _connectives;
 
  Stack<bool> _bools;
 
  Stack<int> _ints;
 
  Stack<Formula::VarList*> _varLists;
 
  Stack<Formula::VarList*> _bindLists;
 
  Stack<Tag> _tags;
 
  Stack<Formula*> _formulas;
 
  Stack<Literal*> _literals;
 
  Stack<TermList> _termLists;
 
  IntNameTable _vars;
 
  Stack<Type*> _types;
 
  Stack<TypeTag> _typeTags;
  typedef List<unsigned> SortList;
 
  Map<int,SortList*> _variableSorts;
 
  Set<string> _overflow;
 
  Color _currentColor;

 
  inline char getChar(int pos)
  {
    CALL("TPTP::getChar");

    while (_cend <= pos) {
      int c = _in->get();
      
      _chars[_cend++] = c == -1 ? 0 : c;
    }
    return _chars[pos];
  } 

 
  inline void shiftChars(int n)
  {
    CALL("TPTP::shiftChars");
    ASS(n > 0);
    ASS(n <= _cend);

    for (int i = 0;i < _cend-n;i++) {
      _chars[i] = _chars[n+i];
    }
    _cend -= n;
    _gpos += n;
  } 

 
  inline void resetChars()
  {
    _gpos += _cend;
    _cend = 0;
  } 

 
  inline Token& getTok(int pos)
  {
    CALL("TPTP::getTok");

    while (_tend <= pos) {
      Token& tok = _tokens[_tend++];
      readToken(tok);
    }
    return _tokens[pos];
  } 

 
  inline void shiftToks(int n)
  {
    CALL("TPTP::shiftToks");

    ASS(n > 0);
    ASS(n <= _tend);

    for (int i = 0;i < _tend-n;i++) {
      _tokens[i] = _tokens[n+i];
    }
    _tend -= n;
  } 

 
  inline void resetToks()
  {
    _tend = 0;
  } 

  
  bool readToken(Token& t);
  void skipWhiteSpacesAndComments();
  void readName(Token&);
  void readReserved(Token&);
  void readString(Token&);
  void readAtom(Token&);
  Tag readNumber(Token&);
  int decimal(int pos);
  int positiveDecimal(int pos);
  static string toString(Tag);

  
  static Formula* makeJunction(Connective c,Formula* lhs,Formula* rhs);
  void unitList();
  void fof(bool fo);
  void tff();
  void vampire();
  void consumeToken(Tag);
  string name();
  void formula();
  void atom();
  void simpleFormula();
  void simpleType();
  void args();
  void varList();
  void term();
  void endTerm();
  void endArgs();
  Literal* createEquality(bool polarity,TermList& lhs,TermList& rhs);
  void makeTerm(TermList& ts,Token& tok);
  void midAtom();
  void endEquality();
  void midEquality();
  void endFormula();
  void endType();
  void tag();
  void endFof();
  void endTff();
  void include();
  void type();
  void itef();
  void itet();
  void endItef();
  void endItet();
  void letff();
  void lettf();
  void endLetff();
  void endLetft();
  void endLettf();
  void endLettt();
  void addTagState(Tag);
  static void checkFlat(const TermList& t);
  static void checkFlat(const Term* t);
  static void checkFlat(const Literal* t);
  static void reportNonFlat(string);

  unsigned readSort(bool newSortExpected);
  void bindVariable(int var,unsigned sortNumber);
  void unbindVariables();
  void skipToRPAR();
  unsigned addFunction(string name,int arity,bool& added,TermList& someArgument);
  int addPredicate(string name,int arity,bool& added,TermList& someArgument);
  unsigned addOverloadedFunction(string name,int arity,int symbolArity,bool& added,TermList& arg,
				 Theory::Interpretation integer,Theory::Interpretation rational,
				 Theory::Interpretation real);
  unsigned addOverloadedPredicate(string name,int arity,int symbolArity,bool& added,TermList& arg,
				  Theory::Interpretation integer,Theory::Interpretation rational,
				  Theory::Interpretation real);
  unsigned addIntegerConstant(const string&);
  unsigned addRationalConstant(const string&);
  unsigned addRealConstant(const string&);
  unsigned addUninterpretedConstant(const string& name,bool& added);
  unsigned sortOf(TermList& term);
  static bool higherPrecedence(int c1,int c2);

private:
 
  static DHMap<unsigned, string> _axiomNames;

#if VDEBUG
  void printStates(string extra);
  void printInts(string extra);
  const char* toString(State s);
#endif
}; 

}

#endif

