
/*
 * File ClauseDisposer.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __ClauseDisposer__
#define __ClauseDisposer__

#include "Forwards.hpp"

#include "SATClause.hpp"
#include "TWLSolver.hpp"

namespace SAT {

class ClauseDisposer
{
public:
  typedef SATClause::ActivityType ActivityType;

  ClauseDisposer(TWLSolver& solver) : _solver(solver) {}
  virtual ~ClauseDisposer() {}

 
  virtual void onRestart() {}
 
  virtual void onSafeSpot() {}
  virtual void onNewInputClause(SATClause* cl) {}
  virtual void onConflict() {}
  virtual void onClauseInConflict(SATClause* cl) {}
protected:
  unsigned decisionLevel();
  unsigned varCnt() const;
  SATClauseStack& getLearntStack();
  DArray<WatchStack>& getWatchedStackArray();
  SATClause* getAssignmentPremise(unsigned var);

  void markAllRemovableUnkept();
  void removeUnkept();
  void keepMostActive(size_t numberOfKept, ActivityType minActivity);
  void keepBinary();

  TWLSolver& _solver;
};

class DecayingClauseDisposer : public ClauseDisposer
{
public:
  DecayingClauseDisposer(TWLSolver& solver, ActivityType decayFactor = 1.001)
   : ClauseDisposer(solver), _decayFactor(decayFactor), _inc(1e-30) {}

  virtual void onConflict();

  virtual void onClauseInConflict(SATClause* cl) {
    cl->activity()+=_inc;
  }
protected:
  ActivityType _decayFactor;
  ActivityType _inc;
};

class MinisatClauseDisposer : public DecayingClauseDisposer
{
public:
  MinisatClauseDisposer(TWLSolver& solver, ActivityType decayFactor = 1.001f)
   : DecayingClauseDisposer(solver, decayFactor), _phaseIdx(0), _phaseLen(100), _clauseCntAcc(0), _survivorCnt(0) {}

  virtual void onNewInputClause(SATClause* cl);
  virtual void onSafeSpot();
  virtual void onConflict();
protected:

  size_t _phaseIdx;
  size_t _phaseLen;

  unsigned _clauseCntAcc;
  size_t _survivorCnt;
};

class GrowingClauseDisposer : public DecayingClauseDisposer
{
public:
  GrowingClauseDisposer(TWLSolver& solver, ActivityType decayFactor = 1.001f)
   : DecayingClauseDisposer(solver, decayFactor), _phaseIdx(0), _phaseLen(100), _clauseCntAcc(0), _survivorCnt(0) {}

  virtual void onNewInputClause(SATClause* cl);
  virtual void onRestart();
  virtual void onConflict();
protected:

  size_t _phaseIdx;
  size_t _phaseLen;

  unsigned _clauseCntAcc;
  size_t _survivorCnt;
};

}

#endif 
