
/*
 * File ISSatSweeping.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#include "Lib/Random.hpp"

#include "ISSatSweeping.hpp"

namespace SAT
{





























ISSatSweeping::ISSatSweeping(unsigned varCnt, SATSolver& solver, IntIterator interestingVarIterator,
    bool doRandomSimulation, unsigned conflictLimit, bool collectImplications)
: _doRandomSimulation(doRandomSimulation),
  _conflictUpperLimit(conflictLimit),
  _collectImplications(collectImplications),
  _varCnt(varCnt),
  _interestingVarsSet(varCnt),
  _probingGroupIndex(0),
  _probingElementIndex(0),
  _conflictCountLimit(0),
  _reachedUpperConflictLimit(false),
  _candidateVarPolarities(varCnt),
  _candidateGroupIndexes(varCnt),
  _equivalentVars(varCnt),
  _lastSweepingImplicationCnt(0),
  _solver(solver)
{
  CALL("ISSatSweeping::ISSatSweeping");
  ASS_EQ(solver.getStatus(),SATSolver::SATISFIABLE);
  ASS(!solver.hasAssumptions());

  if(interestingVarIterator.isInvalid()) {
    interestingVarIterator = pvi(getRangeIterator(1,static_cast<int>(varCnt)));
  }

  _interestingVars.loadFromIterator(interestingVarIterator);
  _interestingVarsSet.insertFromIterator(Stack<unsigned>::Iterator(_interestingVars));

  run();
}


bool ISSatSweeping::sameCandGroup(unsigned v1, unsigned v2)
{
  CALL("ISSatSweeping::sameCandGroup");

  return _candidateGroupIndexes.get(v1, _varCnt+1)==_candidateGroupIndexes.get(v2, _varCnt+2);
}

void ISSatSweeping::addTrueLit(SATLiteral lit)
{
  CALL("ISSatSweeping::addTrueLit");

  ALWAYS(_trueVarSet.insert(lit.var()));
  _trueLits.push(lit);
}

void ISSatSweeping::createCandidates()
{
  CALL("ISSatSweeping::createCandidates");
  ASS_EQ(_solver.getStatus(),SATSolver::SATISFIABLE);
  ASS(_candidateGroups.isEmpty());


  _candidateGroups.push(SATLiteralStack());
  SATLiteralStack& candGrp = _candidateGroups.top();
  Stack<unsigned>::Iterator ivit(_interestingVars);
  while(ivit.hasNext()) {
    unsigned i = ivit.next();
    SATSolver::VarAssignment asgn = _solver.getAssignment(i);
    bool candPolarity;
    switch(asgn) {
    case SATSolver::TRUE:
      candPolarity = true;
      break;
    case SATSolver::FALSE:
      candPolarity = false;
      break;
    case SATSolver::DONT_CARE:
      
      continue;
    case SATSolver::NOT_KNOWN: 
    default:
      ASSERTION_VIOLATION;
    }
    _candidateVarPolarities[i] = candPolarity;
    SATLiteral candLit = SATLiteral(i, candPolarity);
    if(_solver.isZeroImplied(i)) {
      addTrueLit(candLit);
      LOG("sat_iss_init_cands", "discovered true literal: "<<candLit);
    }
    else {
      LOG("sat_iss_init_cands", "added candidate literal: "<<candLit);
      candGrp.push(candLit);
      _candidateGroupIndexes.insert(i, 0);
    }
  }

  if(candGrp.size()<2) {
    _candidateGroups.pop();
    ASS(_candidateGroups.isEmpty());
    _unfinishedAmount = 0;
  }
  else {
    _unfinishedAmount = candGrp.size()-1;
  }
}

void ISSatSweeping::splitByCurrAssignment(SATLiteralStack& orig, SATLiteralStack& newGrp)
{
  CALL("ISSatSweeping::splitByCurrAssignment");
  ASS_GE(orig.size(),2);
  ASS_EQ(newGrp.size(),0);
  ASS_EQ(_solver.getStatus(),SATSolver::SATISFIABLE);

  SATLiteralStack::DelIterator oit(orig);
  while(oit.hasNext()) {
    SATLiteral lit = oit.next();
    SATSolver::VarAssignment asgn = _solver.getAssignment(lit.var());
    bool asgnPol;
    switch(asgn) {
    case SATSolver::TRUE:
      asgnPol = true;
      break;
    case SATSolver::FALSE:
      asgnPol = false;
      break;
    case SATSolver::DONT_CARE:
      
      oit.del();
      continue;
    case SATSolver::NOT_KNOWN: 
    default:
      ASSERTION_VIOLATION;
    }
    
    if(asgnPol!=lit.polarity()) {
      oit.del();
      newGrp.push(lit);
    }
  }

  if(newGrp.size()>orig.size()) {
    std::swap(orig,newGrp);
  }
  if(newGrp.size()<2) {
    newGrp.reset();
    if(orig.size()<2) {
      orig.reset();
    }
  }
}

void ISSatSweeping::splitGroupsByCurrAssignment()
{
  CALL("ISSatSweeping::splitGroupsByCurrAssignment");
  ASS(_candidateGroups.isNonEmpty());

  static SATLiteralStack auxNew;
  ASS(auxNew.isEmpty());

  unsigned gi = 0;
  while(gi<_candidateGroups.size()) {
    {
      SATLiteralStack& currGrp = _candidateGroups[gi];

      ASS(auxNew.isEmpty());
      splitByCurrAssignment(currGrp, auxNew);
      ASS_GE(currGrp.size(), auxNew.size());
    }

    if(auxNew.isNonEmpty()) {
      unsigned addedGrpIdx = _candidateGroups.size();
      _candidateGroups.push(SATLiteralStack()); 
      if(gi+1!=addedGrpIdx) {
	ASS_L(gi+1,addedGrpIdx);
	std::swap(_candidateGroups[gi+1],_candidateGroups[addedGrpIdx]);
      }
      std::swap(_candidateGroups[gi+1],auxNew);
      gi += 2;
    }
    else if(_candidateGroups[gi].isEmpty()) {
      swap(_candidateGroups[gi], _candidateGroups.top());
      _candidateGroups.pop();
    }
    else {
      gi++;
    }
  }

  _unfinishedAmount = 0;
  _candidateGroupIndexes.reset();
  unsigned groupCnt = _candidateGroups.size();
  for(gi=0; gi<groupCnt; gi++) {
    LOG("sat_iss_grps", "Group "<<gi<<" (size "<<_candidateGroups[gi].size()<<")");
    SATLiteralStack& grp = _candidateGroups[gi];
    ASS_GE(grp.size(),2);
    _unfinishedAmount += grp.size()-1;
    SATLiteralStack::ConstIterator git(grp);
    while(git.hasNext()) {
      SATLiteral lit = git.next();
      LOG("sat_iss_grps_content", "  "<<lit);
      _candidateGroupIndexes.insert(lit.var(), gi);
    }
  }
}

void ISSatSweeping::tryRandomSimulation()
{
  CALL("ISSatSweeping::tryRandomSimulation");
  ASS(!_solver.hasAssumptions());

  if(_candidateGroups.isEmpty()) {
    return;
  }

  if(_solver.getStatus()!=SATSolver::SATISFIABLE) {
    _solver.addClauses(SATClauseIterator::getEmpty());
  }
  ASS_EQ(_solver.getStatus(), SATSolver::SATISFIABLE);

  unsigned initLives = 3;
  unsigned lives = initLives;

  do {
    unsigned oldUnfinished = _unfinishedAmount;

    LOG("sat_iss_rand_sim", "group cnt: "<<_candidateGroups.size()<<"  unfinishedAmont: "<<_unfinishedAmount);

    _solver.randomizeAssignment();
    splitGroupsByCurrAssignment();

    if(oldUnfinished==_unfinishedAmount) {
      lives--;
    }
    else {
      lives = initLives;
    }
  } while(_unfinishedAmount && lives>0);
  LOG("sat_iss_rand_sim", "random simulation finished at group cnt: "<<_candidateGroups.size()<<"  unfinishedAmont: "<<_unfinishedAmount);
}

void ISSatSweeping::addImplication(Impl imp, bool& foundEquivalence)
{
  CALL("ISSatSweeping::addImplication");
  LOG("sat_iss_impl", "discovered implication: "<<imp.first<<" -> "<<imp.second);

  Impl rev = Impl(imp.second, imp.first);

  if(_implications.find(rev)) {
    foundEquivalence = _equivalentVars.doUnion(imp.first.var(), imp.second.var());
    COND_LOG("sat_iss_equiv", foundEquivalence, "discovered equivalence: "<<imp.first<<" <-> "<<imp.second);
    COND_LOG("sat_iss_impl", !foundEquivalence, "equivalence discovered again: "<<imp.first<<" <-> "<<imp.second);
    _implications.remove(rev);
    onRedundantImplicationDiscovered();
  }
  else {
    _implications.insert(imp);
  }
}

void ISSatSweeping::onRedundantImplicationDiscovered()
{
  CALL("ISSatSweeping::onRedundantImplicationDiscovered");

  if(_implications.size()>1000 && _implications.size()>_lastSweepingImplicationCnt*2) {
    doRedundantImplicationSweeping();
    _lastSweepingImplicationCnt = _implications.size();
  }
}

void ISSatSweeping::doRedundantImplicationSweeping()
{
  CALL("ISSatSweeping::doRedundantImplicationSweeping");

  size_t removedCnt = 0;
  DHSet<Impl>::DelIterator implIt(_implications);
  while(implIt.hasNext()) {
    Impl imp = implIt.next();
    unsigned v1 = imp.first.var();
    unsigned v2 = imp.second.var();
    if(_equivalentVars.root(v1)!=_equivalentVars.root(v2)) {
      

      
      
      if(_collectImplications || sameCandGroup(v1,v2)) {
	continue;
      }
    }
    LOG("sat_iss_redundant_impl_removed","removed impl: "<<imp.first<<" -> "<<imp.second);
    implIt.del();
    removedCnt++;
  }

  LOG("sat_iss_redundant_impl_removal","removed "<<removedCnt<<" redundant implications, remained "<<_implications.size());
}

void ISSatSweeping::lookForImplications(SATLiteral probedLit, bool assignedOpposite,
    bool& foundEquivalence)
{
  CALL("ISSatSweeping::lookForImplications");
  ASS_EQ(probedLit.polarity(),_candidateVarPolarities[probedLit.var()]); 
  LOG("sat_iss_impl_scan","looking for implications with "<<probedLit
			<<" assigned as "<<(assignedOpposite ? "opposite" : "normal"));

  unsigned probedVar = probedLit.var();

  static SATLiteralStack ziStack;
  ziStack.reset();

  _solver.collectZeroImplied(ziStack);

  SATLiteralStack::Iterator ziit(ziStack);
  while(ziit.hasNext()) {
    SATLiteral lit = ziit.next();
    unsigned litVar = lit.var();

    ASS(_solver.isZeroImplied(litVar));
    ASS(_solver.getStatus()==SATSolver::UNKNOWN || _solver.trueInAssignment(lit));

    if(!_interestingVarsSet.find(litVar)) {
      continue;
    }
    if(!_collectImplications && !sameCandGroup(litVar,probedVar)) {
      continue;
    }
    if(litVar==probedVar) {
      ASS_EQ(lit.polarity(),probedLit.polarity()^assignedOpposite);
      continue;
    }
    SATLiteral candLit = SATLiteral(litVar, _candidateVarPolarities[litVar]);
    bool candidateLitTrue = lit==candLit;
    LOG("sat_iss_impl_scan","have propagation implied: "<<lit<<" candidate was "<<candLit);
    if(candidateLitTrue==assignedOpposite) {
      
      
      
      
      
      ASS_REP2(_solver.getStatus()==SATSolver::UNKNOWN || !sameCandGroup(litVar,probedVar), lit, probedLit);
      Impl imp;
      if(assignedOpposite) {
	
	imp = Impl(probedLit.opposite(), candLit);
      }
      else {
	
	imp = Impl(probedLit, candLit.opposite());
      }
      bool foundEq = false;
      addImplication(imp, foundEq);
      ASS(!foundEq);
      continue;
    }
    else {
      Impl imp;
      if(assignedOpposite) {
	  
	  imp = Impl(candLit, probedLit);
      }
      else {
	  
	  imp = Impl(probedLit, candLit);
      }
      addImplication(imp, foundEquivalence);
    }
  }
}

bool ISSatSweeping::tryProvingImplication(Impl imp, bool& foundEquivalence)
{
  CALL("ISSatSweeping::tryProvingImplication");

  if(_implications.find(imp)) {
    LOG("sat_iss_try_impl","implication found in look-up: "<<imp.first<<" -> "<<imp.second);
    return true;
  }
  LOG("sat_iss_try_impl","attempting to prove implication "<<imp.first<<" -> "<<imp.second);
  bool res = tryProvingImplicationInner(imp, foundEquivalence);
  _solver.retractAllAssumptions();
  LOG("sat_iss_try_impl","implication "<<imp.first<<" -> "<<imp.second<<" "<<(res ? "proved" : "disproved"));
  return res;
}

bool ISSatSweeping::tryProvingImplicationInner(Impl imp, bool& foundEquivalence)
{
  CALL("ISSatSweeping::tryProvingImplication");
  ASS(!_solver.hasAssumptions());
  ASS(sameCandGroup(imp.first.var(),imp.second.var()));

  _solver.addAssumption(imp.second.opposite(), _conflictCountLimit);
  SATSolver::Status status = _solver.getStatus();
  if(status==SATSolver::UNSATISFIABLE) {
    addTrueLit(imp.second);
    foundEquivalence = true;
    return false;
  }

  if(status==SATSolver::SATISFIABLE) {
    splitGroupsByCurrAssignment();
  }
  lookForImplications(imp.second, true, foundEquivalence);

  if(foundEquivalence) {
    return false;
  }
  if(_implications.find(imp)) {
    return true;
  }

  if(status==SATSolver::SATISFIABLE && _solver.trueInAssignment(imp.first)) {
    
    
    ASS(!sameCandGroup(imp.first.var(),imp.second.var()));
    return false;
  }

  _solver.addAssumption(imp.first, _conflictCountLimit);
  status = _solver.getStatus();
  if(status==SATSolver::UNSATISFIABLE) {
    LOG("sat_iss_try_impl","assumption of "<<imp.second.opposite()<<" & "<<imp.first<<" unsatisfiable");
    addImplication(imp, foundEquivalence);
    return true;
  }

  if(status==SATSolver::SATISFIABLE) {
    splitGroupsByCurrAssignment();
    ASS(!sameCandGroup(imp.first.var(),imp.second.var()));
  }
  return false;
}

bool ISSatSweeping::doOneProbing()
{
  CALL("ISSatSweeping::doOneProbing");

  SATLiteral cand1, cand2;
  if(!getProbingCandidates(cand1, cand2)) {
    return false;
  }
  ASS_NEQ(_equivalentVars.root(cand1.var()), _equivalentVars.root(cand2.var()));

  bool foundEquivalence = false;
  if(tryProvingImplication(Impl(cand1, cand2), foundEquivalence) && !foundEquivalence) {
    tryProvingImplication(Impl(cand2, cand1), foundEquivalence);
  }
  return true;
}

bool ISSatSweeping::getProbingCandidates(SATLiteral& cand1, SATLiteral& cand2)
{
  CALL("ISSatSweeping::getProbingCandidates");

  if(getProbingCandidatesWithinRotation(cand1, cand2)) {
    return true;
  }
  if(!nextRotation()) {
    return false;
  }
  if(!getProbingCandidatesWithinRotation(cand1, cand2)) {
    
    
    ASS(!nextRotation());
    return false;
  }
  return true;
}

bool ISSatSweeping::nextRotation()
{
  CALL("ISSatSweeping::nextRotation");

  if(_candidateGroups.isEmpty()) {
    return false;
  }
  _probingGroupIndex = 0;
  _probingElementIndex = 0;

  LOG("sat_iss_rot","rotation with "<<_conflictCountLimit<<" finished, "
      <<_candidateGroups.size()<<" candidate groups left, unfinished number is "<<_unfinishedAmount);

  if(_conflictCountLimit) {
    if(_conflictCountLimit<UINT_MAX/2) {
      _conflictCountLimit*=2;
    }
    else {
      _conflictCountLimit=UINT_MAX;
    }
  }
  else {
    _conflictCountLimit++;
  }
  if(_conflictCountLimit>_conflictUpperLimit) {
    _reachedUpperConflictLimit = true;
  }
  return true;
}


bool ISSatSweeping::getProbingCandidatesFromGroup(SATLiteralStack& grp, unsigned firstIndex,
    SATLiteral& cand1, SATLiteral& cand2)
{
  CALL("ISSatSweeping::getProbingCandidatesFromGroup");
  ASS_G(grp.size(),1);

  if(firstIndex>=grp.size()) {
    return false;
  }

  while(_trueLits.find(grp[firstIndex])) {
    if(firstIndex<grp.size()-1) {
      grp[firstIndex] = grp.pop();
    }
    else {
      ASS_EQ(firstIndex,grp.size()-1);
      grp.pop();
      if(grp.size()==1) {
	
	grp.pop();
      }
      return false;
    }
  }
  ASS_GE(grp.size(),2);

  SATLiteral c1 = grp[firstIndex];
  int c1Root = _equivalentVars.root(c1.var());

  
  
  
  
  
  

  unsigned secondIndex = firstIndex+1;
  if(secondIndex>=grp.size()) {
    ASS_EQ(c1,grp.top());
    swap(grp[0], grp.top());
    firstIndex = 0;
    secondIndex = 1;
  }
  while(_trueLits.find(grp[secondIndex]) || _equivalentVars.root(grp[secondIndex].var())==c1Root) {
    if(secondIndex<grp.size()-1) {
      grp[secondIndex] = grp.pop();
    }
    else {
      ASS_EQ(secondIndex,grp.size()-1);
      grp.pop();
      ASS(!grp.isEmpty());
      if(grp.size()==1) {
	grp.pop();
	return false;
      }
      
      
      
      
      ASS_EQ(c1,grp.top());
      swap(grp[0], grp.top());
      firstIndex = 0;
      secondIndex = 1;
    }
  }

  cand1 = c1;
  cand2 = grp[secondIndex];
  ASS_NEQ(cand1,cand2);
  ASS_NEQ(_equivalentVars.root(cand1.var()),_equivalentVars.root(cand2.var()));
  ASS(!_trueLits.find(cand1));
  ASS(!_trueLits.find(cand2));
  ASS(grp.find(cand1));
  ASS(grp.find(cand2));
  return true;
}

bool ISSatSweeping::getProbingCandidatesWithinRotation(SATLiteral& cand1, SATLiteral& cand2)
{
  CALL("ISSatSweeping::getProbingCandidates");

  cand1 = SATLiteral::dummy();

  for(;;) {
    if(_probingGroupIndex>=_candidateGroups.size()) {
      return false;
    }

    SATLiteralStack& currGrp = _candidateGroups[_probingGroupIndex];
    bool found = getProbingCandidatesFromGroup(currGrp, _probingElementIndex, cand1, cand2);
    ASS_NEQ(currGrp.size(), 1); 
    if(found) {
      _probingElementIndex++;
      return true;
    }
    _probingElementIndex = 0;
    if(currGrp.isEmpty()) {
      if(_probingGroupIndex!=_candidateGroups.size()-1) {
	ASS_L(_probingGroupIndex,_candidateGroups.size()-1);
	swap(currGrp, _candidateGroups.top());
      }
      _candidateGroups.pop();
    }
    else {
      _probingGroupIndex++;
    }
  }
}

void ISSatSweeping::run()
{
  CALL("ISSatSweeping::run");

  createCandidates();

  if(_doRandomSimulation) {
    tryRandomSimulation();
  }

  

  while(!_reachedUpperConflictLimit && doOneProbing()) { }

  

  ASS(_reachedUpperConflictLimit || _candidateGroups.isEmpty());

  ASS(_equivStack.isEmpty());

  _equivalentVars.evalComponents();
  IntUnionFind::ComponentIterator cit(_equivalentVars);
  while(cit.hasNext()) {
    IntUnionFind::ElementIterator eit(cit.next());
    ALWAYS(eit.hasNext());
    unsigned v1 = eit.next();
    if(v1==0) {
      
      
      ASS(!eit.hasNext());
      continue;
    }
    unsigned compSz = 1;
    SATLiteral lit1(v1, _candidateVarPolarities[v1]);
    while(eit.hasNext()) {
      unsigned v2 = eit.next();
      SATLiteral lit2(v2, _candidateVarPolarities[v2]);
      _equivStack.push(Equiv(lit1, lit2));
      compSz++;
    }
    COND_LOG_INT("sat_iss_equiv_class_sizes",compSz>1, compSz);
  }
}




}
