
/*
 * File SingleWatchSAT.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#ifndef __SingleWatchSAT__
#define __SingleWatchSAT__

#include "Forwards.hpp"

#include "Lib/DArray.hpp"
#include "Lib/Stack.hpp"
#include "Lib/List.hpp"
#include "Lib/Int.hpp"
#include "Lib/BinaryHeap.hpp"
#include "Lib/Comparison.hpp"
#include "Lib/BucketSorter.hpp"

#include "SATClause.hpp"

namespace SAT {

class SingleWatchSAT
{
public:
 
  enum TerminationReason {
   
    REFUTATION,
   
    SATISFIABLE,
   
    TIME_LIMIT,
   
    MEMORY_LIMIT
  };

  SingleWatchSAT(unsigned varCnt);
  ~SingleWatchSAT();

  bool loadClauses(SATClauseIterator clauses);
  void satisfy(int timeLimit=0);
  void reset();

  SATClauseIterator learntClauses();

  SATClause* refutation;
  TerminationReason termination;


  bool earlyTermination();

  void nextTimeStamp()
  {
    _currTimeStamp++;
    if(_currTimeStamp==0xFFFFFFFF) {
      resetTimeStamp();
    }
  }

  void resetTimeStamp();
  int evalTail(SATClause* cl);
  SATClause* resolveClauses(SATClause* c1, SATClause* c2);
  void incorporate(SATClause* cl);


  typedef DArray<SATClause*> ClauseArray;

  unsigned _currTimeStamp;
  unsigned _varCnt;

  DArray<bool> _asgnVals;

  struct LevelInfo
  {
    LevelInfo() { reset(); }
    void reset()
    {
      resetTimeStamp();
      determinant=0;
    }
    void resetTimeStamp()
    {
      visitTime=0xFFFFFFFF;
      changeTime=0xFFFFFFFF;
      determiningTime=0xFFFFFFFF;
    }
    unsigned visitTime;
    unsigned changeTime;
    unsigned determiningTime;
    SATClause* determinant;
  };
  DArray<LevelInfo> _levelInfos;
  DArray<unsigned> _conflicts;

  typedef Stack<SATClause*> ClauseStack;

  ClauseStack* _watched;
  ClauseStack* _reeval;
  ClauseStack _unitClauses;


  bool isDetermined(unsigned var)
  {
    SATClause* determinant=_levelInfos[var].determinant;
    if(!determinant) {
      return false;
    }
    if(determinant->length()==1) {
      return true;
    }
    unsigned determiningLevel=(*determinant)[1].var();
    if(determiningLevel>_level) {
      return false;
    }
    return _levelInfos[var].determiningTime>=_levelInfos[determiningLevel].visitTime;
  }

  SATClause* handleUnitConflict(SATClause* cl);
  void performDetermination(SATClause* cl);

  void putForReevaluation(SATClause* cl);

  unsigned _level;
  unsigned _lowestUnevaluated;
  int _startTime;
  int _timeLimit;
  int _unusedRemovalCnt;
};

};

#endif
