
/*
 * File TWLSolver.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#ifndef __TWLSolver__
#define __TWLSolver__

#include "Forwards.hpp"

#include "Lib/Array.hpp"
#include "Lib/ArrayMap.hpp"
#include "Lib/DArray.hpp"
#include "Lib/Deque.hpp"
#include "Lib/Exception.hpp"
#include "Lib/ScopedPtr.hpp"
#include "Lib/Stack.hpp"

#include "SATLiteral.hpp"
#include "SATClause.hpp"
#include "SATSolver.hpp"

namespace SAT {

using namespace Lib;
using namespace Shell;

struct Watch
{
  Watch() {}
  Watch(SATClause* cl, SATLiteral blocker) : blocker(blocker), cl(cl)
  {
    CALL("Watch::Watch/2");
    ASS((*cl)[0]==blocker || (*cl)[1]==blocker);
  }
  SATLiteral blocker;
  SATClause* cl;
};

typedef Stack<Watch> WatchStack;


class TWLSolver : public SATSolver {
  friend class ClauseDisposer;
  friend class VariableSelector;
  friend class RLCVariableSelector;
public:
  TWLSolver(const Options& opt, bool generateProofs=false);
  ~TWLSolver();

  virtual void addClauses(SATClauseIterator cit, bool onlyPropagate);
  virtual Status getStatus() { return _status; };
  virtual void ensureVarCnt(unsigned newVarCnt);
  virtual VarAssignment getAssignment(unsigned var);
  virtual bool isZeroImplied(unsigned var);
  virtual void collectZeroImplied(SATLiteralStack& acc);
  virtual SATClause* getZeroImpliedCertificate(unsigned var);

  virtual void addAssumption(SATLiteral lit, unsigned conflictCountLimit);
  virtual void retractAllAssumptions();
  virtual bool hasAssumptions() const { return _assumptionsAdded; }

  virtual SATClause* getRefutation() {
    CALL("TWLSolver::getRefutation");
    ASS_EQ(getStatus(),SATSolver::UNSATISFIABLE);
    return _refutation;
  }

  virtual void randomizeAssignment();

  void assertValid();
  void printAssignment();
private:

  void doSolving(unsigned conflictNumberLimit);

  enum AsgnVal {
    
    
    AS_FALSE = 0u,
    AS_TRUE = 1u,
    AS_UNDEFINED = 2u
  };
  typedef char PackedAsgnVal;

  WatchStack& getWatchStack(SATLiteral lit);
  WatchStack& getWatchStack(unsigned var, unsigned polarity);
  WatchStack& getTriggeredWatchStack(unsigned var, PackedAsgnVal assignment);

  bool isTrue(SATLiteral lit) const;
  bool isFalse(SATLiteral lit) const;
  bool isUndefined(SATLiteral lit) const;

 
  bool isUndefined(unsigned var) const {
    return _assignment[var] == AS_UNDEFINED;
  }
 
  bool isTrue(unsigned var) const {
    return _assignment[var] == AS_TRUE;
  }

  bool isFalse(SATClause* cl) const;
  bool isTrue(SATClause* cl) const;


  unsigned getAssignmentLevel(SATLiteral lit) const;
  unsigned getAssignmentLevel(unsigned var) const;


  unsigned selectTwoNonFalseLiterals(SATClause* cl) const;
  void addClause(SATClause* cl);
  void addUnitClause(SATClause* cl);

  void handleTopLevelConflict(SATClause* cl);
  void handleConflictingAssumption(SATLiteral assumpt);

  void backtrack(unsigned tgtLevel);

  void doBaseLevelPropagation();
  enum SatLoopResult {
    SLR_UNKNOWN,
    SLR_SATISFIABLE,
    SLR_CONFLICT_LIMIT_REACHED
  };
  SatLoopResult runSatLoop(unsigned conflictCountLimit);

  void setAssignment(unsigned var, unsigned polarity);

  void makeAssumptionAssignment(SATLiteral lit);
  void makeChoiceAssignment(unsigned var, unsigned polarity);
  void makeForcedAssignment(SATLiteral lit, SATClause* premise);
  void undoAssignment(unsigned var);

  enum ClauseVisitResult {
   
    VR_CONFLICT,
   
    VR_NONE,
   
    VR_PROPAGATE,
   
    VR_CHANGE_WATCH
  };

  ClauseVisitResult visitWatchedClause(Watch watch, unsigned var, unsigned& litIndex);

  SATClause* propagate(unsigned var);

  void getTwoHighestAssignmentLevels(SATClause* cl, unsigned& highestAL, unsigned& secondHighestAL);
  unsigned getBacktrackLevel(SATClause* conflictClause);
  unsigned getLearntBacktrackLevel(SATClause* conflictClause);

  void doSubsumptionResolution(SATLiteralStack& lits, SATClauseList*& premises);
  void doShallowMinimize(SATLiteralStack& lits, ArraySet& seenVars);
  void doDeepMinimize(SATLiteralStack& lits, ArraySet& seenVars, SATClauseList*& premises);
  bool isRedundant(SATLiteral lit, ArraySet& seenVars, SATClauseList*& premises);
  SATClause* getLearntClause(SATClause* conflictClause);

  void insertIntoWatchIndex(SATClause* cl);

  void recordClauseActivity(SATClause* cl);

  void recordVariableActivity(unsigned var);
  bool chooseVar(unsigned& var);

  void schedulePropagation(unsigned var);
  void resetPropagation();
  bool anythingToPropagate() { return _toPropagate.isNonEmpty(); }
  unsigned pickForPropagation();

 
  struct USRec
  {
    unsigned var;
    unsigned choice : 1;
    unsigned assumption : 1;

    USRec() {}
    USRec(unsigned var, bool choice, bool assumption=false)
    : var(var), choice(choice), assumption(assumption)
    {
      CALL("TWLSolver::USRec::USRec");
      ASS(!assumption || !choice);
    }

  };

  bool _doLearntMinimization;
  bool _doLearntSubsumptionResolution;

  bool _generateProofs;
  SATClause* _refutation;

  Status _status;

  DArray<PackedAsgnVal> _assignment;
  DArray<unsigned> _assignmentLevels;

 
  DArray<SATClause*> _assignmentPremises;

 
  Stack<USRec> _unitStack;
 
  DArray<WatchStack> _windex;

  unsigned _varCnt;

 
  unsigned _level;

 
  bool _assumptionsAdded;
 
  unsigned _assumptionCnt;
 
  bool _unsatisfiableAssumptions;

 
  DArray<PackedAsgnVal> _lastAssignments;

 
  SATClauseStack _learntClauses;

  ArrayMap<EmptyStruct> _propagationScheduled;
  Deque<unsigned> _toPropagate;

  VariableSelectorSCP _variableSelector;
  RestartStrategySCP _restartStrategy;
  ClauseDisposerSCP _clauseDisposer;

  struct UnsatException : public Exception
  {
    UnsatException(SATClause* refutation=0) : refutation(refutation) {}
    SATClause* refutation;
  };

};

}

#endif
