
/*
 * File BSplitter.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/DHSet.hpp"
#include "Lib/Environment.hpp"
#include "Lib/IntUnionFind.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/SharedSet.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/TermIterators.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "BSplitter.hpp"
#include "SaturationAlgorithm.hpp"

#define SP_REPORTS 0

namespace Saturation
{

using namespace Lib;
using namespace Kernel;

BSplitter::BSplitter()
: _nextLev(1)
{
}

void BSplitter::init(SaturationAlgorithm* sa)
{
  CALL("BSplitter::init");

  Splitter::init(sa);
  _addGroundNegation=getOptions().splitAddGroundNegation();
}

bool BSplitter::doSplitting(Clause* cl)
{
  CALL("BSplitter::doSplitting");

  if(!splittingAllowed(cl)) {
    return false;
  }

  Clause* comp;
  comp=getComponent(cl);
  if(!comp) {
    return false;
  }
#if VDEBUG
  assertSplitLevelsExist(cl->splits());
#endif

  SplitLevel lev=_nextLev++;
  SplitRecord* sr=new SplitRecord(cl, comp);

  
  SplitSet::Iterator bsit(*cl->splits());
  while(bsit.hasNext()) {
    SplitLevel blev=bsit.next();
    _db[blev]->dependent.push(lev);
  }

  ASS(!_db[lev]);
  _db[lev]=sr;


  assignClauseSplitSet(comp, SplitSet::getSingleton(lev));
  _sa->addNewClause(comp);

#if SP_REPORTS
  cout<<"Splitting "<<(*cl)<<" into level "<<lev<<", first component "<<(*comp)<<endl;
#endif


  env.statistics->backtrackingSplits++;
  return true;
}

void BSplitter::onClauseReduction(Clause* cl, ClauseIterator premises, Clause* replacement)
{
  CALL("BSplitter::onClauseReduction");
  ASS(cl);

  if(!premises.hasNext()) {
    ASS(!replacement || cl->splits()==replacement->splits());
    return;
  }

  SplitSet* diff=premises.next()->splits();
  while(premises.hasNext()) {
    Clause* premise=premises.next();
    ASS(premise);
    diff=diff->getUnion(premise->splits());
  }
  if(replacement) {
    diff=diff->getUnion(replacement->splits());
  }
  diff=diff->subtract(cl->splits());
  
#if VDEBUG
  assertSplitLevelsExist(diff);
#endif

#if SP_REPORTS==2
  cout<<"Reduced "<<(*cl)<<". Added to reduced stack on levels {"<<diff->toString()<<"}."<<endl;
#endif

  if(diff->isEmpty()) {
    return;
  }

  cl->incReductionTimestamp();
  
  ASS(BDD::instance()->isFalse(cl->prop()));
  SplitSet::Iterator dit(*diff);
  while(dit.hasNext()) {
    SplitLevel slev=dit.next();
    _db[slev]->addReduced(cl);
  }
}

void BSplitter::onNewClause(Clause* cl)
{
  CALL("BSplitter::onNewClause");

  if(!cl->splits()) {
    SplitSet* splits=getNewClauseSplitSet(cl);
    assignClauseSplitSet(cl, splits);
#if SP_REPORTS==2
    cout<<"New clause assigned levels: "<<(*cl)<<endl;
#endif
  }

#if VDEBUG
  assertSplitLevelsExist(cl->splits());
#endif
}

void BSplitter::SplitRecord::addReduced(Clause* cl)
{
  CALL("BSplitter::SplitRecord::addReduced");

  cl->incRefCnt(); 
  reduced.push(ReductionRecord(cl->getReductionTimestamp(),cl));
}

Clause* BSplitter::getComponent(Clause* cl)
{
  CALL("BSplitter::getComponent");

  unsigned clen=cl->length();
  if(clen<=1) {
    return 0;
  }

  unsigned posLits=0;

  
  
  static DHMap<unsigned, unsigned, IdentityHash> varMasters;
  varMasters.reset();
  IntUnionFind components(clen);

  

  for(unsigned i=0;i<clen;i++) {
    Literal* lit=(*cl)[i];
    if(lit->isPositive()) {
      posLits++;
    }
    VariableIterator vit(lit);
    while(vit.hasNext()) {
      unsigned master=varMasters.findOrInsert(vit.next().var(), i);
      if(master!=i) {
	components.doUnion(master, i);
      }
    }
  }
  components.evalComponents();

  int compCnt=components.getComponentCount();

  if(compCnt==1 || (splitPositive() && posLits<=1)) {
    return 0;
  }

  static Stack<Literal*> lits;

  IntUnionFind::ComponentIterator cit(components);


compAssemblyStart:
  if(!cit.hasNext()) {
    return 0;
  }
  lits.reset();

  IntUnionFind::ElementIterator elit=cit.next();

  unsigned compPosLits=0;
  while(elit.hasNext()) {
    int litIndex=elit.next();
    Literal* lit=(*cl)[litIndex];
    lits.push(lit);
    if(lit->isPositive()) {
      compPosLits++;
    }
    if(isAnswerLiteral(lit) || lit->color()!=COLOR_TRANSPARENT) {
      
      goto compAssemblyStart;
    }
  }
  if(splitPositive()) {
    if(compPosLits==posLits) {
      return 0;
    }
    if(compPosLits==0) {
      goto compAssemblyStart; 
    }
  }

  Clause* res=Clause::fromStack(lits, cl->inputType(), new Inference(Inference::BACKTRACKING_SPLITTING_COMPONENT));
  res->setAge(cl->age());
  res->initProp(BDD::instance()->getFalse());
  return res;
}

SplitSet* BSplitter::getNewClauseSplitSet(Clause* cl)
{
  CALL("BSplitter::getNewClauseSplitSet");

  SplitSet* res;
  Inference* inf=cl->inference();
  Inference::Iterator it=inf->iterator();

  if(!stackSplitting()) {
    res=SplitSet::getEmpty();

    while(inf->hasNext(it)) {
      Unit* premu=inf->next(it);
      if(!premu->isClause()) {
	
	continue;
      }
      Clause* prem=static_cast<Clause*>(premu);
      if(!prem->splits()) {
	
	continue;
      }

      res=res->getUnion(prem->splits());
    }
  }
  else {
    SplitLevel maxLev=0;

    while(inf->hasNext(it)) {
      Unit* premu=inf->next(it);
      if(!premu->isClause()) {
	
	continue;
      }
      Clause* prem=static_cast<Clause*>(premu);
      if(!prem->splits()) {
	
	continue;
      }

      if(!prem->splits()->isEmpty()) {
	maxLev=max(maxLev,prem->splits()->sval());
      }
    }

    if(maxLev) {
      res=SplitSet::getSingleton(maxLev);
    }
    else {
      res=SplitSet::getEmpty();
    }
  }
  return res;
}

void BSplitter::assignClauseSplitSet(Clause* cl, SplitSet* splits)
{
  CALL("BSplitter::assignClauseSplitSet");
  ASS(!cl->splits());

  cl->setSplits(splits);

  
  SplitSet::Iterator bsit(*splits);
  while(bsit.hasNext()) {
    SplitLevel slev=bsit.next();
    _db[slev]->children.push(cl);
    cl->incRefCnt();
  }
}

bool BSplitter::handleEmptyClause(Clause* cl)
{
  CALL("BSplitter::handleEmptyClause");

  if(cl->splits()->isEmpty()) {
    return false;
  }
  ASS(BDD::instance()->isFalse(cl->prop()));

  _splitRefutations.push(cl);
  return true;
}

void BSplitter::onAllProcessed()
{
  CALL("BSplitter::onAllProcessed");
  ASS(_sa->clausesFlushed());

  if(_splitRefutations.isNonEmpty()) {
    backtrack(pvi( RCClauseStack::Iterator(_splitRefutations) ));
    _splitRefutations.reset();
  }
}

void BSplitter::backtrack(ClauseIterator emptyClauses)
{
  CALL("BSplitter::backtrack(ClauseIterator)");
  ASS(emptyClauses.hasNext());
  ASS(_sa->clausesFlushed());

  static ClauseStack toDo;
  toDo.reset();
  toDo.loadFromIterator(emptyClauses);

  Clause::requestAux();
  static RCClauseStack trashed;
  static RCClauseStack restored;
  ASS(restored.isEmpty());

start:
  ASS(trashed.isEmpty());

  Clause* cl=toDo.pop();
  ASS(cl->isEmpty());
  LOG_UNIT("bspl_refutations",cl);

  SplitSet* refSplits=cl->splits();
  ASS_G(refSplits->size(),0);

  SplitLevel refLvl=refSplits->maxval(); 
#if SP_REPORTS
  cout<<"Level "<<refLvl<<" refuted\n";
#endif

  SplitSet* backtracked;
  if(stackSplitting()) {
    NOT_IMPLEMENTED;
    backtracked=SplitSet::getRange(refLvl,_nextLev);
  }
  else {
    backtracked=getTransitivelyDependentLevels(refLvl);
  }


  ClauseStack::Iterator tdit(toDo);
  while(tdit.hasNext()) {
    Clause* ecl=tdit.next();
    if(ecl->splits()->hasIntersection(backtracked)) {
      tdit.del();
    }
  }

#if SP_REPORTS
  cout<<"Backtracking on "<<(*cl)<<endl;
  cout<<"base "<<(*_db[refLvl]->base)<<endl;
  cout<<"comp "<<(*_db[refLvl]->component)<<endl;
  cout<<"to be backtracked: "<<backtracked->toString()<<endl;
#endif

  SplitSet* altSplitSet;
  
  getAlternativeClauses(_db[refLvl]->base, _db[refLvl]->component, cl, refLvl, restored, altSplitSet);


  SplitSet::Iterator blit(*backtracked);
  while(blit.hasNext()) {
    SplitLevel bl=blit.next();
    SplitRecord* sr=_db[bl];

    while(sr->children.isNonEmpty()) {
      Clause* ccl=sr->children.pop();
      if(!ccl->hasAux()) {
	ASS(ccl->splits()->member(bl));
	if(ccl->store()!=Clause::BACKTRACKED) {
	  trashed.push(ccl);
	  LOG_UNIT("bspl_rm_backtracked",ccl);
	}
	ccl->setAux(0);
      }
      ccl->decRefCnt();
    }
  }

  SplitSet::Iterator blit2(*backtracked);
  while(blit2.hasNext()) {
    SplitLevel bl=blit2.next();
    SplitRecord* sr=_db[bl];

    Clause* bcl=sr->base;
    if(bl!=refLvl) {
      restored.push(bcl);
    }
    else {
      
      
      
      SplitSet* difSet=altSplitSet->subtract(bcl->splits());
      if(!difSet->isEmpty()) {
	SplitSet::Iterator dsit(*difSet);
	bcl->incReductionTimestamp();
	while(dsit.hasNext()) {
	  SplitLevel slev=dsit.next();
	  ASS(_db[slev]);
	  _db[slev]->addReduced(bcl);
	}
      }
    }

    while(sr->reduced.isNonEmpty()) {
      ReductionRecord rrec=sr->reduced.pop();
      Clause* rcl=rrec.clause;
      if(rrec.timestamp==rcl->getReductionTimestamp()) {
	restored.push(rcl);
      }
      LOG_UNIT("bspl_rm_restored",rcl);
      rcl->decRefCnt(); 
    }

    _db[bl]=0;
    delete sr;
  }

  while(trashed.isNonEmpty()) {
    Clause* tcl=trashed.popWithoutDec();
    if(tcl->store()!=Clause::NONE) {
      _sa->removeActiveOrPassiveClause(tcl);
      ASS_EQ(tcl->store(), Clause::NONE);
    }
    tcl->setStore(Clause::BACKTRACKED);
    tcl->decRefCnt(); 
  }

  if(toDo.isNonEmpty()) {
    goto start; 
  }

  while(restored.isNonEmpty()) {
    Clause* rcl=restored.popWithoutDec();
    if(!rcl->hasAux() && rcl->store()!=Clause::BACKTRACKED) {
      ASS(!rcl->splits()->hasIntersection(backtracked));
      rcl->setAux(0);
      ASS_EQ(rcl->store(), Clause::NONE);
      rcl->incReductionTimestamp();
      rcl->setProp(BDD::instance()->getFalse()); 
      _sa->addNewClause(rcl);
  #if VDEBUG
      
      assertSplitLevelsExist(rcl->splits());
  #endif
#if SP_REPORTS==2
      cout<<"Restored clause "<<(*rcl)<<endl;
#endif

    }
    rcl->decRefCnt(); 
  }

  Clause::releaseAux();

#if SP_REPORTS
  cout<<"-- backtracking done --"<<"\n";
#endif
}

Inference* BSplitter::getAlternativeClauseInference(Clause* base, Clause* firstComp, Clause* refutation)
{
  CALL("BSplitter::getAlternativeClauseInference");
  ASS_EQ(firstComp->splits()->size(), 1); 
  ASS_EQ(refutation->length(), 0); 

  UnitList* premises=0;
  UnitList::push(refutation, premises);
  UnitList::push(firstComp, premises);
  UnitList::push(base, premises);

  return new InferenceMany(Inference::BACKTRACKING_SPLIT_REFUTATION, premises);
}

void BSplitter::getAlternativeClauses(Clause* base, Clause* firstComp, Clause* refutation, SplitLevel refLvl,
    RCClauseStack& acc, SplitSet*& altSplitSet)
{
  CALL("BSplitter::getAlternativeClauses");

  Unit::InputType inp=max(base->inputType(), refutation->inputType());

  SplitSet* resSplits=base->splits()->getUnion(refutation->splits())->subtract(SplitSet::getSingleton(refLvl));
  BDDNode* resProp=BDD::instance()->getFalse(); 
  int resAge=base->age();

  altSplitSet=resSplits;

  env.statistics->backtrackingSplitsRefuted++;
  if(resSplits->isEmpty()) {
    env.statistics->backtrackingSplitsRefutedZeroLevel++;
  }


  static DHSet<Literal*> firstLits;
  static Stack<Literal*> secLits;
  firstLits.reset();
  secLits.reset();

  firstLits.loadFromIterator(Clause::Iterator(*firstComp));
  Clause::Iterator bit(*base);
  while(bit.hasNext()) {
    Literal* l=bit.next();
    if(!firstLits.find(l)) {
      secLits.push(l);
    }
  }
  Inference* sinf=getAlternativeClauseInference(base, firstComp, refutation);
  Clause* scl=Clause::fromStack(secLits, inp, sinf);
  scl->setAge(resAge);
  scl->initProp(resProp);
  assignClauseSplitSet(scl, resSplits);
  acc.push(scl);
  _sa->onParenthood(scl, base);
  _sa->onParenthood(scl, refutation);
#if SP_REPORTS
  cout<<"sp add "<<(*scl)<<endl;
#endif

  if(_addGroundNegation && firstComp->isGround()) {
    
    SplitSet* gndResSplits=refutation->splits()->subtract(SplitSet::getSingleton(refLvl));
    Clause::Iterator fcit(*firstComp);
    while(fcit.hasNext()) {
      Literal* glit=fcit.next();
      Inference* ginf=getAlternativeClauseInference(base, firstComp, refutation);
      Clause* gcl=Clause::fromIterator(getSingletonIterator(Literal::complementaryLiteral(glit)), inp, ginf);
      gcl->setAge(resAge);
      gcl->initProp(resProp);
      assignClauseSplitSet(gcl, gndResSplits);
      acc.push(gcl);
      _sa->onParenthood(gcl, base);
      _sa->onParenthood(gcl, refutation);
#if SP_REPORTS
      cout<<"sp add "<<(*gcl)<<endl;
#endif
    }
  }
}

SplitSet* BSplitter::getTransitivelyDependentLevels(SplitLevel lev)
{
  CALL("BSplitter::getTransitivelyDependentLevels");

  static DHSet<SplitLevel> depSet; 
  static LevelStack depStack;
  static LevelStack toDo;
  depSet.reset();
  depStack.reset();
  toDo.reset();

  toDo.push(lev);
  depSet.insert(lev);
  while(toDo.isNonEmpty()) {
    SplitLevel l=toDo.pop();
    if(!_db[l]) {
      continue;
    }
    depStack.push(l);
    LevelStack::Iterator dit(_db[l]->dependent);
    while(dit.hasNext()) {
      SplitLevel dl=dit.next();
      if(!depSet.find(dl)) {
	toDo.push(dl);
	depSet.insert(dl);
      }
    }
  }
  return SplitSet::getFromArray(depStack.begin(), depStack.size());
}

#if VDEBUG

void BSplitter::assertSplitLevelsExist(SplitSet* s)
{
  CALL("BSplitter::assertSplitLevelsExist");

  SplitSet::Iterator sit(*s);
  while(sit.hasNext()) {
    SplitLevel lev=sit.next();
    ASS_REP(_db[lev]!=0, lev);
  }
}

#endif

}
