
/*
 * File SWBSplitter.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/DArray.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/Environment.hpp"
#include "Lib/IntUnionFind.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "Indexing/TermSharing.hpp"
#include "Inferences/PropositionalToBDDISE.hpp"

#include "SaturationAlgorithm.hpp"

#include "SWBSplitter.hpp"

namespace Saturation
{

using namespace Lib;
using namespace Kernel;
using namespace Inferences;

bool SWBSplitter::doSplitting(Clause* cl)
{
  CALL("SWBSplitter::doSplitting");
  ASS(cl->noSplits());

  if(!splittingAllowed(cl)) {
    return false;
  }

  unsigned clen=cl->length();
  ASS_G(clen,0);

  if(clen<=1) {
    return handleNoSplit(cl);
  }

  
  
  static DHMap<unsigned, unsigned, IdentityHash> varMasters;
  varMasters.reset();
  IntUnionFind components(clen);
  
  int coloredMaster=-1;

  for(unsigned i=0;i<clen;i++) {
    Literal* lit=(*cl)[i];
    if( !canSplitOut(lit) ) {
      if(coloredMaster==-1) {
	coloredMaster=i;
      } else {
	
	components.doUnion(coloredMaster, i);
      }
    }
    VariableIterator vit(lit);
    while(vit.hasNext()) {
      unsigned master=varMasters.findOrInsert(vit.next().var(), i);
      if(master!=i) {
	components.doUnion(master, i);
      }
    }
  }
  components.evalComponents();

  unsigned compCnt=components.getComponentCount();

  if(standAloneObligations() && compCnt>1) {

    
    
    

    IntUnionFind::ComponentIterator cit(components);

    int someCompEl=-1;
    bool someCompOK=false;
    while(cit.hasNext()) {
      IntUnionFind::ElementIterator elit=cit.next();

      int compEl=elit.next();
      if(someCompEl==-1) {
	someCompEl=compEl;
      }
      bool saok=false; 
      for(;;) {
	if(canStandAlone((*cl)[compEl])) {
	  saok=true;
	  break;
	}
	if(!elit.hasNext()) {
	  break;
	}
	compEl=elit.next();
      }
      if(!saok || !someCompOK) {
	components.doUnion(compEl, someCompEl);
	if(saok) {
	  someCompOK=true;
	}
      }
    }

    
    components.evalComponents();
    compCnt=components.getComponentCount();
  }

  if(compCnt==1) {
    return handleNoSplit(cl);
  }

  env.statistics->splitClauses++;
  env.statistics->splitComponents+=compCnt;

  static DArray<Literal*> lits;
  lits.ensure(clen);
  unsigned litsNextAvail=0;

  static DArray<CompRec> comps;
  comps.ensure(compCnt);
  unsigned compi=0;

  int masterIndex=-1;

  IntUnionFind::ComponentIterator cit(components);
  ASS(cit.hasNext());
  while(cit.hasNext()) {
    IntUnionFind::ElementIterator elit=cit.next();

    bool master=false;
    Literal** first=&lits[litsNextAvail];
    int compLen=0;

    while(elit.hasNext()) {
      int litIndex=elit.next();
      lits[litsNextAvail++]=(*cl)[litIndex];
      compLen++;
      if(litIndex==coloredMaster) {
	master=true;
      }
    }

    if(master) {
      masterIndex=compi;
    }
    comps[compi++]=CompRec(first, compLen);
  }
  ASS_EQ(compi,compCnt);
  ASS_EQ(litsNextAvail,clen);
  bool haveMaster=masterIndex!=-1;
  if(haveMaster) {
    swap(comps[0], comps[masterIndex]);
  }

  buildAndInsertComponents(cl, comps.array(), compCnt, haveMaster);
  return true;
}

bool SWBSplitter::canSplitOut(Literal* lit)
{
  Signature::Symbol* predSym = env.signature->getPredicate(lit->functor());

  return PropositionalToBDDISE::colorOkForBDD(lit->color()) &&
    (!getOptions().showSymbolElimination() || lit->skip()) &&
    (!predSym->cfName()) && (!isAnswerLiteral(lit));
}

}
