
/*
 * File SWBSplitterWithBDDs.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Term.hpp"

#include "Indexing/TermSharing.hpp"

#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "SaturationAlgorithm.hpp"

#include "SWBSplitterWithBDDs.hpp"

namespace Saturation
{

void SWBSplitterWithBDDs::buildAndInsertComponents(Clause* cl, const CompRec* comps, unsigned compCnt, bool firstIsMaster)
{
  CALL("SWBSplitterWithBDDs::buildAndInsertComponents");

#if REPORT_SPLITS
  cout<<"####Split####\n";
  cout<<(*cl)<<endl;
  cout<<"vvv Into vvv\n";
#endif

  BDD* bdd=BDD::instance();

  static Stack<Clause*> unnamedComponentStack(16);
  unnamedComponentStack.reset();

  typedef pair<int, Clause*> CompNameRec;
  
  
  
  static Stack<CompNameRec> namedComponents(16);
  namedComponents.reset();



  int remainingComps=compCnt;


  Clause* masterComp=0;
  for(unsigned ci=0;ci<compCnt;ci++) {
    CompRec cr=comps[ci];

    remainingComps--;

    int compName;
    bool compNew;
    Clause* comp=getComponent(cl, cr, compName, compNew);
    ASS(!compNew || !compName); 

    bool isMaster= (ci==0 && firstIsMaster) || 
      (!masterComp &&			       
	  (compNew ||				
	   (!remainingComps && unnamedComponentStack.isEmpty())	
	  )
      );

    if(isMaster) {
      masterComp=comp;
    }
    else if(compName) {
#if REPORT_SPLITS
      cout<<compName<<": "<<(*comp)<<endl;
#endif
      namedComponents.push(CompNameRec(compName, comp));
    }
    else {
      unnamedComponentStack.push(comp);
    }
  }

  if(!masterComp) {
    ASS(unnamedComponentStack.isNonEmpty());
    masterComp=unnamedComponentStack.pop();
  }


  Stack<Clause*>::Iterator unnamedIt(unnamedComponentStack);
  while(unnamedIt.hasNext()) {
    Clause* comp=unnamedIt.next();
    if(comp==masterComp) {
      
      
      
      continue;
    }

    
    int compName=nameComponent(comp);
    if(!compName) {
      
      
      continue;
    }

    namedComponents.push(CompNameRec(compName, comp));


#if REPORT_SPLITS
    cout<<'n'<<compName<<": "<<(*comp)<<endl;
#endif
  }

  InferenceStore::SplittingRecord* srec=new InferenceStore::SplittingRecord(cl);

  static Stack<UnitSpec> masterPremises;
  masterPremises.reset();
  masterPremises.push(UnitSpec(cl));

  BDDNode* newMasterProp=cl->prop();

  while(namedComponents.isNonEmpty()) {
    CompNameRec cr=namedComponents.pop();
    int compName=cr.first;
    Clause* comp=cr.second;

    newMasterProp=bdd->disjunction(newMasterProp, getNameProp(compName));
    if(bdd->isTrue(newMasterProp)) {
      
      
      
      delete srec;
      _sa->onClauseReduction(cl, 0, 0);
#if REPORT_SPLITS
      cout<<"obtained tautology"<<endl;
      cout<<"^^^^^^^^^^^^\n";
#endif
      return;
    }

    
    
    
    UnitSpec compDefPremise=UnitSpec(comp, getNameProp(-compName));
    masterPremises.push(compDefPremise);
    srec->namedComps.push(make_pair(compName, comp));
  }

  UnitSpec splittedMCompCS(masterComp, newMasterProp);

  srec->result=splittedMCompCS;
  InferenceStore::instance()->recordSplitting(srec, masterPremises.size(),
	  masterPremises.begin());

  BDDNode* oldProp=masterComp->prop();
  masterComp->setProp( bdd->conjunction(oldProp, newMasterProp) );
  if(oldProp!=masterComp->prop() && newMasterProp!=masterComp->prop()) {
    InferenceStore::instance()->
      recordMerge(masterComp, oldProp, newMasterProp, masterComp->prop());
  }

  ASS(!bdd->isTrue(masterComp->prop()));

#if REPORT_SPLITS
  cout<<(*masterComp)<<endl;
  cout<<"^^^^^^^^^^^^\n";
#endif

  if(oldProp!=masterComp->prop()) {
    _sa->addNewClause(masterComp);
    _sa->onParenthood(masterComp, cl);
  }
  while(unnamedComponentStack.isNonEmpty()) {
    Clause* comp=unnamedComponentStack.pop();
    _sa->onParenthood(comp, cl);
  }

  _sa->onClauseReduction(cl, 0, 0);
}


bool SWBSplitterWithBDDs::handleNoSplit(Clause* cl)
{
  CALL("SWBSplitterWithBDDs::handleNoSplit");

  return _sa->getSharing()->doSharing(cl);
}

Clause* SWBSplitterWithBDDs::getComponent(Clause* cl, const CompRec& cr, int& name, bool& newComponent)
{
  CALL("SWBSplitterWithBDDs::getComponent");

  Literal* const * lits=cr.array();
  unsigned compLen=cr.size();

  Clause* comp=_sa->getSharing()->tryGet(lits,compLen);
  newComponent=!comp;
  if(comp) {
    if(!_clauseNames.find(comp, name)) {
      name=0;
    }
#if VDEBUG
    else {
      ASS_NEQ(name,0);
    }
#endif
  } else {
    name=0;

    Inference* inf=new Inference(Inference::TAUTOLOGY_INTRODUCTION);
    Unit::InputType inpType = cl->inputType();
    comp = new(compLen) Clause(compLen, inpType, inf);
    for(unsigned i=0;i<compLen;i++) {
      (*comp)[i]=lits[i];
    }

    comp->setAge(cl->age());
    comp->initProp(BDD::instance()->getTrue());
    InferenceStore::instance()->recordNonPropInference(comp);

    _sa->getSharing()->insertNew(comp);
  }
  return comp;
}

int SWBSplitterWithBDDs::nameComponent(Clause* comp)
{
  CALL("SWBSplitterWithBDDs::nameComponent");

  int *pname;

  if(!_clauseNames.getValuePtr(comp, pname)) {
    
    return 0;
  }

  int compName=0;
  bool newlyIntroduced=true;

  if(comp->length()==1 && (*comp)[0]->ground()) {
    

    Literal* lit=(*comp)[0];
    Literal* opLit=env.sharing->tryGetOpposite(lit);
    
    if(opLit && _groundNames.pop(opLit, compName)) {
      
      

      
      compName=-compName;
      newlyIntroduced=false;
    }
    else {
      unsigned newVar = BDD::instance()->getNewVar();
      compName = (lit->polarity()?1:-1) * newVar;
      env.statistics->splittingNamesIntroduced++;
      _groundNames.insert(lit, compName);
      _sa->getProblem().addBDDVarMeaning(newVar, Problem::BDDMeaningSpec(Literal::positiveLiteral(lit), 0));
    }
  }
  else {
    compName=BDD::instance()->getNewVar();
    env.statistics->splittingNamesIntroduced++;
    comp->incRefCnt();
    _sa->getProblem().addBDDVarMeaning(compName, Problem::BDDMeaningSpec(0, comp));
  }

  *pname=compName;
  BDDNode* oldCompProp=comp->prop();
  BDDNode* nameProp=BDD::instance()->getAtomic(abs(compName), compName<0);
  InferenceStore::instance()->recordIntroduction(comp, nameProp, Inference::SPLITTING_COMPONENT);

  BDDNode* newCompProp=BDD::instance()->conjunction(oldCompProp, nameProp);
  if(newCompProp!=nameProp && newCompProp!=oldCompProp) {
    InferenceStore::instance()->recordMerge(comp, oldCompProp, nameProp, newCompProp);
  }
  comp->setProp(newCompProp);

  _sa->addNewClause(comp);

  if(getOptions().showDefinitions() && newlyIntroduced) {
    env.beginOutput();
    env.out() << "Definition: ";
    if(compName<0) {
      env.out() << "~";
    }
    env.out() << BDD::instance()->getPropositionalPredicateName(abs(compName))
      << " <=> (" << comp->nonPropToString() << ")" << endl;
    env.endOutput();
  }
  return compName;
}

BDDNode* SWBSplitterWithBDDs::getNameProp(int name)
{
  CALL("SWBSplitterWithBDDs::getNameProp");
  ASS_NEQ(name,0);

  return BDD::instance()->getAtomic(abs(name), name>0);
}

}
