
/*
 * File SaturationAlgorithm.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "Lib/DHSet.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/SharedSet.hpp"
#include "Lib/Stack.hpp"
#include "Lib/Timer.hpp"
#include "Lib/VirtualIterator.hpp"

#include "Indexing/LiteralIndexingStructure.hpp"

#include "Inferences/BDDMarkingSubsumption.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/BDDConjunction.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/ColorHelper.hpp"
#include "Kernel/EqHelper.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/KBO.hpp"
#include "Kernel/LiteralSelector.hpp"
#include "Kernel/MLVariant.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/SubformulaIterator.hpp"
#include "Kernel/Unit.hpp"

#include "Inferences/InferenceEngine.hpp"
#include "Inferences/BackwardDemodulation.hpp"
#include "Inferences/BackwardSubsumptionResolution.hpp"
#include "Inferences/BDDMarkingSubsumption.hpp"
#include "Inferences/BinaryResolution.hpp"
#include "Inferences/CTFwSubsAndRes.hpp"
#include "Inferences/EqualityFactoring.hpp"
#include "Inferences/EqualityResolution.hpp"
#include "Inferences/Factoring.hpp"
#include "Inferences/ForwardDemodulation.hpp"
#include "Inferences/ForwardLiteralRewriting.hpp"
#include "Inferences/ForwardSubsumptionAndResolution.hpp"
#include "Inferences/GlobalSubsumption.hpp"
#include "Inferences/HyperSuperposition.hpp"
#include "Inferences/InterpretedSimplifier.hpp"
#include "Inferences/RefutationSeekerFSE.hpp"
#include "Inferences/SLQueryForwardSubsumption.hpp"
#include "Inferences/SLQueryBackwardSubsumption.hpp"
#include "Inferences/Superposition.hpp"
#include "Inferences/URResolution.hpp"

#include "Shell/AnswerExtractor.hpp"
#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "BSplitter.hpp"
#include "SSplitter.hpp"
#include "SWBSplitterWithBDDs.hpp"
#include "SWBSplitterWithoutBDDs.hpp"

#include "ConsequenceFinder.hpp"
#include "Splitter.hpp"
#include "SymElOutput.hpp"
#include "SaturationAlgorithm.hpp"
#include "AWPassiveClauseContainer.hpp"
#include "Discount.hpp"
#include "LRS.hpp"
#include "Otter.hpp"

using namespace Lib;
using namespace Kernel;
using namespace Shell;
using namespace Saturation;

#define REPORT_CONTAINERS 0
#define REPORT_FW_SIMPL 0
#define REPORT_BW_SIMPL 0


SaturationAlgorithm* SaturationAlgorithm::s_instance = 0;

SaturationAlgorithm::SaturationAlgorithm(Problem& prb, const Options& opt)
  : MainLoop(prb, opt),
    _mergedBddEmptyClause(0),
    _limits(opt),
    _clauseActivationInProgress(false),
    _fwSimplifiers(0), _bwSimplifiers(0), _splitter(0),
    _propToBDDConv(this),
    _consFinder(0), _symEl(0), _bddMarkingSubsumption(0), _answerLiteralManager(0),
    _generatedClauseCount(0)
{
  CALL("SaturationAlgorithm::SaturationAlgorithm");
  ASS_EQ(s_instance, 0);  

  _ordering = OrderingSP(Ordering::create(prb, opt));
  if(!Ordering::trySetGlobalOrdering(_ordering)) {
    
    cerr << "SaturationAlgorithm cannot set its ordering as global" << endl;
  }
  _selector = LiteralSelector::getSelector(*_ordering, opt, opt.selection());

  _propToBDD = opt.propositionalToBDD();
  _completeOptionSettings = opt.complete(prb);

  _unprocessed = new UnprocessedClauseContainer();
  _passive = new AWPassiveClauseContainer(opt);
  _active = new ActiveClauseContainer(opt);

  _active->attach(this);
  _passive->attach(this);

  _active->addedEvent.subscribe(this, &SaturationAlgorithm::onActiveAdded);
  _active->removedEvent.subscribe(this, &SaturationAlgorithm::activeRemovedHandler);
  _passive->addedEvent.subscribe(this, &SaturationAlgorithm::onPassiveAdded);
  _passive->removedEvent.subscribe(this, &SaturationAlgorithm::passiveRemovedHandler);
  _passive->selectedEvent.subscribe(this, &SaturationAlgorithm::onPassiveSelected);
  _unprocessed->addedEvent.subscribe(this, &SaturationAlgorithm::onUnprocessedAdded);
  _unprocessed->removedEvent.subscribe(this, &SaturationAlgorithm::onUnprocessedRemoved);
  _unprocessed->selectedEvent.subscribe(this, &SaturationAlgorithm::onUnprocessedSelected);

  if(opt.maxWeight()) {
    _limits.setLimits(0,opt.maxWeight());
  }

  s_instance=this;
}

SaturationAlgorithm::~SaturationAlgorithm()
{
  CALL("SaturationAlgorithm::~SaturationAlgorithm");
  ASS_EQ(s_instance,this);

  s_instance=0;

  if(_splitter) {
    delete _splitter;
  }
  if(_consFinder) {
    delete _consFinder;
  }
  if(_symEl) {
    delete _symEl;
  }
  if(_bddMarkingSubsumption) {
    delete _bddMarkingSubsumption;
  }

  _active->detach();
  _passive->detach();

  if(_generator) {
    _generator->detach();
  }
  if(_immediateSimplifier) {
    _immediateSimplifier->detach();
  }

  while(_fwSimplifiers) {
    ForwardSimplificationEngine* fse = FwSimplList::pop(_fwSimplifiers);
    fse->detach();
    delete fse;
  }
  while(_bwSimplifiers) {
    BackwardSimplificationEngine* bse = BwSimplList::pop(_bwSimplifiers);
    bse->detach();
    delete bse;
  }

  delete _unprocessed;
  delete _active;
  delete _passive;
}

void SaturationAlgorithm::tryUpdateFinalClauseCount()
{
  CALL("SaturationAlgorithm::tryUpdateFinalClauseCount");

  SaturationAlgorithm* inst = tryGetInstance();
  if(!inst) {
    return;
  }
  env.statistics->finalActiveClauses = inst->_active->size();
  env.statistics->finalPassiveClauses = inst->_passive->size();
}

bool SaturationAlgorithm::isComplete()
{
  return _completeOptionSettings;
}

ClauseIterator SaturationAlgorithm::activeClauses()
{
  CALL("SaturationAlgorithm::activeClauses");

  LiteralIndexingStructure* gis=getIndexManager()->getGeneratingLiteralIndexingStructure();
  return pvi( getMappingIterator(gis->getAll(), SLQueryResult::ClauseExtractFn()) );
}

ClauseIterator SaturationAlgorithm::passiveClauses()
{
  return _passive->iterator();
}

size_t SaturationAlgorithm::activeClauseCount()
{
  return _active->size();
}

size_t SaturationAlgorithm::passiveClauseCount()
{
  return _passive->size();
}


void SaturationAlgorithm::onActiveAdded(Clause* c)
{
  LOG_UNIT("sa_active_added", c);
  LOG_INT("sa_active_size", _active->size());
}

void SaturationAlgorithm::onActiveRemoved(Clause* c)
{
  CALL("SaturationAlgorithm::onActiveRemoved");

  LOG_UNIT("sa_active_removed", c);
  LOG_INT("sa_active_size", _active->size());

  ASS(c->store()==Clause::ACTIVE || c->store()==Clause::REACTIVATED ||
      c->store()==Clause::SELECTED_REACTIVATED);

  if(c->store()==Clause::ACTIVE) {
    c->setStore(Clause::NONE);
    
  } else if(c->store()==Clause::REACTIVATED) {
    c->setStore(Clause::PASSIVE);
  } else if(c->store()==Clause::SELECTED_REACTIVATED) {
    c->setStore(Clause::SELECTED);
  }
#if VDEBUG
  else {
    ASSERTION_VIOLATION;
  }
#endif
}

void SaturationAlgorithm::onAllProcessed()
{
  CALL("SaturationAlgorithm::onAllProcessed");
  ASS(clausesFlushed());

  if(_symEl) {
    _symEl->onAllProcessed();
  }

  if(_splitter) {
    _splitter->onAllProcessed();
  }

  if(_bddMarkingSubsumption) {
    _bddMarkingSubsumption->onAllProcessed();
  }

  if(_consFinder) {
    _consFinder->onAllProcessed();
  }
}

void SaturationAlgorithm::onPassiveAdded(Clause* c)
{
  LOG_UNIT("sa_passive_added", c);
  LOG_INT("sa_passive_size", _passive->size());

  
  
  onNonRedundantClause(c);
}

void SaturationAlgorithm::onPassiveRemoved(Clause* c)
{
  CALL("SaturationAlgorithm::onPassiveRemoved");

  LOG_UNIT("sa_passive_removed", c);
  LOG_INT("sa_passive_size", _passive->size());

  ASS(c->store()==Clause::PASSIVE || c->store()==Clause::REACTIVATED)

  if(c->store()==Clause::PASSIVE) {
    c->setStore(Clause::NONE);
    
  } else if(c->store()==Clause::REACTIVATED) {
    c->setStore(Clause::ACTIVE);
  }
#if VDEBUG
  else {
    ASSERTION_VIOLATION;
  }
#endif
}

void SaturationAlgorithm::onPassiveSelected(Clause* c)
{
  LOG_UNIT("sa_passive_selected", c);
  LOG_INT("sa_passive_size", _passive->size());
}

void SaturationAlgorithm::onUnprocessedAdded(Clause* c)
{
  LOG_UNIT("sa_unprocessed_added", c);
}

void SaturationAlgorithm::onUnprocessedRemoved(Clause* c)
{
  LOG_UNIT("sa_unprocessed_removed", c);
}

void SaturationAlgorithm::onUnprocessedSelected(Clause* c)
{
  LOG_UNIT("sa_unprocessed_selected", c);
}

void SaturationAlgorithm::onNewClause(Clause* cl)
{
  CALL("SaturationAlgorithm::onNewClause");

  if(_splitter) {
    _splitter->onNewClause(cl);
  }

  if(!cl->prop()) {
    BDD* bdd=BDD::instance();
    BDDNode* prop=bdd->getFalse();

    Inference* inf=cl->inference();
    Inference::Iterator it=inf->iterator();
    while(inf->hasNext(it)) {
      Unit* premu=inf->next(it);
      if(!premu->isClause()) {
	
	continue;
      }
      Clause* prem=static_cast<Clause*>(premu);
      if(!prem->prop()) {
	
	continue;
      }

      prop=bdd->disjunction(prop, prem->prop());
    }

    cl->initProp(prop);
    if(!bdd->isTrue(prop)) {
      InferenceStore::instance()->recordNonPropInference(cl);
    }
  }

  LOG_UNIT("sa_new_clause", cl);

  if(!_propToBDD && cl->isPropositional()) {
    onNewUsefulPropositionalClause(cl);
  }

  if(_answerLiteralManager) {
    _answerLiteralManager->onNewClause(cl);
  }
}

void SaturationAlgorithm::onNewUsefulPropositionalClause(Clause* c)
{
  CALL("SaturationAlgorithm::onNewUsefulPropositionalClause");
  ASS(c->isPropositional());

  LOG_UNIT("sa_new_prop_clause", c);

  if(_bddMarkingSubsumption) {
    _bddMarkingSubsumption->onNewPropositionalClause(c);
  }

  if(_consFinder) {
    _consFinder->onNewPropositionalClause(c);
  }
}

void SaturationAlgorithm::onClauseRetained(Clause* cl)
{
  CALL("SaturationAlgorithm::onClauseRetained");

  LOG_UNIT("sa_retained_clause", cl);
}

void SaturationAlgorithm::onClauseReduction(Clause* cl, Clause* replacement,
    Clause* premise, Clause* reductionPremise, bool forward)
{
  CALL("SaturationAlgorithm::onClauseReduction/5");
  ASS(cl);

  ClauseIterator premises;
  
  if(reductionPremise) {
    ASS(premise);
    premises = pvi( getConcatenatedIterator(getSingletonIterator(premise), getSingletonIterator(reductionPremise)) );
  }
  else if(premise) {
    premises = pvi( getSingletonIterator(premise) );
  }
  else {
    premises=ClauseIterator::getEmpty();
  }

  onClauseReduction(cl, replacement, premises, forward);
}

void SaturationAlgorithm::onClauseReduction(Clause* cl, Clause* replacement,
    ClauseIterator premises, bool forward)
{
  CALL("SaturationAlgorithm::onClauseReduction/4");
  ASS(cl);

  if(replacement && BDD::instance()->isTrue(replacement->prop())) {
    
    
    replacement=0;
  }

  static ClauseStack premStack;
  premStack.reset();
  premStack.loadFromIterator(premises);

  if(_splitter) {
    _splitter->onClauseReduction(cl, pvi( ClauseStack::Iterator(premStack) ), replacement);
  }

  if(replacement) {
    onParenthood(replacement, cl);
    while(premStack.isNonEmpty()) {
      onParenthood(replacement, premStack.pop());
    }
  }
}


void SaturationAlgorithm::onNonRedundantClause(Clause* c)
{
  CALL("SaturationAlgorithm::onNonRedundantClause");

  if(_symEl) {
    _symEl->onNonRedundantClause(c);
  }
}

void SaturationAlgorithm::onParenthood(Clause* cl, Clause* parent)
{
  CALL("SaturationAlgorithm::onParenthood");

  if(_symEl) {
    _symEl->onParenthood(cl, parent);
  }
}

void SaturationAlgorithm::activeRemovedHandler(Clause* cl)
{
  CALL("SaturationAlgorithm::activeRemovedHandler");

  onActiveRemoved(cl);
}

void SaturationAlgorithm::passiveRemovedHandler(Clause* cl)
{
  CALL("SaturationAlgorithm::passiveRemovedHandler");

  onPassiveRemoved(cl);
}

int SaturationAlgorithm::elapsedTime()
{
  return env.timer->elapsedMilliseconds()-_startTime;
}

void SaturationAlgorithm::addInputClause(Clause* cl)
{
  CALL("SaturationAlgorithm::addInputClause");
  ASS_EQ(cl->prop(),0);

  cl->markInput();
  cl->initProp(BDD::instance()->getFalse());

  if(_symEl) {
    _symEl->onInputClause(cl);
  }

  if(_propToBDD) {
    
    cl=_propToBDDConv.simplify(cl);
    if(!cl) {
      return;
    }
  }

  if(_opt.sos() != Options::SOS_OFF && cl->inputType()==Clause::AXIOM) {
    addInputSOSClause(cl);
  } else {
    addNewClause(cl);
  }

  env.statistics->initialClauses++;
}

LiteralSelector& SaturationAlgorithm::getSosLiteralSelector()
{
  CALL("SaturationAlgorithm::getSosLiteralSelector");

  if(_opt.sos() == Options::SOS_ALL) {
    if(!_sosLiteralSelector) {
      _sosLiteralSelector = new TotalLiteralSelector(getOrdering(), getOptions());
    }
    return *_sosLiteralSelector;
  }
  else {
    return *_selector;
  }
}

void SaturationAlgorithm::addInputSOSClause(Clause* cl)
{
  CALL("SaturationAlgorithm::addInputSOSClause");
  ASS_EQ(cl->inputType(),Clause::AXIOM);

  
  
  cl->incRefCnt();

  onNewClause(cl);

simpl_start:

  Clause* simplCl=_immediateSimplifier->simplify(cl);
  if(simplCl!=cl) {
    if(!simplCl) {
      onClauseReduction(cl, 0, 0);
      goto fin;
    }

    simplCl->incRefCnt();
    cl->decRefCnt(); 

    onNewClause(simplCl);
    onClauseReduction(cl, simplCl, 0);
    cl=simplCl;
    goto simpl_start;
  }

  if(cl->isEmpty()) {
    addNewClause(cl);
    goto fin;
  }

  ASS(!cl->selected());
  {
    LiteralSelector& sosSelector = getSosLiteralSelector();
    sosSelector.select(cl);
  }

  cl->setStore(Clause::ACTIVE);
  env.statistics->activeClauses++;
  _active->add(cl);

  onSOSClauseAdded(cl);

fin:
  cl->decRefCnt();
}


void SaturationAlgorithm::init()
{
  CALL("SaturationAlgorithm::init");

  ClauseIterator toAdd = _prb.clauseIterator();

  while(toAdd.hasNext()) {
    Clause* cl=toAdd.next();
    addInputClause(cl);
  }

  _sharing.init(this);
  if(_splitter) {
    _splitter->init(this);
  }
  if(_consFinder) {
    _consFinder->init(this);
  }
  if(_symEl) {
    _symEl->init(this);
  }
  if(_bddMarkingSubsumption) {
    _bddMarkingSubsumption->init(this);
  }

  _startTime=env.timer->elapsedMilliseconds();
}

class SaturationAlgorithm::TotalSimplificationPerformer
: public ForwardSimplificationPerformer
{
public:
  TotalSimplificationPerformer(SaturationAlgorithm* sa, Clause* cl) : _sa(sa), _cl(cl) {}

  void perform(ClauseIterator premises, Clause* replacement)
  {
    CALL("TotalSimplificationPerformer::perform");
    ASS(_cl);

    BDD* bdd=BDD::instance();

    BDDNode* oldClProp=_cl->prop();

    LOG_UNIT("sa_fw_simpl_red_clause",_cl);
    TRACE("sa_fw_simpl",
	tout << "->>--------\n";
	ClauseList* lst=0;
	while(premises.hasNext()) {
	  Clause* premise=premises.next();
	  ASS(willPerform(premise));
	  ClauseList::push(premise, lst);
	  tout << ":" << (*premise) << endl;
	}
	cout << "-" << (*_cl) << endl;
	premises=pvi( ClauseList::DestructiveIterator(lst) );
    );

    if(replacement) {
      replacement->initProp(oldClProp);
      InferenceStore::instance()->recordNonPropInference(replacement);
      _sa->addNewClause(replacement);
    }
    _sa->onClauseReduction(_cl, replacement, premises);

    _cl->setProp(bdd->getTrue());
    InferenceStore::instance()->recordPropReduce(_cl, oldClProp, bdd->getTrue());
    _cl=0;

    TRACE("sa_fw_simpl",
	if(replacement) {
	  tout << "+" << (*replacement) << endl;
	}
	tout << "removed\n";
	tout << "^^^^^^^^^^^^\n";
    );
  }

  bool willPerform(Clause* premise)
  {
    CALL("TotalSimplificationPerformer::willPerform");
    ASS(_cl);

    if(!premise) {
      return true;
    }

    if( !ColorHelper::compatible(_cl->color(), premise->color()) ) {
      return false;
    }

    BDD* bdd=BDD::instance();
    if(!bdd->isXOrNonYConstant(_cl->prop(), premise->prop(), true)) {
      return false;
    }
    return true;
  }

  bool clauseKept()
  { return _cl; }
private:
  SaturationAlgorithm* _sa;
  Clause* _cl;
};


Clause* SaturationAlgorithm::doImmediateSimplification(Clause* cl0)
{
  CALL("SaturationAlgorithm::doImmediateSimplification");
  ASS(cl0->prop());

  Clause* cl=cl0;

  Clause* simplCl=_immediateSimplifier->simplify(cl);
  if(simplCl!=cl) {
    if(simplCl) {
      addNewClause(simplCl);
    }
    onClauseReduction(cl, simplCl, 0);
    return 0;
  }

  if(cl!=cl0 && cl0->isInput()) {
    
    cl->markInput();
  }

  return cl;
}

void SaturationAlgorithm::addNewClause(Clause* cl)
{
  CALL("SaturationAlgorithm::addNewClause");

  
  
  
  
  cl->incRefCnt();

  if(_opt.abstraction() && cl->number()%100000==(24788+(env.statistics->inputClauses%50000)) && env.statistics->inputClauses%3==0) {
    Clause* newCl = 0;
    if(cl->length()>1) {
      LiteralStack lits;
      lits.loadFromIterator(Clause::Iterator(*cl));
      newCl = Clause::fromStack(lits, cl->inputType(), cl->inference());
    }
    else if(cl->length()==1) {
      Literal* lit = (*cl)[0];
      if(lit->arity()) {
	TermList orig = *lit->nthArgument(0);
	Literal* newLit = EqHelper::replace(lit, orig, TermList(lit->vars(), false));
	newCl = Clause::fromIterator(getSingletonIterator(newLit), cl->inputType(), cl->inference());
      }
    }
    if(newCl) {
      cl->incRefCnt();
      newCl->incRefCnt();
      cl=newCl;
    }
  }

  onNewClause(cl);

  ASS(cl->prop());

  _newClauses.push(cl);
  
  
  cl->decRefCnt();
}

void SaturationAlgorithm::newClausesToUnprocessed()
{
  CALL("SaturationAlgorithm::newClausesToUnprocessed");

  while(_newClauses.isNonEmpty()) {
    Clause* cl=_newClauses.popWithoutDec();

    if(BDD::instance()->isTrue(cl->prop())) {
      continue;
    }

    switch(cl->store())
    {
    case Clause::BACKTRACKED:
    case Clause::UNPROCESSED:
      break;
    case Clause::PASSIVE:
    case Clause::REACTIVATED:
      onNonRedundantClause(cl);
      break;
    case Clause::ACTIVE:
      ASS(!cl->isEmpty());
      reanimate(cl);
      break;
    case Clause::NONE:
      addUnprocessedClause(cl);
      break;
#if VDEBUG
    case Clause::SELECTED:
    case Clause::SELECTED_REACTIVATED:
      
      ASSERTION_VIOLATION;
#endif
    }
    cl->decRefCnt(); 
  }
}

bool SaturationAlgorithm::clausesFlushed()
{
  return _unprocessed->isEmpty() && _newClauses.isEmpty();
}


void SaturationAlgorithm::addUnprocessedClause(Clause* cl)
{
  CALL("SaturationAlgorithm::addUnprocessedClause");
  ASS(cl->prop());

  _generatedClauseCount++;
  env.statistics->generatedClauses++;

  BDD* bdd=BDD::instance();
  ASS(!bdd->isTrue(cl->prop()));

  env.checkTimeSometime<64>();

  if(_bddMarkingSubsumption && _bddMarkingSubsumption->subsumed(cl)) {
    return;
  }

  cl=doImmediateSimplification(cl);
  if(!cl) {
    return;
  }
  ASS(!bdd->isTrue(cl->prop()));

  if(cl->isEmpty()) {
    handleEmptyClause(cl);
    return;
  }

  cl->setStore(Clause::UNPROCESSED);
  _unprocessed->add(cl);
}

void SaturationAlgorithm::handleEmptyClause(Clause* cl)
{
  CALL("SaturationAlgorithm::handleEmptyClause");
  ASS(cl->isEmpty());

  if(isRefutation(cl)) {
    onNonRedundantClause(cl);
    throw RefutationFoundException(cl);
  }

  if(_splitter && _splitter->handleEmptyClause(cl)) {
    return;
  }

  BDD* bdd=BDD::instance();

  ASS(!bdd->isFalse(cl->prop()));
  env.statistics->bddPropClauses++;

  if(_opt.satSolverForEmptyClause()) {
    static BDDConjunction ecProp(_opt);

    onNonRedundantClause(cl);
    onNewUsefulPropositionalClause(cl);
    ecProp.addNode(cl->prop());
    if(ecProp.isFalse()) {
      InferenceStore::instance()->recordMerge(cl, cl->prop(), _bddSatSolverEmptyClauses.begin(),
	  _bddSatSolverEmptyClauses.size(), bdd->getFalse());
      cl->setProp(bdd->getFalse());
      onNewUsefulPropositionalClause(cl);
      onNonRedundantClause(cl);
      throw RefutationFoundException(cl);
    } else {
      cl->incRefCnt();
      _bddSatSolverEmptyClauses.push(UnitSpec(cl));
      return;
    }
  } else {
    if(_mergedBddEmptyClause==0) {
      cl->incRefCnt();
      onNonRedundantClause(cl);
      _mergedBddEmptyClause=cl;
      onNewUsefulPropositionalClause(cl);
      return;
    }
    BDDNode* newProp=bdd->conjunction(_mergedBddEmptyClause->prop(), cl->prop());
    if(newProp!=_mergedBddEmptyClause->prop()) {
      onNonRedundantClause(cl);
      onNewUsefulPropositionalClause(cl);
    }
    if(bdd->isFalse(newProp)) {
      InferenceStore::instance()->recordMerge(cl, cl->prop(), _mergedBddEmptyClause, newProp);
      cl->setProp(newProp);
      onNonRedundantClause(cl);
      onNewUsefulPropositionalClause(cl);
      throw RefutationFoundException(cl);
    }
    if(newProp!=_mergedBddEmptyClause->prop()) {
      InferenceStore::instance()->recordMerge(_mergedBddEmptyClause, _mergedBddEmptyClause->prop(), cl, newProp);
      _mergedBddEmptyClause->setProp(newProp);
      onNonRedundantClause(_mergedBddEmptyClause);
    } else {
      env.statistics->subsumedEmptyClauses++;
      if(_opt.emptyClauseSubsumption()) {
	performEmptyClauseParentSubsumption(cl, _mergedBddEmptyClause->prop());
      }
    }
  }
}

void SaturationAlgorithm::performEmptyClauseParentSubsumption(Clause* cl0, BDDNode* emptyClauseProp)
{
  CALL("SaturationAlgorithm::performEmptyClauseSubsumption");
  ASS(cl0->isEmpty());

  BDD* bdd=BDD::instance();

  static Stack<Clause*> parentsToCheck;
  ASS(parentsToCheck.isEmpty());

  Clause* cl=cl0;
  for(;;) {
    UnitSpecIterator parents=
	InferenceStore::instance()->getParents(UnitSpec(cl, false));

    while(parents.hasNext()) {
      Clause* par=parents.next().cl();
      if(par->store()!=Clause::ACTIVE &&
	  par->store()!=Clause::PASSIVE &&
	  par->store()!=Clause::REACTIVATED) {
	continue;
      }
      if(!par->prop() || bdd->isTrue(par->prop())) {
	continue;
      }
      
      if(!bdd->isXOrNonYConstant(par->prop(), emptyClauseProp, true)) {
	continue;
      }
      par->setProp(bdd->getTrue());
      removeActiveOrPassiveClause(par);
      env.statistics->emptyClauseSubsumptions++;
      
      
      
      parentsToCheck.push(par);
    }

    if(parentsToCheck.isEmpty()) {
      break;
    }
    cl=parentsToCheck.pop();
  }
}


void SaturationAlgorithm::reanimate(Clause* cl)
{
  CALL("SaturationAlgorithm::reanimate");
  ASS_EQ(cl->store(), Clause::ACTIVE);

  ASS(!BDD::instance()->isTrue(cl->prop()));

  cl->setStore(Clause::REACTIVATED);
  _passive->add(cl);
}

bool SaturationAlgorithm::forwardSimplify(Clause* cl)
{
  CALL("SaturationAlgorithm::forwardSimplify");

  if(cl->store()==Clause::REACTIVATED || cl->store()==Clause::SELECTED_REACTIVATED) {
    return true;
  }

  if(!getLimits()->fulfillsLimits(cl)) {
    RSTAT_CTR_INC("clauses discarded by weight limit in forward simplification");
    env.statistics->discardedNonRedundantClauses++;
    return false;
  }

  TotalSimplificationPerformer performer(this, cl);

  FwSimplList::Iterator fsit(_fwSimplifiers);

  while(fsit.hasNext()) {
    ForwardSimplificationEngine* fse=fsit.next();

    fse->perform(cl, &performer);
    if(!performer.clauseKept()) {
      break;
    }
  }

  
  if(performer.clauseKept()) {
    cl->incRefCnt();
  }

  if(!performer.clauseKept()) {
    return false;
  }


  if( _splitter && !_opt.splitAtActivation() ) {
    if(_splitter->doSplitting(cl)) {
      return false;
    }
  }

  return true;
}

void SaturationAlgorithm::backwardSimplify(Clause* cl)
{
  CALL("SaturationAlgorithm::backwardSimplify");

  BDD* bdd=BDD::instance();

  BwSimplList::Iterator bsit(_bwSimplifiers);
  while(bsit.hasNext()) {
    BackwardSimplificationEngine* bse=bsit.next();

    BwSimplificationRecordIterator simplifications;
    bse->perform(cl,simplifications);
    while(simplifications.hasNext()) {
      BwSimplificationRecord srec=simplifications.next();
      Clause* redundant=srec.toRemove;
      ASS_NEQ(redundant, cl);

      BDDNode* oldRedundantProp=redundant->prop();

      if( !bdd->isXOrNonYConstant(oldRedundantProp, cl->prop(), true) ) {
	
	continue;
      }

      Clause* replacement=srec.replacement;

      if(replacement) {
	addNewClause(replacement);
      }
      LOG_UNIT("sa_bw_simpl_red_clause",redundant);
      onClauseReduction(redundant, replacement, cl, 0, false);


      
      
      

      redundant->incRefCnt(); 

      removeActiveOrPassiveClause(redundant);

      redundant->setProp(bdd->getTrue());
      InferenceStore::instance()->recordPropReduce(redundant, oldRedundantProp, bdd->getTrue());


      TRACE("sa_bw_simpl",
	tout << "-<<--------\n";
	tout << ":" << (*cl) << endl;
	tout << "-" << (*redundant) << endl;
	if(replacement) {
	  tout << "+" << (*replacement) << endl;
	}
	tout << "removed\n";
	tout << "^^^^^^^^^^^\n";
      );

      redundant->decRefCnt();
    }
  }
}

void SaturationAlgorithm::removeActiveOrPassiveClause(Clause* cl)
{
  CALL("SaturationAlgorithm::removeBackwardSimplifiedClause");

  if(_clauseActivationInProgress) {
    
    
    _postponedClauseRemovals.push(cl);
    return;
  }

  switch(cl->store()) {
  case Clause::PASSIVE:
    _passive->remove(cl);
    break;
  case Clause::ACTIVE:
  case Clause::SELECTED_REACTIVATED:
    _active->remove(cl);
    break;
  case Clause::REACTIVATED:
    _passive->remove(cl);
    _active->remove(cl);
    break;
  default:
    ASS_REP2(false, cl->store(), *cl);
  }
  
}

void SaturationAlgorithm::addToPassive(Clause* cl)
{
  CALL("SaturationAlgorithm::addToPassive");
  ASS_EQ(cl->store(), Clause::UNPROCESSED);

  cl->setStore(Clause::PASSIVE);
  env.statistics->passiveClauses++;

  _passive->add(cl);
}

bool SaturationAlgorithm::activate(Clause* cl)
{
  CALL("SaturationAlgorithm::activate");

  if(_bddMarkingSubsumption && _bddMarkingSubsumption->subsumed(cl)) {
    return false;
  }

  if(_consFinder && _consFinder->isRedundant(cl)) {
    return false;
  }

  if(_splitter && _opt.splitAtActivation()) {
    if(_splitter->doSplitting(cl)) {
      return false;
    }
  }

  _clauseActivationInProgress=true;

  if(!cl->selected()) {
    _selector->select(cl);
  }

  if(cl->store()==Clause::SELECTED_REACTIVATED) {
    cl->setStore(Clause::ACTIVE);
    env.statistics->reactivatedClauses++;
    LOG_UNIT("sa_reanimated", cl);
  } else {
    ASS_EQ(cl->store(), Clause::SELECTED);
    cl->setStore(Clause::ACTIVE);
    env.statistics->activeClauses++;

    _active->add(cl);
  }

  ClauseIterator toAdd=_generator->generateClauses(cl);

  while(toAdd.hasNext()) {
    Clause* genCl=toAdd.next();

    addNewClause(genCl);

    LOG_UNIT("sa_generated_clause", genCl);

    Inference::Iterator iit=genCl->inference()->iterator();
    while(genCl->inference()->hasNext(iit)) {
      Unit* premUnit=genCl->inference()->next(iit);
      ASS(premUnit->isClause());
      Clause* premCl=static_cast<Clause*>(premUnit);

      onParenthood(genCl, premCl);
    }
  }

  _clauseActivationInProgress=false;

  
  while(_postponedClauseRemovals.isNonEmpty()) {
    Clause* cl=_postponedClauseRemovals.pop();
    if(cl->store()!=Clause::ACTIVE &&
	cl->store()!=Clause::PASSIVE &&
	cl->store()!=Clause::REACTIVATED &&
	cl->store()!=Clause::SELECTED_REACTIVATED) {
      continue;
    }
    removeActiveOrPassiveClause(cl);
  }

  return true;
}

void SaturationAlgorithm::doUnprocessedLoop()
{
  CALL("SaturationAlgorithm::doUnprocessedLoop");

start:

  newClausesToUnprocessed();

  while (! _unprocessed->isEmpty()) {
    Clause* c = _unprocessed->pop();
    ASS(!isRefutation(c));

    if(forwardSimplify(c)) {
      onClauseRetained(c);
      addToPassive(c);
      ASS_EQ(c->store(), Clause::PASSIVE);
    }
    else {
      ASS_EQ(c->store(), Clause::UNPROCESSED);
      c->setStore(Clause::NONE);
    }

    newClausesToUnprocessed();

    if(env.timeLimitReached()) {
      throw TimeLimitExceededException();
    }
  }

  ASS(clausesFlushed());
  onAllProcessed();
  if(!clausesFlushed()) {
    
    goto start;
  }

}

void SaturationAlgorithm::handleUnsuccessfulActivation(Clause* cl)
{
  CALL("SaturationAlgorithm::handleUnsuccessfulActivation");

  if(cl->store()==Clause::SELECTED_REACTIVATED) {
    cl->setStore(Clause::ACTIVE);
  }
  else {
    ASS_EQ(cl->store(), Clause::SELECTED);
    cl->setStore(Clause::NONE);
  }
}

bool SaturationAlgorithm::handleClauseBeforeActivation(Clause* c)
{
  return true;
}

void SaturationAlgorithm::initAlgorithmRun()
{
  CALL("SaturationAlgorithm::initAlgorithmRun");

  init();
}


UnitList* SaturationAlgorithm::collectSaturatedSet()
{
  CALL("SaturationAlgorithm::collectSaturatedSet");

  LiteralIndexingStructure* gis=getIndexManager()->getGeneratingLiteralIndexingStructure();

  UnitList* res = 0;
  SLQueryResultIterator qrit = gis->getAll();
  while(qrit.hasNext()) {
    SLQueryResult qres = qrit.next();
    UnitList::push(qres.clause, res);
    qres.clause->incRefCnt();
  }

  if(getOptions().propositionalToBDD()) {
    BDD& _bdd = *BDD::instance();
    if(getOptions().satSolverForEmptyClause()) {
      Stack<UnitSpec>::ConstIterator ecit(_bddSatSolverEmptyClauses);
      while(ecit.hasNext()) {
	UnitSpec us = ecit.next();
	Clause* cl = Clause::fromClause(us.cl());
	cl->setProp(us.prop());

	InferenceStore::FullInference* finf = new(1) InferenceStore::FullInference(1);
	finf->rule = Inference::REORDER_LITERALS;
	finf->premises[0] = us;
	finf->increasePremiseRefCounters();
	InferenceStore::instance()->recordInference(UnitSpec(cl), finf);

	UnitList::push(cl, res);
	cl->incRefCnt();
      }
    }
    else {
      if(_mergedBddEmptyClause) {
	UnitList::push(_mergedBddEmptyClause, res);
	_mergedBddEmptyClause->incRefCnt();
      }
    }

    const Problem::BDDVarMeaningMap& bvm = getProblem().getBDDVarMeanings();
    Problem::BDDVarMeaningMap::Iterator bvmIt(bvm);
    while(bvmIt.hasNext()) {
      unsigned var;
      Problem::BDDMeaningSpec spec;
      bvmIt.next(var, spec);

      Literal* namedLit;

      if(spec.first) {
        ASS(!spec.second);

        namedLit = spec.first;
      }
      else {
	unsigned newPred = env.signature->addFreshPredicate(0, "bddAux");
	namedLit = Literal::create(newPred, 0, true, false, 0);

	Formula* clForm = spec.second->getFormula(_bdd.getFalse());
	Formula* namingForm = new BinaryFormula(IFF, new AtomicFormula(namedLit), clForm);
	FormulaUnit* namingFU = new FormulaUnit(namingForm, new Inference(Inference::PREDICATE_DEFINITION), Unit::AXIOM);
	UnitList::push(namingFU, res);
      }
      Clause* posNm = Clause::fromIterator(getSingletonIterator(namedLit),
	  Unit::AXIOM, new Inference(Inference::PREDICATE_DEFINITION));
      posNm->setProp(_bdd.getAtomic(var, true));
      UnitList::push(posNm, res);
      posNm->incRefCnt();

      Clause* negNm = Clause::fromIterator(
	  getSingletonIterator(Literal::complementaryLiteral(namedLit)),
	  Unit::AXIOM, new Inference(Inference::PREDICATE_DEFINITION));
      negNm->setProp(_bdd.getAtomic(var, false));
      UnitList::push(negNm, res);
      negNm->incRefCnt();
    }
  }

  return res;
}

void SaturationAlgorithm::doOneAlgorithmStep()
{
  CALL("SaturationAlgorithm::doOneAlgorithmStep");

  doUnprocessedLoop();

  if(_passive->isEmpty()) {
    MainLoopResult::TerminationReason termReason =
	isComplete() ? Statistics::SATISFIABLE : Statistics::REFUTATION_NOT_FOUND;
    MainLoopResult res(termReason);
    if(termReason == Statistics::SATISFIABLE && getOptions().proof()!=Options::PROOF_OFF) {
      res.saturatedSet = collectSaturatedSet();
    }
    throw MainLoopFinishedException(res);
  }

  Clause* cl = _passive->popSelected();

  if(cl->store()==Clause::REACTIVATED) {
    cl->setStore(Clause::SELECTED_REACTIVATED);
  }
  else {
    ASS_EQ(cl->store(),Clause::PASSIVE);
    cl->setStore(Clause::SELECTED);
  }

  if(!handleClauseBeforeActivation(cl)) {
    return;
  }

  bool isActivated=activate(cl);
  if(!isActivated) {
    handleUnsuccessfulActivation(cl);
  }
}


MainLoopResult SaturationAlgorithm::runImpl()
{
  CALL("SaturationAlgorithm::runImpl");

  try
  {
    for (;;) {
      doOneAlgorithmStep();

      Timer::syncClock();
      if(env.timeLimitReached()) {
        throw TimeLimitExceededException();
      }
    }
  }
  catch(ThrowableBase&)
  {
    tryUpdateFinalClauseCount();
    throw;
  }

}

void SaturationAlgorithm::setGeneratingInferenceEngine(GeneratingInferenceEngine* generator)
{
  CALL("SaturationAlgorithm::setGeneratingInferenceEngine");

  ASS(!_generator);
  _generator=generator;
  _generator->attach(this);
}

void SaturationAlgorithm::setImmediateSimplificationEngine(ImmediateSimplificationEngine* immediateSimplifier)
{
  CALL("SaturationAlgorithm::setImmediateSimplificationEngine");

  ASS(!_immediateSimplifier);
  _immediateSimplifier=immediateSimplifier;
  _immediateSimplifier->attach(this);
}

void SaturationAlgorithm::addForwardSimplifierToFront(ForwardSimplificationEngine* fwSimplifier)
{
  FwSimplList::push(fwSimplifier, _fwSimplifiers);
  fwSimplifier->attach(this);
}

void SaturationAlgorithm::addBackwardSimplifierToFront(BackwardSimplificationEngine* bwSimplifier)
{
  BwSimplList::push(bwSimplifier, _bwSimplifiers);
  bwSimplifier->attach(this);
}

SaturationAlgorithm* SaturationAlgorithm::createFromOptions(Problem& prb, const Options& opt, IndexManager* indexMgr)
{
  CALL("SaturationAlgorithm::createFromOptions");

  SaturationAlgorithm* res;
  switch(opt.saturationAlgorithm()) {
  case Shell::Options::DISCOUNT:
    res=new Discount(prb, opt);
    break;
  case Shell::Options::LRS:
    res=new LRS(prb, opt);
    break;
  case Shell::Options::OTTER:
    res=new Otter(prb, opt);
    break;
  default:
    NOT_IMPLEMENTED;
  }
  if(indexMgr) {
    res->_imgr = SmartPtr<IndexManager>(indexMgr, true);
    indexMgr->setSaturationAlgorithm(res);
  }
  else {
    res->_imgr = SmartPtr<IndexManager>(new IndexManager(res));
  }

  
  CompositeGIE* gie=new CompositeGIE();
  if(prb.hasEquality()) {
    gie->addFront(new EqualityFactoring());
    gie->addFront(new EqualityResolution());
    gie->addFront(new Superposition());
  }
  gie->addFront(new Factoring());
  if(opt.binaryResolution()) {
    gie->addFront(new BinaryResolution());
  }
  if(opt.unitResultingResolution()!=Options::URR_OFF) {
    gie->addFront(new URResolution());
  }
  res->setGeneratingInferenceEngine(gie);

  res->setImmediateSimplificationEngine(createISE(prb, opt));

  
  if(opt.hyperSuperposition()) {
    res->addForwardSimplifierToFront(new HyperSuperposition());
  }
  if(opt.globalSubsumption()) {
    res->addForwardSimplifierToFront(new GlobalSubsumption());
  }
  if(opt.forwardLiteralRewriting()) {
    res->addForwardSimplifierToFront(new ForwardLiteralRewriting());
  }
  if(prb.hasEquality()) {
    switch(opt.forwardDemodulation()) {
    case Options::DEMODULATION_ALL:
    case Options::DEMODULATION_PREORDERED:
      res->addForwardSimplifierToFront(new ForwardDemodulation());
      break;
    case Options::DEMODULATION_OFF:
      break;
#if VDEBUG
    default:
      ASSERTION_VIOLATION;
#endif
    }
  }
  if (opt.forwardSubsumption()) {
    if(opt.forwardSubsumptionResolution()) {
      res->addForwardSimplifierToFront(new CTFwSubsAndRes(true));
    }
    else {
      res->addForwardSimplifierToFront(new CTFwSubsAndRes(false));
    }
  }
  else if (opt.forwardSubsumptionResolution()) {
    USER_ERROR("Forward subsumption resolution requires forward subsumption to be enabled.");
  }

  
  if(prb.hasEquality()) {
    switch(opt.backwardDemodulation()) {
    case Options::DEMODULATION_ALL:
    case Options::DEMODULATION_PREORDERED:
      res->addBackwardSimplifierToFront(new BackwardDemodulation());
      break;
    case Options::DEMODULATION_OFF:
      break;
#if VDEBUG
    default:
      ASSERTION_VIOLATION;
#endif
    }
  }
  if(opt.backwardSubsumption()!=Options::SUBSUMPTION_OFF) {
    bool byUnitsOnly=opt.backwardSubsumption()==Options::SUBSUMPTION_UNIT_ONLY;
    res->addBackwardSimplifierToFront(new SLQueryBackwardSubsumption(byUnitsOnly));
  }
  if(opt.backwardSubsumptionResolution()!=Options::SUBSUMPTION_OFF) {
    bool byUnitsOnly=opt.backwardSubsumptionResolution()==Options::SUBSUMPTION_UNIT_ONLY;
    res->addBackwardSimplifierToFront(new BackwardSubsumptionResolution(byUnitsOnly));
  }

  if(opt.mode()==Options::MODE_CONSEQUENCE_ELIMINATION) {
    res->_consFinder=new ConsequenceFinder();
  }
  if(opt.showSymbolElimination()) {
    res->_symEl=new SymElOutput();
  }

  switch(opt.splitting()) {
  case Options::SM_OFF:
    break;
  case Options::SM_BACKTRACKING:
    res->_splitter=new BSplitter();
    break;
  case Options::SM_NOBACKTRACKING:
    if(opt.propositionalToBDD()) {
      res->_splitter=new SWBSplitterWithBDDs();
    }
    else {
      res->_splitter=new SWBSplitterWithoutBDDs();
    }
    break;
  case Options::SM_SAT:
    res->_splitter = new SSplitter();
    break;
  default:
    ASSERTION_VIOLATION;
  }


  if(opt.bddMarkingSubsumption()) {
    res->_bddMarkingSubsumption=new BDDMarkingSubsumption();
  }

  if(opt.questionAnswering()==Options::QA_ANSWER_LITERAL) {
    res->_answerLiteralManager = AnswerLiteralManager::getInstance();
  }
  return res;
} 
