
/*
 * File Splitter.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/IntUnionFind.hpp"
#include "Lib/Metaiterators.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"

#include "Shell/AnswerExtractor.hpp"
#include "Shell/Options.hpp"
#include "Shell/Statistics.hpp"

#include "SaturationAlgorithm.hpp"

#include "Splitter.hpp"

namespace Saturation
{

Splitter::CompRec::CompRec(Literal** lits, unsigned len)
{
  CALL("Splitter::CompRec::CompRec/2");

  _lits.loadFromIterator(getArrayishObjectIterator(lits, len));
}

void Splitter::init(SaturationAlgorithm* sa)
{
  CALL("Splitter::init");

  _sa = sa;
  _ansLitMgr = _sa->getAnswerLiteralManager();
}

const Options& Splitter::getOptions() const
{
  CALL("Splitter::getOptions");
  ASS(_sa);

  return _sa->getOptions();
}

void Splitter::onClauseReduction(Clause* cl, Clause* premise, Clause* replacement)
{
  CALL("Splitter::onClauseReduction(Clause*,Clause*,Clause*)");
  
  if(!premise) {
    ASS(!replacement);
    return;
  }

  onClauseReduction(cl, pvi( getSingletonIterator(premise) ), replacement);
}

bool Splitter::splitPositive()
{
  return getOptions().splitPositive();
}

bool Splitter::splittingAllowed(Clause* cl)
{
  CALL("Splitter::splittingAllowed");

  if(getOptions().splitInputOnly() && !cl->isInput()) {
    return false;
  }

  if(getOptions().splitGoalOnly() && cl->inputType()!=Unit::CONJECTURE) {
    return false;
  }

  return true;
}

bool Splitter::isAnswerLiteral(Literal* lit)
{
  CALL("Splitter::isAnswerLiteral");

  return _ansLitMgr && _ansLitMgr->isAnswerLiteral(lit);
}

bool Splitter::isSpecial(Literal* lit)
{
  CALL("Splitter::isSpecial");

  Signature::Symbol* predSym = env.signature->getPredicate(lit->functor());

  return lit->color()==COLOR_TRANSPARENT &&
    (!getOptions().showSymbolElimination() || lit->skip()) &&
    (!predSym->cfName()) && (!isAnswerLiteral(lit));
}

bool Splitter::canStandAlone(Literal* lit)
{
  if(lit->isNegative() && splitPositive()) {
    return false;
  }
  return true;
}

bool Splitter::standAloneObligations()
{
  return splitPositive();
}


bool Splitter::getComponents(Clause* cl, Stack<CompRec>& acc, bool putSpecialsTogether)
{
  CALL("Splitter::doSplitting");
  ASS_EQ(acc.size(), 0);

  if(!splittingAllowed(cl)) {
    return false;
  }

  unsigned clen=cl->length();
  ASS_G(clen,0);

  if(clen<=1) {
    return false;
  }

  
  
  static DHMap<unsigned, unsigned, IdentityHash> varMasters;
  varMasters.reset();
  IntUnionFind components(clen);
  
  int coloredMaster=-1;

  for(unsigned i=0;i<clen;i++) {
    Literal* lit=(*cl)[i];
    if( putSpecialsTogether && isSpecial(lit) ) {
      if(coloredMaster==-1) {
	coloredMaster=i;
      } else {
	
	components.doUnion(coloredMaster, i);
      }
    }
    VariableIterator vit(lit);
    while(vit.hasNext()) {
      unsigned master=varMasters.findOrInsert(vit.next().var(), i);
      if(master!=i) {
	components.doUnion(master, i);
      }
    }
  }
  components.evalComponents();

  unsigned compCnt=components.getComponentCount();

  if(standAloneObligations() && compCnt>1) {

    
    
    

    IntUnionFind::ComponentIterator cit(components);

    int someCompEl=-1;
    bool someCompOK=false;
    while(cit.hasNext()) {
      IntUnionFind::ElementIterator elit=cit.next();

      int compEl=elit.next();
      if(someCompEl==-1) {
	someCompEl=compEl;
      }
      bool saok=false; 
      for(;;) {
	if(canStandAlone((*cl)[compEl])) {
	  saok=true;
	  break;
	}
	if(!elit.hasNext()) {
	  break;
	}
	compEl=elit.next();
      }
      if(!saok || !someCompOK) {
	components.doUnion(compEl, someCompEl);
	if(saok) {
	  someCompOK=true;
	}
      }
    }

    
    components.evalComponents();
    compCnt=components.getComponentCount();
  }

  if(compCnt==1) {
    return false;
  }

  env.statistics->splitClauses++;
  env.statistics->splitComponents+=compCnt;

  IntUnionFind::ComponentIterator cit(components);
  ASS(cit.hasNext());
  while(cit.hasNext()) {
    IntUnionFind::ElementIterator elit=cit.next();

    acc.push(CompRec());

    while(elit.hasNext()) {
      int litIndex=elit.next();
      Literal* lit = (*cl)[litIndex];

      acc.top().add(lit);

      if(litIndex==coloredMaster) {
	acc.top().markSpecial();
      }
    }
  }
  ASS_EQ(acc.size(),compCnt);
  return true;
}

}
