
/*
 * File BFNTMainLoop.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <cerrno>
#include <csignal>

#include "Lib/Portability.hpp"

#if !COMPILER_MSVC

#include <sys/types.h>
#include <sys/wait.h>

#endif


#include "Lib/Environment.hpp"
#include "Lib/Int.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/ScopedPtr.hpp"
#include "Lib/System.hpp"
#include "Lib/Timer.hpp"

#include "Lib/Sys/Multiprocessing.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Unit.hpp"

#include "Shell/Property.hpp"
#include "Shell/TPTP.hpp"
#include "Shell/UIHelper.hpp"

#include "Statistics.hpp"

#include "BFNTMainLoop.hpp"

#define BFNT_CHILD_RESULT_SAT 0
#define BFNT_CHILD_RESULT_UNSAT 6
#define BFNT_CHILD_RESULT_UNKNOWN 7

namespace Shell
{
#if COMPILER_MSVC

BFNTMainLoop::BFNTMainLoop(Problem& prb, const Options& opt)
: MainLoop(prb, opt)
{}

void BFNTMainLoop::init()
{
  CALL("BFNTMainLoop::init");

  USER_ERROR("BFNT not supported on Windows");
}

MainLoopResult BFNTMainLoop::runImpl()
{
  CALL("BFNTMainLoop::runImpl");

  USER_ERROR("BFNT not supported on Windows");
}

#else

BFNTMainLoop::BFNTMainLoop(Problem& prb, const Options& opt)
: MainLoop(prb, opt),
  _childOpts(opt),
  _bfnt(prb.getProperty())
{
  CALL("BFNTMainLoop::BFNTMainLoop");

  
  
  
  _childOpts.setBfnt(false);
}


void BFNTMainLoop::init()
{
  CALL("BFNTMainLoop::init");

  _hasSorts = _prb.getProperty()->hasNonDefaultSorts();
  if(_hasSorts) {
    return;
  }

  
  
  
  
  
  
  _bfnt.apply(_prb.units());
}

void BFNTMainLoop::runChild(size_t modelSize)
{
  CALL("BFNTMainLoop::runChild");

  ScopedPtr<Problem> childPrb(_bfnt.createProblem(modelSize));

  TRACE("bfnt_loop",
      tout << "Flattenned problem:" << endl;
      UnitList::Iterator puit(childPrb->units());
      while(puit.hasNext()) {
	Unit* u = puit.next();
	TRACE_OUTPUT_UNIT(u);
      }
  );

  ScopedPtr<MainLoop> childMainLoop(MainLoop::createFromOptions(*childPrb, _childOpts));
  MainLoopResult innerRes = childMainLoop->run();
  innerRes.updateStatistics();

  LOG("bfnt_loop","Child termination reason: "
      << ((innerRes.terminationReason==Statistics::SATISFIABLE) ? "Satisfiable" :
	  (innerRes.terminationReason==Statistics::REFUTATION) ? "Unsatisfiable" : "Unknown") );
  COND_TRACE("bfnt_loop", env.statistics->model!="",
      tout << "Model: " << endl
	   << env.statistics->model;
  );

  if(env.options->mode()!=Options::MODE_SPIDER) {
    if(innerRes.terminationReason==Statistics::SATISFIABLE || innerRes.terminationReason==Statistics::TIME_LIMIT) {
      env.beginOutput();
      env.statistics->print(env.out());
      env.endOutput();
    }
  }

  if(env.options->mode()!=Options::MODE_SPIDER && innerRes.terminationReason==Statistics::SATISFIABLE) {
    env.beginOutput();
    UIHelper::outputSatisfiableResult(env.out());
    env.endOutput();
    UIHelper::satisfiableStatusWasAlreadyOutput = true;
  }

  switch(innerRes.terminationReason) {
  case Statistics::SATISFIABLE:
    exit(BFNT_CHILD_RESULT_SAT);
  case Statistics::REFUTATION:
    exit(BFNT_CHILD_RESULT_UNSAT);
  default:
    exit(BFNT_CHILD_RESULT_UNKNOWN);
  }
  exit(BFNT_CHILD_RESULT_UNKNOWN);
}


MainLoopResult BFNTMainLoop::spawnChild(size_t modelSize)
{
  CALL("BFNTMainLoop::spawnChild");

  pid_t childPid=Multiprocessing::instance()->fork();

  if(!childPid) {
    runChild(modelSize);
    ASSERTION_VIOLATION;
  }


  System::ignoreSIGINT();

  int status;
  errno=0;
  pid_t res=waitpid(childPid, &status, 0);
  if(res==-1) {
    SYSTEM_FAIL("Error in waiting for forked process.",errno);
  }

  System::heedSIGINT();

  Timer::syncClock();

  if(res!=childPid) {
    INVALID_OPERATION("Invalid waitpid return value: "+Int::toString(res)+"  pid of forked Vampire: "+Int::toString(childPid));
  }

  ASS(!WIFSTOPPED(status));

  if( (WIFSIGNALED(status) && WTERMSIG(status)==SIGINT) ||
      (WIFEXITED(status) && WEXITSTATUS(status)==VAMP_RESULT_STATUS_SIGINT) )  {
    
    
    

    raise(SIGINT);
  }

  if(WIFEXITED(status)) {
    switch(WEXITSTATUS(status)) {
    case BFNT_CHILD_RESULT_SAT:
      UIHelper::satisfiableStatusWasAlreadyOutput = true;
      return MainLoopResult(Statistics::SATISFIABLE);
    case BFNT_CHILD_RESULT_UNSAT:
      return MainLoopResult(Statistics::REFUTATION);

    case VAMP_RESULT_STATUS_OTHER_SIGNAL:
    case VAMP_RESULT_STATUS_UNHANDLED_EXCEPTION:
      INVALID_OPERATION("error in the child process");

    case BFNT_CHILD_RESULT_UNKNOWN:
    default: 
      return MainLoopResult(Statistics::UNKNOWN);
    }
  }
  else {
    return MainLoopResult(Statistics::UNKNOWN);
  }
}


MainLoopResult BFNTMainLoop::runImpl()
{
  CALL("BFNTMainLoop::runImpl");

  if(_hasSorts) {
    return MainLoopResult(Statistics::UNKNOWN);
  }


  env.timer->makeChildrenIncluded();

  size_t modelSize = 1;
  for(;;) {
    Timer::syncClock();
    if(env.timeLimitReached()) { return MainLoopResult(Statistics::TIME_LIMIT); }
    LOG("bfnt_loop","Trying model size "<<modelSize);
    env.statistics->maxBFNTModelSize = modelSize;
    MainLoopResult childResult = spawnChild(modelSize);

    if(childResult.terminationReason == Statistics::SATISFIABLE) {
      return childResult;
    }

    modelSize++;
  }
}
#endif

}
