
/*
 * File Flattening.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"
#include "Debug/Tracer.hpp"

#include "Kernel/Inference.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Unit.hpp"

#include "Flattening.hpp"

namespace Shell
{

Formula* Flattening::getFlattennedNegation(Formula* f)
{
  CALL("Flattening::getFlattennedNegation");

  switch(f->connective()) {
  case NOT:
    return f->uarg();
  case TRUE:
    return Formula::falseFormula();
  case FALSE:
    return Formula::trueFormula();
  default:
    return new NegatedFormula(f);
  }

}

FormulaUnit* Flattening::flatten (FormulaUnit* unit)
{
  CALL("Flattening::flatten (Unit*)");
  ASS(! unit->isClause());

  LOG_UNIT("pp_flt_inp", unit);
  Formula* f = unit->formula();
  Formula* g = flatten(f);
  if (f == g) { 
    return unit;
  }

  FormulaUnit* res = new FormulaUnit(g,
			 new Inference1(Inference::FLATTEN,unit),
			 unit->inputType());
  if(unit->included()) {
    res->markIncluded();
  }
  LOG_UNIT("pp_flt_out", res);
  return res;
} 

Formula* Flattening::flatten (Formula* f)
{
  CALL("Flattening::flatten(Formula*)");

  Connective con = f->connective();
  switch (con) {
  case LITERAL:
  case TRUE:
  case FALSE:
    return f;

  case AND:
  case OR: 
    {
      FormulaList* args = flatten(f->args(),con);
      if (args == f->args()) { 
	return f;
      }
      return new JunctionFormula(con,args);
    }

  case IMP:
  case IFF:
  case XOR: 
    {
      Formula* left = flatten(f->left());
      Formula* right = flatten(f->right());
      if (left == f->left() && right == f->right()) {
	return f;
      }
      return new BinaryFormula(con,left,right);
    }
    
  case NOT: 
    {
      Formula* arg = flatten(f->uarg());

      if(arg->connective()==NOT) {
	
	return arg->uarg();
      }
      if(arg->connective()==LITERAL) {
	return new AtomicFormula(Literal::complementaryLiteral(arg->literal()));
      }
      if (arg == f->uarg()) {
	return f;
      }
      return new NegatedFormula(arg);
    }
    
  case FORALL:
  case EXISTS: 
    {
      Formula* arg = flatten(f->qarg());
      if (arg->connective() != con) {
	if (arg == f->qarg()) {
	  return f;
	}
	return new QuantifiedFormula(con,f->vars(),arg);
      }

      
      return new QuantifiedFormula(con,
				   f->vars()->append(arg->vars()),
				   arg->qarg());
    }

  case ITE:
    {
      Formula* c = flatten(f->condArg());
      Formula* t = flatten(f->thenArg());
      Formula* e = flatten(f->elseArg());
      if (c == f->condArg() && t == f->thenArg() && e == f->elseArg()) {
	return f;
      }
      return new IteFormula(c,t,e);
    }

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }

} 


FormulaList* Flattening::flatten (FormulaList* fs, 
				  Connective con)
{
  CALL("Flattening::flatten (FormulaList*...)");
  ASS(con == OR || con == AND);

#if 1
  if(!fs) {
    return 0;
  }

  FormulaList* fs0 = fs;

  bool modified = false;
  Stack<Formula*> res(8);
  Stack<FormulaList*> toDo(8);
  for(;;) {
    if(fs->head()->connective()==con) {
      modified = true;
      if(fs->tail()) {
	toDo.push(fs->tail());
      }
      fs = fs->head()->args();
      continue;
    }
    Formula* hdRes = flatten(fs->head());
    if(hdRes!=fs->head()) {
      modified = true;
    }
    res.push(hdRes);
    fs = fs->tail();
    if(!fs) {
      if(toDo.isEmpty()) {
	break;
      }
      fs = toDo.pop();
    }
  }
  if(!modified) {
    return fs0;
  }
  FormulaList* resLst = 0;
  FormulaList::pushFromIterator(Stack<Formula*>::TopFirstIterator(res), resLst);
  return resLst;
#else
  if (fs->isEmpty()) {
    return fs;
  }
  Formula* head = flatten(fs->head());
  FormulaList* tail = flatten(fs->tail(),con);

  if (head->connective() == con) {
    return head->args()->append(tail);
  }

  if (head == fs->head() && tail == fs->tail()) {
    return fs;
  }
  return new FormulaList(head,tail);
#endif
} 


bool TopLevelFlatten::apply(Problem& prb)
{
  CALL("TopLevelFlatten::apply(Problem&)");

  if(apply(prb.units())) {
    prb.invalidateProperty();
    return true;
  }
  else {
    return false;
  }
}

bool TopLevelFlatten::apply(UnitList*& units)
{
  CALL("TopLevelFlatten::apply(UnitList*&)");

  Stack<FormulaUnit*> flAcc;

  bool modified = false;
  UnitList::DelIterator uit(units);
  while(uit.hasNext()) {
    Unit* u = uit.next();
    if(u->isClause()) {
      continue;
    }
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    flAcc.reset();
    if(!apply(fu, flAcc)) {
      continue;
    }
    modified = true;
    ASS_GE(flAcc.size(),2);
    while(flAcc.isNonEmpty()) {
      uit.insert(flAcc.pop());
    }
    uit.del();
  }
  return modified;
}

bool TopLevelFlatten::apply(FormulaUnit* fu, Stack<FormulaUnit*>& acc)
{
  CALL("TopLevelFlatten::apply(FormulaUnit*,Stack<FormulaUnit*>&)");

  Formula* f = fu->formula();

  Formula::VarList* vars;
  if(f->connective()==FORALL) {
    vars = f->vars();
    f = f->qarg();
  }
  else {
    vars = 0;
  }
  if(f->connective()!=AND) {
    return false;
  }
  RSTAT_CTR_INC("top level flattened formulas")
  LOG_UNIT("pp_tlf_processed", fu);
  FormulaList::Iterator fit(f->args());
  while(fit.hasNext()) {
    Formula* sf = fit.next();
    bool needsFlattening = vars && sf->connective()==FORALL;
    if(vars) {
      sf = new QuantifiedFormula(FORALL, vars, sf);
    }
    Inference* inf = new Inference1(Inference::FLATTEN, fu);
    FormulaUnit* sfu = new FormulaUnit(sf, inf, fu->inputType());
    if(fu->included()) {
      sfu->markIncluded();
    }
    if(needsFlattening) {
      sfu = Flattening::flatten(sfu);
    }
    acc.push(sfu);
    LOG_UNIT("pp_tlf_res", sfu);
    RSTAT_CTR_INC("top level flattenning results")
  }
  return true;
}



}
