
/*
 * File FormulaIteExpander.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <algorithm>

#include "Lib/Environment.hpp"
#include "Lib/List.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Unit.hpp"

#include "FormulaIteExpander.hpp"

namespace Shell
{

using namespace Lib;
using namespace Kernel;

void FormulaIteExpander::apply(Problem& prb)
{
  CALL("FormulaIteExpander::apply(Problem&)");

  if(apply(prb.units())) {
    prb.invalidateProperty();
  }
}

bool FormulaIteExpander::apply(UnitList*& units)
{
  CALL("FormulaIteExpander::apply(UnitList*&)");
  ASS(!_defs);

  bool modified = false;
  UnitList::DelIterator us(units);
  while(us.hasNext()) {
    Unit* u = us.next();
    if(u->isClause()) {
      continue;
    }
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    FormulaUnit* v = apply(fu);
    if(v != fu) {
      us.replace(v);
      modified = true;
    }
  }
  ASS(_defs==0 || modified);
  units = UnitList::concat(_defs, units);
  _defs=0;
#if 0
  UnitList::Iterator puit(units); 
  while(puit.hasNext()) {
    Unit* u = puit.next();
    cout<<u->toString()<<endl;
  }
#endif
  return modified;
}

FormulaUnit* FormulaIteExpander::apply(FormulaUnit* unit, UnitList*& defs)
{
  CALL("FormulaIteExpander::apply(Unit*,UnitList*&)");
  ASS(!_defs);
  ASS(!unit->isClause());

  FormulaUnit* res=apply(unit);

  defs=_defs;
  _defs=0;
  return res;
}

FormulaUnit* FormulaIteExpander::apply(FormulaUnit* unit)
{
  CALL("FormulaIteExpander::apply(Unit*)");
  ASS(! unit->isClause());

  Formula* f = unit->formula();
  Formula* g = apply(f);
  if (f == g) { 
    return unit;
  }

  FormulaUnit* res = new FormulaUnit(g,
			 new Inference1(Inference::FORMULA_ITE_EXPANSION,unit),
			 unit->inputType());
  if(unit->included()) {
    res->markIncluded();
  }
  return res;
}


Formula* FormulaIteExpander::apply(Formula* f)
{
  CALL("FormulaIteExpander::apply(Formula*)");

  Connective con = f->connective();
  switch (con) {
  case LITERAL:
  case TRUE:
  case FALSE:
    return f;

  case AND:
  case OR:
    {
      FormulaList* args = apply(f->args());
      if (args == f->args()) {
	return f;
      }
      return new JunctionFormula(con,args);
    }

  case IMP:
  case IFF:
  case XOR:
    {
      Formula* left = apply(f->left());
      Formula* right = apply(f->right());
      if (left == f->left() && right == f->right()) {
	return f;
      }
      return new BinaryFormula(con,left,right);
    }

  case NOT:
    {
      Formula* arg = apply(f->uarg());
      if (arg == f->uarg()) {
	return f;
      }
      return new NegatedFormula(arg);
    }

  case FORALL:
  case EXISTS:
    {
      Formula* arg = apply(f->qarg());
      if (arg->connective() != con) {
	if (arg == f->qarg()) {
	  return f;
	}
	return new QuantifiedFormula(con,f->vars(),arg);
      }

      
      return new QuantifiedFormula(con,
				   f->vars()->append(arg->vars()),
				   arg->qarg());
    }

  case ITE:
    {
      Formula* c = apply(f->condArg());
      Formula* t = apply(f->thenArg());
      Formula* e = apply(f->elseArg());

      while(c->connective()==NOT) {
	c = c->uarg();
        std::swap(t,e);
      }
      if(c->connective()!=LITERAL) {
	c = introduceDefinition(c);
	LOG("pp_fite","processing ite "<<(*f)<<" with introduced definition "<<(*c));
      }
      else {
	LOG("pp_fite","processing ite "<<(*f)<<" without definition introduction");
      }
      ASS(c->connective()==LITERAL)

      return makeBinaryJunction(OR, makeBinaryJunction(AND,c,t),
	  makeBinaryJunction(AND,new NegatedFormula(c), e));
    }

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }

}

FormulaList* FormulaIteExpander::apply(FormulaList* fs)
{
  CALL("FormulaIteExpander::apply(FormulaList*)");

  if (fs->isEmpty()) {
    return fs;
  }
  Formula* head = apply(fs->head());
  FormulaList* tail = apply(fs->tail());

  if (head == fs->head() && tail == fs->tail()) {
    return fs;
  }
  return new FormulaList(head,tail);
}

Formula* FormulaIteExpander::makeBinaryJunction(Connective con, Formula* f1, Formula* f2)
{
  CALL("FormulaIteExpander::makeBinaryJunction");
  ASS(con == OR || con == AND);

  FormulaList* args = 0;
  FormulaList::push(f1, args);
  FormulaList::push(f2, args);
  return new JunctionFormula(con, args);
}

Formula* FormulaIteExpander::introduceDefinition(Formula* f)
{
  CALL("FormulaIteExpander::introduceDefinition");

  Formula::VarList* fv=f->freeVariables();
  unsigned fvLen=fv->length();
  unsigned dpred=env.signature->addNamePredicate(fvLen);

  static Stack<TermList> args;
  Formula::VarList::Iterator fvit(fv);
  while(fvit.hasNext()) {
    unsigned var=fvit.next();
    args.push(TermList(var, false));
  }
  Literal* dlit=Literal::create(dpred, fvLen, true, false, args.begin());
  Formula* datom=new AtomicFormula(dlit);

  Formula* def=new BinaryFormula(IFF, datom, f);
  if(fv) {
    def=new QuantifiedFormula(FORALL, fv, def);
  }
  Unit* dunit=new FormulaUnit(def, new Inference(Inference::PREDICATE_DEFINITION), Unit::AXIOM);
  UnitList::push(dunit, _defs);

  return datom;
}


}
