
/*
 * File InterpolantMinimizer.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#ifndef __InterpolantMinimizer__
#define __InterpolantMinimizer__

#include "Lib/DHSet.hpp"
#include "Lib/Int.hpp"
#include "Lib/MapToLIFO.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/InferenceStore.hpp"

#include "SMTFormula.hpp"

namespace Shell {

using namespace Lib;
using namespace Kernel;

class InterpolantMinimizer
{
public:
  enum OptimizationTarget
  {
    OT_WEIGHT,
    OT_COUNT,
    OT_QUANTIFIERS
  };

  InterpolantMinimizer(OptimizationTarget target=OT_WEIGHT, bool noSlicing=false,
      bool showStats=false, string statsPrefix="");
  ~InterpolantMinimizer();

  typedef List<UnitSpec> USList;

  Formula* getInterpolant(Unit* refutation);

private:
  

  void traverse(Unit* refutationUnit);

  struct TraverseStackEntry;

  enum UnitState {
    TRANSPARENT_PARENTS,
    HAS_LEFT_PARENT,
    HAS_RIGHT_PARENT
  };

  struct UnitInfo
  {
    UnitInfo() : state(TRANSPARENT_PARENTS), isRefutation(false),
	isParentOfLeft(false), isParentOfRight(false), leadsToColor(false),
	leftSuccessors(0), rightSuccessors(0), transparentSuccessors(0) {}

    Color color;
    Color inputInheritedColor;

    UnitState state;

    bool isRefutation;
    bool isParentOfLeft;
    bool isParentOfRight;

   
    bool leadsToColor;

    
    List<UnitSpec>* leftSuccessors;
    List<UnitSpec>* rightSuccessors;
    List<UnitSpec>* transparentSuccessors;
  };

  typedef DHMap<UnitSpec,UnitInfo> InfoMap;

  InfoMap _infos;

private:
  

  struct ParentSummary
  {
    void reset() {
      rParents.reset();
      bParents.reset();
      gParents.reset();
    }

    Stack<string> rParents;
    Stack<string> bParents;
    Stack<string> gParents;
  };

  enum PredType
  {
   
    R,
   
    B,
   
    G,
   
    S,
   
    RC,
   
    BC,
   
    RF,
   
    BF,
   
    D,
   
    V
  };

  SMTConstant pred(PredType t, string node);
  SMTConstant costFunction();

  void addDistinctColorsFormula(string n);

  void addLeafNodePropertiesFormula(string n);
  void addGreyNodePropertiesFormula(string n, ParentSummary& parents);
  void addColoredParentPropertiesFormulas(string n, ParentSummary& parents);
  void addNodeFormulas(string n, ParentSummary& parents);

  void addFringeFormulas(UnitSpec u);
private:
  

  void addAtomImplicationFormula(UnitSpec u);
  void addCostFormula();

  void collectAtoms(FormulaUnit* f, Stack<string>& atoms);
  string getComponentId(Clause* cl);
  void collectAtoms(UnitSpec u, Stack<string>& atoms);

  class ClauseSplitter;

  DHMap<Clause*, string> _atomIds;
  DHMap<string, string> _formulaAtomIds;

  typedef DHMap<string, unsigned> WeightMap;
  WeightMap _atomWeights;

  DHMap<string,UnitSpec> _unitsById;

  ClauseSplitter* _splitter;
private:
  

  void collectSlicedOffNodes(SMTSolverResult& solverResult, DHSet<UnitSpec>& acc);

  string getUnitId(UnitSpec u);

  void addNodeFormulas(UnitSpec u);

  void addAllFormulas();

  OptimizationTarget _optTarget;
  bool _noSlicing;
  bool _showStats;
  string _statsPrefix;
  SMTBenchmark _resBenchmark;
};

}

#endif 
