
/*
 * File Interpolants.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/DHMap.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/ColorHelper.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/InferenceStore.hpp"
#include "Kernel/SubstHelper.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Unit.hpp"

#include "Flattening.hpp"
#include "SimplifyFalseTrue.hpp"

#include "Interpolants.hpp"

#define ALLOW_SURPRISING_COLORS 1

namespace Shell
{

using namespace Lib;
using namespace Kernel;

typedef pair<Formula*, Formula*> UIPair; 
typedef List<UIPair> UIPairList;

VirtualIterator<UnitSpec> Interpolants::getParents(UnitSpec u)
{
  CALL("Interpolants::getParents");

  if(!_slicedOff) {
    return InferenceStore::instance()->getParents(u);
  }

  static Stack<UnitSpec> toDo;
  static Stack<UnitSpec> parents;

  toDo.reset();
  parents.reset();

  for(;;) {
    VirtualIterator<UnitSpec> pit = InferenceStore::instance()->getParents(u);
    while(pit.hasNext()) {
      UnitSpec par = pit.next();
      if(_slicedOff->find(par)) {
	toDo.push(par);
      }
      else {
	parents.push(par);
      }
    }
    if(toDo.isEmpty()) {
      break;
    }
    u = toDo.pop();
  }

  return getPersistentIterator(Stack<UnitSpec>::BottomFirstIterator(parents));
}

struct ItemState
{
  ItemState() {}

  ItemState(UnitSpec us) : parCnt(0), inheritedColor(COLOR_TRANSPARENT), interpolant(0),
      leftInts(0), rightInts(0), processed(false), _us(us)
  {
    CALL("ItemState::ItemState");
    _usColor = us.unit()->getColor();
  }

  void destroy()
  {
    CALL("ItemState::destroy");

    leftInts->destroy();
    rightInts->destroy();
  }

  UnitSpec us() const { return _us; }
  Color usColor() const { return _usColor; }
 
  VirtualIterator<UnitSpec> pars;
 
  int parCnt;
 
  Color inheritedColor;
 
  Formula* interpolant;
 
  List<UIPair>* leftInts;
 
  List<UIPair>* rightInts;
 
  bool processed;
private:
 
  UnitSpec _us;
  Color _usColor;
};

Comparison compareUIP(const UIPair& a, const UIPair& b)
{
  CALL("compareUIP");

  Comparison res = Int::compare(a.first, b.first);
  if(res==EQUAL) {
    res = Int::compare(a.second, b.second);
  }
  return res;
}

void mergeCopy(UIPairList*& tgt, UIPairList* src)
{
  CALL("mergeCopy");
  if(!tgt) {
    tgt = src->copy();
    return;
  }

  UIPairList** curr = &tgt;
  UIPairList::Iterator sit(src);
  while(sit.hasNext()) {
    UIPair spl = sit.next();
  retry_curr:
    if(*curr) {
      Comparison cmp = compareUIP((*curr)->head(), spl);
      if(cmp!=GREATER) {
	curr = (*curr)->tailPtr();
	if(cmp==EQUAL) {
	  continue;
	}
	else {
	  goto retry_curr;
	}
      }
    }
    *curr = new UIPairList(spl, *curr);
    curr = (*curr)->tailPtr();
  }
}

void generateInterpolant(ItemState& st);

Formula* Interpolants::getInterpolant(Unit* unit)
{
  CALL("Interpolants::getInterpolant");

  typedef DHMap<UnitSpec,ItemState> ProcMap;
  ProcMap processed;

  Stack<ItemState> sts;

  UnitSpec curr=UnitSpec(unit);

  Formula* resultInterpolant = 0;

  int ctr=0;

  for(;;) {
    ItemState st;

    if(processed.find(curr)) {
      st = processed.get(curr);
      ASS(st.us()==curr);
      ASS(st.processed);
      ASS(!st.pars.hasNext());
    }
    else {
      st = ItemState(curr);
      st.pars = getParents(curr);
    }

    if(curr.unit()->inheritedColor()!=COLOR_INVALID) {
      
      st.inheritedColor=ColorHelper::combine(curr.unit()->inheritedColor(), st.usColor());
      ASS_NEQ(st.inheritedColor, COLOR_INVALID);
    }
#if ALLOW_SURPRISING_COLORS
    
    
    
    
    
    
    
    st.inheritedColor=st.usColor();
#else
    else if(!st.processed && !st.pars.hasNext()) {
      
      
      
      st.inheritedColor=st.usColor();
    }
#endif

    if(sts.isNonEmpty()) {
      
      ItemState& pst=sts.top(); 
      pst.parCnt++;
      if(pst.inheritedColor==COLOR_TRANSPARENT) {
        pst.inheritedColor=st.usColor();
      }
#if VDEBUG
      else {
        Color clr=st.usColor();
        ASS_REP2(pst.inheritedColor==clr || clr==COLOR_TRANSPARENT, pst.us().toString(), curr.toString());
      }
      ASS_EQ(curr.unit()->getColor(),st.usColor());
#endif
    }

    sts.push(st);

    for(;;) {
      if(sts.top().pars.hasNext()) {
        curr=sts.top().pars.next();
        break;
      }
      
      st=sts.pop();
      ctr++;
      Color color = st.usColor();
      if(!st.processed) {
	if(st.inheritedColor!=color || sts.isEmpty()) {
	  

	  ASS_REP2(color==COLOR_TRANSPARENT, st.us().toString(), st.inheritedColor);
	  generateInterpolant(st);
	  LOG("itp_sub","intermediate interpolant for "<<st.us().toString()<<endl<<
	      "  "<<(*st.interpolant));
	}
	st.processed = true;
	processed.insert(st.us(), st);
      }
      
      if(sts.isNonEmpty()) {
	
        if(color!=COLOR_LEFT) {
          mergeCopy(sts.top().leftInts, st.leftInts);
        } 
        if(color!=COLOR_RIGHT) {
          mergeCopy(sts.top().rightInts, st.rightInts);
        }
      } 
      else {
	
	
	
        resultInterpolant = st.interpolant;
        goto fin;
      }
    }
  }

fin:

  
  ProcMap::Iterator destrIt(processed);
  while(destrIt.hasNext()) {
    destrIt.next().destroy();
  }

  
  return Flattening::flatten(SimplifyFalseTrue::simplify(resultInterpolant));

}

void generateInterpolant(ItemState& st)
{
  CALL("generateInterpolant");

  Unit* u=st.us().unit();
  Color color=st.usColor();
  ASS_EQ(color, COLOR_TRANSPARENT);

  Formula* interpolant;
  Formula* unitFormula=u->getFormula(st.us().prop());

  

  if(st.parCnt) {
    
    FormulaList* conj=0;
    List<UIPair>* src= (st.inheritedColor==COLOR_LEFT) 
        ? st.rightInts
        : st.leftInts;
    
    List<UIPair>::Iterator sit(src);
    while(sit.hasNext()) {
      UIPair uip=sit.next();
      FormulaList* disj=0;
      FormulaList::push(uip.first, disj);
      FormulaList::push(uip.second, disj);
      FormulaList::push(JunctionFormula::generalJunction(OR, disj), conj);
    }

    if(st.inheritedColor==COLOR_LEFT) {
      
      FormulaList* innerConj=0;
      List<UIPair>::Iterator sit2(src);
      while(sit2.hasNext()) {
        UIPair uip=sit2.next();
        FormulaList::push(uip.first, innerConj);
      }
      FormulaList::push(new NegatedFormula(JunctionFormula::generalJunction(AND, innerConj)), conj);
    }
    else {
      
    }
    interpolant=JunctionFormula::generalJunction(AND, conj);
  }
  else {
    
    if(st.inheritedColor==COLOR_RIGHT) {
      interpolant=new NegatedFormula(unitFormula); 
    }
    else {
      
      interpolant=unitFormula; 
    }
  }
  st.interpolant=interpolant;



  UIPair uip=make_pair(unitFormula, interpolant);
  if(st.inheritedColor==COLOR_LEFT) {
    st.leftInts->destroy();
    st.leftInts=0;
    List<UIPair>::push(uip,st.leftInts);
  }
  else if(st.inheritedColor==COLOR_RIGHT) {
    st.rightInts->destroy();
    st.rightInts=0;
    List<UIPair>::push(uip,st.rightInts);
  }
}

}
