
/*
 * File LaTeX.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <fstream>






#include "Options.hpp"
#include "LaTeX.hpp"


#include "Debug/Tracer.hpp"

#include "Lib/Environment.hpp"
#include "Lib/Exception.hpp"
#include "Lib/Int.hpp"
#include "Lib/List.hpp"

#include "Kernel/BDD.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Unit.hpp"



namespace Shell
{

using namespace Lib;
using namespace Kernel;









string LaTeX::refutationToString(Unit* ref)
{
  CALL("LaTeX::refutationToString(Unit* ref)");

  string res = "\\documentclass[fleqn]{article}\n"
    "\\usepackage{fullpage,latexsym}\n"

    "\\newenvironment{VampireProof}{%\n"
    "	\\section{Proof}}{}\n"
    "\\newenvironment{VampireInference}{%\n"
    "	\\begin{array}{c}}{\\end{array}}\n"
    "\\newenvironment{VampireInferencePremises}{}{}\n"
    "\\newenvironment{VampirePremise}%\n"
    "	{\\begin{array}{l}}%\n"
    "	{\\end{array}}\n"
    "\\newenvironment{VampireConclusion}%\n"
    "	{\\begin{array}{l}}%\n"
    "	{\\end{array}}\n"
    "\\newcommand{\\VampireUnit}[3]{%\n"
    "	#1.~#2~[#3]}\n"

    "\\newcommand{\\VPremiseSeparator}{\\\\}\n"
    "\\newcommand{\\VConclusionSeparator}{\\\\ \\hline}\n"

    "\\newcommand{\\Vor}{\\vee}\n"
    "\\newcommand{\\Vand}{\\wedge}\n"
    "\\newcommand{\\Vimp}{\\rightarrow}\n"
    "\\newcommand{\\Viff}{\\leftrightarrow}\n"
    "\\newcommand{\\Vxor}{\\not\\Viff}\n"

    "\\newcommand{\\VEmptyClause}{\\Box}\n"

    "\\begin{document}\n"
    "\\begin{VampireProof}\n";


  BDD* bdd=BDD::instance();
  InferenceStore* is=InferenceStore::instance();

  Stack<UnitSpec> outKernel;
  Set<UnitSpec> handledKernel;
  Stack<Unit*> outShell;
  Set<Unit*> handledShell;

  if( ref->isClause() && static_cast<Clause*>(ref)->prop() ) {
    Clause* refCl=static_cast<Clause*>(ref);
    ASS( bdd->isFalse(refCl->prop()) );
    UnitSpec cs=UnitSpec(refCl);
    outKernel.push(cs);
    handledKernel.insert(cs);
  } else {
    outShell.push(ref);
    handledShell.insert(ref);
  }

  while(outKernel.isNonEmpty()) {
    UnitSpec cs=outKernel.pop();
    InferenceStore::FullInference* finf;
    if(bdd->isTrue(cs.prop())) {
      
      ASSERTION_VIOLATION;
    } else if(is->findInference(cs, finf)) {
      InferenceStore::SplittingRecord* srec;
      if(finf->rule==Inference::SPLITTING && is->findSplitting(cs, srec)) {
	if(!bdd->isTrue(srec->premise.prop()) && !handledKernel.contains(srec->premise)) {
	  handledKernel.insert(srec->premise);
	  outKernel.push(srec->premise);
	}
	res+=splittingToString(srec);
	continue;
      }

      res+=toStringAsInference(cs, finf);

      for(unsigned i=0;i<finf->premCnt;i++) {
	UnitSpec prem=finf->premises[i];
	ASS(prem!=cs);
	Clause* premCl=prem.cl();
	ASS(premCl->prop());

	if(!bdd->isTrue(prem.prop()) && !handledKernel.contains(prem)) {
	  handledKernel.insert(prem);
	  outKernel.push(prem);
	}
      }

    } else {
      Clause* cl=cs.cl();
      Inference* inf = cl->inference();

      res+=toStringAsInference(cl);
      Inference::Iterator it = inf->iterator();
      bool first=true;
      while (inf->hasNext(it)) {
	Unit* prem=inf->next(it);
	first=false;
	if(prem->isClause() && static_cast<Clause*>(prem)->prop()) {
	  
	  UnitSpec premCS=UnitSpec(prem, bdd->getFalse());

	  if(!handledKernel.contains(premCS)) {
	    handledKernel.insert(premCS);
	    outKernel.push(premCS);
	  }
	} else {
	  if(!handledShell.contains(prem)) {
	    handledShell.insert(prem);
	    outShell.push(prem);
	  }
	}
      }
    }
  }

  while(outShell.isNonEmpty()) {
    Unit* unit=outShell.pop();
    Inference* inf = unit->inference();

    res+=toStringAsInference(unit);
    Inference::Iterator it = inf->iterator();
    bool first=true;
    while (inf->hasNext(it)) {
      Unit* prem=inf->next(it);
      first=false;
      if(!handledShell.contains(prem)) {
	handledShell.insert(prem);
	outShell.push(prem);
      }
    }
  }

  if(definitionStack.isNonEmpty()) {
    res+="\\section{BDD Definitions}\n";
    for(unsigned i=0;i<definitionStack.length();i++) {
      res+=definitionStack[i];
    }
  }

  return res + "\\end{VampireProof}\n"
    "\\end{document}\n";
}


string LaTeX::toString(Unit* u)
{
  CALL("LaTeX::toString(Unit* u)");

  if(u->isClause()) {
    return toString(static_cast<Clause*>(u));
  } else {
    return toString(static_cast<FormulaUnit*>(u)->formula());
  }
}

string LaTeX::toString(Unit* u, BDDNode* prop)
{
  CALL("LaTeX::toString(Unit* u, BDDNode* prop)");

  if(u->isClause()) {
    return toString(static_cast<Clause*>(u), prop);
  } else {
    ASS(prop==0 || BDD::instance()->isFalse(prop));
    return toString(static_cast<FormulaUnit*>(u)->formula());
  }
}

string LaTeX::toString (Formula* f) const
{
  CALL("LaTeX::toString(Formula* f)");

  static string names [] =
  { "", " \\Vand ", " \\Vor ", " \\Vimp ", " \\Viff ", " \\Vxor ",
	  "\\neg ", "\\forall ", "\\exists ", "\bot", "\top"};

  Connective c = f->connective();
  string con = names[(int)c];
  switch (c) {
  case LITERAL:
    return toString(f->literal());

  case AND:
  case OR:
  {
    FormulaList::Iterator arg(f->args());
    ASS(arg.hasNext());
    string result = toString(arg.next(),c);
    while (arg.hasNext()) {
      result += con + toString(arg.next(),c);
    }
    return result;
  }

  case IMP:
  case IFF:
  case XOR:
    return toString(f->left(),c) + con + toString(f->right(),c);

  case NOT:
    if (f->uarg()->connective() == LITERAL &&
	    f->uarg()->literal()->isEquality()) {
      Literal* l = f->uarg()->literal();
      return toString(l->nthArgument(0)) + " \\neq " + toString(l->nthArgument(1));
    }
    return con + toString(f->uarg(),c);

  case FORALL:
  case EXISTS:
  {
    string result("(");
    Formula::VarList::Iterator vs(f->vars());
    while (vs.hasNext()) {
      result += con + varToString(vs.next()) + string(" ");
    }
    return result + ")" + toString(f->qarg(),c);
  }

  case FALSE:
  case TRUE:
    return con;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 

string LaTeX::toString (Formula* f, Connective outer) const
{
  CALL("LaTeX::toString (Formula* f, Connective outer)");

  return f->parenthesesRequired(outer) ?
	  string("(") + toString(f) + ")" :
	  toString(f);
} 


string LaTeX::toString (Clause* c)
{
  CALL("LaTeX::toString (Clause* c)");

  return toString(c,c->prop());
} 
string LaTeX::toString (Clause* c, BDDNode* prop)
{
  CALL("LaTeX::toString (Clause* c)");

  string result;

  if (c->isEmpty()) {
    result = "\\VEmptyClause";
  }
  else {
    result = toString((*c)[0]);

    unsigned clen=c->length();
    for(unsigned i=1;i<clen;i++) {
      result += string(" \\Vor ") + toString((*c)[i]);
    }
  }

  if(prop && !BDD::instance()->isFalse(prop)) {
    result += string(" \\Vor ") + toString(prop);
  }

  return result;
} 


string LaTeX::toString (Literal* l) const
{
  CALL("LaTeX::toString (Literal* l)");

  if (l->isEquality()) {
    if (l->isNegative()) {
      return toString(l->nthArgument(0)) + " \\neq " + toString(l->nthArgument(1));
    }
    else {
      return toString(l->nthArgument(0)) + "=" + toString(l->nthArgument(1));
    }
  }








  string res;

  if (l->isNegative()) {
    res="\\neg ";
  }
  res+=symbolToString(l->functor(), true) + toString(l->args());

  return res;
} 


string LaTeX::symbolToString (unsigned num, bool pred) const
{
  CALL("LaTeX::symbolToString (unsigned num, bool pred)");

  string symbolName; 

  if(pred) {
    symbolName = env.signature->predicateName(num);
  }
  else {



    symbolName = env.signature->functionName(num);
  }

  
#define LENGTH 8004
  char newName[LENGTH]; 
  char* name = newName;
  const char* nm = symbolName.substr(0,8000).c_str();
  
  const char* end = nm;
  while (*end) {
    end++;
  }
  
  
  const char* digits = end;
  while (nm != digits) {
    if (digits[-1] >= '0' && digits[-1] <= '9') {
      digits--;
    }
    else {
      break;
    }
  }
  if (digits == nm) { 
    return symbolName;
  }
  while (nm < digits) {
    switch (*nm) {
    case '$':
    case '_':
      *name++ = '\\';
      *name++ = *nm;
      break;
    default:
      *name++ = *nm;
      break;
    }
    nm++;
  }
  if (digits == end) {
    *name = 0;
    return string("\\mathit{") + newName + '}';
  }
  
  *name++ = '_';
  *name++ = '{';
  while (digits != end) {
    *name++ = *digits++;
  }
  *name++ = '}';
  *name = 0;
  return string("\\mathit{") + newName + '}';
}


string LaTeX::toString (TermList* terms) const
{
  CALL("LaTeX::toString (TermList* terms)");

  if (terms->isEmpty()) {
    return "";
  }

  string result = string("(");
  bool first=true;
  TermList* t=terms;
  while(t->isNonEmpty()) {






    if(t->isVar()) {
      ASS(t->isOrdinaryVar());

    }
    else {
      ASS(t->isTerm());
      Term* trm=t->term();
      result += symbolToString(trm->functor(), false) + toString(trm->args());
    }

    t=t->next();
    if(first) {
      first=false;
    }
    else if(t->isNonEmpty()){
      result += ",";
    }
  }
  return result + ")";
}























string LaTeX::getClauseLatexId(UnitSpec cs)
{
  return Int::toString(cs.cl()->number())+"_{"+InferenceStore::instance()->getClauseIdSuffix(cs)+"}";
}

string LaTeX::toStringAsInference(UnitSpec cs, InferenceStore::FullInference* inf)
{
  CALL("LaTeX::toStringAsInference(ClauseSpec,FullInference*)");

  string res("[$");

  bool hasParents=inf->premCnt;
  for(unsigned i=0;i<inf->premCnt;i++) {
    UnitSpec prem=inf->premises[i];
    res += getClauseLatexId(prem);
    if(i+1<inf->premCnt) {
	res += ",";
    }
  }
  if(hasParents) {
    res += "\\rightarrow ";
  }
  res += getClauseLatexId(cs)
    +"$, "+Inference::ruleName(inf->rule)+"]\\\\*[-2ex]\n";

  res += "\\[\\begin{VampireInference}\n";

  if(hasParents) {
    for(unsigned i=0;i<inf->premCnt;i++) {
      UnitSpec prem=inf->premises[i];
      res += "\\begin{VampirePremise}%\n~~";
      res += toString(prem.cl(),prem.prop());
      res += "\n\\end{VampirePremise}\n";
      if(i+1<inf->premCnt) {
	res += "\\VPremiseSeparator\n";
      }
    }
    res += "\\VConclusionSeparator\n";
  }

  res += "\\begin{VampireConclusion}\n~~";

  res += toString(cs.cl(),cs.prop());

  return res + "\n\\end{VampireConclusion}\n\\end{VampireInference}\n\\]\n";
}

string LaTeX::toStringAsInference(Unit* unit)
{
  CALL("LaTeX::toStringAsInference(Unit* unit)");

  Inference* inf = unit->inference();

  string res("[$");

  bool hasParents=false;
  Inference::Iterator it = inf->iterator();
  while (inf->hasNext(it)) {
    hasParents=true;
    Unit* prem=inf->next(it);
    res += Int::toString(prem->number());
    if(inf->hasNext(it)) {
	res += ",";
    }
  }
  if(hasParents) {
    res += "\\rightarrow ";
  }
  res += Int::toString(unit->number())+"$, "+Inference::ruleName(inf->rule())+"]\\\\*[-2ex]\n";

  res += "\\[\\begin{VampireInference}\n";

  if(hasParents) {
    it = inf->iterator();
    while (inf->hasNext(it)) {
	Unit* prem=inf->next(it);
      res += "\\begin{VampirePremise}%\n~~";
      res += toString(prem,0);
      res += "\n\\end{VampirePremise}\n";
	if(inf->hasNext(it)) {
	  res += "\\VPremiseSeparator\n";
	}
    }
    res += "\\VConclusionSeparator\n";
  }

  res += "\\begin{VampireConclusion}\n~~";

  res += toString(unit,0);

  return res + "\n\\end{VampireConclusion}\n\\end{VampireInference}\n\\]\n";
}

string LaTeX::splittingToString(InferenceStore::SplittingRecord* sr)
{
  CALL("LaTeX::splittingToString");

  string res("[$");
  res += getClauseLatexId(sr->premise);


  Stack<pair<int,Clause*> >::Iterator ncit(sr->namedComps);
  while(ncit.hasNext()) {
    res += string(",")+Int::toString(ncit.next().second->number())+"_D";
  }
  res += "\\rightarrow ";
  res += getClauseLatexId(sr->result)
    +"$, "+Inference::ruleName(Inference::SPLITTING)+"]\\\\*[-2ex]\n";


  res += "\\[\\begin{VampireInference}\n";

  res += "\\begin{VampirePremise}%\n~~";
  res += toString(sr->premise.cl(),sr->premise.prop());
  res += "\n\\end{VampirePremise}\n";

  Stack<pair<int,Clause*> >::Iterator ncit2(sr->namedComps);
  while(ncit2.hasNext()) {
    pair<int,Clause*> nrec=ncit2.next();
    res += "\\VPremiseSeparator\n";
    res += "\\begin{VampirePremise}%\n~~";
    if(nrec.first>0) {
      res += getBDDVarName(nrec.first);
    }
    else {
      res += "\\neg " + getBDDVarName(-nrec.first);
    }
    res += "\\Viff" + toString(nrec.second,0);
    res += "\n\\end{VampirePremise}\n";
  }
  res += "\\VConclusionSeparator\n";

  res += "\\begin{VampireConclusion}\n~~";

  res += toString(sr->result.cl(),sr->result.prop());

  return res + "\n\\end{VampireConclusion}\n\\end{VampireInference}\n\\]\n";
}



string LaTeX::varToString (unsigned num) const
{
  CALL("LaTeX::varToString (unsigned num)");






















  return string("x_{") + Int::toString(num) + "}";
} 
































string LaTeX::getBDDVarName(int var)
{
  CALL("LaTeX::getBDDVarName(int var)");

  string name;
  if(BDD::instance()->getNiceName(var, name)) {
    return string("\\mathit{") + name + '}';
  }
  return string("\\mathit{b_{") + Int::toString(var) + "}}";
}

string LaTeX::toString(BDDNode* node)
{
  CALL("LaTeX::toString(BDDNode*)");

  BDD* inst=BDD::instance();

  
  if(inst->isTrue(node)) {
    return "\\top";
  }
  if(inst->isFalse(node)) {
    return "\\bot";
  }

  string name;
  if(_nodeNames.find(node, name)) {
    return name;
  }


  string propPred=getBDDVarName(node->_var);
  if(inst->isTrue(node->_pos) && inst->isFalse(node->_neg)) {
    return propPred;
  }
  else if(inst->isFalse(node->_pos) && inst->isTrue(node->_neg)) {
    return "\\neg "+propPred;
  }
  else if(inst->isTrue(node->_pos)) {
    return "("+propPred+" \\Vor "+toString(node->_neg)+")";
  }
  else if(inst->isFalse(node->_neg)) {
    return "("+propPred+" \\Vand "+toString(node->_pos)+")";
  }
  else if(inst->isFalse(node->_pos)) {
    return "(\\neg "+propPred+" \\Vand "+toString(node->_neg)+")";
  }
  else if(inst->isTrue(node->_neg)) {
    return "(\\neg "+propPred+" \\Vor "+toString(node->_pos)+")";
  }
  else {
    string posDef=toString(node->_pos);
    string negDef=toString(node->_neg);

    string name=string("\\mathit{n_{") + Int::toString(_nextNodeNum++) + "}}";
    string report="$"+name+" \\Viff ("+propPred+" ? "+posDef+" : "+negDef+")$\\\\\n";
    definitionStack.push(report);
    ALWAYS(_nodeNames.insert(node, name));

    return name;
  }

}


}
