
/*
 * File Lexer.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"

#include "Lib/Int.hpp"

#include "Lexer.hpp"

using namespace Shell;
using namespace Lib;

Lexer::Lexer (istream& in)
  : _charBuffer(512),
    _charCursor(0),
    _stream(in),
    _eof(false),
    _lineNumber(1),
    _lookAheadChar(0)
{
  readNextChar();
} 


bool Lexer::readNextChar ()
{
  CALL("Lexer::readNextChar");

  if (_lookAheadChar) {
    _lastCharacter = _lookAheadChar;
    _lookAheadChar = 0;
    if (_lastCharacter == -1) {
      _eof = true;
      return false;
    }
    return true;
  }

  if (_eof) {
    return false;
  }

  _lastCharacter = _stream.get();
  if (_lastCharacter == -1) {
    _eof = true;
    return false;
  }

  if (_lastCharacter == '\n') {
    _lineNumber++;
  }
  return true;
} 


void Lexer::readNumber (Token& token)
{
  CALL("Lexer::readNumber");

  if (_lastCharacter == '-') {
    saveLastChar();
    readNextChar();
  }
  readUnsignedInteger();
  if (_lastCharacter == '.') {
    saveLastChar();
    if (readNextChar()) {
      if (isDigit(_lastCharacter)) {
	readUnsignedInteger();
	token.tag = TT_REAL;
	saveTokenText(token);
	return;
      }
    }
    saveTokenText(token);
    throw LexerException((string)"incorrect number format in " + token.text,
			 *this);
  }
  token.tag = TT_INTEGER;
  saveTokenText(token);
} 


void Lexer::saveLastChar ()
{
  CALL("Lexer::saveLastChar");
  _charBuffer[_charCursor++]  = (char)_lastCharacter;
} 


void Lexer::saveChar (int character)
{
  CALL("Lexer::saveChar");

  _charBuffer[_charCursor++]  = (char)character;
} 


void Lexer::saveTokenText (Token& token)
{
  CALL("Lexer::saveTokenText");

  _charBuffer[_charCursor] = 0;
  token.text = _charBuffer.content();
} 


void Lexer::readUnsignedInteger ()
{
  CALL("TPTPLexer::readUnsignedInteger");

  saveLastChar();

  while (readNextChar() && isDigit(_lastCharacter)) {
    saveLastChar();
  }
} 

LexerException::LexerException (string message,const Lexer& lexer)
  : _message (message)
{
  if (lexer.isAtEndOfFile()) {
    _message += " at end of input";
    return;
  }
  int line = lexer.lineNumber();
  if (lexer.lastCharacter() == '\n') {
    line--;
  }
  _message += " in line ";
  _message += Int::toString(line);
} 


void LexerException::cry (ostream& out)
{
  out << "Lexer exception: " << _message << '\n';
} 


void Lexer::readSequence (const char* cs)
{
  while (*cs) {
    readNextChar();
    if (lastCharacter() != *cs) {
      throw LexerException((string)cs + 
			   " expected",*this);
    }
    cs++;
  }
  readNextChar();
} 


int Lexer::lookAhead()
{
  CALL("Lexer::lookAhead");
  ASS(! _lookAheadChar); 

  _lookAheadChar = _stream.get();
  return _lookAheadChar;
} 

