
/*
 * File NNF.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"

#include "Kernel/Inference.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/Unit.hpp"

#include "Indexing/TermSharing.hpp"

#include "NNF.hpp"

using namespace Kernel;
using namespace Shell;

FormulaUnit* NNF::ennf(FormulaUnit* unit)
{
  CALL("NNF::ennf(Unit* u)");
  ASS(! unit->isClause());

  Formula* f = static_cast<FormulaUnit*>(unit)->formula();
  Formula* g = ennf(f,true);
  if (f == g) { 
    return unit;
  }

  return new FormulaUnit(g,
			 new Inference1(Inference::ENNF,unit),
			 unit->inputType());
} 


FormulaUnit* NNF::nnf(FormulaUnit* unit)
{
  CALL("NNF::nnf(Unit*)");
  ASS(! unit->isClause());

  Formula* f = static_cast<FormulaUnit*>(unit)->formula();
  Formula* g = nnf(f,true);
  if (f == g) { 
    return unit;
  }

  return new FormulaUnit(g,
			 new Inference1(Inference::NNF,unit),
			 unit->inputType());
} 


Formula* NNF::ennf (Formula* f, bool polarity)
{
  CALL("NNF::ennf(Formula*...)");

  Connective c = f->connective();
  switch (c) {
  case LITERAL:
    if (! polarity) {
      Literal* lit = f->literal();
      Literal* newLit = Literal::complementaryLiteral(lit);
      return new AtomicFormula(newLit);
    }
    return f;

  case AND: 
  case OR: 
    {
      FormulaList* fs = f->args();
      FormulaList* gs = ennf(fs,polarity);
      if (fs == gs) {
	ASS(polarity);
	return f;
      }
      if (polarity) {
 	return new JunctionFormula(c,gs);
      }
      return new JunctionFormula(c == AND ? OR : AND,gs);
    }
      
  case IMP: 
    {
      Formula* l = ennf(f->left(),! polarity);
      Formula* r = ennf(f->right(),polarity);
      FormulaList* args = new FormulaList(l,new FormulaList(r));
      return new JunctionFormula(polarity ? OR : AND,args);
    }
    
  case IFF: 
  case XOR: 
    {
      Formula* l = f->left();
      Formula* r = f->right();
      Formula* ll = ennf(l,true);
      Formula* rr = ennf(r,true);
      if (polarity) {
	if (l == ll && r == rr) { 
	  return f;
	}
	return new BinaryFormula(c, ll, rr);
      }
      return new BinaryFormula(c == XOR ? IFF : XOR, ll, rr);
    }

    case NOT:
      return ennf(f->uarg(),!polarity);

  case FORALL:
  case EXISTS: 
    {
      Formula* g = f->qarg();
      Formula* gg = ennf(g,polarity);
      if (g == gg) {
	ASS(polarity);
	return f;
      }
      if (polarity) {
	return new QuantifiedFormula(c,f->vars(),gg);
      }
      return new QuantifiedFormula(c == EXISTS ? FORALL : EXISTS,
				   f->vars(),gg);
    }

  case TRUE:
  case FALSE:
    if(polarity) {
      return f;
    }
    else {
      if(c==TRUE) {
	return Formula::falseFormula();
      }
      else {
	return Formula::trueFormula();
      }
    }
#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 


FormulaList* NNF::ennf (FormulaList* fs, bool polarity)
{
  CALL("NNF::ennf(FormulaList*...)");

  if (fs->isEmpty()) {
    return fs;
  }

  FormulaList* result = FormulaList::empty();
  FormulaList::FIFO stack(result);
  bool changed = false;
  FormulaList::Iterator it(fs);
  while (it.hasNext()) {
    Formula* f = it.next();
    Formula* g = ennf(f,polarity);
    stack.push(g);
    if (f != g) {
      changed = true;
    }
  }
  if (changed) {
    return result;
  }
  result->destroy();
  return fs;
} 


Formula* NNF::nnf (Formula* f, bool polarity)
{
  CALL("NNF::nnf(Formula*...)");

  Connective c = f->connective();
  switch (c) {
  case LITERAL:
    if (! polarity) {
      Literal* lit = f->literal();
      Literal* newLit = Literal::complementaryLiteral(lit);
      return new AtomicFormula(newLit);
    }
    return f;

  case AND: 
  case OR: 
    {
      FormulaList* fs = f->args();
      FormulaList* gs = nnf(fs,polarity);
      if (fs == gs) {
	return f;
      }
      if (polarity) {
	return new JunctionFormula(c,gs);
      }
      return new JunctionFormula(c == AND ? OR : AND,gs);
    }
      
  case IMP: 
    {
      Formula* l = nnf(f->left(),! polarity);
      Formula* r = nnf(f->right(),polarity);
      FormulaList* args = new FormulaList(l,new FormulaList(r));
      return new JunctionFormula(polarity ? OR : AND,args);
    }
    
  case IFF: 
  case XOR: 
    {
      Formula* l = f->left();
      Formula* r = f->right();
      Formula* g;

      if (polarity ? c == IFF : c == XOR) {
	
	
	g = new JunctionFormula(AND,
				new FormulaList(new BinaryFormula(IMP,l,r),
						new FormulaList(new BinaryFormula(IMP,r,l))));
      }
      else {
	
	
	g = new JunctionFormula(AND,
				new FormulaList(new JunctionFormula(OR,
								    new FormulaList(l,
										    new FormulaList(r))),
						new FormulaList(new JunctionFormula(OR,
										    new FormulaList(new NegatedFormula(l),
												    new FormulaList(new NegatedFormula(r)))))));
      }
      return nnf(g,true);
    }

    case NOT:
      return nnf(f->uarg(),!polarity);

  case FORALL:
  case EXISTS: 
    {
      Formula* g = f->qarg();
      Formula* gg = nnf(g,polarity);

      if (g == gg) {
	ASS(polarity);
	return f;
      }
      if (polarity) {
	return new QuantifiedFormula(c,f->vars(),gg);
      }
      return new QuantifiedFormula(c == EXISTS ? FORALL : EXISTS,
				   f->vars(),gg);
    }

  case TRUE:
  case FALSE:
    return f;
#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 


FormulaList* NNF::nnf (FormulaList* fs, bool polarity)
{
  CALL("NNF::nnf(FormulaList*...)");

  if (fs->isEmpty()) {
    return fs;
  }

  FormulaList* result = FormulaList::empty();
  FormulaList::FIFO stack(result);
  bool changed = false;
  FormulaList::Iterator it(fs);
  while (it.hasNext()) {
    Formula* f = it.next();
    Formula* g = nnf(f,polarity);
    stack.push(g);
    if (f != g) {
      changed = true;
    }
  }
  if (changed) {
    return result;
  }
  result->destroy();
  return fs;
} 


