
/*
 * File Naming.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/RuntimeStatistics.hpp"

#include "Lib/Allocator.hpp"
#include "Lib/DHMap.hpp"
#include "Lib/Int.hpp"
#include "Lib/Environment.hpp"

#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/SubformulaIterator.hpp"
#include "Kernel/Term.hpp"

#include "Shell/Statistics.hpp"

#include "Indexing/TermSharing.hpp"

#include "Naming.hpp"

using namespace Lib;
using namespace Kernel;
using namespace Shell;

Naming::Naming (int threshold, bool preserveEpr)
  : _threshold(threshold+1), _preserveEpr(preserveEpr), _varsInScope(false)
{
  ASS(threshold < 32768);
} 


FormulaUnit* Naming::apply (FormulaUnit* unit,UnitList*& defs)
{
  CALL("Naming::apply(Unit*)");
  ASS(! unit->isClause());
  ASS_REP(unit->formula()->freeVariables()==0, *unit);
  ASS(!_varsInScope); 

  LOG_UNIT("pp_naming_args",unit);

  Formula* f = unit->formula();
  switch (f->connective()) {
  case TRUE:
  case FALSE:
    defs = UnitList::empty();
    return unit;
  default:
    break;
  }

  int pos;
  int neg;
  _defs = UnitList::empty();
  Formula* g = apply(f,ON_TOP,pos,neg);
  if (f == g) { 
    ASS(_defs->isEmpty());
    defs = UnitList::empty();
    return unit;
  }
  ASS(! _defs->isEmpty());
  defs = _defs;
  UnitList* premises = _defs->copy();
  UnitList::push(unit, premises);
  return new FormulaUnit(g,
			 new InferenceMany(Inference::DEFINITION_FOLDING,premises),
			 unit->inputType());
} 



Formula* Naming::apply (Formula* f,Where where,int& pos,int& neg)
{
  CALL("Naming::apply(Formula* ...)");

  switch (f->connective()) {
  case LITERAL:
    pos = 1;
    neg = 1;
    return f;

  case AND: {
    FormulaList* fs = f->args();
    int length = fs->length();
    void* mem = ALLOC_UNKNOWN(length*sizeof(int),"Naming::apply");
    int* cls = array_new<int>(mem, length);
    int* negCls = 0;
    if (where == UNDER_IFF) {
      mem = ALLOC_UNKNOWN(length*sizeof(int),"Naming::apply");
      negCls = array_new<int>(mem, length);
    }
    fs = apply(fs,where,cls,negCls);
    bool split = false;
    
    Formula** gs = 0;
    
    for (;;) {
      int sum = 0; 
      int product = 1;
      int maxPos = 0;
      int maxNeg = 0;
      int maxPosIndex = 0;
      int maxNegIndex = 0;
      FormulaList* currArg = f->args();
      for (int i = 0;i < length;i++) {
	int c = cls[i];
	sum = Int::min(_threshold,sum+c);
	bool canBeDefEvaluated=false;
	bool canBeDef;
	if (c > maxPos) {
	  canBeDefEvaluated=true;
	  canBeDef=canBeInDefinition(currArg->head(), where);
	  if(canBeDef) {
	    maxPos = c;
	    maxPosIndex=i;
	  }
	}
	if (where == UNDER_IFF) {
	  int d = negCls[i];
	  product = Int::min(_threshold,product*d);
	  if (d > maxNeg) {
	    if(!canBeDefEvaluated) {
	      canBeDef=canBeInDefinition(currArg->head(), where);
	    }
	    if(canBeDef) {
	      maxNeg = d;
	      maxNegIndex=i;
	    }
	  }
	}
	currArg=currArg->tail();
      }
      ASS(currArg==0);
      ASS(maxPos > 0 || _preserveEpr);
      ASS(where != UNDER_IFF || maxNeg > 0);
      
      
      int splitWhat = 0;
      if (where != UNDER_IFF || product > 1) { 
	if (where != ON_TOP && maxPos > 1 && sum >= _threshold) {
	  splitWhat |= 1;
	}
	if (where == UNDER_IFF && product >= _threshold && maxNeg > 1) {
	  splitWhat |= 2;
	}
      }
      if (! splitWhat) { 
	if (split) {
	  FormulaList* rs = FormulaList::empty();
	  for (int i = length-1;i >= 0;i--) {
	    FormulaList::push(gs[i],rs);
	  }
	  f = new JunctionFormula(AND,rs);
 	  DEALLOC_UNKNOWN(gs,"Naming::apply");
	}
	else if (fs != f->args()) {
	  f = new JunctionFormula(AND,fs);
	}
 	DEALLOC_UNKNOWN(cls,"Naming::apply");
	if (negCls) {
 	  DEALLOC_UNKNOWN(negCls,"Naming::apply");
	}
	neg = product;
	pos = sum;
	return f;
      }

      
      split = true;
      if (! gs) {
	void* mem = ALLOC_UNKNOWN(length*sizeof(Formula*),"Naming::apply");
	gs = array_new<Formula*>(mem, length);
	int j = 0;
	FormulaList::Iterator hs(fs);
	while (hs.hasNext()) {
	  gs[j] = hs.next();
	  j++;
	}
      }
      
      int maxIndex = splitWhat == 1 ||
                     (splitWhat == 3 && sum > product) ||
                     maxNeg == 1
                       ? maxPosIndex
	               : maxNegIndex;
      Formula* nm = introduceDefinition(gs[maxIndex],where==UNDER_IFF);
      gs[maxIndex] = nm;
      cls[maxIndex] = 1;
      if (where == UNDER_IFF) {
	negCls[maxIndex] = 1;
      }
    } 
  } 

  case OR: {
    FormulaList* fs = f->args();
    int length = fs->length();
    void* mem = ALLOC_UNKNOWN(length*sizeof(int),"Naming::apply");
    int* cls = array_new<int>(mem, length);
    int* negCls = 0;
    if (where == UNDER_IFF) {
      mem = ALLOC_UNKNOWN(length*sizeof(int),"Naming::apply");
      negCls = array_new<int>(mem, length);
    }
    if(where==ON_TOP) {
      where=OTHER;
    }
    fs = apply(fs,where,cls,negCls);
    bool split = false;
    
    Formula** gs = 0;
    
    for (;;) {
      int sum = 0; 
      int product = 1;
      int maxPos = 0;
      int maxNeg = 0;
      int maxPosIndex = 0;
      int maxNegIndex = 0;
      FormulaList* currArg = f->args();
      for (int i = 0;i < length;i++) {
	int c = cls[i];
	product = Int::min(_threshold,product*c);
	bool canBeDefEvaluated=false;
	bool canBeDef;
	if (c > maxPos) {
	  canBeDefEvaluated=true;
	  canBeDef=canBeInDefinition(currArg->head(), where);
	  if(canBeDef) {
	    maxPos = c;
	    maxPosIndex=i;
	  }
	}
	if (where == UNDER_IFF) {
	  int d = negCls[i];
	  sum = Int::min(_threshold,sum+d);
	  if (d > maxNeg) {
	    if(!canBeDefEvaluated) {
	      canBeDef=canBeInDefinition(currArg->head(), where);
	    }
	    if(canBeDef) {
	      maxNeg = d;
	      maxNegIndex=i;
	    }
	  }
	}
	currArg = currArg->tail();
      }
      ASS(currArg==0);
      ASS(maxPos > 0 || _preserveEpr);
      ASS(where != UNDER_IFF || maxNeg > 0);
      
      
      int splitWhat = 0;
      if ( maxPos>0 && (where != UNDER_IFF || product > 1) ) { 
	if (product >= _threshold && maxPos>1) {
	  splitWhat |= 1;
	}
	if (where == UNDER_IFF && sum >= _threshold && maxNeg>1) {
	  splitWhat |= 2;
	}
      }
      if (! splitWhat) { 
	if (split) {
	  FormulaList* rs = FormulaList::empty();
	  for (int i = length-1;i >= 0;i--) {
	    FormulaList::push(gs[i],rs);
	  }
	  f = new JunctionFormula(OR,rs);
 	  DEALLOC_UNKNOWN(gs,"Naming::apply");
	}
	else if (fs != f->args()) {
	  f = new JunctionFormula(OR,fs);
	}
 	DEALLOC_UNKNOWN(cls,"Naming::apply");
 	if (negCls) {
 	  DEALLOC_UNKNOWN(negCls,"Naming::apply");
	}
	neg = sum;
	pos = product;
	return f;
      }

      
      split = true;
      if (! gs) {
	void* mem = ALLOC_UNKNOWN(length*sizeof(Formula*),"Naming::apply");
	gs = array_new<Formula*>(mem, length);
	int j = 0;
	FormulaList::Iterator hs(fs);
	while (hs.hasNext()) {
	  gs[j] = hs.next();
	  j++;
	}
      }
      
      int maxIndex = splitWhat == 1 ||
	             (splitWhat == 3 && product > sum) ||
	             maxNeg == 1
                       ? maxPosIndex
	               : maxNegIndex;
      Formula* nm = introduceDefinition(gs[maxIndex],where==UNDER_IFF);
      gs[maxIndex] = nm;
      cls[maxIndex] = 1;
      if (where == UNDER_IFF) {
	negCls[maxIndex] = 1;
      }
    } 
  } 

  case IFF:
  case XOR: {
    
    

    int negl;
    int posl;
    int negr;
    int posr;
    Connective con = f->connective();
    Formula* l;
    Formula* r;
    if (con == IFF) {
      l = apply(f->left(),UNDER_IFF,posl,negl);
      r = apply(f->right(),UNDER_IFF,posr,negr);
    }
    else {
      l = apply(f->left(),UNDER_IFF,negl,posl);
      r = apply(f->right(),UNDER_IFF,negr,posr);
    }
    if (where == ON_TOP && ((posl == 1 && posr == 1) ||
			    (negl == 1 && negr == 1))) {
      if (l != f->left() || r != f->right()) {
 	f = new BinaryFormula(con,l,r);
      }
      pos = Int::min(_threshold,Int::max(posl,posr));
      return f;
    }
    pos = Int::min(negl*posr + negr*posl,_threshold);
    neg = Int::min(posl*posr+negl*negr,_threshold);
    bool left; 
    if (pos < _threshold) {
      if (where != UNDER_IFF || neg < _threshold) {
	if (l != f->left() || r != f->right()) {
	  f = new BinaryFormula(con,l,r);
	}
	return f;
      }
      
      left = posl*posr > negl*negr
             ? (posl >= posr)
	     : (negl >= negr);
    }
    else { 
      left = negl*posr > negr*posl
	     ? (negl >= posr)
	     : (posl >= negr);
      
      bool splitBoth = left
	               ? (posr+negr >= _threshold)
	               : (posl+negl >= _threshold);
      if (splitBoth) {
	Formula* newl = introduceDefinition(l,true);
	Formula* newr = introduceDefinition(r,true);
	f = new BinaryFormula(con,newl,newr);
	neg = 2;
	pos = 2;
	return f;
      }
    }

    if (left) {
      Formula* newl = introduceDefinition(l,true);
      f = new BinaryFormula(con,newl,r);
      neg = Int::min(posr+negr,_threshold);
      pos = neg;
      return f;
    }

    Formula* newr = introduceDefinition(r,true);
    f = new BinaryFormula(con,l,newr);
    neg = Int::min(posl+negl,_threshold);
    pos = neg;
    return f;
  }

  case FORALL:
  case EXISTS: {
    bool varFlagSet=f->connective()==FORALL && !_varsInScope;
    if(varFlagSet) {
      _varsInScope=true;
    }
    Formula* g = apply(f->qarg(),where,pos,neg);
    ASS (pos <= _threshold || _preserveEpr);
    if (g != f->qarg()) {
      f = new QuantifiedFormula(f->connective(),f->vars(),g);
    }
    if(varFlagSet) {
      _varsInScope=false;
    }
    return f;
  }

#if VDEBUG
  default:
    ASSERTION_VIOLATION_REP(*f);
#endif
  }
} 

bool Naming::canBeInDefinition(Formula* f, Where where)
{
  CALL("Naming::canBeInDefinition");

  if(!_preserveEpr) {
    return true;
  }

  bool unQuant=false;
  bool exQuant=false;
  SubformulaIterator sfit(f);
  while(sfit.hasNext()) {
    Formula* sf=sfit.next();
    if(sf->connective()==FORALL) {
      unQuant=true;
    }
    if(sf->connective()==EXISTS) {
      exQuant=true;
    }
  }


  Formula::VarList* fvars=f->freeVariables();
  bool freeVars=fvars;
  fvars->destroy();

  if(!_varsInScope && freeVars && (exQuant|| (unQuant && where==UNDER_IFF))) {
    return false;
  }

  return true;
}

Literal* Naming::getDefinitionLiteral(Formula* f, Formula::VarList* freeVars)
{
  CALL("Naming::getDefinitionLiteral");

  int length = freeVars->length();
  unsigned pred = env.signature->addNamePredicate(length);
  Signature::Symbol* predSym = env.signature->getPredicate(pred);

  if(env.colorUsed) {
    Color fc=f->getColor();
    if(fc!=COLOR_TRANSPARENT) {
      predSym->addColor(fc);
    }
    if(f->getSkip()) {
      predSym->markSkip();
    }
  }

  static Stack<unsigned> domainSorts;
  static Stack<TermList> predArgs;
  static DHMap<unsigned,unsigned> varSorts;
  domainSorts.reset();
  predArgs.reset();
  varSorts.reset();

  SortHelper::collectVariableSorts(f, varSorts);

  Formula::VarList::Iterator vit(freeVars);
  while(vit.hasNext()) {
    unsigned uvar = vit.next();
    domainSorts.push(varSorts.get(uvar, Sorts::SRT_DEFAULT));
    predArgs.push(TermList(uvar, false));
  }

  predSym->setType(new PredicateType(length, domainSorts.begin()));

  return Literal::create(pred, length, true, false, predArgs.begin());
}

Formula* Naming::introduceDefinition (Formula* f,bool iff)
{
  CALL("Naming::introduceDefinition");
  ASS_NEQ(f->connective(), LITERAL);
  ASS_NEQ(f->connective(), NOT);

  RSTAT_CTR_INC("naming_introduced_defs");

  Formula::VarList* vs;
  vs = f->freeVariables();
  Literal* atom = getDefinitionLiteral(f, vs);
  Formula* name = new AtomicFormula(atom);
  Formula* def;
  if (iff) {
    def = new BinaryFormula(IFF,name,f);
  }
  
  else {
    FormulaList* fs = f->connective() == OR
  		      ? f->args()
                      : new FormulaList(f);
    Formula* nameFormula = new NegatedFormula(name);
    FormulaList::push(nameFormula,fs);
    def = new JunctionFormula(OR,fs);
  }
  if (vs->isNonEmpty()) {
    def = new QuantifiedFormula(FORALL,vs,def);
  }
  Inference* inf = new Inference(Inference::PREDICATE_DEFINITION);
  Unit* definition = new FormulaUnit(def,inf,Unit::AXIOM);
  env.statistics->formulaNames++;
  UnitList::push(definition,_defs);

  LOG_UNIT("pp_naming_defs",definition);
  return name;
} 


FormulaList* Naming::apply (FormulaList* fs,Where where,
			    int* results,int* negResults)
{
  CALL("Naming::apply (FormulaList*...)");

  if (fs->isEmpty()) {
    return fs;
  }

  int neg;
  Formula* g = apply(fs->head(),where,results[0],neg);
  if (where == UNDER_IFF) {
    negResults[0] = neg;
  }
  FormulaList* gs = apply(fs->tail(),where,results+1,negResults+1);
  if (g == fs->head() && gs == fs->tail()) {
    return fs;
  }
  return new FormulaList(g,gs);
} 


