
/*
 * File Preprocess.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "Lib/ScopedLet.hpp"

#include "Kernel/Unit.hpp"
#include "Kernel/Clause.hpp"
#include "Kernel/Problem.hpp"

#include "AIG.hpp"
#include "AIGCompressor.hpp"
#include "AIGConditionalRewriter.hpp"
#include "AIGDefinitionIntroducer.hpp"
#include "AIGInliner.hpp"
#include "AnswerExtractor.hpp"
#include "CNF.hpp"
#include "DistinctProcessor.hpp"
#include "EPRInlining.hpp"
#include "EPRSkolem.hpp"
#include "EqResWithDeletion.hpp"
#include "EqualityPropagator.hpp"
#include "EqualityProxy.hpp"
#include "EquivalenceDiscoverer.hpp"
#include "Flattening.hpp"
#include "FormulaIteExpander.hpp"
#include "FunctionDefinition.hpp"
#include "GeneralSplitting.hpp"
#include "HornRevealer.hpp"
#include "InequalitySplitting.hpp"
#include "InterpretedNormalizer.hpp"
#include "Naming.hpp"
#include "Normalisation.hpp"
#include "NNF.hpp"
#include "Options.hpp"
#include "PDInliner.hpp"
#include "PDMerger.hpp"
#include "PredicateDefinition.hpp"
#include "PredicateIndexIntroducer.hpp"
#include "Preprocess.hpp"
#include "Property.hpp"
#include "Rectify.hpp"
#include "Skolem.hpp"
#include "SimplifyFalseTrue.hpp"
#include "SineUtils.hpp"
#include "Statistics.hpp"
#include "SpecialTermElimination.hpp"
#include "TheoryAxioms.hpp"
#include "TrivialPredicateRemover.hpp"

#include "UIHelper.hpp"














using namespace Shell;

void Preprocess::preprocess (Problem& prb)
{
  CALL("Preprocess::preprocess");

  
























  LOG("pp","preprocessing started");
  TRACE("pp_input",
      UnitList::Iterator uit(prb.units());
      while(uit.hasNext()) {
	Unit* u = uit.next();
	TRACE_OUTPUT_UNIT(u);
      }
      );

  
  
  
  prb.getProperty();

  if(prb.hasInterpretedOperations()) {
    env.interpretedOperationsUsed = true;
  }

  if(prb.hasSpecialTermsOrLets()) {
    SpecialTermElimination().apply(prb);
  }

  
  if (_options.normalize()) {
    env.statistics->phase=Statistics::NORMALIZATION;
    Normalisation().normalise(prb);
  }

  if(_options.sineSelection()!=Options::SS_OFF) {
    env.statistics->phase=Statistics::SINE_SELECTION;
    SineSelector(_options).perform(prb);
  }

  if(_options.questionAnswering()==Options::QA_ANSWER_LITERAL) {
    env.statistics->phase=Statistics::UNKNOWN_PHASE;
    AnswerLiteralManager::getInstance()->addAnswerLiterals(prb);
  }

  if(prb.hasInterpretedOperations() && _options.theoryAxioms()) {
    InterpretedNormalizer().apply(prb);
    env.statistics->phase=Statistics::INCLUDING_THEORY_AXIOMS;
    TheoryAxioms().apply(prb);
  }

  if(prb.mayHaveFormulas()) {
    preprocess1(prb);
  }

  



  if(prb.hasFormulaItes()){
    FormulaIteExpander().apply(prb);
  }

  if ( _options.equalityPropagation() && prb.mayHaveEquality()) {
    
    
    
    
    env.statistics->phase=Statistics::EQUALITY_PROPAGATION;
    EqualityPropagator().apply(prb);
  }

  if (_options.predicateIndexIntroduction()) {
    PredicateIndexIntroducer().apply(prb);
  }

  if (_options.predicateDefinitionInlining()!=Options::INL_OFF) {
    env.statistics->phase=Statistics::PREDICATE_DEFINITION_INLINING;
    PDInliner pdInliner(_options.predicateDefinitionInlining()==Options::INL_AXIOMS_ONLY, false,
	_options.predicateDefinitionInlining()==Options::INL_NON_GROWING);
    pdInliner.apply(prb);

    if(_options.flattenTopLevelConjunctions()) {
        if(TopLevelFlatten().apply(prb)) {
          PDInliner pdInliner2(_options.predicateDefinitionInlining()==Options::INL_AXIOMS_ONLY, false,
              _options.predicateDefinitionInlining()==Options::INL_NON_GROWING);
          pdInliner2.apply(prb);
        }
    }
  }
  else if(_options.flattenTopLevelConjunctions()) {
    TopLevelFlatten().apply(prb);
  }

  if(_options.distinctProcessor()) {
    DistinctProcessor().apply(prb);
  }

  if(_options.predicateEquivalenceDiscovery()!=Options::PED_OFF) {
    EquivalenceDiscoveringTransformer(_options).apply(prb);
  }

  if(_options.aigFormulaSharing()) {
    AIGFormulaSharer().apply(prb);
  }

  if (_options.predicateDefinitionMerging()) {
    env.statistics->phase=Statistics::PREDIACTE_DEFINITION_MERGING;
    PDMerger().apply(prb);
  }

  if (_options.eprPreservingSkolemization()) {
    env.statistics->phase=Statistics::EPR_PRESERVING_SKOLEMIZATION;
    EPRSkolem().apply(prb);
  }

  if (_options.eprRestoringInlining()) {
    env.statistics->phase=Statistics::PREDICATE_DEFINITION_INLINING;
    EPRInlining().apply(prb);
  }

  if(_options.aigBddSweeping()) {
    AIGCompressingTransformer().apply(prb);
  }

  if(_options.flattenTopLevelConjunctions()) {
    TopLevelFlatten().apply(prb);
  }

  if(_options.aigInliner()) {
    AIGInliner().apply(prb);
  }

  if(_options.aigConditionalRewriting()) {
    AIGConditionalRewriter().apply(prb);
  }

  if(_options.aigDefinitionIntroduction()) {
    AIGDefinitionIntroducer(_options.aigDefinitionIntroductionThreshold()).apply(prb);
  }

  if (_options.unusedPredicateDefinitionRemoval()) {
    env.statistics->phase=Statistics::UNUSED_PREDICATE_DEFINITION_REMOVAL;
    PredicateDefinition pdRemover;
    pdRemover.removeUnusedDefinitionsAndPurePredicates(prb);
  }

  if (prb.mayHaveFormulas()) {
    preprocess2(prb);
  }

  if (_options.equalityPropagation() && prb.mayHaveEquality()) {
    env.statistics->phase=Statistics::EQUALITY_PROPAGATION;
    EqualityPropagator().apply(prb);
  }

  if (prb.mayHaveFormulas() && _options.naming()) {
    naming(prb);
  }

  if (prb.mayHaveFormulas()) {
    preprocess3(prb);
  }

  
  if (prb.mayHaveFormulas() && _options.naming() && _options.eprPreservingNaming()) {
    secondStageEprPreservingNaming(prb);
  }











  if (prb.mayHaveFormulas()) {
    clausify(prb);
  }

  if(prb.mayHaveFunctionDefinitions()) {
    env.statistics->phase=Statistics::FUNCTION_DEFINITION_ELIMINATION;
    if(_options.functionDefinitionElimination() == Options::FDE_ALL) {
      FunctionDefinition fd;
      fd.removeAllDefinitions(prb);
    }
    else if(_options.functionDefinitionElimination() == Options::FDE_UNUSED) {
      FunctionDefinition::removeUnusedDefinitions(prb);
    }
  }


  if (prb.mayHaveEquality() && _options.inequalitySplitting() != 0) {
    env.statistics->phase=Statistics::INEQUALITY_SPLITTING;
    InequalitySplitting is(_options);
    is.perform(prb);
  }



















   if (_options.equalityResolutionWithDeletion()!=Options::RA_OFF &&
	   prb.mayHaveInequalityResolvableWithDeletion() ) {
     env.statistics->phase=Statistics::EQUALITY_RESOLUTION_WITH_DELETION;
     EqResWithDeletion resolver;
     resolver.apply(prb);
   }

   if(_options.trivialPredicateRemoval()) {
     env.statistics->phase=Statistics::UNKNOWN_PHASE;
     TrivialPredicateRemover().apply(prb);
   }

   if (_options.generalSplitting()!=Options::RA_OFF) {
     env.statistics->phase=Statistics::GENERAL_SPLITTING;
     GeneralSplitting gs;
     gs.apply(prb);
   }

   if(_options.hornRevealing()) {
     env.statistics->phase=Statistics::HORN_REVEALING;
     HornRevealer hr(_options);
     hr.apply(prb);
   }

   if(_options.equalityProxy()!=Options::EP_OFF && prb.mayHaveEquality() &&
	   (prb.mayHaveXEqualsY() || _options.equalityProxy()!=Options::EP_ON) ) {
     env.statistics->phase=Statistics::EQUALITY_PROXY;
     EqualityProxy proxy(_options.equalityProxy());
     proxy.apply(prb);
   }

   TRACE("pp_final",
       UnitList::Iterator uit(prb.units());
       while(uit.hasNext()) {
 	Unit* u = uit.next();
 	TRACE_OUTPUT_UNIT(u);
       }
       );

   if(_options.printClausifierPremises()) {
     UIHelper::outputAllPremises(cerr, prb.units());
   }

   LOG("pp","preprocessing finished");
} 


void Preprocess::preprocess1 (Problem& prb)
{
  CALL("Preprocess::preprocess1");

  ScopedLet<Statistics::ExecutionPhase> epLet(env.statistics->phase, Statistics::PREPROCESS_1);

  bool formulasSimplified = false;

  UnitList*& units = prb.units();

  UnitList::DelIterator us(units);
  while (us.hasNext()) {
    Unit* u = us.next();
    if(u->isClause()) {
      continue;
    }

    
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    
    fu = Rectify::rectify(fu);
    FormulaUnit* rectFu = fu;
    
    fu = SimplifyFalseTrue::simplify(fu);
    if(fu!=rectFu) {
      formulasSimplified = true;
    }
    fu = Flattening::flatten(fu);

    if (fu != u) {
      us.replace(fu);
    }
  }

  if(formulasSimplified) {
    prb.invalidateByRemoval();
  }
}


void Preprocess::preprocess2(Problem& prb)
{
  CALL("Preprocess::preprocess2");

  env.statistics->phase=Statistics::PREPROCESS_2;

  UnitList::DelIterator us(prb.units());
  while (us.hasNext()) {
    Unit* u = us.next();
    LOG_UNIT("pp_pre_ennf",u);
    if(u->isClause()) {
	continue;
    }
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    FormulaUnit* fu0 = fu;

    fu = NNF::ennf(fu);
    fu = Flattening::flatten(fu);


    if (fu != fu0) {
      us.replace(fu);
    }
  }
} 

void Preprocess::naming(Problem& prb)
{
  CALL("Preprocess::naming");
  ASS(_options.naming());

  env.statistics->phase=Statistics::NAMING;
  UnitList::DelIterator us(prb.units());
  Naming naming(_options.naming(), _options.eprPreservingNaming());
  while (us.hasNext()) {
    Unit* u = us.next();
    if (u->isClause()) {
      continue;
    }
    UnitList* defs;
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    FormulaUnit* v = naming.apply(fu,defs);
    if (v != fu) {
      ASS(defs);
      us.insert(defs);
      us.replace(v);
    }
  }
  prb.invalidateProperty();
}

void Preprocess::secondStageEprPreservingNaming(Problem& prb)
{
  CALL("Preprocess::secondStageEprPreservingNaming");
  ASS(_options.naming());
  ASS(_options.eprPreservingNaming());

  env.statistics->phase=Statistics::NAMING;

  bool modified = false;

  UnitList::DelIterator us(prb.units());
  Naming naming(min(_options.naming()*2+1,32767), true);
  while (us.hasNext()) {
    Unit* u = us.next();
    if (u->isClause()) {
      continue;
    }
    UnitList* defs;
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    FormulaUnit* v = naming.apply(fu,defs);
    if (v != fu) {
      ASS(defs);
      while(defs) {
	Unit* d=preprocess3(UnitList::pop(defs));
	us.insert(new UnitList(d,0));
      }
      us.replace(v);
      modified = true;
    }
  }

  if(modified) {
    prb.invalidateProperty();
  }
}

Unit* Preprocess::preprocess3 (Unit* u)
{
  CALL("Preprocess::preprocess3(Unit*)");

  if (u->isClause()) {
    return u;
  }

  FormulaUnit* fu = static_cast<FormulaUnit*>(u);
  
  fu = NNF::nnf(fu);
  
  fu = Flattening::flatten(fu);





  fu = Skolem::skolemise(fu);
  return fu;
}

void Preprocess::preprocess3 (Problem& prb)
{
  CALL("Preprocess::preprocess3(Problem&)");

  bool modified = false;

  env.statistics->phase=Statistics::PREPROCESS_3;
  UnitList::DelIterator us(prb.units());
  while (us.hasNext()) {
    Unit* u = us.next();
    Unit* v = preprocess3(u);
    if (u!=v) {
      us.replace(v);
      modified = true;
    }
  }

  if(modified) {
    prb.invalidateProperty();
  }
} 

void Preprocess::clausify(Problem& prb)
{
  CALL("Preprocess::clausify");

  env.statistics->phase=Statistics::CLAUSIFICATION;

  
  Unit* emptyClause = 0;

  bool modified = false;

  UnitList::DelIterator us(prb.units());
  CNF cnf;
  Stack<Clause*> clauses(32);
  while (us.hasNext()) {
    Unit* u = us.next();
    LOG_UNIT("pp_pre_cl", u);
    if (u->isClause()) {
      if(static_cast<Clause*>(u)->isEmpty()) {
	emptyClause = u;
	break;
      }
      continue;
    }
    modified = true;
    cnf.clausify(u,clauses);
    while (! clauses.isEmpty()) {
      Unit* u = clauses.pop();
      if(static_cast<Clause*>(u)->isEmpty()) {
	emptyClause = u;
	goto fin;
      }
      us.insert(u);
    }
    us.del();
  }
fin:
  if(emptyClause) {
    prb.units()->destroy();
    prb.units() = 0;
    UnitList::push(emptyClause, prb.units());
  }
  if(modified) {
    prb.invalidateProperty();
  }
  prb.reportFormulasEliminated();
}

