
/*
 * File Property.hpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#ifndef __Property__
#define __Property__

#include <string>

#include "Lib/DArray.hpp"
#include "Kernel/Unit.hpp"
#include "Kernel/Theory.hpp"

namespace Lib {
  class MultiCounter;
}

namespace Kernel {
  class Clause;
  class FormulaUnit;
  class Literal;
  class Term;
  class TermList;
  class Formula;
}

namespace Shell {

using namespace std;
using namespace Kernel;
using namespace Lib;

class Property
{
public:
 
  enum Category {
    NEQ,
    HEQ,
    PEQ,
    HNE,
    NNE,
    FEQ,
    FNE,
    EPR,
    UEQ
  };

  
 
  static const unsigned PR_HAS_X_EQUALS_Y = 1u; 
 
  static const unsigned PR_HAS_FUNCTION_DEFINITIONS = 2u; 
 
  static const unsigned PR_HAS_SUBSET = 4u; 
 
  static const unsigned PR_HAS_EXTENSIONALITY = 8u; 
 
  static const unsigned PR_GROUP = 16u; 
 
  static const unsigned PR_RING = 32u; 
 
  static const unsigned PR_ROBBINS_ALGEBRA = 64u; 
 
  static const unsigned PR_NA_RING = 128u; 
 
  static const unsigned PR_BOOLEAN_ALGEBRA = 256u; 
 
  static const unsigned PR_LATTICE = 512u; 
 
  static const unsigned PR_LO_GROUP = 1024u; 
 
  static const unsigned PR_COMBINATOR_B = 2048u; 
 
  static const unsigned PR_COMBINATOR = 4096u; 
 
  static const unsigned PR_HAS_CONDENSED_DETACHMENT1 = 8192u; 
 
  static const unsigned PR_HAS_CONDENSED_DETACHMENT2 = 16384u; 
 
  static const unsigned PR_HAS_FLD1 = 32768u; 
 
  static const unsigned PR_HAS_FLD2 = 65536u; 
 
  static const unsigned PR_HAS_INEQUALITY_RESOLVABLE_WITH_DELETION = 131072u; 
 
  static const unsigned PR_HAS_STRINGS = 262144u; 
 
  static const unsigned PR_HAS_INTEGERS = 524288u; 
 
  static const unsigned PR_HAS_RATS = 1048576u; 
 
  static const unsigned PR_HAS_REALS = 2097152u; 
 
  static const unsigned PR_SORTS = 4194304u; 










 public:
  CLASS_NAME(Property);
  USE_ALLOCATOR(Property);

  static Property* scan(UnitList*);
  void add(UnitList*);
  ~Property();

 
  Category category() const { return _category;}
  string categoryString() const;

  string toString() const;
  string toSpider(const string& problemName) const;

 
  int clauses() const { return _goalClauses + _axiomClauses; }
 
  int formulas() const { return _goalFormulas + _axiomFormulas; }
 
  int unitClauses() const { return _unitGoals + _unitAxioms; }
 
  int hornClauses() const { return _hornGoals + _hornAxioms; }
 
  int atoms() const { return _atoms; }
 
  int equalityAtoms() const { return _equalityAtoms; }
 
  int positiveEqualityAtoms() const { return _positiveEqualityAtoms; }
 
  bool hasFormulas() const { return _axiomFormulas || _goalFormulas; }
 
  int maxFunArity() const { return _maxFunArity; }


 
  bool hasProp(unsigned p) const { return _props & p; }
 
  void addProp(unsigned p) { _props |= p; }
 
  unsigned int props() const { return _props; }

 
  void scanForInterpreted(Term* t);

  bool hasInterpretedOperation(Interpretation i) const { return _interpretationPresence[i]; }
 
  bool hasInterpretedOperations() const { return _hasInterpreted; }
 
  bool hasNonDefaultSorts() const { return _hasNonDefaultSorts; }
  bool hasSpecialTermsOrLets() const { return _hasSpecialTermsOrLets; }
  bool hasFormulaItes() const { return _hasFormulaItes; }
 private:
  
  explicit Property();

  static bool hasXEqualsY(const Clause* c);
  static bool isXEqualsY(const Literal*,bool polarity);
  static bool hasXEqualsY(const Formula*, MultiCounter&, int polarity);

  
  void scan(Unit*);
  void scan(Clause*);
  void scan(FormulaUnit*);
  void scan(Formula*);
  void scan(TermList* ts);
  void scan(Literal* lit);
  void scanSpecialTerm(Term* t);
  void scanSort(unsigned sort);

  char axiomTypes() const;
  char goalTypes() const;
  char equalityContent() const;
  char nonGroundUnitContent() const;
  char groundPositiveContent() const;
  char goalsAreGround() const;
  char setClauseSize() const;
  char setLiteralSize() const;
  char setTermSize() const;
  char maxPredArity() const;

  
  int _goalClauses;
  int _axiomClauses;

  int _positiveEqualityAtoms;
  int _equalityAtoms;
  int _atoms;

  int _goalFormulas;
  int _axiomFormulas;
  int _subformulas;

  int _terms;
  int _unitGoals;
  int _unitAxioms;
  int _hornGoals;
  int _hornAxioms;
  int _equationalClauses;
  int _pureEquationalClauses;
  int _groundUnitAxioms;
  int _positiveAxioms;
  int _groundPositiveAxioms;
  int _groundGoals;
  int _maxFunArity;
  int _maxPredArity;

 
  int _variablesInThisClause;
 
  int _totalNumberOfVariables;
 
  int _maxVariablesInClause;

 
  unsigned _props;

 
  Category _category;

 
  bool _hasInterpreted;
 
  bool _hasNonDefaultSorts;
  DArray<bool> _interpretationPresence;

  bool _hasSpecialTermsOrLets;
  bool _hasFormulaItes;
}; 

}

#endif 
