
/*
 * File Rectify.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/Recycler.hpp"

#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/SortHelper.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"
#include "Kernel/Unit.hpp"

#include "Indexing/TermSharing.hpp"

#include "VarManager.hpp"

#include "Rectify.hpp"

using namespace Shell;

bool Rectify::Renaming::tryGetBoundAndMarkUsed (int var,int& boundTo) const
{
  CALL("Rectify::Renaming::tryGetBoundAndMarkUsed");

  if ((unsigned)var >= _capacity) {
    return false;
  }
  VarUsageTrackingList* vs = _array[var];
  if (vs->isEmpty()) {
    return false;
  }
  boundTo = vs->head().first;
  vs->headPtr()->second = true;
  return true;
} 

Rectify::VarWithUsageInfo Rectify::Renaming::getBoundAndUsage(int var) const
{
  CALL("Rectify::Renaming::getBoundAndUsage");

  ASS_L((unsigned)var,_capacity);

  VarUsageTrackingList* vs = _array[var];
  ASS(!vs->isEmpty());
  return vs->head();
}

FormulaUnit* Rectify::rectify (FormulaUnit* unit0, bool removeUnusedVars)
{
  CALL("Rectify::rectify (Unit*...)");
  ASS(!unit0->isClause());

  FormulaUnit* unit = unit0;

  Formula* f = unit->formula();
  Rectify rect;
  rect._removeUnusedVars = removeUnusedVars;
  Formula* g = rect.rectify(f);

  VarList* vars = rect._free;

  if (f != g) {
    unit = new FormulaUnit(g,
			   new Inference1(Inference::RECTIFY,unit),
			   unit->inputType());
    if(unit0->included()) {
      unit->markIncluded();
    }
  }

  if (vars->isNonEmpty()) {
    unit = new FormulaUnit(new QuantifiedFormula(FORALL,vars,g),
			   new Inference1(Inference::CLOSURE,unit),
			   unit->inputType());
    if(unit0->included()) {
      unit->markIncluded();
    }
  }
  return unit;
} 

void Rectify::rectify(UnitList*& units)
{
  CALL("Rectify::rectify(UnitList*&)");

  UnitList::DelIterator us(units);
  while (us.hasNext()) {
    Unit* u = us.next();
    if(u->isClause()) {
      continue;
    }
    FormulaUnit* fu = static_cast<FormulaUnit*>(u);
    FormulaUnit* v = rectify(fu);
    if (v != fu) {
	us.replace(v);
    }
  }
}

Rectify::Renaming::~Renaming ()
{
  CALL("Rectify::Renaming::~Renaming");

  for (int i = _capacity-1;i >= 0;i--) {
    _array[i]->destroy();
    _array[i] = 0;
  }

  if(_used) {
    Recycler::release(_used);
  }
} 



Term* Rectify::rectifySpecialTerm(Term* t)
{
  CALL("Rectify::rectifySpecialTerm");

  Term::SpecialTermData* sd = t->getSpecialData();
  switch(t->functor()) {
  case Term::SF_TERM_ITE:
  {
    ASS_EQ(t->arity(),2);
    Formula* c = rectify(sd->getCondition());
    TermList th = rectify(*t->nthArgument(0));
    TermList el = rectify(*t->nthArgument(1));
    if(c==sd->getCondition() && th==*t->nthArgument(0) && el==*t->nthArgument(1)) {
	return t;
    }
    return Term::createTermITE(c, th, el);
  }
  case Term::SF_LET_FORMULA_IN_TERM:
  {
    ASS_EQ(t->arity(),1);
    Literal* orig = sd->getLhsLiteral();
    Formula* tgt = sd->getRhsFormula();
    rectifyFormulaLet(orig, tgt);
    TermList body = rectify(*t->nthArgument(0));
    if(orig==sd->getLhsLiteral() && tgt==sd->getRhsFormula() && body==*t->nthArgument(0)) {
      return t;
    }
    return Term::createFormulaLet(orig, tgt, body);
  }
  case Term::SF_LET_TERM_IN_TERM:
  {
    ASS_EQ(t->arity(),1);
    TermList orig = sd->getLhsTerm();
    TermList tgt = sd->getRhsTerm();
    rectifyTermLet(orig, tgt);
    TermList body = rectify(*t->nthArgument(0));
    if(orig==sd->getLhsTerm() && tgt==sd->getRhsTerm() && body==*t->nthArgument(0)) {
      return t;
    }
    return Term::createTermLet(orig, tgt, body);
  }
  default:
    ASSERTION_VIOLATION;
  }
  ASSERTION_VIOLATION;
}

Term* Rectify::rectify (Term* t)
{
  CALL("Rectify::rectify(Term*)");

  if (t->shared() && t->ground()) {
    return t;
  }

  if(t->isSpecial()) {
    return rectifySpecialTerm(t);
  }

  Term* s = new(t->arity()) Term(*t);
  if (rectify(t->args(),s->args())) {
    if(TermList::allShared(s->args())) {
      return env.sharing->insert(s);
    }
    else {
      return s;
    }
  }
  
  s->destroy();
  return t;
} 

Literal* Rectify::rectifyShared(Literal* lit)
{
  CALL("Rectify::rectifyShared");
  ASS(lit->shared());

  return SubstHelper::apply(lit, *this);
}

Literal* Rectify::rectify (Literal* l)
{
  CALL("Rectify::rectify(Literal*)");

  if (l->shared()) {
    if(l->ground()) {
      return l;
    }


  }

  Literal* m = new(l->arity()) Literal(*l);
  if (rectify(l->args(),m->args())) {
    if(TermList::allShared(m->args())) {
      if(l->isEquality() && m->nthArgument(0)->isVar() && m->nthArgument(1)->isVar()) {
	ASS(l->shared());
	unsigned srt = SortHelper::getEqualityArgumentSort(l);
	return env.sharing->insertVariableEquality(m, srt);
      }
      else {
	return env.sharing->insert(m);
      }
    }
    else {
      return m;
    }
  }
  
  m->destroy();
  return l;
} 

bool Rectify::rectify(TermList* from,TermList* to)
{
  CALL("Rectify::rectify(TermList* ...)");

  bool changed = false;
  while (! from->isEmpty()) {
    if (from->isVar()) {
      int v = from->var();
      int newV = rectifyVar(v);
      to->makeVar(newV);
      if (v != newV) { 
	changed = true;
      }
    }
    else { 
      Term* f = from->term();
      Term* t = rectify(f);
      to->setTerm(t);
      if (f != t) {
	changed = true;
      }
    }
    from = from->next();
    ASS(! to->isEmpty());
    to = to->next();
  }
  ASS(to->isEmpty());
  return changed;
} 

unsigned Rectify::rectifyVar(unsigned v)
{
  CALL("Rectify::rectifyVar");

  int newV;
  if (! _renaming.tryGetBoundAndMarkUsed(v,newV)) {
    newV = _renaming.bind(v);
    _free = new VarList(newV,_free);
  }
  return newV;
}


TermList Rectify::rectify(TermList t)
{
  CALL("Rectify::rectify");

  if(t.isTerm()) {
    return TermList(rectify(t.term()));
  }
  ASS(t.isOrdinaryVar());
  return TermList(rectifyVar(t.var()), false);
}

void Rectify::rectifyTermLet(TermList& lhs, TermList& rhs)
{
  CALL("Rectify::rectifyTermLet");

  
  
  VarList* vs = 0;
  VariableIterator vit(lhs);
  VarList::pushFromIterator(getMappingIterator(
	getUniquePersistentIteratorFromPtr(&vit), OrdVarNumberExtractorFn()), vs);
  
  bindVars(vs);
  lhs = rectify(lhs);
  rhs = rectify(rhs);
  unbindVars(vs);
  vs->destroy();
}

void Rectify::rectifyFormulaLet(Literal*& lhs, Formula*& rhs)
{
  CALL("Rectify::rectifyFormulaLet");

  
  
  VarList* vs = 0;
  VariableIterator vit(lhs);
  VarList::pushFromIterator(getMappingIterator(
	getUniquePersistentIteratorFromPtr(&vit), OrdVarNumberExtractorFn()), vs);
  
  bindVars(vs);
  lhs = rectify(lhs);
  rhs = rectify(rhs);
  unbindVars(vs);
  vs->destroy();
}


Formula* Rectify::rectify (Formula* f)
{
  CALL("Rectify::rectify (Formula*)");

  switch (f->connective()) {
  case LITERAL: 
  {
    Literal* lit = rectify(f->literal());
    return lit == f->literal() ? f : new AtomicFormula(lit);
  }

  case AND: 
  case OR: 
  {
    FormulaList* newArgs = rectify(f->args());
    if (newArgs == f->args()) {
      return f;
    }
    return new JunctionFormula(f->connective(), newArgs);
  }

  case IMP: 
  case IFF: 
  case XOR:
  {
    Formula* l = rectify(f->left());
    Formula* r = rectify(f->right());
    if (l == f->left() && r == f->right()) {
      return f;
    }
    return new BinaryFormula(f->connective(), l, r);
  }

  case NOT:
  {
    Formula* arg = rectify(f->uarg());
    if (f->uarg() == arg) {
      return f;
    }
    return new NegatedFormula(arg);
  }

  case FORALL: 
  case EXISTS:
  {
    bindVars(f->vars());
    Formula* arg = rectify(f->qarg());
    VarList* vs = rectifyBoundVars(f->vars());
    unbindVars(f->vars());
    if (vs == f->vars() && arg == f->qarg()) {
      return f;
    }
    if(vs->isEmpty()) {
      return arg;
    }
    return new QuantifiedFormula(f->connective(),vs,arg);
  }

  case ITE:
  {
    Formula* c = rectify(f->condArg());
    Formula* t = rectify(f->thenArg());
    Formula* e = rectify(f->elseArg());
    if (c == f->condArg() && t == f->thenArg() && e == f->elseArg()) {
      return f;
    }
    return new IteFormula(c,t,e);
  }

  case TERM_LET:
  {
    TermList o = f->termLetLhs();
    TermList t = f->termLetRhs();
    rectifyTermLet(o, t);
    Formula* b = rectify(f->letBody());
    if(o==f->termLetLhs() && t==f->termLetRhs() && b==f->letBody()) {
      return f;
    }
    return new TermLetFormula(o,t,b);
  }

  case FORMULA_LET:
  {
    Literal* o = f->formulaLetLhs();
    Formula* t = f->formulaLetRhs();
    rectifyFormulaLet(o, t);
    Formula* b = rectify(f->letBody());
    if(o==f->formulaLetLhs() && t==f->formulaLetRhs() && b==f->letBody()) {
      return f;
    }
    return new FormulaLetFormula(o, t, b);
  }

  case TRUE:
  case FALSE:
    return f;

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 

void Rectify::Renaming::undoBinding (int var)
{
  CALL("Rectify::Renaming::undoBinding");

  ASS(var < (int)_capacity);
  ASS(var >= 0);

  VarUsageTrackingList::pop(_array[var]);
} 

int Rectify::Renaming::bind (int var)
{
  CALL("Rectify::Renaming::bind");

  int result;

  if(VarManager::varNamePreserving()) {
    if(!_used) {
      Recycler::get(_used);
      _used->reset();
    }
    if(_used->insert(var)) {
      result=var;
    }
    else {
      result=VarManager::getVarAlias(var);
    }
  }
  else {
    result = _nextVar++;
  }
  VarUsageTrackingList::push(make_pair(result,false), get(var));

  return result;
} 


void Rectify::bindVars(VarList* vs)
{
  CALL ("Rectify::bindVars (VarList*)");

  VarList::Iterator vit(vs);
  while(vit.hasNext()) {
    int v = vit.next();
    _renaming.bind(v);
  }
}

void Rectify::unbindVars(VarList* vs)
{
  CALL ("Rectify::unbindVars (VarList*)");

  VarList::Iterator vit(vs);
  while(vit.hasNext()) {
    int v = vit.next();
    _renaming.undoBinding(v);
  }
}

Rectify::VarList* Rectify::rectifyBoundVars (VarList* vs)
{
  CALL ("Rectify::rectifyBoundVars(VarList*)");

  if (vs->isEmpty()) {
    return vs;
  }
  VarList* vtail = vs->tail();
  VarList* ws = rectifyBoundVars(vtail);

  int v = vs->head();
  int w;
  VarWithUsageInfo wWithUsg = _renaming.getBoundAndUsage(v);
  if(!wWithUsg.second && _removeUnusedVars) {
    return ws;
  }
  w = wWithUsg.first;
  if (v == w && vtail == ws) {
    return vs;
  }
  return new VarList(w,ws);
} 


FormulaList* Rectify::rectify (FormulaList* fs)
{
  CALL ("Rectify::rectify (FormulaList*)");

  Stack<FormulaList*>* els;
  Recycler::get(els);
  els->reset();

  FormulaList* el = fs;
  while(el) {
    els->push(el);
    el = el->tail();
  }

  FormulaList* res = 0;

  bool modified = false;
  while(els->isNonEmpty()) {
    FormulaList* el = els->pop();
    Formula* f = el->head();
    Formula* g = rectify(f);
    if(!modified && f!=g) {
      modified = true;
    }
    if(modified) {
      FormulaList::push(g, res);
    }
    else {
      res = el;
    }
  }

  Recycler::release(els);
  return res;













} 

void Rectify::Renaming::fillInterval (size_t start,size_t end)
{
  for (unsigned i = start;i < end;i++) {
    _array[i] = 0;
  }
} 
