
/*
 * File SimplifyFalseTrue.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Tracer.hpp"

#include "Lib/DArray.hpp"

#include "Kernel/Inference.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Unit.hpp"

#include "SimplifyFalseTrue.hpp"

using namespace Kernel;
using namespace Shell;
using namespace Lib;

FormulaUnit* SimplifyFalseTrue::simplify (FormulaUnit* unit)
{
  CALL("SimplifyFalseTrue::simplify(Unit*)");
  ASS(! unit->isClause());

  Formula* f = unit->formula();
  Formula* g = simplify(f);
  if (f == g) { 
    return unit;
  }

  FormulaUnit* res = new FormulaUnit(g,
			 new Inference1(Inference::REDUCE_FALSE_TRUE,unit),
			 unit->inputType());
  if(unit->included()) {
    res->markIncluded();
  }
  return res;
} 


Formula* SimplifyFalseTrue::simplify (Formula* f)
{
  CALL("SimplifyFalseTrue::simplify(Formula*)");

  Connective con = f->connective();
  switch (con) {
  case LITERAL:
  case TRUE:
  case FALSE:
    return f;

  case AND:
  case OR: 
    {
      int length = 0;  
      bool changed = false;
      FormulaList* fs = f->args();
      DArray<Formula*> gs(fs->length());

      FormulaList::Iterator it(fs);
      while (it.hasNext()) {
	Formula* h = it.next();
	Formula* g = simplify(h);
	switch (g->connective()) {
	case TRUE:
	  if (con == OR) {
	    return g;
	  }
	  if (con == AND) {
	    changed = true;
	    break;
	  }
	  gs[length++] = g;
	  if (h != g) {
	    changed = true;
	  }
	  break;

	case FALSE:
	  if (con == AND) {
	    return g;
	  }
	  if (con == OR) {
	    changed = true;
	    break;
	  }
	  gs[length++] = g;
	  if (h != g) {
	    changed = true;
	  }
	  break;

	default:
	  gs[length++] = g;
	  if (h != g) {
	    changed = true;
	  }
	  break;
	}
      }
      if (! changed) {
	return f;
      }
      switch (length) {
      case 0:
	return new Formula(con == OR ? false : true);
      case 1:
	return gs[0];
      default:
	FormulaList* res = FormulaList::empty();
	for (int l = length-1;l >= 0;l--) {
	  FormulaList::push(gs[l],res);
	}
	return new JunctionFormula(con,res);
      }
    }

  case IMP:
    {
      Formula* right = simplify(f->right());
      if (right->connective() == TRUE) {
	return right;
      }
      Formula* left = simplify(f->left());

      switch (left->connective()) {
      case TRUE: 
	return right;
      case FALSE: 
	return new Formula(true);
      default: 
	break;
      }

      if (right->connective() == FALSE) {
	return new NegatedFormula(left);
      }
      if (left == f->left() && right == f->right()) {
	return f;
      }
      return new BinaryFormula(con,left,right);
    }

  case IFF:
  case XOR: 
    {
      Formula* left = simplify(f->left());
      Formula* right = simplify(f->right());

      Connective lc = left->connective();
      Connective rc = right->connective();

      switch (lc) {
      case FALSE: 
	switch (rc) {
	case FALSE: 
	  return con == XOR
	         ? right
	         : new Formula(true);
	case TRUE: 
	  return con == XOR
	         ? right
     	         : left;
	default: 
	  return con == XOR
	         ? right
 	         : new NegatedFormula(right);
	}
      case TRUE: 
	switch (rc) {
	case FALSE: 
	  return con == XOR
	         ? left
	         : right;
	case TRUE: 
	  return con == XOR
 	         ? new Formula(false)
     	         : left;
	default: 
	  return con == XOR
 	         ? new NegatedFormula(right)
     	         : right;
	}
      default: 
	switch (rc) {
	case FALSE: 
	  return con == XOR
	         ? left
 	         : new NegatedFormula(left);
	case TRUE:  
	  return con == XOR
 	         ? new NegatedFormula(left)
     	         : left;
	default:    
	  if (left == f->left() && right == f->right()) {
	    return f;
	  }
	  return new BinaryFormula(con,left,right);
	}
      }
    }
    
  case NOT: 
    {
      Formula* arg = simplify(f->uarg());
      switch (arg->connective()) {
      case FALSE:
	return new Formula(true);
      case TRUE:
	return new Formula(false);
      default:
	return arg == f->uarg() ? f : new NegatedFormula(arg);
      }
    }
    
  case FORALL:
  case EXISTS: 
    {
      Formula* arg = simplify(f->qarg());
      switch (arg->connective()) {
      case FALSE:
      case TRUE:
	return arg;
      default:
	return arg == f->qarg()
               ? f
               : new QuantifiedFormula(con,f->vars(),arg);
      }
    }

  case ITE:
    {
      Formula* c = simplify(f->condArg());
      switch (c->connective()) {
      case TRUE:
	return simplify(f->thenArg());
      case FALSE:
	return simplify(f->elseArg());
      default:
	break;
      }
      Formula* t = simplify(f->thenArg());
      Formula* e = simplify(f->elseArg());

      switch (t->connective()) {
      case TRUE:
	switch (e->connective()) {
	case TRUE:
	  return new Formula(true);
	case FALSE:
	  return c;
	default:
	  return new BinaryFormula(IMP, new NegatedFormula(c), e);
	}
      case FALSE:
	switch (e->connective()) {
	case TRUE:
	  return new NegatedFormula(c);
	case FALSE:
	  return new Formula(false);
	default: {
	  FormulaList* args = 0;
	  FormulaList::push(new NegatedFormula(c), args);
	  FormulaList::push(e, args);
	  return new JunctionFormula(AND, args);
	}
	}
      default:
	switch (e->connective()) {
	case TRUE:
	  return new BinaryFormula(IMP, c, t);
	case FALSE: {
	  FormulaList* args = 0;
	  FormulaList::push(c, args);
	  FormulaList::push(t, args);
	  return new JunctionFormula(AND, args);
	}
	default:
	  break;
	}
      }

      if (c == f->condArg() && t == f->thenArg() && e == f->elseArg()) {
	return f;
      }
      return new IteFormula(c,t,e);
    }

#if VDEBUG
  default:
    ASSERTION_VIOLATION;
#endif
  }
} 


