
/*
 * File SineUtils.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include <cmath>

#include "Lib/Deque.hpp"
#include "Lib/DHSet.hpp"
#include "Lib/DHMultiset.hpp"
#include "Lib/Environment.hpp"
#include "Lib/List.hpp"
#include "Lib/Metaiterators.hpp"
#include "Lib/Set.hpp"
#include "Lib/TimeCounter.hpp"
#include "Lib/VirtualIterator.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/SubformulaIterator.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"

#include "Options.hpp"
#include "Statistics.hpp"

#include "SineUtils.hpp"

#define SINE_PRINT_SELECTED 0

namespace Shell
{

using namespace Lib;
using namespace Kernel;

SineSymbolExtractor::SymId SineSymbolExtractor::getSymIdBound()
{
  return max(env.signature->predicates()*2-1, env.signature->functions()*2);
}

void SineSymbolExtractor::addSymIds(Literal* lit, int polarity, Stack<SymId>& ids) const
{
  CALL("SineSymbolExtractor::addSymIds");

  SymId predId=lit->functor()*2;
  ids.push(predId);

  if(!lit->shared()) {
    
    return;
  }

  NonVariableIterator nvi(lit);
  while(nvi.hasNext()) {
    Term* t=nvi.next().term();

    SymId funId=t->functor()*2+1;
    ids.push(funId);
  }
}

void SineSymbolExtractor::decodeSymId(SymId s, bool& pred, unsigned& functor)
{
  pred = (s%2)==0;
  functor = s/2;
}

bool SineSymbolExtractor::validSymId(SymId s)
{
  CALL("SineSymbolExtractor::validSymId");

  bool pred;
  unsigned functor;
  decodeSymId(s, pred, functor);
  if(pred) {
    if(functor>=static_cast<unsigned>(env.signature->predicates())) {
      return false;
    }
  }
  else {
    if(functor>=static_cast<unsigned>(env.signature->functions())) {
      return false;
    }
  }
  return true;
}

void SineSymbolExtractor::extractFormulaSymbols(Formula* f,int polarity,Stack<SymId>& itms)
{
  CALL("SineSymbolExtractor::extractFormulaSymbols");

  switch (f->connective()) {
    case LITERAL:
      addSymIds(f->literal(), polarity, itms);
      return;

    case AND:
    case OR: {
      FormulaList::Iterator fs(f->args());
      while (fs.hasNext()) {
	extractFormulaSymbols (fs.next(),polarity,itms);
      }
      return;
    }

    case IMP:
      extractFormulaSymbols (f->left(), -polarity, itms);
      extractFormulaSymbols (f->right(), polarity, itms);
      return;

    case NOT:
      extractFormulaSymbols (f->uarg(), -polarity, itms);
      return;

    case IFF:
    case XOR:
      extractFormulaSymbols (f->left(), 0, itms);
      extractFormulaSymbols (f->right(), 0, itms);
      return;

    case FORALL:
    case EXISTS:
      extractFormulaSymbols (f->qarg(), polarity, itms);
      return;

    case ITE:
      extractFormulaSymbols (f->condArg(), 0, itms);
      extractFormulaSymbols (f->thenArg(), polarity, itms);
      extractFormulaSymbols (f->elseArg(), polarity, itms);
      return;
    case TERM_LET:
      
      extractFormulaSymbols (f->letBody(), polarity, itms);
      return;
    case FORMULA_LET:
      extractFormulaSymbols (f->letBody(), polarity, itms);
      addSymIds(f->formulaLetLhs(), 0, itms);
      extractFormulaSymbols (f->formulaLetRhs(), 0, itms);
      return;

    case TRUE:
    case FALSE:
      return;
#if VDEBUG
    default:
      ASSERTION_VIOLATION;
      return;
#endif
  }
}

SineSymbolExtractor::SymIdIterator SineSymbolExtractor::extractSymIds(Unit* u)
{
  CALL("SineSymbolExtractor::extractSymIds");

  static Stack<SymId> itms;
  itms.reset();

  if(u->isClause()) {
    Clause* cl=static_cast<Clause*>(u);
    unsigned clen=cl->length();
    for(unsigned i=0;i<clen;i++) {
      Literal* lit=(*cl)[i];
      addSymIds(lit, 1, itms);
    }
  } else {
    FormulaUnit* fu=static_cast<FormulaUnit*>(u);
    extractFormulaSymbols(fu->formula(), 1, itms);
  }
  return pvi( getUniquePersistentIterator(Stack<SymId>::Iterator(itms)) );
}

void SineBase::initGeneralityFunction(UnitList* units)
{
  CALL("SineBase::initGeneralityFunction");

  SymId symIdBound=_symExtr.getSymIdBound();
  _gen.init(symIdBound,0);

  UnitList::Iterator uit(units);
  while(uit.hasNext()) {
    Unit* u=uit.next();
    SymIdIterator sit=_symExtr.extractSymIds(u);
    while(sit.hasNext()) {
      SymId sid=sit.next();
      _gen[sid]++;
    }
  }
}

SineSelector::SineSelector(const Options& opt)
: _onIncluded(opt.sineSelection()==Options::SS_INCLUDED),
  _genThreshold(opt.sineGeneralityThreshold()),
  _tolerance(opt.sineTolerance()),
  _depthLimit(opt.sineDepth())
{
  CALL("SineSelector::SineSelector/0");

  init();
}

SineSelector::SineSelector(bool onIncluded, float tolerance, unsigned depthLimit, unsigned genThreshold)
: _onIncluded(onIncluded),
  _genThreshold(genThreshold),
  _tolerance(tolerance),
  _depthLimit(depthLimit)
{
  CALL("SineSelector::SineSelector/4");

  init();
}

void SineSelector::init()
{
  CALL("SineSelector::init");
  ASS(_tolerance>=1.0f || _tolerance==-1);

  _strict=_tolerance==1.0f;
}

void SineSelector::updateDefRelation(Unit* u)
{
  CALL("SineSelector::updateDefRelation");

  SymIdIterator sit=_symExtr.extractSymIds(u);

  if(!sit.hasNext()) {
    _unitsWithoutSymbols.push(u);
    return;
  }

  static Stack<SymId> equalGenerality;
  equalGenerality.reset();

  SymId leastGenSym=sit.next();
  unsigned leastGenVal=_gen[leastGenSym];

  
  if(leastGenVal<=_genThreshold) {
    UnitList::push(u,_def[leastGenSym]);
  }

  while(sit.hasNext()) {
    SymId sym=sit.next();
    unsigned val=_gen[sym];
    ASS_G(val,0);

    
    if(val<=_genThreshold) {
      UnitList::push(u,_def[sym]);
    }

    if(val<leastGenVal) {
      leastGenSym=sym;
      leastGenVal=val;
      equalGenerality.reset();
    } else if(val==leastGenVal) {
      equalGenerality.push(sym);
    }
  }


  if(_strict) {
    
    if(leastGenVal>_genThreshold) {
      UnitList::push(u,_def[leastGenSym]);
      while(equalGenerality.isNonEmpty()) {
        UnitList::push(u,_def[equalGenerality.pop()]);
      }
    }
  }
  else {
    unsigned generalityLimit=static_cast<int>(leastGenVal*_tolerance);
    if(_tolerance==-1.0f) {
      generalityLimit = UINT_MAX;
    }

    
    if(generalityLimit>_genThreshold) {
      sit=_symExtr.extractSymIds(u);
      while(sit.hasNext()) {
	SymId sym=sit.next();
	unsigned val=_gen[sym];
	
	if(val>_genThreshold && val<=generalityLimit) {
	  UnitList::push(u,_def[sym]);
	}
      }
    }
  }

}

void SineSelector::perform(Problem& prb)
{
  CALL("SineSelector::perform");

  perform(prb.units());
  prb.invalidateByRemoval();
}

void SineSelector::perform(UnitList*& units)
{
  CALL("SineSelector::perform");

  TimeCounter tc(TC_SINE_SELECTION);

  initGeneralityFunction(units);

  SymId symIdBound=_symExtr.getSymIdBound();

  Set<Unit*> selected;
  Stack<Unit*> selectedStack; 
  Deque<Unit*> newlySelected;

  
  _def.init(symIdBound,0);
  UnitList::Iterator uit2(units);
  while(uit2.hasNext()) {
    Unit* u=uit2.next();
    bool performSelection= _onIncluded ? u->included() : (u->inputType()==Unit::AXIOM);
    if(performSelection) {
      updateDefRelation(u);
    }
    else {
      selected.insert(u);
      selectedStack.push(u);
      newlySelected.push_back(u);
    }
  }

  unsigned depth=0;
  newlySelected.push_back(0);

  
  while(newlySelected.isNonEmpty()) {
    Unit* u=newlySelected.pop_front();

    if(!u) {
      
      depth++;
      if(_depthLimit && depth==_depthLimit) {
	break;
      }
      ASS(!_depthLimit || depth<_depthLimit);

      if(newlySelected.isNonEmpty()) {
	
	newlySelected.push_back(0);
      }
      continue;
    }

    SymIdIterator sit=_symExtr.extractSymIds(u);
    while(sit.hasNext()) {
      SymId sym=sit.next();
      UnitList::Iterator defUnits(_def[sym]);
      while(defUnits.hasNext()) {
	Unit* du=defUnits.next();
	if(selected.contains(du)) {
	  continue;
	}
	selected.insert(du);
	selectedStack.push(du);
	newlySelected.push_back(du);
      }
      
      
      _def[sym]->destroy();
      _def[sym]=0;
    }
  }

  env.statistics->sineIterations=depth;

  env.statistics->selectedBySine=_unitsWithoutSymbols.size() + selectedStack.size();

  units->destroy();
  units=0;
  UnitList::pushFromIterator(Stack<Unit*>::Iterator(_unitsWithoutSymbols), units);
  while(selectedStack.isNonEmpty()) {
    UnitList::push(selectedStack.pop(), units);
  }

#if SINE_PRINT_SELECTED
  UnitList::Iterator selIt(units);
  while(selIt.hasNext()) {
    cout<<'#'<<selIt.next()->toString()<<endl;
  }
#endif
}





SineTheorySelector::SineTheorySelector(const Options& opt)
: _genThreshold(opt.sineGeneralityThreshold()), _opt(opt)
{
  CALL("SineTheorySelector::SineTheorySelector");
}

void SineTheorySelector::handlePossibleSignatureChange()
{
  CALL("SineTheorySelector::handlePossibleSignatureChange");

  size_t symIdBound=_symExtr.getSymIdBound();
  size_t oldSize=_def.size();
  ASS_EQ(_gen.size(), oldSize);

  if(symIdBound==oldSize) {
    return;
  }
  ASS_G(symIdBound, oldSize);

  _gen.expand(symIdBound);
  _def.expand(symIdBound);
  for(size_t i=oldSize;i<symIdBound;i++) {
    _gen[i]=0;
    _def[i]=0;
  }
}

void SineTheorySelector::updateDefRelation(Unit* u)
{
  CALL("SineTheorySelector::updateDefRelation");

  SymIdIterator sit0=_symExtr.extractSymIds(u);

  if(!sit0.hasNext()) {
    _unitsWithoutSymbols.push(u);
    return;
  }

  static Stack<SymId> symIds;
  symIds.reset();
  symIds.loadFromIterator(sit0);

  Stack<SymId>::Iterator sit(symIds);

  ALWAYS(sit.hasNext());
  unsigned leastGenVal=_gen[sit.next()];

  while(sit.hasNext()) {
    SymId sym=sit.next();
    unsigned val=_gen[sym];
    ASS_G(val,0);

    if(val<leastGenVal) {
      leastGenVal=val;
    }
  }

  unsigned generalityLimit=leastGenVal*(maxTolerance/strictTolerance);

  Stack<SymId>::Iterator sit2(symIds);
  while(sit2.hasNext()) {
    SymId sym=sit2.next();
    unsigned val=_gen[sym];

    if(val<=_genThreshold) {
      
      DEntryList::push(DEntry(strictTolerance,u),_def[sym]);
    }
    else if(val<=generalityLimit) {
      unsigned short minTolerance=(val*strictTolerance)/leastGenVal;
      
      DEntryList::push(DEntry(minTolerance,u),_def[sym]);
    }
  }

}

void SineTheorySelector::initSelectionStructure(UnitList* units)
{
  CALL("SineTheorySelector::initSelectionStructure");

  TimeCounter tc(TC_SINE_SELECTION);

  initGeneralityFunction(units);

  SymId symIdBound=_symExtr.getSymIdBound();

  
  _def.init(symIdBound,0);
  UnitList::Iterator uit(units);
  while(uit.hasNext()) {
    Unit* u=uit.next();
    updateDefRelation(u);
  }
}


void SineTheorySelector::perform(UnitList*& units)
{
  CALL("SineTheorySelector::perform");

  TimeCounter tc(TC_SINE_SELECTION);

  handlePossibleSignatureChange();

  UnitList::Iterator uit(units);
  while(uit.hasNext()) {
    Unit* u=uit.next();
    SymIdIterator sit=_symExtr.extractSymIds(u);
    while(sit.hasNext()) {
      SymId sid=sit.next();
      _gen[sid]++;
    }
  }

  UnitList* res=0;
  DHSet<SymId> addedSymIds;
  DHSet<Unit*> selected;
  Deque<Unit*> newlySelected;

  bool sineOnIncluded=_opt.sineSelection()==Options::SS_INCLUDED;

  
  UnitList::Iterator uit2(units);
  while(uit2.hasNext()) {
    Unit* u=uit2.next();
    bool performSelection= sineOnIncluded ? u->included() : (u->inputType()==Unit::AXIOM);
    if(performSelection) {
      updateDefRelation(u);
    }
    else {
      selected.insert(u);
      newlySelected.push_back(u);
      UnitList::push(u,res);
    }
  }


  unsigned short intTolerance=static_cast<unsigned short>(ceil(_opt.sineTolerance()*10));

  unsigned depthLimit=_opt.sineDepth();
  unsigned depth=0;
  newlySelected.push_back(0);

  
  while(newlySelected.isNonEmpty()) {
    Unit* u=newlySelected.pop_front();

    if(!u) {
      
      depth++;
      if(depthLimit && depth==depthLimit) {
	break;
      }
      ASS(!depthLimit || depth<depthLimit);

      if(newlySelected.isNonEmpty()) {
	
	newlySelected.push_back(0);
      }
      continue;
    }

    SymIdIterator sit=_symExtr.extractSymIds(u);
    while(sit.hasNext()) {
      SymId sym=sit.next();
      if(!addedSymIds.insert(sym)) {
	
	continue;
      }
      DEntryList::Iterator defUnits(_def[sym]);
      while(defUnits.hasNext()) {
	DEntry de=defUnits.next();

	if(de.minTolerance>intTolerance || !selected.insert(de.unit)) {
	  continue;
	}
	UnitList::push(de.unit,res);
	newlySelected.push_back(de.unit);
      }
    }
  }

  UnitList::pushFromIterator(Stack<Unit*>::Iterator(_unitsWithoutSymbols), res);

  units->destroy();

  units=res;

  env.statistics->sineIterations=depth;
  env.statistics->selectedBySine=_unitsWithoutSymbols.size() + selected.size();

#if SINE_PRINT_SELECTED
  UnitList::Iterator selIt(units);
  while(selIt.hasNext()) {
    cout<<'#'<<selIt.next()->toString()<<endl;
  }
#endif
}

}
