
/*
 * File TPLexer.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Debug/Assertion.hpp"
#include "Debug/Tracer.hpp"

#include "Lib/Int.hpp"
#include "TPLexer.hpp"

using namespace Shell;
using namespace Lib;

TPLexer::TPLexer (istream& in)
  : _in(in),
    _chars(32),
    _tokens(32)
    
    
    
    
{
} 


void TPLexer::parse()
{
  CALL("TPLexer::parse");

  
  for (int i = 0;;i++) {
    int c = _in.get();
    if (c == -1) {
      _chars[i] = 0;
      break;
    }
    if (c == '\n') {
      _eols.push(i);
    }
    _chars[i] = c;
  }

  
  _cpos = 0;
  _tpos = 0;
  while (readToken()) {}
} 

bool TPLexer::readToken()
{
  CALL("TPLexer::readToken");
  skipWhiteSpacesAndComments();
  Token& tok = _tokens[_tpos++];
  tok.start = _cpos;
  switch (_chars[_cpos]) {
  case 0:
    tok.tag = TEOF;
    return false;
  case 'a':
  case 'b':
  case 'c':
  case 'd':
  case 'e':
  case 'f':
  case 'g':
  case 'h':
  case 'i':
  case 'j':
  case 'k':
  case 'l':
  case 'm':
  case 'n':
  case 'o':
  case 'p':
  case 'q':
  case 'r':
  case 's':
  case 't':
  case 'u':
  case 'v':
  case 'w':
  case 'x':
  case 'y':
  case 'z':
  case '$':
    tok.tag = NAME;
    readName();
    return true;
  case 'A':
  case 'B':
  case 'C':
  case 'D':
  case 'E':
  case 'F':
  case 'G':
  case 'H':
  case 'I':
  case 'J':
  case 'K':
  case 'L':
  case 'M':
  case 'N':
  case 'O':
  case 'P':
  case 'Q':
  case 'R':
  case 'S':
  case 'T':
  case 'U':
  case 'V':
  case 'W':
  case 'X':
  case 'Y':
  case 'Z':
  case '_':
    tok.tag = VAR;
    readName();
    return true;
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    tok.tag = readNumber();
    return true;
  case '"':
    tok.tag = STRING;
    readString();
    return true;
  case '\'':
    tok.tag = NAME;
    readAtom();
    return true;
  case '(':
    tok.tag = LPAR;
    _cpos++;
    return true;
  case ')':
    tok.tag = RPAR;
    _cpos++;
    return true;
  case '[':
    tok.tag = LBRA;
    _cpos++;
    return true;
  case ']':
    tok.tag = RBRA;
    _cpos++;
    return true;
  case ',':
    tok.tag = COMMA;
    _cpos++;
    return true;
  case ':':
    if (_chars[_cpos+1] == '=') {
      tok.tag = ASS;
      _cpos += 2;
      return true;
    }
    tok.tag = COLON;
    _cpos++;
    return true;
  case '~':
    if (_chars[_cpos+1] == '&') {
      tok.tag = NOT_AND;
      _cpos += 2;
      return true;
    }
    if (_chars[_cpos+1] == '|') {
      tok.tag = NOT_OR;
      _cpos += 2;
      return true;
    }
    tok.tag = NOT;
    _cpos++;
    return true;
  case '=':
    if (_chars[_cpos+1] == '>') {
      tok.tag = IMPLY;
      _cpos += 2;
      return true;
    }
    tok.tag = EQUAL;
    _cpos++;
    return true;
  case '&':
    tok.tag = AND;
    _cpos++;
    return true;
  case '^':
    tok.tag = LAMBDA;
    _cpos++;
    return true;
  case '@':
    if (_chars[_cpos+1] == '+') {
      tok.tag = APP_PLUS;
      _cpos += 2;
      return true;
    }
    if (_chars[_cpos+1] == '-') {
      tok.tag = APP_MINUS;
      _cpos += 2;
      return true;
    }
    tok.tag = APP;
    _cpos++;
    return true;
  case '*':
    tok.tag = STAR;
    _cpos++;
    return true;
  case '>':
    tok.tag = ARROW;
    _cpos++;
    return true;
  case '!':
    if (_chars[_cpos+1] == '=') {
      tok.tag = NEQ;
      _cpos += 2;
      return true;
    }
    if (_chars[_cpos+1] == '>') {
      tok.tag = THF_QUANT_ALL;
      _cpos += 2;
      return true;
    }
    if (_chars[_cpos+1] == '!') {
      tok.tag = SIGMA;
      _cpos += 2;
      return true;
    }
    tok.tag = FORALL;
    _cpos++;
    return true;
  case '?':
    if (_chars[_cpos+1] == '?') {
      tok.tag = PI;
      _cpos += 2;
      return true;
    }
    if (_chars[_cpos+1] == '*') {
      tok.tag = THF_QUANT_SOME;
      _cpos += 2;
      return true;
    }
    tok.tag = EXISTS;
    _cpos++;
    return true;
  case '<':
    if (_chars[_cpos+1] == '<') {
      tok.tag = SUBTYPE;
      _cpos += 2;
      return true;
    }
    if (_chars[_cpos+1] == '~' && _chars[_cpos+2] == '>') {
      tok.tag = XOR;
      _cpos += 3;
      return true;
    }
    if (_chars[_cpos+1] != '=') {
      throw Exception(*this,"unrecognized symbol");
    }
    if (_chars[_cpos+2] == '>') {
      tok.tag = IFF;
      _cpos += 3;
      return true;
    }
    tok.tag = REVERSE_IMP;
    _cpos += 2;
    return true;
  case '.':
    tok.tag = DOT;
    _cpos ++;
    return true;
  case '|':
    tok.tag = OR;
    _cpos ++;
    return true;
  case '-':
    if (_chars[_cpos+1] == '-' && _chars[_cpos+2] == '>') {
      tok.tag = SEQUENT;
      _cpos += 3;
      return true;
    }
    tok.tag = readNumber();
    return true;
  case '+':
    if (_chars[_cpos+1] < '0' || _chars[_cpos+1] > '9') {
      tok.tag = UNION;
      _cpos ++;
      return true;
    }
    tok.tag = readNumber();
    return true;
  default:
    throw ::Exception(string("Bad character: ") + _chars[_cpos]);
  }
} 

void TPLexer::skipWhiteSpacesAndComments()
{
  CALL("TPLexer::skipWhiteSpacesAndComments");

  for (;;) {
    switch (_chars[_cpos]) {
    case 0: 
      return;

    case ' ':
    case '\t':
    case '\r':
    case '\f':
    case '\n':
      _cpos++;
      break;

    case '%': 
    _cpos++;
    while (_chars[_cpos] != '\n') {
      if (!_chars[_cpos]) {
	return;
      }
      _cpos++;
    }
    _cpos++;
    break;

    case '/': 
      if (_chars[_cpos+1] != '*') {
	return;
      }
      
      _cpos += 2;
      
      for (;;) {
	int c = _chars[_cpos];
	if (!c) {
	  return;
	}
	if (c != '*') {
	  _cpos++;
	  continue;
	}
	
	_cpos++;
	if (_chars[_cpos] != '/') {
	  continue;
	}
	_cpos++;
	break;
      }
      break;
    
    default:
      return;
    }
  }
} 

void TPLexer::readName()
{
  CALL("TPLexer::readName");
  for (;;) {
    switch (_chars[++_cpos]) {
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'E':
    case 'F':
    case 'G':
    case 'H':
    case 'I':
    case 'J':
    case 'K':
    case 'L':
    case 'M':
    case 'N':
    case 'O':
    case 'P':
    case 'Q':
    case 'R':
    case 'S':
    case 'T':
    case 'U':
    case 'V':
    case 'W':
    case 'X':
    case 'Y':
    case 'Z':
    case '_':
    case 'a':
    case 'b':
    case 'c':
    case 'd':
    case 'e':
    case 'f':
    case 'g':
    case 'h':
    case 'i':
    case 'j':
    case 'k':
    case 'l':
    case 'm':
    case 'n':
    case 'o':
    case 'p':
    case 'q':
    case 'r':
    case 's':
    case 't':
    case 'u':
    case 'v':
    case 'w':
    case 'x':
    case 'y':
    case 'z':
    case '$':
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      break;
    default:
      return;
    }
  }
} 

void TPLexer::readString()
{
  CALL("TPLexer::readString");
  for (;;) {
    int c = _chars[++_cpos];
    if (!c) {
      throw Exception(*this,"non-terminated string");
    }
    if (c == '\\') { 
      char c1 = _chars[++_cpos];
      if (!c1) {
	throw Exception(*this,"non-terminated string");
      }
      continue;
    }
    if (c == '"') {
      _cpos++;
      return;
    }
  }
} 

void TPLexer::readAtom()
{
  CALL("TPLexer::readAtom");
  for (;;) {
    int c = _chars[++_cpos];
    if (!c) {
      throw Exception(*this,"non-terminated quoted atom");
    }
    if (c == '\\') { 
      char c1 = _chars[++_cpos];
      if (!c1) {
	throw Exception(*this,"non-terminated quoted atom");
      }
      continue;
    }
    if (c == '\'') {
      _cpos++;
      return;
    }
  }
} 

TPLexer::Exception::Exception(TPLexer& lex,const char* message)
{
  int start = lex._tokens[lex._tpos-1].start;
  const char* chars = lex._chars.content()+start;
  _message = (string)message + " at\n" + chars + '\n';
} 

void TPLexer::Exception::cry(ostream& str)
{
  str << _message << "\n";
}

TPLexer::Tag TPLexer::readNumber()
{
  CALL("TPLexer::readNumber");

  
  switch (_chars[_cpos]) {
  case '+':
  case '-':
    _cpos++;
  default:
    break;
  }
  decimal();
  switch (_chars[_cpos]) {
  case '/':
    _cpos++;
    positiveDecimal();
    return RAT;
  case 'E':
  case 'e':
    _cpos++;
    if (_chars[_cpos] == '+' || _chars[_cpos] == '-') {
      _cpos++;
    }
    decimal();
    return REAL;
  case '.': 
    {
      int pos = _cpos;
      int c;
      do {
	c = _chars[++_cpos];
      }
      while (c >= '0' && c <= '9');
      if (_cpos == pos+1) {
	throw Exception(*this,"wrong number format");
      }
      if (c == 'e' || c =='E') {
	_cpos++;
	if (_chars[_cpos] == '+' || _chars[_cpos] == '-') {
	  _cpos++;
	}
	decimal();
      }
    }
    return REAL;
  default:
    return INT;
  }
} 

void TPLexer::decimal()
{
  CALL("TPLexer::decimal");

  switch (_chars[_cpos]) {
  case '0':
    _cpos++;
    return;
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    break;
  default:
    throw Exception(*this,"wrong number format");
  }


  int c;
  do {
    c = _chars[++_cpos];
  }
  while (c >= '0' && c <= '9');
} 

void TPLexer::positiveDecimal()
{
  CALL("TPLexer::positiveDecimal");
  switch (_chars[_cpos]) {
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    break;
  default:
    throw Exception(*this,"wrong number format");
  }


  int c;
  do {
    c = _chars[++_cpos];
  }
  while (c >= '0' && c <= '9');
} 

