
/*
 * File TheoryAxioms.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */


#include "Lib/Environment.hpp"
#include "Lib/Stack.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/EqHelper.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Inference.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"
#include "Kernel/TermIterators.hpp"
#include "Kernel/Theory.hpp"

#include "AxiomGenerator.hpp"
#include "Property.hpp"
#include "SymCounter.hpp"

#include "TheoryAxioms.hpp"


namespace Shell
{
using namespace Lib;
using namespace Kernel;

void TheoryAxioms::addTheoryUnit(Literal* lit, UnitList*& units)
{
  CALL("TheoryAxioms::addTheoryUnit");

  Clause* unit = Clause::fromIterator(getSingletonIterator(lit), Unit::AXIOM, new Inference(Inference::THEORY));
  UnitList::push(unit, units);
  LOG_UNIT("arith_axioms",unit);
}

void TheoryAxioms::addTheoryClause(UnitList*& units, Literal* lit1, Literal* lit2, Literal* lit3)
{
  CALL("TheoryAxioms::addTheoryClause");

  static LiteralStack lits;
  lits.reset();
  ASS(lit1);
  lits.push(lit1);
  ASS(lit2);
  lits.push(lit2);
  if(lit3) {
    lits.push(lit3);
  }

  Clause* cl = Clause::fromStack(lits, Unit::AXIOM, new Inference(Inference::THEORY));
  UnitList::push(cl, units);
  LOG_UNIT("arith_axioms",cl);
}


void TheoryAxioms::addCommutativity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addCommutativity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned func = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList v1(0,false);
  TermList v2(1,false);
  TermList f12(Term::create2(func, v1, v2));
  TermList f21(Term::create2(func, v2, v1));
  Literal* eq = Literal::createEquality(true, f12, f21, srt);

  addTheoryUnit(eq, units);
}

void TheoryAxioms::addAssociativity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addCommutativity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned func = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList v1(0,false);
  TermList v2(1,false);
  TermList v3(2,false);
  TermList f12(Term::create2(func, v1, v2));
  TermList f23(Term::create2(func, v2, v3));
  TermList f1f23(Term::create2(func, v1, f23));
  TermList ff12_3(Term::create2(func, f12, v3));
  Literal* eq = Literal::createEquality(true, f1f23, ff12_3, srt);

  addTheoryUnit(eq, units);
}

void TheoryAxioms::addIdentity(Interpretation op, TermList idElement, UnitList*& units)
{
  CALL("TheoryAxioms::addIdentity");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned func = env.signature->getInterpretingSymbol(op);
  unsigned srt = theory->getOperationSort(op);
  TermList v1(0,false);
  TermList f1I(Term::create2(func, v1, idElement));
  Literal* eq = Literal::createEquality(true, f1I, v1, srt);

  addTheoryUnit(eq, units);
}

void TheoryAxioms::addCommutativeGroupAxioms(Interpretation op, Interpretation inverse, TermList idElement, UnitList*& units)
{
  CALL("TheoryAxioms::addCommutativeGroupAxioms");
  ASS(theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);
  ASS(theory->isFunction(inverse));
  ASS_EQ(theory->getArity(inverse),1);

  addCommutativity(op, units);
  addAssociativity(op, units);
  addIdentity(op, idElement, units);

  

  unsigned opFunc = env.signature->getInterpretingSymbol(op);
  unsigned invFunc = env.signature->getInterpretingSymbol(inverse);
  unsigned srt = theory->getOperationSort(op);
  ASS_EQ(srt, theory->getOperationSort(inverse));
  TermList v1(0,false);
  TermList v2(1,false);
  TermList f12(Term::create2(opFunc, v1, v2));
  TermList nv1(Term::create1(invFunc, v1));
  TermList nv2(Term::create1(invFunc, v2));
  TermList nf12(Term::create1(invFunc, f12));
  TermList fn1n2(Term::create2(opFunc, nv2, nv1));
  Literal* eq1 = Literal::createEquality(true, nf12, fn1n2, srt);
  addTheoryUnit(eq1, units);

  
  TermList f1n1(Term::create2(opFunc, v1, nv1));
  Literal* eq2 = Literal::createEquality(true, f1n1, idElement, srt);
  addTheoryUnit(eq2, units);
}

void TheoryAxioms::addReflexivity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addReflexivity");

  ASS(!theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned opPred = env.signature->getInterpretingSymbol(op);
  TermList v1(0,false);
  Literal* l11 = Literal::create2(opPred, true, v1, v1);
  addTheoryUnit(l11, units);
}

void TheoryAxioms::addTransitivity(Interpretation op, UnitList*& units)
{
  CALL("TheoryAxioms::addTransitivity");
  ASS(!theory->isFunction(op));
  ASS_EQ(theory->getArity(op),2);

  unsigned opPred = env.signature->getInterpretingSymbol(op);
  TermList v1(0,false);
  TermList v2(1,false);
  TermList v3(2,false);

  Literal* nonL12 = Literal::create2(opPred, false, v1, v2);
  Literal* nonL23 = Literal::create2(opPred, false, v2, v3);
  Literal* l13 = Literal::create2(opPred, true, v1, v3);

  addTheoryClause(units, nonL12, nonL23, l13);
}

void TheoryAxioms::addOrderingTotality(Interpretation lessEqual, UnitList*& units)
{
  CALL("TheoryAxioms::addOrderingTotality");
  ASS(!theory->isFunction(lessEqual));
  ASS_EQ(theory->getArity(lessEqual),2);

  unsigned opPred = env.signature->getInterpretingSymbol(lessEqual);
  TermList v1(0,false);
  TermList v2(1,false);

  Literal* l12 = Literal::create2(opPred, true, v1, v2);
  Literal* l21 = Literal::create2(opPred, true, v2, v1);

  addTheoryClause(units, l12, l21);
}

void TheoryAxioms::addTotalOrderAxioms(Interpretation lessEqual, UnitList*& units)
{
  CALL("TheoryAxioms::addTotalOrderAxioms");

  addReflexivity(lessEqual, units);
  addTransitivity(lessEqual, units);
  addOrderingTotality(lessEqual, units);
}

void TheoryAxioms::addMonotonicity(Interpretation lessEqual, Interpretation addition, UnitList*& units)
{
  CALL("TheoryAxioms::addMonotonicity");
  ASS(!theory->isFunction(lessEqual));
  ASS_EQ(theory->getArity(lessEqual),2);
  ASS(theory->isFunction(addition));
  ASS_EQ(theory->getArity(addition),2);

  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned addFun = env.signature->getInterpretingSymbol(addition);
  TermList v1(0,false);
  TermList v2(1,false);
  TermList v3(2,false);
  TermList v1Pv3(Term::create2(addFun, v1,v3));
  TermList v2Pv3(Term::create2(addFun, v2,v3));
  Literal* nonLe = Literal::create2(lePred, false, v1, v2);
  Literal* leAdded = Literal::create2(lePred, true, v1Pv3, v2Pv3);

  addTheoryClause(units, nonLe, leAdded);
}

void TheoryAxioms::addAdditionAndOrderingAxioms(Interpretation plus, Interpretation unaryMinus,
    TermList zeroElement, TermList oneElement, Interpretation lessEqual, UnitList*& units)
{
  CALL("TheoryAxioms::addAdditionAndOrderingAxioms");

  addCommutativeGroupAxioms(plus, unaryMinus, zeroElement, units);
  addTotalOrderAxioms(lessEqual, units);
  addMonotonicity(lessEqual, plus, units);

  
  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  Literal* nonLeOneZero = Literal::create2(lePred, false, oneElement, zeroElement);
  addTheoryUnit(nonLeOneZero, units);

  
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  TermList v1(0,false);
  TermList v2(1,false);
  Literal* nonLe21 = Literal::create2(lePred, false, v2, v1);
  TermList v1POne(Term::create2(plusFun, v1, oneElement));
  Literal* nonLt1POne2 = Literal::create2(lePred, false, v1POne, v2);
  addTheoryClause(units, nonLe21, nonLt1POne2);

  
  

  unsigned varSort = theory->getOperationSort(lessEqual);
  Literal* v1EqV2 = Literal::createEquality(true, v1, v2, varSort);
  Literal* nonLe12 = Literal::create2(lePred, false, v1, v2);
  addTheoryClause(units, nonLe21, nonLe12, v1EqV2);
}

void TheoryAxioms::addAdditionOrderingAndMultiplicationAxioms(Interpretation plus, Interpretation unaryMinus,
    TermList zeroElement, TermList oneElement, Interpretation lessEqual, Interpretation multiply,
    UnitList*& units)
{
  CALL("TheoryAxioms::addAdditionOrderingAndMultiplicationAxioms");

  unsigned srt = theory->getOperationSort(plus);
  ASS_EQ(srt, theory->getOperationSort(unaryMinus));
  ASS_EQ(srt, theory->getOperationSort(lessEqual));
  ASS_EQ(srt, theory->getOperationSort(multiply));

  addAdditionAndOrderingAxioms(plus, unaryMinus, zeroElement, oneElement, lessEqual, units);

  addCommutativity(multiply, units);
  addAssociativity(multiply, units);
  addIdentity(multiply, oneElement, units);

  
  unsigned mulFun = env.signature->getInterpretingSymbol(multiply);
  TermList v1(0,false);
  TermList v1MulZero(Term::create2(mulFun, v1, zeroElement));
  Literal* v1EqV1MulZero = Literal::createEquality(true, v1MulZero, zeroElement, srt);
  addTheoryUnit(v1EqV1MulZero, units);

  
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  TermList v2(1,false);
  TermList v2POne(Term::create2(plusFun, v2, oneElement));
  TermList v1MulV2POne(Term::create2(mulFun, v1, v2POne));
  TermList v1MulV2(Term::create2(mulFun, v1, v2));
  TermList v1MulV2PV1(Term::create2(plusFun, v1MulV2, v1));
  Literal* succDistrEq = Literal::createEquality(true, v1MulV2POne, v1MulV2PV1, srt);
  addTheoryUnit(succDistrEq, units);

  
  TermList v3(2,false);
  TermList v4(3,false);
  TermList v1Pv2(Term::create2(plusFun, v1, v2));
  TermList v3Pv4(Term::create2(plusFun, v3, v4));
  TermList distrLhs(Term::create2(mulFun, v1Pv2, v3Pv4));
  TermList v1Mv3(Term::create2(mulFun, v1, v3));
  TermList v1Mv4(Term::create2(mulFun, v1, v4));
  TermList v2Mv3(Term::create2(mulFun, v2, v3));
  TermList v2Mv4(Term::create2(mulFun, v2, v4));
  TermList add1(Term::create2(plusFun, v1Mv3, v1Mv4));
  TermList add2(Term::create2(plusFun, v2Mv3, v2Mv4));
  TermList distrRhs(Term::create2(plusFun, add1, add2));
  Literal* distrEq = Literal::createEquality(true, distrLhs, distrRhs, srt);
  addTheoryUnit(distrEq, units);
}

void TheoryAxioms::addExtraIntegerOrderingAxiom(Interpretation plus, TermList oneElement, Interpretation lessEqual,
    UnitList*& units)
{
  CALL("TheoryAxioms::addExtraIntegerOrderingAxiom");

  
  unsigned lePred = env.signature->getInterpretingSymbol(lessEqual);
  unsigned plusFun = env.signature->getInterpretingSymbol(plus);
  TermList v1(0,false);
  TermList v2(1,false);
  Literal* le21 = Literal::create2(lePred, true, v2, v1);
  TermList v1POne(Term::create2(plusFun, v1, oneElement));
  Literal* lt1POne2 = Literal::create2(lePred, true, v1POne, v2);
  addTheoryClause(units, le21, lt1POne2);
}










void TheoryAxioms::apply(Problem& prb)
{
  CALL("TheoryAxioms::apply(Problem&)");

  Property* prop = prb.getProperty();
  if(apply(prb.units(), prop)) {
    prb.invalidateProperty();
    prb.reportEqualityAdded(false);
  }
}

bool TheoryAxioms::apply(UnitList*& units, Property* prop)
{
  CALL("TheoryAxioms::apply");

  bool modified = false;
  {
    bool haveIntPlus =
	prop->hasInterpretedOperation(Theory::INT_PLUS) ||
	prop->hasInterpretedOperation(Theory::INT_UNARY_MINUS) ||
	prop->hasInterpretedOperation(Theory::INT_LESS_EQUAL) ||
	prop->hasInterpretedOperation(Theory::INT_MULTIPLY);
    bool haveIntMultiply =
	prop->hasInterpretedOperation(Theory::INT_MULTIPLY);
    if(haveIntPlus) {
      TermList zero(theory->representConstant(IntegerConstantType(0)));
      TermList one(theory->representConstant(IntegerConstantType(1)));
      if(haveIntMultiply) {
	addAdditionOrderingAndMultiplicationAxioms(Theory::INT_PLUS, Theory::INT_UNARY_MINUS, zero, one,
	    Theory::INT_LESS_EQUAL, Theory::INT_MULTIPLY, units);
      }
      else {
	addAdditionAndOrderingAxioms(Theory::INT_PLUS, Theory::INT_UNARY_MINUS, zero, one,
	    Theory::INT_LESS_EQUAL, units);
      }
      addExtraIntegerOrderingAxiom(Theory::INT_PLUS, one, Theory::INT_LESS_EQUAL, units);
      modified = true;
    }
  }

  {
    bool haveRatPlus =
	prop->hasInterpretedOperation(Theory::RAT_PLUS) ||
	prop->hasInterpretedOperation(Theory::RAT_UNARY_MINUS) ||
	prop->hasInterpretedOperation(Theory::RAT_LESS_EQUAL) ||
	prop->hasInterpretedOperation(Theory::RAT_MULTIPLY);
    bool haveRatMultiply =
	prop->hasInterpretedOperation(Theory::RAT_MULTIPLY);
    if(haveRatPlus) {
      TermList zero(theory->representConstant(RationalConstantType(0, 1)));
      TermList one(theory->representConstant(RationalConstantType(1, 1)));
      if(haveRatMultiply) {
	addAdditionOrderingAndMultiplicationAxioms(Theory::RAT_PLUS, Theory::RAT_UNARY_MINUS, zero, one,
	    Theory::RAT_LESS_EQUAL, Theory::RAT_MULTIPLY, units);
      }
      else {
	addAdditionAndOrderingAxioms(Theory::RAT_PLUS, Theory::RAT_UNARY_MINUS, zero, one,
	    Theory::RAT_LESS_EQUAL, units);
      }
      modified = true;
    }
  }

  {
    bool haveRealPlus =
	prop->hasInterpretedOperation(Theory::REAL_PLUS) ||
	prop->hasInterpretedOperation(Theory::REAL_UNARY_MINUS) ||
	prop->hasInterpretedOperation(Theory::REAL_LESS_EQUAL) ||
	prop->hasInterpretedOperation(Theory::REAL_MULTIPLY);
    bool haveRealMultiply =
	prop->hasInterpretedOperation(Theory::REAL_MULTIPLY);
    if(haveRealPlus) {
      TermList zero(theory->representConstant(RealConstantType(RationalConstantType(0, 1))));
      TermList one(theory->representConstant(RealConstantType(RationalConstantType(1, 1))));
      if(haveRealMultiply) {
	addAdditionOrderingAndMultiplicationAxioms(Theory::REAL_PLUS, Theory::REAL_UNARY_MINUS, zero, one,
	    Theory::REAL_LESS_EQUAL, Theory::REAL_MULTIPLY, units);
      }
      else {
	addAdditionAndOrderingAxioms(Theory::REAL_PLUS, Theory::REAL_UNARY_MINUS, zero, one,
	    Theory::REAL_LESS_EQUAL, units);
      }
      modified = true;
    }
  }
  return modified;
}

}
