
/*
 * File vampire.cpp.
 *
 * This file is part of the source code of the software program
 * Vampire as used in CASC-J6. It is protected by applicable
 * copyright laws.
 *
 * Posession of, or access to, this program does not
 * give you or anybody else a right or licence to
 * distribute, modify, copy, compile, create derivatives,
 * or use in any form or for any purpose, this program,
 * or any part thereof.
 * Any licence to use Vampire shall only be obtained
 * as described on Vampire's home page http://www.vprover.org.
 *
 * This file was included in the Vampire source code
 * for CASC-J6 only because so is required by the CASC-J6
 * rules, and for no other reason.
 */



#include <string>
#include <iostream>
#include <ostream>
#include <fstream>
#include <csignal>

#include "Debug/Tracer.hpp"

#include "Lib/Exception.hpp"
#include "Lib/Environment.hpp"
#include "Lib/Int.hpp"
#include "Lib/MapToLIFO.hpp"
#include "Lib/Random.hpp"
#include "Lib/Set.hpp"
#include "Lib/Stack.hpp"
#include "Lib/TimeCounter.hpp"
#include "Lib/Timer.hpp"

#include "Lib/List.hpp"
#include "Lib/Vector.hpp"
#include "Lib/System.hpp"
#include "Lib/Metaiterators.hpp"

#include "Kernel/Clause.hpp"
#include "Kernel/Formula.hpp"
#include "Kernel/FormulaUnit.hpp"
#include "Kernel/Problem.hpp"
#include "Kernel/Signature.hpp"
#include "Kernel/Term.hpp"

#include "Indexing/TermSharing.hpp"

#include "Inferences/InferenceEngine.hpp"
#include "Inferences/TautologyDeletionISE.hpp"

#include "InstGen/IGAlgorithm.hpp"

#include "SAT/DIMACS.hpp"

#include "Shell/CASC/CASCMode.hpp"
#include "Shell/CASC/CLTBMode.hpp"
#include "Shell/CASC/CMZRMode.hpp"
#include "Shell/CASC/SimpleLTBMode.hpp"
#include "Shell/CParser.hpp"
#include "Shell/CommandLine.hpp"
#include "Shell/EqualityProxy.hpp"
#include "Shell/Grounding.hpp"
#include "Shell/Normalisation.hpp"
#include "Shell/Options.hpp"
#include "Shell/Property.hpp"
#include "Saturation/ProvingHelper.hpp"
#include "Shell/Preprocess.hpp"
#include "Shell/Refutation.hpp"
#include "Shell/TheoryFinder.hpp"
#include "Shell/TPTP.hpp"
#include "Parse/TPTP.hpp"
#include "Shell/SpecialTermElimination.hpp"
#include "Shell/Statistics.hpp"
#include "Shell/UIHelper.hpp"

#include "Saturation/SaturationAlgorithm.hpp"


#if CHECK_LEAKS
#include "Lib/MemoryLeak.hpp"
#endif

#define SPIDER 0
#define SAVE_SPIDER_PROPERTIES 0

using namespace Shell;
using namespace SAT;
using namespace Saturation;
using namespace Inferences;
using namespace InstGen;

Problem* globProblem=0;

int vampireReturnValue = VAMP_RESULT_STATUS_UNKNOWN;

Problem* getPreprocessedProblem()
{
  CALL("getInputClauses");
  
  Problem* prb=UIHelper::getInputProblem(*env.options);

  TimeCounter tc2(TC_PREPROCESSING);

  Preprocess prepro(*env.options);
  
  prepro.preprocess(*prb);
  globProblem=prb;

  return prb;
}

void doProving()
{
  CALL("doProving()");
  ScopedPtr<Problem> prb(getPreprocessedProblem());
  ProvingHelper::runVampireSaturation(*prb, *env.options);
}

void explainException (Exception& exception)
{
  env.beginOutput();
  exception.cry(env.out());
  env.endOutput();
} 

int main(int argc, char* argv [])
{
  CALL ("main");

  System::registerArgv0(argv[0]);
  System::setSignalHandlers();
   
  Lib::Random::setSeed(123456);

  try {
    
    Shell::CommandLine cl(argc,argv);
    cl.interpret(*env.options);

    PROCESS_TRACE_SPEC_STRING(env.options->traceSpecString());
    env.options->enableTracesAccordingToOptions();

    if(env.options->showOptions()) {
      env.beginOutput();
      env.options->output(env.out());
      env.endOutput();
    }

    Allocator::setMemoryLimit(env.options->memoryLimit()*1048576ul);
    Lib::Random::setSeed(env.options->randomSeed());

    switch (env.options->mode())
    {
    case Options::MODE_CASC:
      if(Shell::CASC::CASCMode::perform(argc,argv)) {	
	vampireReturnValue = VAMP_RESULT_STATUS_SUCCESS;
      }
      break;
    case Options::MODE_CASC_SAT:
      Shell::CASC::CASCMode::makeSat();
      if(Shell::CASC::CASCMode::perform(argc,argv)) {
	
	vampireReturnValue = VAMP_RESULT_STATUS_SUCCESS;
      }
      break;
    case Options::MODE_CASC_EPR:
      Shell::CASC::CASCMode::makeEPR();
      if(Shell::CASC::CASCMode::perform(argc,argv)) {
	
	vampireReturnValue = VAMP_RESULT_STATUS_SUCCESS;
      }
      break;
    case Options::MODE_CASC_SIMPLE_LTB:
    {
      Shell::CASC::SimpleLTBMode sltbm;
      sltbm.perform();
      
      vampireReturnValue = VAMP_RESULT_STATUS_SUCCESS;
      break;
    }
    case Options::MODE_CASC_LTB:
    {
      Shell::CASC::CLTBMode::perform();
      
      vampireReturnValue = VAMP_RESULT_STATUS_SUCCESS;
      break;
    }
    case Options::MODE_CASC_MZR:
    {
      Shell::CASC::CMZRMode::perform();
      
      vampireReturnValue = VAMP_RESULT_STATUS_SUCCESS;
      break;
    }
    default:
      USER_ERROR("Unsupported mode");
    }
#if CHECK_LEAKS
    if (globUnitList) {
      MemoryLeak leak;
      leak.release(globUnitList);
    }
    delete env.signature;
    env.signature = 0;
#endif
  }
#if VDEBUG
  catch (Debug::AssertionViolationException& exception) {
    vampireReturnValue = VAMP_RESULT_STATUS_UNHANDLED_EXCEPTION;
    reportSpiderFail();
#if CHECK_LEAKS
    MemoryLeak::cancelReport();
#endif
  }
#endif
  catch (UserErrorException& exception) {
    vampireReturnValue = VAMP_RESULT_STATUS_UNHANDLED_EXCEPTION;
    reportSpiderFail();
#if CHECK_LEAKS
    MemoryLeak::cancelReport();
#endif
    explainException(exception);
  }
  catch (Exception& exception) {
    vampireReturnValue = VAMP_RESULT_STATUS_UNHANDLED_EXCEPTION;
    reportSpiderFail();
#if CHECK_LEAKS
    MemoryLeak::cancelReport();
#endif
    env.beginOutput();
    explainException(exception);
    env.statistics->print(env.out());
    env.endOutput();
  }
  catch (std::bad_alloc& _) {
    vampireReturnValue = VAMP_RESULT_STATUS_UNHANDLED_EXCEPTION;
    reportSpiderFail();
#if CHECK_LEAKS
    MemoryLeak::cancelReport();
#endif
    env.beginOutput();
    env.out() << "Insufficient system memory" << '\n';
    env.endOutput();
  }


  return vampireReturnValue;
} 

